/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger;

import org.antlr.runtime.RecognitionException;
import org.ow2.dsrg.fm.badger.reference.TranslationArchitecture;
import org.ow2.dsrg.fm.badger.reference.TranslationReference;
import org.ow2.dsrg.fm.badger.simulation.transition.BehaviorInterpreter;
import org.ow2.dsrg.fm.badger.traversal.Traversal;
import org.ow2.dsrg.fm.badger.util.CheckingChain;
import org.ow2.dsrg.fm.tbplib.TBPLibLog;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.architecture.Interface;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.ParserResult;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.node.visitor.SymbolCollectingVisitor;
import org.ow2.dsrg.fm.tbplib.parser.ParserFacade;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.util.TransformationChain;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import org.ow2.dsrg.fm.badger.simulation.TBPPropertyViolationException;

/**
 * Main class for analysis of a closed TBP specification using simple architectural description.
 * Parses TBP specification, transforms it to LTS specification and than
 * generates Java stack machine and uses it to state space traversal.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class CheckArch {



    public static class ArchitectureReader {

        Architecture arch;
        List<TBPSpecification<String>> tbps;

        String dir;

        public Architecture getArch() {
            return arch;
        }

        public List<TBPSpecification<String>> getTbps() {
            return tbps;
        }

        int lineCount = -1;



        public class ArchParseException extends Exception{
            public ArchParseException(String msg){
                super(msg+ " at " + lineCount);
            }
        }




        public ArchitectureReader(String fname) throws IOException, RecognitionException, ArchParseException {

            BadgerLog.logger.info("Parsing architecture file:" + fname);

            File f = new File(fname);
            dir = f.getParent();
            if (dir==null) {
                dir=".";
            }

            BufferedReader r = new BufferedReader(new FileReader(fname));
            String line;

            while ((line=r.readLine())!=null){
                lineCount++;
                line = line.trim();
                if (line.startsWith("#")||line.isEmpty()) continue;
                String[] tokens = line.split(" ");
                String keyword = tokens[0];
                if (keyword.equals("architecture")){
                    createArchitecture(tokens[1]);
                } else if (keyword.equals("component")) {
                    addComponent(tokens[1], tokens[2]);
                } else if (keyword.equals("bind")) {
                    bind(tokens[1],tokens[2]);
                }
            }
        }


        void createArchitecture(String name) throws ArchParseException {
            if (arch!=null) throw new ArchParseException("Multiple architectures in file");
            arch = new TranslationArchitecture(name);
            tbps = new ArrayList<TBPSpecification<String>>(10);
        }


        void addComponent(String instanceName, String filename) throws IOException, RecognitionException {
            TBPLibLog.logger.info("Parsing file: \"" + filename + "\"");

            String absFileName = dir+File.separator+filename;

            ParserResult res = ParserFacade.parseFile(absFileName);
            TBPSpecification<String> tbp = res.getSpecification();
            tbp.setComponentInstance(instanceName);
            tbps.add(tbp);

            Component c = new Component(instanceName,res.getTypes());
            arch.addComponent(c);

            SymbolCollectingVisitor visitor = new SymbolCollectingVisitor(arch);
            visitor.setComponent(c);
            tbp.visit(visitor);
        }


        Component getComponent(String n) throws ArchParseException {
             Component ret = arch.getComponent(n);
            if (ret==null){
                throw new ArchParseException("Component "+ n + " not found");
            }
            return ret;
        }


        void bind(String from, String to) throws ArchParseException {
            String[] fromTokens = from.split("\\.");
            String[] toTokens = to.split("\\.");

            if (fromTokens.length!=2) {
                throw new ArchParseException("Wrong format of required interface");
            }

            if (toTokens.length!=2) {
                throw new ArchParseException("Wrong format of provided interface");
            }

            Component fromComponent = getComponent(fromTokens[0]);
            Interface requiredIface = fromComponent.getRequiredInterface(fromTokens[1]);
            if (requiredIface==null){
                throw new ArchParseException("Required interface "+ fromTokens[0]+"."+fromTokens[1] + " not found");
            }


            String toComponentStr = toTokens[0];
            Component toComponent = getComponent(toComponentStr);
            Interface providedIface = toComponent.getProvidedInterface(toTokens[1]);
            if (providedIface==null){
                throw new ArchParseException("ProvidedIface interface "+ toTokens[0]+"."+toTokens[1] + " not found");
            }

            requiredIface.bindTo(providedIface);

        }
    }



    public static void main(String[] args) throws Exception {

        if (args.length != 1) {
            BadgerLog.logger.severe("No input files!");
            return;
        }

        ArchitectureReader ar = new ArchitectureReader(args[0]);

        EventTable table = new EventTableImpl();

        List<TBPSpecification<String>> tbps = ar.getTbps();

        Architecture arch =ar.getArch();

        TranslationArchitecture tArch =
                new TranslationArchitecture(arch);

        BadgerLog.logger.info("Implicit binding");
        TransformationChain.performImplicitBinding(tArch);

        BadgerLog.logger.info("Resolving symbols");
        List<TBPSpecification<Reference>> ref_tbps =
                TransformationChain.resolveArchitecture(tbps, tArch);
        TransformationChain.fillEventTable(arch, table);

        BadgerLog.logger.info("Transformation of provisions");
        List<Observer> provisions = TransformationChain.transformProvisions(ref_tbps, table);



        BadgerLog.logger.info("Creating composition of components");
        TBPSpecification tbpComp =
                TransformationChain.performComposition(ref_tbps, null);
        TransformationChain.replaceArbitraryValuesInImperativePartsByAlternatives(tbpComp);

        BadgerLog.logger.info("Transformation to LTS represenation");
        LTSSpecification<TranslationReference> ltsSpecification =
                CheckingChain.transformToLTS(tbpComp, tArch);

        BadgerLog.logger.info("Transformation to Java");
        BehaviorInterpreter interpreter =
            CheckingChain.transformFromLTSToJava(ltsSpecification, provisions, tArch, table);

        if (interpreter == null) {
            BadgerLog.logger.severe("Transformation to Java has failed.");
            return;
        }

        BadgerLog.logger.info("Traversing the composite state space");
        Traversal tr = new Traversal(interpreter, table);
        try{
        tr.traverse();
        } catch (TBPPropertyViolationException ex) {
            ex.printError(System.out);
        }

    }
}