/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger;

import org.ow2.dsrg.fm.badger.reference.TranslationArchitecture;
import org.ow2.dsrg.fm.badger.reference.TranslationReference;
import org.ow2.dsrg.fm.badger.simulation.transition.BehaviorInterpreter;
import org.ow2.dsrg.fm.badger.traversal.Traversal;
import org.ow2.dsrg.fm.badger.util.CheckingChain;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.util.TransformationChain;

import java.util.Arrays;
import java.util.List;
import org.ow2.dsrg.fm.badger.simulation.TBPPropertyViolationException;

/**
 * Main class for analysis of a single closed TBP specification using implicit
 * bindings.
 * Parses TBP specification, transforms it to LTS specification and than
 * generates Java stack machine and uses it to state space traversal.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class CheckSingleLTSRepresentation {




    private static BehaviorInterpreter createInterpreterFromParsedArchitecture( Architecture arch, List<TBPSpecification<String>> tbps, EventTable table) throws Exception {
        TransformationChain.collectSymbolsFromArchitecture(tbps, arch);

        TranslationArchitecture tArch =
                new TranslationArchitecture(arch);

        BadgerLog.logger.info("Implicit binding");
        TransformationChain.performImplicitBinding(tArch);

        BadgerLog.logger.info("Resolving symbols");
        List<TBPSpecification<Reference>> ref_tbps =
                TransformationChain.resolveArchitecture(tbps, tArch);
        TransformationChain.fillEventTable(arch, table);

        BadgerLog.logger.info("Transformation of provisions");
        List<Observer> provisions = TransformationChain.transformProvisions(ref_tbps, table);

        BadgerLog.logger.info("Creating composition of components");
        TBPSpecification tbpComp =
                TransformationChain.performComposition(ref_tbps, null);
        TransformationChain.replaceArbitraryValuesInImperativePartsByAlternatives(tbpComp);

        BadgerLog.logger.info("Transformation to LTS represenation");
        LTSSpecification<TranslationReference> ltsSpecification =
                CheckingChain.transformToLTS(tbpComp, tArch);

        BadgerLog.logger.info("Transformation to Java");
        BehaviorInterpreter interpreter =
            CheckingChain.transformFromLTSToJava(ltsSpecification, provisions, tArch, table);
        return interpreter;
    }

    public static BehaviorInterpreter createInterpreterFromFiles(String args[], EventTable table) throws Exception {
        BadgerLog.logger.info("Parsing files: " + Arrays.toString(args));


        Architecture arch = new Architecture("testing-arch");
        List<TBPSpecification<String>> tbps = TransformationChain.parseFiles(arch, args);
        return createInterpreterFromParsedArchitecture(arch, tbps, table);
        
    }

    public static BehaviorInterpreter createInterpreterFromFile(String filename, EventTable table) throws Exception {
        Architecture arch = new Architecture("testing-arch");
        List<TBPSpecification<String>> tbps =
                TransformationChain.parseFiles(arch, filename);

        return createInterpreterFromParsedArchitecture(arch, tbps, table);

    }

    public static void main(String[] args) throws Exception {

        if (args.length < 1) {
            BadgerLog.logger.severe("No input files!");
            return;
        }

        /*
        InputStream in = null;
        AudioStream as = null;
        try {
            in = new FileInputStream("data/badgerbadgerbadger.wav");
            as = new AudioStream(in);
            AudioPlayer.player.start(as);
        } catch (Exception e) {
        }*/

        EventTable table = new EventTableImpl();

        BehaviorInterpreter interpreter = createInterpreterFromFiles(args, table);
        
        if (interpreter == null) {
            BadgerLog.logger.severe("Transformation to Java has failed.");
            /*
            if (as != null) {
                AudioPlayer.player.stop(as);
                AudioPlayer.player.interrupt();
            }
            System.exit(0);
             */
            return;
        }

        BadgerLog.logger.info("Traversing the composite state space");
        Traversal tr = new Traversal(interpreter, table);
        try{
        tr.traverse();
        } catch (TBPPropertyViolationException ex) {
            ex.printError(System.out);
        }
        
        
        
        
        
        
    }
}
