/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C)      2011  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca;

import java.util.LinkedList;
import java.util.List;

/** Wraps parameters and options available for TBP checking.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class Parameters {
	private static class ParametersException extends Exception {
		private static final long serialVersionUID = 3791787737064626842L;
		private String param;
		public ParametersException(String opt) {
			this(opt, null);
		}
		public ParametersException(String opt, String message) {
			super(message);
			this.param = opt;
		}
		public String getOpt() {
			return param;
		}
	}
	
	private static final int DEFAULT_TREE_FRAGMENT_WIDTH = 100;
	private static final String DEFAULT_STATS_FILE_NAME = "stats.xml";
	
	private String[] fileNames;
	/** Print debugging information &larr; flush important data structures to DOT and text files. */
	private boolean printDI = false; // -debug
	/** Prefix debugging information file names. */
	private String diFileNamePrefix = "DI_";
	/** Use observers. */
	private boolean forceFiniteAutomata = false; // -fa
	/** Use observing automata. */
	private boolean forceCounterAutomata = false; // -pa
	/** Continue tree construction after protocol violation was found. */
	private boolean continueAfterViolation = false; // -c
	
	/** Print stats. */
	private boolean printStats = false; // -s
	
	///** Check provisions actuated with actuators generated from the provisions. */
	//private boolean checkActuatedProvisions = false;
	//** Use variation of provision automata that decrements counters. */
	//public boolean useDecrementingProvisionAutomata = false;
	//** Track unbound counters of provision automata. */
	//public boolean trackUnboundCounters = true;
	//** Print logs of current memory usage. */
	//public boolean logMemoryUsage = true;
	
	/** Print Karp-Miller Tree. */
	private boolean printKMTree = false; // -t
	/** Print Karp-Miller Tree Fragments. */
	private boolean printKMTreeFragments = false; // -f
	/** Width of Karp-Miller Tree Fragments. */
	private int KMTFragmentWidth = 0; // -f=N
	/** Karp-Miller Tree Fragments dir. */
	private String KMTFragmentsDir = "TreeFragments";
	
	public String[] getFileNames() {
		return this.fileNames;
	}
	public boolean getPrintDebugInfo() {
		return printDI;
	}
	public String getDebugInfoFileNamePrefix() {
		return diFileNamePrefix;
	}
	public boolean getForceFiniteAutomata() {
		return forceFiniteAutomata;
	}
	public boolean getForceCounterAutomata() {
		return forceCounterAutomata;
	}
	public boolean getPrintKMTree() {
		return printKMTree;
	}
	public boolean getPrintKMTreeFragments() {
		return printKMTreeFragments;
	}
	public int getKMTreeFragmentWidth() {
		return KMTFragmentWidth;
	}
	public String getKMTreeFragmentsDir() {
		return KMTFragmentsDir;
	}
	public boolean getContinueAfterVioloation() {
		return continueAfterViolation;
	}
	public String getStatsFileName() {
		return DEFAULT_STATS_FILE_NAME;
	}
	public boolean getPrintStats() {
		return printStats;
	}
	
	
	public static Parameters parse(String[] args) {
		Parameters opts = new Parameters();
		int i = 0;
		try {
			// parse options
			while(i < args.length && args[i] != null && args[i].startsWith("-")) {
				String o = args[i];
				if (o.equals("-c")) {
					opts.continueAfterViolation = true;
				} else if (o.equals("-fa")) {
					opts.forceFiniteAutomata = true;
					opts.forceCounterAutomata = false;
				} else if (o.equals("-ca")) {
					opts.forceFiniteAutomata = false;
					opts.forceCounterAutomata = true;
				} else if (o.equals("-s")) {
					opts.printStats = true;
				} else if (o.equals("-debug")) {				
					opts.printDI = true;					
				} else if (o.equals("-t")) {
					if (!opts.printDI) {
						throw new ParametersException(o);
					}
					opts.printKMTree = true;
				} else if (o.equals("-f")) {
					if (!opts.printDI) {
						throw new ParametersException(o);
					}
					opts.printKMTreeFragments = true;
					opts.KMTFragmentWidth = DEFAULT_TREE_FRAGMENT_WIDTH;
				} else if (o.startsWith("-f=")) {
					if (!opts.printDI) {
						throw new ParametersException(o);
					}
					opts.printKMTreeFragments = true;
					String n = o.substring("-f=".length());
					try {
						opts.KMTFragmentWidth = Integer.parseInt(n);
						if (opts.KMTFragmentWidth < 0) {
							throw new Exception();
						}
					} catch (Exception e) {
						throw new ParametersException(o);
					}
				} else {
					throw new ParametersException(o);
				}
				++i;
			}
			// parse arguments
			List<String> tmpArgs = new LinkedList<String>();
			while(i < args.length && args[i] != null) {
				String a = args[i];

				if (a.startsWith("-")) {
					throw new ParametersException(a);
				}
				tmpArgs.add(a);
				++i;
			}
			if (tmpArgs.isEmpty()) {
				throw new ParametersException(null, "Missing input file.");
			}
			opts.fileNames = new String[tmpArgs.size()];
			opts.fileNames = tmpArgs.toArray(opts.fileNames);
			
		} catch (ParametersException pe) {
			opts = null;
			String opt = pe.getOpt();
			String msg = pe.getMessage();
			if (opt == null && msg != null) {
				System.err.println(msg);
			} else if (opt != null){
				if (!opt.equals("-?")) {
					System.err.printf("Invalid option \"%s\".\n", opt);
				}
			} else { // msg == null && opt == null
				System.err.println("Unknown error.");
			}
			// print usage
			System.out.println(getUsage());
		}
		return opts;
	}
	
	public static String getUsage() {
		String programName = "CheckSingle";
		return String.format(
			  "Usage: %s [-c] [-fa|-ca] [-s] [-debug [-t] [-f|-f=N]] component-1.tbp ...\n"
			+ "   -c  continue after protocol violation was reached; this options allows\n"
			+ "       to find more paths that lead to protocol errors\n"
			+ "   -fa use finite automata exclusively for checking provision protocols\n"
			+ "   -ca use counter automata exclusively for checking provision protocols\n"
			+ "       If neither -fa nor -ca is specified, then the best-fit strategy\n"
			+ "       will be used for each provision separately. I.e. counter automata\n"
			+ "       will be used for provisions with unlimited re-entrancy. Finite automata\n"
			+ "       will be used for other provisions.\n"
			+ "\n"
			+ "   -s  print stats to %s\n"
			+ "\n"
			+ "   -debug  print debugging information and utilized data structures\n"
			+ "       to auxilliary files\n"
			+ "   -t  print Karp-Miller Tree to DOT file\n"
			+ "   -f  print fragments of Karp-Miller Tree to DOT files continuously\n"
			+ "       One can controll width of the fragment using -f=N option. Option -f is\n"
			+ "       equivalent to -f=%d\n"
			  ,programName, DEFAULT_STATS_FILE_NAME, DEFAULT_TREE_FRAGMENT_WIDTH);
	}
}
