/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2011       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller;

import java.util.Map;

import org.ow2.dsrg.fm.badger.ca.statespace.GlobalState;
import org.ow2.dsrg.fm.badger.ca.statespace.LocalState;

/** Represents <i>configuration</i> of model execution point.
 * 
 * <p>The formal object consists of shared (global) state <i>s</i>,
 *    set of occupied local states <i>L</i> and valuation of counters
 *    of particular local states <i>f</i>: <i>L</i> &rarr; <b>N</b> &cup; {&omega;}.
 *    Note that <i>f(l)</i> &isin; <b>N</b> gives the number of threads that reside
 *    in <i>l</i>. <i>f(l)</i> = &omega; means, that in <i>l</i> can potentialy
 *    reside any number of threads.  
 * </p>
 * 
 * <p>Since <i>f</i> is limited to <i>L</i>, a {@link Map} is used to represent
 *    both <i>L</i> and <i>f</i>.  
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public interface Configuration<NAME, VAL> {
		
	/** Returns shared state <i>s</i>.
	 * 
	 * @return {@link GlobalState}
	 */
	public GlobalState<NAME, VAL> getShared();
	
	/** Returns {@link Map} representing set of local states <i>L</i>
	 *  and counter valuation function <i>f</i>.
	 * 
	 * @return {@link Map Map&ltLocalState, Counter&gt;}
	 */
	public Map<LocalState<NAME, VAL>, Counter> getLocals();
	
	/** Performs <b><i>omega</i></b> function which introduces &omega; into counters.
	 * 
	 * <p>Adjusts counter values of this.</p> 
	 * 
	 * @param predecessors all predecessors of this till the root 
	 */
	public void omega(Iterable<Configuration<NAME, VAL>> predecessors);
	
	/** Determines whether <b>this</b> <i>covers</i> <b>that</b>.
	 * 
	 *  <p>Configuration <i>A covers</i> configuration <i>B</i> if all
	 *     <b>shared</b> states of <i>A</i> and <i>B</i> are equal
	 *     and <b>local</b> states
	 *     occupied in <i>B</i> are also occupied in <i>A</i> and
	 *     for each such state the value of counter in <i>A</i> is at least the
	 *     same as in in <i>B</i>.
	 *  </p> 
	 * 
	 * @param that {@link Configuration} which will be examined whehter covered by <b>this</b>
	 * @return <code>true</code> if and only if <b>this</b> covers <b>that</b>
	 */
	public boolean covers(Configuration<NAME, VAL> that);
	
	/** Determines whether <b>this</b> <i>covers</i> <b>that</b> in <i>locally</i>.
	 * 
	 *  <p>Configuration <i>A covers</i> configuration <i>B</i>
	 *     <i>locally</i> if all <b>local</b> states
	 *     occupied in <i>B</i> are also occupied in <i>A</i> and
	 *     for each such state the value of counter in <i>A</i> is at least the
	 *     same as in in <i>B</i>.
	 *  </p> 
	 * 
	 * @param that {@link Configuration} which will be examined whehter covered by <b>this</b>
	 * @return <code>true</code> if and only if <b>this</b> covers <b>that</b>
	 */	
	public boolean coversLocally(Configuration<NAME, VAL> that);
	
	/** Creates new configuration from <b>this</b> one using transition <b>t</b>.
	 * 
	 * @param t {@link Transition} which transforms this into another {@link Configuration}
	 * @return {@link Configuration} or null
	 */
	public Configuration<NAME, VAL> transform(Transition<NAME, VAL> t);
	
	/** Marks this configuration with error flag.
	 * 
	 */
	public void setError();
	
	/** Has this configuration been marked with error flag?
	 * 
	 * @return boolean
	 */
	public boolean isError();	
}
