/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2011       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller;

import java.util.Comparator;

import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.provision.CounterLimit;

/** Represents counter value &isin; <b>N</b> &cup; {&omega;}.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class Counter implements Comparable<Counter>, Cloneable {
	
	private static final int OMEGA_VALUE = LTSAThread.INFINITY;
	public static final Comparator<Counter> COMPARATOR = new Comparator<Counter>() {
		@Override
		public int compare(Counter o1, Counter o2) {
			// roughly equivalent to o1.compareTo(o2);
			// o1 = o2 -> 0; o1 < o2 -> -1; o1 > o2 -> 1
			if (o1 == o2) {
				return 0;
			} else if (o1 == null) {
				return -1;
			} else if (o2 == null) {
				return 1;
			} else {
				return o1.compareTo(o2);
			}
		}
	};
	
	private int value;
	
	public Counter(int value) {
		this.value = value;
		assert(value > 0 || value == OMEGA_VALUE);
	}
	
	public int getIntValue() {
		return value;
	}
	public boolean isOmega() {
		return (value == OMEGA_VALUE);
	}
	
	public void setOmega() {
		value = OMEGA_VALUE;
	}
	
	public void increment() {
		if (value != OMEGA_VALUE) {
			value++;
		}
		assert(value > 0 || value == OMEGA_VALUE);		
	}
	public void decrement() {
		if (value > 0) {
			value--; 
		}
		assert(value > 0 || value == OMEGA_VALUE);
	}
	
	@Override
	public int hashCode() {
		return value;
	}
	
	@Override
	public boolean equals(Object o) {
		if (o != null && o instanceof Counter) {
			return (this == o) || ((Counter)o).value == this.value;
		} else {
			return false;
		}
	}
	
	@Override
	public int compareTo(Counter o) {
		if (this.equals(o)) {
			return 0;
		}
		if ( o == null
				|| (this.value == OMEGA_VALUE && o.value != OMEGA_VALUE)
				|| (this.value > o.value && o.value != OMEGA_VALUE)) {
			return 1;
		}
		if ((this.value != OMEGA_VALUE)
				&& (o != null)
				&& (this.value < o.value || o.value == OMEGA_VALUE)) {
			return -1;
		}
		throw new RuntimeException("Counter.compareTo() failed.");
	}

	/** Returns true if <code>this</code> is larger than
	 *  {@link CounterLimit#getUpperLimit() upper limit} of given
	 *  {@link CounterLimit} <b>limit</b>.
	 * 
	 * @param limit {@link CounterLimit}
	 * @return boolean
	 */
	public boolean exceedsUpperLimit(CounterLimit limit) {
		int tmpUpperLimit = limit != null ? limit.getUpperLimit() : CounterLimit.INFINITY;
		return (this.isOmega() && tmpUpperLimit != CounterLimit.INFINITY)
			|| (!this.isOmega() && tmpUpperLimit != CounterLimit.INFINITY && tmpUpperLimit < this.value);			
	}
	
	@Override
	public Counter clone() {
		return new Counter(value);
	}
	
	@Override
	public String toString() {
		return this.isOmega()
			? "&omega;"
			: String.format("%d", this.value);

	}
}
