/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2011       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller;

import java.util.Collection;
import java.util.Deque;
import java.util.LinkedList;

import org.ow2.dsrg.fm.badger.ca.statespace.LocalState;

/** Wraps basic Karp-Miller tree construction algorithm.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 * @param <NAME> variable name type
 * @param <VAL> variable value type
 */
public class TreeConstruction<NAME, VAL> {
	
	public enum SearchStrategy {
		DepthFirst, BreadthFirst
	}
	
	public enum StopStrategy {
		StopOnError, DoNotStop
	}
	
	public static class Settings {
		public static int DEPTH_UNLIMITED = 0;
		public SearchStrategy searchStrategy;
		public StopStrategy stopStrategy;
		public int depthLimit;
		
		public Settings(	SearchStrategy searchStrategy,
							StopStrategy stopStrategy,
							int depthLimit) {
			this.searchStrategy = searchStrategy;
			this.stopStrategy = stopStrategy;
			this.depthLimit = depthLimit > 0 ? depthLimit : DEPTH_UNLIMITED;
		}
		public Settings() {
			this(SearchStrategy.DepthFirst, StopStrategy.DoNotStop, DEPTH_UNLIMITED);
		}
	}
	
	/** Settings of this construction provider. */
	protected Settings settings;
	/** Container for end nodes. Note that this is client-provided. */
	protected Collection<ConfigurationNode<NAME, VAL>> endNodesContainer;
	/** Queue of unexplored nodes utilized during construction algorithm. */
	protected Deque<ConfigurationNode<NAME, VAL>> unexploredNodes;
	/** Collection for nodes with error label. */
	protected Collection<ConfigurationNode<NAME, VAL>> errorNodes;
	
	/** Creates instance of this without container for end nodes.
	 * 
	 * <p>Note that reached <i>end</i> nodes without <i>error</i> label
	 *    will <b>NOT</b> be stored anywhere.</p>
	 * 
	 * @param settings algorithm modifiers
	 */
	public TreeConstruction(Settings settings) {
		this(settings, null);
	}
	
	/** Creates instance of this with container for end nodes <b>endNodesContainer</b>.
	 * 
	 * <p>The <b>endNodesContainer</b> will be filled with reached <i>end</i> nodes
	 *    during tree construction.</p>
	 * 
	 * @param settings algorithm modifiers
	 * @param endNodesContainer
	 */
	public TreeConstruction(Settings settings, Collection<ConfigurationNode<NAME, VAL>> endNodesContainer) {
		this.settings = settings;
		this.endNodesContainer = endNodesContainer;		
	}
	
	/** Initializes the construction.
	 * 
	 * <p>Should initialize {@link #unexploredNodes} and {@link #errorNodes}
	 *    collections and add initial configuration (see {@link Context#getInitialConfiguration()})
	 *    to {@link #unexploredNodes}.
	 * </p>
	 * 
	 * @param ctx
	 */
	protected void init(Context<NAME, VAL> ctx) {
		this.unexploredNodes = new LinkedList<ConfigurationNode<NAME,VAL>>();
		this.unexploredNodes.add(new org.ow2.dsrg.fm.badger.ca.karpmiller.impl.ConfigurationNode<NAME, VAL>(ctx.getInitialConfiguration()));
		this.errorNodes = new LinkedList<ConfigurationNode<NAME,VAL>>();
	}
	
	/** Determines whether <b>node</b> is <i>end</i> node of tree (leaf).
	 * 
	 * <p>Note that this function does not reflect successors of <b>node</b>,
	 *    thus this method can return <code>false</code> even if <b>node</b>
	 *    has no successors. Also any method that overrides this should not
	 *    reflect successors of <b>node</b>.
	 * </p>
	 * 
	 * @param node
	 * @return true if and only if <b>node</b> is <i>end</i> nodes regardless of successors
	 */
	protected boolean isEndNode(ConfigurationNode<NAME, VAL> node) {
		return	(this.settings.depthLimit != Settings.DEPTH_UNLIMITED && node.depth() > this.settings.depthLimit) ||
			node.containsPredecessor(node.getConfiguration());
	}
		
	/** Adds <b>node</b> to the que of unexplored nodes.
	 * 
	 * @param node
	 */
	protected void addUnexplored(ConfigurationNode<NAME, VAL> node) {
		if (settings.searchStrategy == SearchStrategy.DepthFirst) {
			unexploredNodes.addFirst(node);
		} else {
			unexploredNodes.addLast(node);
		}		
	}
	
	/** Registers end node.
	 * 
	 * <p>This method should notify that <i>end</i> node was reached.
	 *    Should add it to {@link #endNodesContainer} if that is not null.
	 * </p>
	 * 
	 * @param node
	 */
	protected void registerEndNode(ConfigurationNode<NAME, VAL> node) {
		if (this.endNodesContainer != null) {
			this.endNodesContainer.add(node);
		}
	}
	
	/** Adds error node to {@link #errorNodes} collection.
	 * 
	 * @param node
	 */
	protected void addErrorNode(ConfigurationNode<NAME, VAL> node) {
		this.errorNodes.add(node);
	}
	
	/** Decides whether to stop construction due to error node <b>node</b>.
	 * 
	 * @param node error node
	 * @return boolean
	 */
	protected boolean doStopConstructionDueToErrorNode(ConfigurationNode<NAME, VAL> node) {
		return this.settings.stopStrategy == StopStrategy.StopOnError;
	}
	
	/** Gets resulting error nodes.
	 * 
	 * <p>Note that as this method is called, the construction ends.
	 *    Only valid calls after this are {@link #init(Context)}
	 *    and {@link #getEndNodes()}.
	 * </p> 
	 * 
	 * @return {@link Collection} of {@link ConfigurationNode}s that correspond to errors
	 */
	protected Collection<ConfigurationNode<NAME, VAL>> getResultNodes() {
		Collection<ConfigurationNode<NAME, VAL>> tmp = errorNodes;
		this.errorNodes = null;
		return tmp;
	}
	
	/** Gets end nodes reached during tree construction.
	 * 
	 * @return  {@link Collection} of <i>end</i> {@link ConfigurationNode nodes}.
	 */
	public Collection<ConfigurationNode<NAME, VAL>> getEndNodes() {
		return endNodesContainer;
	}
	
	/** Performs Karp-Miller tree construction.
	 * 
	 * @param ctx {@link Context}
	 * @return {@link Collection} of {@link ConfigurationNode}s labelled as error.
	 */
	public Collection<ConfigurationNode<NAME, VAL>> constructTree(Context<NAME, VAL> ctx) {
		init(ctx);
		
		while(!unexploredNodes.isEmpty()) {
			ConfigurationNode<NAME, VAL> cNode = unexploredNodes.removeFirst();
			boolean anyNewUnexploredTransition = false;
			if (!isEndNode(cNode)) {
				Configuration<NAME, VAL> c = cNode.getConfiguration();
				for(LocalState<NAME, VAL> l : c.getLocals().keySet()) {
					Collection<Transition<NAME, VAL>> transitions = ctx.execute(c.getShared(), l);
					for(Transition<NAME, VAL> t: transitions) {
						Configuration<NAME, VAL> nextC = c.transform(t);
						if (nextC != null) {
							ConfigurationNode<NAME, VAL> nextCNode = new org.ow2.dsrg.fm.badger.ca.karpmiller.impl.ConfigurationNode<NAME, VAL>(nextC, cNode);
							nextC.omega(cNode.getAllPredecessors());
							
							if (ctx.hitsError(nextC)) {
								nextC.setError();
								addErrorNode(nextCNode);
								if (doStopConstructionDueToErrorNode(nextCNode)) {
									return getResultNodes();
								}
							} else {
								addUnexplored(nextCNode);
								anyNewUnexploredTransition = true;							
							}
						}
					}
				}
			}
			
			if (!anyNewUnexploredTransition) {
				registerEndNode(cNode);
			}
		}
		return getResultNodes();
		
	}	
}
