/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2011       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller.impl;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.ow2.dsrg.fm.badger.ca.karpmiller.Counter;
import org.ow2.dsrg.fm.badger.ca.karpmiller.Transition;
import org.ow2.dsrg.fm.badger.ca.statespace.GlobalState;
import org.ow2.dsrg.fm.badger.ca.statespace.LocalState;

/**
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class Configuration<NAME, VAL> implements
		org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<NAME, VAL> {
	
	private GlobalState<NAME, VAL> shared;
	private Map<LocalState<NAME, VAL>, Counter> locals;
	private boolean isError; 
	
	public Configuration(GlobalState<NAME, VAL> shared, Map<LocalState<NAME, VAL>, Counter> locals) {
		assert(shared != null && locals != null);
		this.shared = shared;
		this.locals = new HashMap<LocalState<NAME,VAL>, Counter>(locals);
		this.isError = false;
	}
	
	@Override
	public GlobalState<NAME, VAL> getShared() {
		return shared;
	}

	@Override
	public Map<LocalState<NAME, VAL>, Counter> getLocals() {
		return Collections.unmodifiableMap(this.locals);
	}

	/** Introduces &omega; counters to the valuation of counters for occupied local states <i>f</i>.
	 * 
	 * <p>The value of <i>f(l)</i> will be &omega; after this call whenever
	 *    there exists configuration <i>p</i> &isin; <b>predecessors</b> of <i>this</i>
	 *    for which 1. and 2. holds:
	 *    <ol>
	 *    	<li><i>this</i>.{@link #covers(org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration) covers}<i>(p)</i></li>
	 *    	<li>0 &lt; <i>g(l)</i> &lt; <i>f(l)</i> and both <i>g(l)</i>, <i>f(l)</i> are defined for <i>l</i>
	 *    		<ul><li><i>g</i> stands for <i>p</i>'s valuation of counters of occupied local states</li></ul>
	 *    	</li>
	 *    </ol>
	 * </p>
	 */
	@Override
	public void omega(
			Iterable<org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<NAME, VAL>> predecessors) {
		
		for(org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<NAME, VAL> p : predecessors) {
			if (this.covers(p)) {
				for(Map.Entry<LocalState<NAME, VAL>, Counter> localEntry: this.locals.entrySet()) {
					Counter predCounter = p.getLocals().get(localEntry.getKey());
					Counter myCounter = localEntry.getValue();
					if (predCounter == null || predCounter.compareTo(myCounter) < 0) {
						localEntry.getValue().setOmega();
					}
				}
			}
		}
	}
	
	@Override
	public boolean covers(
			org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<NAME, VAL> that) {
		return
			this.shared.equals(that.getShared()) && this.coversLocally(that);
	}

	@Override
	public boolean coversLocally(
			org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<NAME, VAL> that) {
		if (that != null && that.getLocals() != null) {
			for(Map.Entry<LocalState<NAME, VAL>, Counter> thatEntry: that.getLocals().entrySet()) {
				if (thatEntry.getValue() != null) {
					Counter thisCounter = this.locals.get(thatEntry.getKey());
					if (thisCounter == null || thisCounter.compareTo(thatEntry.getValue()) < 0) {
						return false;
					}
				}
			}
			return true;
		}
		return false;
	}

	/** Transforms this configuration to another one using {@link Transition} <b>t</b>.
	 * 
	 * <p>The process of transformation consists of 3 crucial steps
	 * <ol><li>Create copy of {@link #locals} map of <code>this</code>.
	 *         Note that the {@link Counter}s <b>MUST</b> be {@link Counter#clone() cloned} since they
	 *         might be modified in next steps.</li>
	 *     <li>Decrement or remove counters of abandoned local states.
	 *         More formally, existing mapping <i>l</i> &rarr; <i>c</i> of each abandoned local states <i>l</i> should be
	 *         <ul><li>changed to <i>l</i> &rarr; <i>(c - 1)</i> if <i>c &gt; 1</i></li>
	 *             <li>removed if <i>c = 1</i></li>
	 *             <li>kept as it is if <i>c</i> = &omega;</li>
	 *         </ul></li>
	 *     <li>Increment or add counters of taken local states.
	 *         More formally, existing mapping <i>l</i> &rarr; <i>c</i> of taken local state <i>l</i> should be
	 *         <ul><li>changed to <i>l</i> &rarr; <i>(c + 1)</i> if <i>c</i> &ne; &omega;</li>
	 *             <li>kept as it is if <i>c</i> = &omega;</li>
	 *         </ul>
	 *         Moreover, if a taken state <i>l</i> was not mapped in {@link #locals} mapping of this,
	 *         it should be added as <i>l</i> &rarr; <i>1</i> to newly created {@link Configuration}.    
	 *     </li>
	 * </ol>
	 * The newly created {@link Configuration} will have {@link #locals} mapping as described above and
	 * {@link #shared} state taken from {@link Transition#getGlobalState()}.
	 * </p>
	 * 
	 * <p>
	 * Note also, that given transition could lead to invalid configuration
	 * (e.g. with negative counter(s) for some local state(s)).
	 * <code>Null</code> is returned in such case.
	 * </p>
	 * 
	 */
	@Override
	public org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<NAME, VAL> transform(
			Transition<NAME, VAL> t) {
		
		// create copy of mapping (note that counters MUST be cloned)
		// (step 1)
		Map<LocalState<NAME, VAL>, Counter> newMapping = new HashMap<LocalState<NAME,VAL>, Counter>();
		for(Map.Entry<LocalState<NAME, VAL>, Counter> e: locals.entrySet()) {
			newMapping.put(e.getKey(), e.getValue().clone());
		}
		
		// decrement or remove abandoned states
		// (step 2)
		for(LocalState<NAME, VAL> l: t.getAbandonedLocalStates()) {
			Counter c = newMapping.get(l);
			if (c == null) {
				// Abandoned state currently not occupied 
				// => INVALID TRANSITION
				return null;
			} else {
				if (c.isOmega() || c.getIntValue() > 1) {
					c.decrement();
				} else {
					newMapping.remove(l);
				}
			}
		}
				
		// insert or increment taken states
		// (step 3)
		for(LocalState<NAME, VAL> takenState : t.getTakenLocalStates()) {
			Counter counter = newMapping.get(takenState);
			if (counter == null) {
				newMapping.put(takenState, new Counter(1));
			} else {
				counter.increment();
			}
		}
		
		return new org.ow2.dsrg.fm.badger.ca.karpmiller.impl.Configuration<NAME, VAL>(t.getGlobalState(), newMapping);
	}

	@Override
	public void setError() {
		this.isError = true;
	}

	@Override
	public boolean isError() {
		return this.isError;
	}
	
	@Override
	public boolean equals(Object o) {
		if (o instanceof org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<?, ?>) {
			return this.equals((org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<?, ?>)o);
		} else {
			return false;
		}
	}
	
	public <N, V> boolean equals(org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration<N, V> c) {
		if (c == null) {
			return false;
		} else if (this == c) {
			return true;
		} else {
			Map<LocalState<N, V>, Counter> thatLocals = c.getLocals();
			int thatLocalsSize = thatLocals != null ? thatLocals.size() : 0;
			
			if (this.shared.equals(c.getShared()) && this.locals.size() == thatLocalsSize) {
				// shared states are equal, sizes of locals are equal
				if (thatLocalsSize == 0) {
					return true;
				} else {
					for(Map.Entry<LocalState<NAME, VAL>, Counter> e: this.locals.entrySet()) {
						Counter thatCounter = thatLocals.get(e.getKey());
						if (thatCounter == null || !thatCounter.equals(e.getValue())) {
							return false;
						}
					}
					return true;
				}
			} else {
				return false;
			}
		}
	}
	
	@Override
	public int hashCode() {
		return this.locals != null ? this.locals.size() : 0; 
	}
}
