/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller.opt;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

/** Represents minimal closure in relation &rarr;.
 * 
 * <p>The relation &rarr; must be <i>transitive</i> and <i>antisymetric</i>,
 *    it is represented with {@link PartialOrder}.
 * </p>
 * 
 * <p>Minimal closure of <i>R</i> set of elements in relation &rarr;
 *    is such subset <i>C</i> of <i>R</i>, that following hold:
 *    <ul><li>for all <i>r</i> &isin; <i>R</i> there is <i>c</i> &isin; <i>C</i> s.t. <i>c</i> &rarr; <i>r</i> (<i>C</i> &rarr; <i>R</i>)</li>
 *        <li>for all <i>c, d</i> &isin; <i>C</i> if <i>c</i> &rarr; <i>d</i> then <i>d</i> &rarr; <i>c</i> which implies <i>c</i> = <i>d</i> (<i>C</i> is minimal)</li>
 *    </ul>
 * </p>
 *     
 * <p>This class represents minimal closure of set <i>R</i> in a iterative manner,
 *    which means that <i>R</i> is never explicitly constructed, but its closure
 *    is built as elements are added to <i>R</i>.
 * </p>
 *     
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class MinimalClosure<E> implements Cover<E> {
	 
	protected Collection<E> elements;
	private PartialOrder<E> relation;
	
	public MinimalClosure(PartialOrder<E> relation) {
		this.relation = relation;
		this.elements = new LinkedList<E>();
	}
	
	protected void addElementToClosure(E e) {
		elements.add(e);
	}
			
	/** Adds element to set enclosed by this closure.
	 * 
	 * <p>Note that the closure itself may but must not be modified during this call.
	 *    This method returns arbitrary <i>c</i> &isin; <i>this</i> closure s.t
	 *    <i>c</i> &rarr; <b>e</b> if such <i>c</i> existed right before the current call
	 *    of this method.
	 * </p>
	 * 
	 * @param e
	 * @return arbitrary <i>c</i> &isin; <i>this</i> cover s.t. <i>c</i> &rarr; <b>e</b>
	 *         if such <i>c</i> existed in <i>this</i> right before the current call modified it
	 */
	public E add(E e) {
		boolean cCovers = false;
		Iterator<E> it = elements.iterator();
		while(it.hasNext()) {
			E c = it.next();
			if (relation.inOrder(e, c)/* e -> c */) {
				cCovers = true;
				if (relation.inOrder(c, e)/* c -> e */) {
					/* once e is covered by a member of this,
					 * no other member can be covered by e
					 * hence we can return
					 */
					return c;
				} else {
					it.remove();
				}
			}
			if (!cCovers) {	// once e covers a member of this, there is not another member that would cover e
				if (relation.inOrder(c, e)/* c -> e */) {
					/* once e is covered by a member of this,
					 * no other member can be covered by e
					 * hence we can return
					 */
					return c;

				}
			}
		}
		addElementToClosure(e);
		
		return null;
	}
}
