/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2011       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller.util;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration;
import org.ow2.dsrg.fm.badger.ca.karpmiller.ConfigurationNode;

/** Represents karp miller tree as {@link Iterable} of its {@link Node nodes}.
 * 
 * <p>Note that iterators returned from {@link Iterable#iterator() iterator()}
 *    iterate trough nodes <i>in-order</i>.</p>
 * 
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class KarpMillerTree<NAME, VAL> implements Iterable<KarpMillerTree<NAME, VAL>.Node> {
	
	public class NodeIterator implements Iterator<KarpMillerTree<NAME, VAL>.Node>
	{
		private LinkedList<Iterator<Node>> stack;
		
		
		public NodeIterator() {
			this(null);
		}		
		public NodeIterator(Node root) {
			this.stack = new LinkedList<Iterator<Node>>();
			if (root != null) {
				List<Node> l = Collections.singletonList(root);
				this.stack.addFirst(l.iterator());
			}
		}
		
		@Override
		public boolean hasNext() {
			return !stack.isEmpty();
		}

		@Override
		public Node next() {
			if (!stack.isEmpty()) {
				Iterator<Node> top = stack.getFirst();
				assert(top.hasNext());
				Node result = top.next();
				if (!result.getChildren().isEmpty()) {
					stack.addFirst(result.getChildren().iterator());
				} else {
					while(!stack.isEmpty() && !stack.getFirst().hasNext())
						stack.removeFirst();
				}
				return result;
			} else {
				return null;
			}
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}
		
	}
	public class Node {
		private Configuration<NAME, VAL> configuration;
		private List<Node> children;
		private Node parent;
		private int id;
		
		public Node(int id, Configuration<NAME, VAL> configuration, Node parent) {
			this.id = id;
			this.configuration = configuration;
			this.parent = parent;
			this.children = new LinkedList<KarpMillerTree<NAME,VAL>.Node>();
		}
		
		private void addChild(Node node) {
			this.children.add(node);
		}
		
		public Collection<Node> getChildren() {
			return Collections.unmodifiableList(this.children);
		}
		
		public Node getParent() {
			return this.parent;
		}
		
		public Configuration<NAME, VAL> getConfiguration() {
			return this.configuration;
		}
		
		public int getId() {
			return this.id;
		}
	}
	
	private Node root;
	private int lastId;
	
	public static <NAME, VAL> Collection<KarpMillerTree<NAME, VAL>> build(Collection<ConfigurationNode<NAME, VAL>> cNodes) {
		LinkedList<KarpMillerTree<NAME, VAL>> trees = new LinkedList<KarpMillerTree<NAME,VAL>>();
		
		for(ConfigurationNode<NAME, VAL> node: cNodes) {
			// make path from node
			LinkedList<Configuration<NAME, VAL>> path = new LinkedList<Configuration<NAME,VAL>>();
			path.add(node.getConfiguration());
			for(Configuration<NAME, VAL> conf: node.getAllPredecessors()) {
				path.addFirst(conf);
			}
			Configuration<NAME, VAL> firstConf = path.removeFirst();
			
			// find or create appropriate tree
			KarpMillerTree<NAME, VAL> tree = null;
			for(KarpMillerTree<NAME, VAL> treeIx: trees) {
				if (treeIx.root.getConfiguration().equals(firstConf)) {
					tree = treeIx;
					break;
				}
			}
			if (tree == null) {
				tree = new KarpMillerTree<NAME, VAL>(firstConf);
				trees.add(tree);
			}
			KarpMillerTree<NAME, VAL>.Node treeNode = tree.root;
			
			while(!path.isEmpty()) {
				firstConf = path.removeFirst();
				KarpMillerTree<NAME, VAL>.Node nextTreeNode = null; 
				for(KarpMillerTree<NAME, VAL>.Node tmpTreeNode: treeNode.getChildren()) {
					if (tmpTreeNode.getConfiguration().equals(firstConf)) {
						nextTreeNode = tmpTreeNode;
						break;
					}
				}
				if (nextTreeNode == null) {
					nextTreeNode = tree.addNode(firstConf, treeNode);
				}
				treeNode = nextTreeNode;
			}
		}
		
		return trees;
	}
	
	private KarpMillerTree(Configuration<NAME, VAL> c) {
		this.lastId = 1;
		this.root = new Node(lastId++, c, null);
	}
	
	private Node addNode(Configuration<NAME, VAL> c, Node parentNode) {
		Node tmp = new Node(lastId++, c, parentNode);
		parentNode.addChild(tmp);
		return tmp;
	}

	@Override
	public Iterator<KarpMillerTree<NAME, VAL>.Node> iterator() {
		return new NodeIterator(this.root);
	}	
}
