/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.karpmiller.util;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Collection;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;

import org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration;
import org.ow2.dsrg.fm.badger.ca.karpmiller.Counter;
import org.ow2.dsrg.fm.badger.ca.statespace.GlobalState;
import org.ow2.dsrg.fm.badger.ca.statespace.LocalState;
import org.ow2.dsrg.fm.tbplib.util.DOTPrintStream;
import org.ow2.dsrg.fm.tbplib.util.DOTPrintStreamWithSubgraphs;
import org.ow2.dsrg.fm.tbplib.util.DOTPrintStreamWithSubgraphs.NodeShape;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/** Provides several ways of writing out Karp-Miller Tree(s).
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class KarpMillerTreeWriter {
	/* DOT
	 */
	
	public static <N, V> void writeAsDot(OutputStream outputStrem, KarpMillerTree<N, V> tree) {
		DOTPrintStreamWithSubgraphs s = new DOTPrintStreamWithSubgraphs(outputStrem);
		writeToDotStream(s, tree);
		s.DOTclose();
	}
	
	public static <N, V> void writeToDotStream(DOTPrintStreamWithSubgraphs s, KarpMillerTree<N, V> tree) {
		NodeShape oldShape = s.getCurrentNodeShape();
		s.changeNodeShape(NodeShape.BOX);		
		appendToDotStream(s, tree);
		s.changeNodeShape(oldShape);		
	}
	public static <N, V> void writeToDotStream(DOTPrintStream s, KarpMillerTree<N, V> tree) {
		appendToDotStream(s, tree);
	}	
	
	private static <N, V> void appendToDotStream(DOTPrintStream dotStream, KarpMillerTree<N, V> t) {
		for(KarpMillerTree<N, V>.Node n: t) {
			boolean tmp = dotStream.filled;
			if (n.getConfiguration().isError()) {
				dotStream.filled = true;
			}
			dotStream.node(n.getId(), nodeToDotNodeLabel(n));				
			dotStream.filled = tmp;
			
			if (n.getParent() != null) {
				dotStream.edge(n.getParent().getId(), n.getId());
			}
		}
	}
	
	private static <N, V> String nodeToDotNodeLabel(KarpMillerTree<N, V>.Node n) {
		StringBuilder sb = new StringBuilder();
		if (n != null) {
			Configuration<N, V> c = n.getConfiguration();
			if (c != null) {
				GlobalState<N, V> s = c.getShared();
				if (s != null) {
					sb.append("{").append(s.toString()).append("}");
				} else {
					sb.append("--- null shared state ---");
				}
				sb.append("\\n");
				Map<LocalState<N, V>, Counter> locals = c.getLocals();
				if (locals != null) {
					for(Map.Entry<LocalState<N, V>, Counter> e: locals.entrySet()) {
						LocalState<N, V> l = e.getKey();
						Counter lv = e.getValue();
						if (l != null && lv != null) {
							sb.append(lv.toString()).append(" &larr; ").append("{").append(l.toString()).append("}");
						} else {
							sb.append("--- null local state or counter value ---");
						}
						sb.append("\\n");
					}
				}
			}
		}
		return sb.toString();
	}
	
	/* XML
	 */
	public static <N, V> void writeAsXml(OutputStream s, KarpMillerTree<N, V> tree) throws IOException {
		Document d = prepareDocument();
		addTreeToXml(d, d, tree);
		flushXml(d, s);
	}
	
	public static <N, V> void writeAsFlatXml(OutputStream s, KarpMillerTree<N, V> tree) throws IOException {
		Document d = prepareDocument();
		addTreeToFlatXml(d, d, tree);
		flushXml(d, s);
	}
	
	public static <N, V> void writeAsXml(OutputStream s, Collection<KarpMillerTree<N, V>> forest) throws IOException {
		Document d = prepareDocument();
		Element root = d.createElement("forest");
		d.appendChild(root);
		for(KarpMillerTree<N, V> t: forest) {
			addTreeToXml(d, root, t);
		}
		flushXml(d, s);		
	}
	
	public static <N, V> void writeFlatAsXml(OutputStream s, Collection<KarpMillerTree<N, V>> forest) throws IOException {
		Document d = prepareDocument();
		Element root = d.createElement("paths");
		d.appendChild(root);
		for(KarpMillerTree<N, V> t: forest) {
			addTreeToFlatXml(d, root, t);
		}
		flushXml(d, s);
	}
	
	private static void flushXml(Document d, OutputStream s) throws IOException {
		try {
			Transformer t = TransformerFactory.newInstance().newTransformer();
			t.setOutputProperty(OutputKeys.INDENT, "yes");
	        t.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
	       	t.transform(	new javax.xml.transform.dom.DOMSource(d),
	        				new javax.xml.transform.stream.StreamResult(s));
		} catch (TransformerException te) {
			throw new IOException("Unable to flush xml.", te);
		}
	}
	private static Document prepareDocument() throws IOException {
		try {
			DocumentBuilder xmlB = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			Document xmlDoc = xmlB.newDocument();
			xmlDoc.setXmlStandalone(true);
			return xmlDoc;
		} catch (ParserConfigurationException pce) {
			throw new IOException("Can not convert to xml.", pce);
		}
	}
	private static <N, V> org.w3c.dom.Node addNodeToXml(Document d, org.w3c.dom.Node parent, KarpMillerTree<N, V>.Node treeNode) {
		Element nodeRoot = d.createElement("node");
		parent.appendChild(nodeRoot);
		nodeRoot.setAttribute("id", Integer.toString(treeNode.getId()));
		if (treeNode.getParent() != null) {
			nodeRoot.setAttribute("parent", Integer.toString(treeNode.getParent().getId()));
		}
		Element conf = d.createElement("configuration");
		nodeRoot.appendChild(conf);
		
		// TODO: add details of node (configuration)
		
		return nodeRoot;
	}
	private static <N, V> void addTreeToXml(Document d, org.w3c.dom.Node parent, KarpMillerTree<N, V> tree) {
		// TODO --- the container element should be named "tree"
	}
	private static <N, V> void addTreeToFlatXml(Document d, org.w3c.dom.Node parent, KarpMillerTree<N, V> tree) {
		Element treeRoot = d.createElement("path");
		parent.appendChild(treeRoot);
		for(KarpMillerTree<N, V>.Node n: tree) {
			addNodeToXml(d, treeRoot, n);
		}
	}	
	
	
	
	/* String (for paths)
	 */
	
	/** Gets tree as string.
	 * 
	 * <p>This way method of converting the <b>tree</b> to string
	 *    is suitable just for <i>paths</i> (i.e. tree where
	 *    each node has at most one child).</p>
	 * 
	 * <p>If the <b>tree</b> is not a <i>path</i>, the result
	 *    may be not well readable.</p>
	 *
	 * @param tree
	 */
	public static <N, V> String pathAsString(KarpMillerTree<N, V> tree) {
		// TODO
		return "TODO: write path";
	}

}
