/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2011       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace;

import java.util.Collection;
import java.util.Map;

import org.ow2.dsrg.fm.badger.ca.karpmiller.Configuration;
import org.ow2.dsrg.fm.badger.ca.karpmiller.Context;
import org.ow2.dsrg.fm.badger.ca.karpmiller.Counter;
import org.ow2.dsrg.fm.badger.ca.karpmiller.Transition;

/** Adapts calls to {@link Context} to {@link StateSpaceExplorer}.
 * 
 * <p>Note that responsibility of {@link StateSpaceExplorer} is to
 *    explore states from given one (couple of global and local state)
 *    by execution of single step of model and return those states.
 *    On the other hand, the responsibility of {@link Context}
 *    is to provide all states that may occur in next step, including
 *    <i>newly created threads</i>.
 *    Hence we have to add transitions representing creation of new threads.
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class ContextAdapter<NAME,VAL> implements Context<NAME, VAL> {
		
	private Configuration<NAME, VAL> initial;
	private StateSpaceExplorer<NAME, VAL> explorer;
	private LTSARepository<?, ?, NAME, VAL> repository;
	
	public ContextAdapter(
			LTSARepository<?, ?, NAME, VAL> repository,
			StateSpaceExplorer<NAME, VAL> explorer,
			GlobalState<NAME, VAL> sharedState,
			Map<LocalState<NAME, VAL>, Counter> locals) {
		assert(repository != null && explorer != null && sharedState != null && locals != null && !locals.isEmpty());
		
		this.repository = repository;
		this.explorer = explorer;
		this.initial = new org.ow2.dsrg.fm.badger.ca.karpmiller.impl.Configuration<NAME, VAL>(
				sharedState,
				locals);
	}
	
	@Override
	public Configuration<NAME, VAL> getInitialConfiguration() {
		return this.initial;
	}

	/**
	 * 
	 */
	@Override
	public Collection<Transition<NAME, VAL>> execute(GlobalState<NAME, VAL> s,
			LocalState<NAME, VAL> l) {
		
		return this.explorer.explore(s, l);
	}

	@Override
	public boolean hitsError(Configuration<NAME, VAL> c) {
		for(Map.Entry<LocalState<NAME, VAL>, Counter> valuation: c.getLocals().entrySet()) {
			LocalState<NAME, VAL> l = valuation.getKey();
			Counter v = valuation.getValue();
			int pc = l.getProgramCounterValue();
			if (	pc == repository.getInvalidStateProgramCounterValue() 
				||	v.exceedsUpperLimit(repository.getCounterLimit(pc))) {
				return true;
			}
		}
		
		return false;
	}
	
}
