/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace;

import org.ow2.dsrg.fm.badger.ca.provision.ThreadObservationSystem;
import org.ow2.dsrg.fm.badger.ca.statespace.encoding.LocalStateEncoder;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.provision.CounterLimit;

/** Provides storage of {@link LTSA} transitions.
 * 
 * <p><b>Storage of LTSA transitions</b><br/>
 *    LTSA transition is accessible using its <i>program counter</i> value,
 *    which is technically an integral number which is unique for
 *    the transition.
 *    See {@link #get(int)}.
 * </p>
 * <p><b>Storage of LocalStateEncoders associated with LTSA transitions</b><br/>
 *    In addition, an implementation of this should provide association
 *    custom data with LTSA transition.
 *    Particularly, one can associate {@link LocalStateEncoder} with
 *    {@link LTSAThread} and it will then be accessible using
 *    <i>program counter</i> value of any LTSA transition of that LTSAThread.
 *    See {@link #getLocalStateEncoder(int)}.
 * </p>
 * <p><b>Storage of CounterLimits associated with LTSA transitions</b><br/>
 *    A {@link CounterLimit} may be associated with LTSA transitions,
 *    that stand as parallel counters (<i>isolated locations</i>).
 *    Note that the default {@link CounterLimit} should be unbounded
 *    (i.e. in case that no {@link CounterLimit} has been associated with
 *     particular LTSA). See {@link #get(int)}.
 * </p>
 * 
 * <p><b>Storage of ThreadObservationSystems</b><br/>
 *    This provides obtaining {@link ThreadObservationSystem}
 *    associated with LTSA transition. In particular,
 *    a ThreadObservationSystem is associated with a
 *    {@link LTSAThread}, thus the same {@link ThreadObservationSystem}
 *    is associated with all LTSA transition of the LTSAThread. 
 * </p>
 * 
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public interface LTSARepository<CODEREF_TYPE, REFERENCE, VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> {
	
	/** Returns {@link LTSA} transition for <i>program counter</i>
	 *  value <b>pc</b>.
	 * 
	 * @param pc <i>program counter</i> value,
	 *           that is valid after execution of LTSA transition
	 *           (this value exclusively identifies LTSA transition)
	 * @return {@link LTSA} transition which, after execution,
	 *         provides state with given program counter <b>pc</b>
	 */
	public LTSA<CODEREF_TYPE, REFERENCE> get(int pc);
	
	/** Returns {@link LocalStateEncoder} associated with {@link LTSA} transition
	 *  identified using <i>program counter</i> value <b>pc</b>.
	 *  
	 * @param pc <i>program counter</i> value
	 * @return <i>custom data</i>
	 */
	public LocalStateEncoder<VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> getLocalStateEncoder(int pc);

	/** Returns {@link CounterLimit} associated with {@link LTSA} transition
	 *  identified using <i>program counter</i> value <b>pc</b>.
	 *  
	 *  <p>Note that if no {@link CounterLimit} is logically associated
	 *     with the identified {@link LTSA}, an <i>unbounded</i> {@link CounterLimit}
	 *     will be returned
	 *     (i.e. {@link CounterLimit} <i>L</i> s.t.
	 *     L.{@link CounterLimit#getUpperLimit() getUpperLimit()} == 
	 *     {@link CounterLimit#INFINITY INFINITY}).
	 *  </p>
	 *
	 * @param pc <i>progam counter</i> value
	 * @return {@link CounterLimit} 
	 */
	public CounterLimit getCounterLimit(int pc);
	
	/** Returns {@link ThreadObservationSystem} associated with a thread,
	 *  that the given <i>program counter</i> <b>pc</b> belongs to.
	 * 
	 * @param pc program counter
	 * @return {@link ThreadObservationSystem}
	 */
	public ThreadObservationSystem<VARIABLE_NAME_TYPE> getThreadObservationSystem(int pc);
	
	/** Returns <i>program counter</i> value for an invalid LTSA transition.
	 * 
	 * @return invalid <i>program counter</i> value
	 */
	public int getInvalidStateProgramCounterValue();
	
	/** Returns maximal value of program counter.
	 * 
	 * <p>Program counter values less or equal to the result may be
	 * <ul><li>associated with a valid {@link LTSA} transition</li>
	 *     <li>invalid state program counter value
	 *         (see {@link #getInvalidStateProgramCounterValue()})</li> 
	 * </p>
	 * 
	 * @return maximal value of program counter
	 */
	public int getMaximalStorableProgramCounterValue();
	
	/** Returns value of <i>program counter</i> after execution
	 *  of the {@link LTSA} transition <b>ltsa</b> .
	 * 
	 * @param ltsa LTSA transition
	 * @return value of <i>program counter</i>
	 */
	public int getProgramCounterValue(LTSA<?, REFERENCE> ltsa);
}
