/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace.impl;

import java.util.*;

import org.ow2.dsrg.fm.badger.ca.karpmiller.Transition;
import org.ow2.dsrg.fm.badger.ca.provision.ThreadObservationSystem;
import org.ow2.dsrg.fm.badger.ca.statespace.*;
import org.ow2.dsrg.fm.badger.ca.statespace.encoding.LocalStateCode;
import org.ow2.dsrg.fm.badger.ca.util.*;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.guard.*;
import org.ow2.dsrg.fm.tbplib.provision.ObservingProvisionAutomaton;
import org.ow2.dsrg.fm.tbplib.provision.Transition.OutputSymbol;

/** Basic abstract implementation of {@link StateSpaceExplorer}.
 * 
 * <p>Based on
 * <ul>
 * <li>{@link LTSARepository} as storage of {@link LTSA} transitions</li>
 * <li>{@link ThreadObservationSystem} within the storage as
 *     {@link ObservingProvisionAutomaton provision automata}
 *     related to individual threads</li>
 * </ul>
 * </p> 
 * 
 * <p>
 * Note that the {@link #getReferenceType(Object)} method is recommended to
 * override, as it could improve performance (see {@link #getReferenceType(Object)}). 
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public abstract class AbstractStateSpaceExplorerImpl<NN, VV, REFERENCE> implements StateSpaceExplorer<NN, VV> {
	
	protected static enum ReferenceType {
		CONSTANT, VARIABLE, UNKNOWN
	}
	
	protected static enum VariableType {
		LOCAL_VARIABLE, GLOBAL_VARIABLE, UNKNOWN
	}
	
	private class InvalidLocalState implements LocalState<NN, VV>, LocalStateCode {
		private int pc;
		private AbstractStateSpaceExplorerImpl<NN, VV, REFERENCE> owner;
		public InvalidLocalState(AbstractStateSpaceExplorerImpl<NN, VV, REFERENCE> owner, int pc) {
			this.pc = pc;
			this.owner = owner;
		}
		@Override
		public VV getValueOfVariable(NN variableName) {
			return null;
		}
		@Override
		public int getProgramCounterValue() {
			return pc;
		}
		@Override
		public Map<NN, VV> getVariableValuation() {
			return Collections.emptyMap();
		}		
		@Override
		public LocalState<NN, VV> modify(
				int newProgramCounterValue) {
			Map<NN, VV> empty = Collections.emptyMap();
			return modify(newProgramCounterValue, empty);
		}
		@Override
		public LocalState<NN, VV> modify(
				Map<NN, VV> newVariableValues) {
			return modify (this.getProgramCounterValue(), newVariableValues);
		}
		@Override
		public LocalState<NN, VV> modify(
				int newProgramCounterValue,
				Map<NN, VV> newVariableValues) {
			throw new UnsupportedOperationException("Unable to modify invalid local state.");
		}
		@Override
		public boolean contains(NN v) {
			return false;
		}
		@Override
		public int hashCode() {
			return pc;
		}
		@Override
		public boolean equals(Object o) {
			return	(o == this)
					||
					(o != null && o.getClass().equals(this.getClass()) && o.hashCode() == this.hashCode());
		}
		@Override
		public String toString() {
			return String.format("PC=%d, %s=1", this.pc, this.owner != null && this.owner.invalidLocalState == this
						? "IS_ERR" : "IS_CTR");
		}
	}
	
	private LTSARepository<?, REFERENCE, NN, VV> repository;
	private LocalState<NN, VV> invalidLocalState;
	protected EqualityComparator<VV> valueComparator;
	
	/** Tries to determine type of given reference <b>var</b>.
	 * 
	 * <p>May return {@link VariableType#UNKNOWN UNKOWN} and in such case the
	 *    type is determined within <i>fall-back procedure</i> (out of this method)
	 *    from context:
	 * <ol><li>In case that <b>var</b> is member of {@link GlobalState},
	 *         it is {@link VariableType#GLOBAL_VARIABLE GLOBAL_VARIABLE}</li> 
	 *     <li>In case that <b>var</b> is member of {@link LocalState},
	 *         it is {@link VariableType#LOCAL_VARIABLE LOCAL_VARIABLE}</li> 
	 *     </ol>
	 *     Note that the <i>fall-back</i> is ineffective
	 *     (the <em>is member of</em> operation).
	 * </p>
	 * 
	 * <p><b>Note</b> that in this general implementation it is not possible to determine
	 *    the type, thus always returns {@link ReferenceType#UNKNOWN}.<br/>
	 *    One can affect performance by overriding this method
	 *    (e.g. successfull decision will disable <i>slow</i>
	 *     fall-back procedures; see above).  
	 * </p>
	 * 
	 * @param var variable name
	 * @return {@link VariableType}
	 */
	protected VariableType getVariableType(NN var) {
		return VariableType.UNKNOWN;
	}
		
	/** Returns <i>constant</i> value of reference.
	 * 
	 * <p>Note that this method may called only if
	 *    {@link #getReferenceType(Object) getReferenceType(r)}
	 *    ==
	 *    {@link ReferenceType#CONSTANT CONSTANT}.
	 *    Otherwise the behavior is undefined.
	 * </p>
	 * 
	 * @param r reference
	 * @return value of <b>r</b>
	 */
	protected abstract VV getConstant(REFERENCE r);
	
	/** Returns name of variable represented by given reference.
	 * 
	 * <p>Note that this method may be called only if
	 *    {@link #getReferenceType(Object) getReferenceType(r)}
	 *    ==
	 *    {@link ReferenceType#VARIABLE VARIABLE}.
	 * </p>
	 * 
	 * @param r reference
	 * @return name of variable represented by <b>r</b>
	 */
	protected abstract NN getVariableName(REFERENCE r);
		
	/** Returns type of given reference.
	 * 
	 * <p>Must determine whether <b>r</b> refers to
	 *    {@link ReferenceType#CONSTANT CONSTANT} or
	 *    {@link ReferenceType#VARIABLE VARIABLE}.
	 *    In case that the reference does not refer
	 *    any of those subjects, {@link ReferenceType#UNKNOWN}
	 *    should be returned which may result in exception
	 *    (out of this method).
	 * </p>
	 * 
	 * @param r reference
	 * @return {@link ReferenceType type} of <b>r</b>
	 */
	protected abstract ReferenceType getReferenceType(REFERENCE r);

	/** Encodes event.
	 * 
	 * @param label represents the event
	 * @param s global state (shared variables valuation)
	 * @param l local state (local variables valuation)
	 * @return event code
	 */
	protected abstract long getEventCode(Label<REFERENCE> label,  GlobalState<NN, VV> s, LocalState<NN, VV> l);
	
	/** Encodes state number (of arbitrary automaton) into valuation of given variables.
	 * 
	 * @param stateNumber state number
	 * @param variableNames names of variables that encode the state number
	 * @return values of variables that encode the state number (in respective order)
	 */
	@Override
	public abstract List<VV> encodeStateNumber(int stateNumber, List<NN> variableNames);
	
	/** Decodes state number (of arbitrary automaton) from given valuation of variables.
	 * 
	 * @param variables variable values
	 * @return state number 
	 */
	protected abstract int decodeStateNumber(List<Pair<NN, VV>> variables);
		
	/** Creates state space explorer.
	 * 
	 * @param ltsaRepository repository of {@link LTSA} transitions
	 * @param valueComparator {@link EqualityComparator equals-comparator} of variable values
	 */
	public AbstractStateSpaceExplorerImpl(
			LTSARepository<?, REFERENCE, NN, VV> ltsaRepository,
			EqualityComparator<VV> valueComparator) {
		this.repository = ltsaRepository;
		this.invalidLocalState = new InvalidLocalState(this, repository.getInvalidStateProgramCounterValue());
		this.valueComparator = valueComparator;
	}

	@Override
	public Collection<Transition<NN, VV>> explore(GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		
		if (l.getProgramCounterValue() == invalidLocalState.getProgramCounterValue()) {
			// invalid state - skip it (return empty set of transitions)
			return Collections.emptyList();
		}
		
		// obtain starting LTSA transition		
		LTSA<?, REFERENCE> ltsaTransition = this.repository.get(l.getProgramCounterValue());
		
		// filter successors by guard satisfaction
		Collection<Transition<NN, VV>> transitions = new ArrayList<Transition<NN, VV>>(ltsaTransition.getTransitions().size());
		for(LTSA<?, REFERENCE> successor: ltsaTransition.getTransitions()) {
			if (evaluateGuard(successor.getGuard(), s, l)) {
				// viable successor - continue on per-successor basis
				Transition<NN, VV> transition = exploreSingleOption(ltsaTransition, successor, s, l);
				if (transition != null) {
					transitions.add(transition);
				}
			}
		}		
		return transitions;
	}
	
	private Transition<NN, VV> exploreSingleOption(
			LTSA<?, REFERENCE> pred, 
			LTSA<?, REFERENCE> transition,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		
		List<LocalState<NN, VV>> extraAbandonedLocalStates = new LinkedList<LocalState<NN, VV>>();
		List<LocalState<NN, VV>> extraTakenLocalStates = new LinkedList<LocalState<NN, VV>>();
		List<LocalState<NN, VV>> takenErrorStates = new LinkedList<LocalState<NN, VV>>();
				
		// execute observers
		Pair<GlobalState<NN, VV>, LocalState<NN, VV>> pObs = executeObservers(
				transition,
				s,
				l,
				extraTakenLocalStates,
				takenErrorStates);
				
		// execute provision automata
		Pair<GlobalState<NN, VV>, LocalState<NN, VV>> pPA = executeProvisionAutomata(
				transition,
				pObs.getFirst(),
				pObs.getSecond(),
				extraAbandonedLocalStates,
				extraTakenLocalStates,
				takenErrorStates);
		
		// perform assignments
		Pair<GlobalState<NN, VV>, LocalState<NN, VV>> pAs = executeAssignments(
				transition, pPA.getFirst(), pPA.getSecond());
		
		// adjust abandoned and taken local states
		extraAbandonedLocalStates.add(l);
		extraTakenLocalStates.add(pAs.getSecond());
		extraTakenLocalStates.addAll(takenErrorStates);
		
		// create and return
		return new ImmutableTransitionImpl<NN, VV>(
				extraAbandonedLocalStates,
				pAs.getFirst(),
				extraTakenLocalStates);
	}
	
	private boolean evaluateGuard(ITBPCondition<REFERENCE> guard, GlobalState<NN, VV> s, LocalState<NN, VV> l) {
		if (guard == null) {
			return true;
		}
		ThreeStateLogic result = evaluateGuard3S(guard, s, l);
		return (result == ThreeStateLogic.BOTH || result == ThreeStateLogic.TRUE);
	}
	//
	// Auxiliary
	//
	
	private Pair<GlobalState<NN, VV>, LocalState<NN, VV>> executeObservers(
			LTSA<?, REFERENCE> transition,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l,
			List<LocalState<NN, VV>> extraTakenLocalStates,
			List<LocalState<NN, VV>> takenErrorStates) {
		
		boolean tmpErrorStateTaken = !takenErrorStates.isEmpty();
		Collection<Pair<Observer, List<NN>>> observers = getObservers(
				this.repository.getProgramCounterValue(transition));
		for(Pair<Observer, List<NN>> op: observers) {
			boolean stateDecoded = false;
			int state = 0;
			Observer o = op.getFirst();
			
			for(Label<REFERENCE> label : transition.getLabels()) {
				long eventCode = getEventCode(label, s, l);
				if (o.observesEvent(eventCode)) {
					if (!stateDecoded) {
						state = decodeAutomatonState(op.getSecond(), s, l);
						stateDecoded = true;
					}
					state = o.nextState(state, eventCode);
					if (state == Observer.NO_SUCH_TRANSITION) {
						if (!tmpErrorStateTaken) {
							takenErrorStates.add(this.invalidLocalState);
							tmpErrorStateTaken = true;
						}
						stateDecoded = false;
						break;
					}
				}
			}
			
			if (stateDecoded) {
				Pair<GlobalState<NN, VV>, LocalState<NN, VV>> p = encodeAutomatonState(op.getSecond(), false, state, s, l);
				s = p.getFirst();
				l = p.getSecond();
			}
		}
		
		return new Pair<GlobalState<NN, VV>, LocalState<NN, VV>>(s, l);
	}
	
	private Pair<GlobalState<NN, VV>, LocalState<NN, VV>> executeProvisionAutomata(
			LTSA<?, REFERENCE> transition,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l,
			List<LocalState<NN, VV>> extraAbandonedLocalStates,
			List<LocalState<NN, VV>> extraTakenLocalStates,
			List<LocalState<NN, VV>> takenErrorStates) {

		Collection<Pair<ObservingProvisionAutomaton<LocationReference>, List<NN>>> provisionAutomata =
			getProvisionAutomataSystem(this.repository.getProgramCounterValue(transition));
		
		boolean tmpErrorStateTaken = !takenErrorStates.isEmpty();
		
		for(Pair<ObservingProvisionAutomaton<LocationReference>, List<NN>> op: provisionAutomata) {
			boolean stateDecoded = false;
			int state = 0;
			ObservingProvisionAutomaton<LocationReference> opa = op.getFirst();
			
			for(Label<REFERENCE> label : transition.getLabels()) {
				long eventCode = getEventCode(label, s, l);
				if (opa.observesEvent(eventCode)) {
					if (!stateDecoded) {
						state = decodeAutomatonState(op.getSecond(), s, l);
						stateDecoded = true;
					}
					org.ow2.dsrg.fm.tbplib.provision.Transition<LocationReference> opaTransition = opa.getTransition(state, eventCode);
					if (opaTransition == null || opaTransition.getTargetState() == Observer.NO_SUCH_TRANSITION) {
						if (!tmpErrorStateTaken) {
							takenErrorStates.add(this.invalidLocalState);
							tmpErrorStateTaken = true;
						}
						stateDecoded = false;
						break;						
					} else {
						state = opaTransition.getTargetState();
						for(OutputSymbol<LocationReference> os : opaTransition.getOutputSymbols()) {
							switch (os.getDirection()) {
							case Up:
								extraTakenLocalStates.add(
										createIsolatedLocalState(
												os.getSymbol().getLocationProgramCounter(),
												s,
												l));
												
								break;
							case Down:
								extraAbandonedLocalStates.add(
										createIsolatedLocalState(
												os.getSymbol().getLocationProgramCounter(),
												s,
												l));
								break;
							default:
								break;
							}
						}
					}
				}
			}
			
			if (stateDecoded) {
				Pair<GlobalState<NN, VV>, LocalState<NN, VV>> p = encodeAutomatonState(op.getSecond(), true, state, s, l);
				s = p.getFirst();
				l = p.getSecond();
			}
		}
		
		return new Pair<GlobalState<NN, VV>, LocalState<NN, VV>>(s, l);
	}
	
	private VV getValueOfVariable(
			NN n,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		VV v = null;
		if (n != null) {
			VariableType t = getVariableTypeWithFallBack(n, s, l);
			switch (t) {
			case GLOBAL_VARIABLE:
				v = s.getValueOfVariable(n);
				break;
			case LOCAL_VARIABLE:
				v = l.getValueOfVariable(n);
				break;
			default:
				HandleUnsupportedType(t);
			}
		}
		return v;
	}
			
	//
	// Handling LTSA assignments
	//
	
	private Pair< GlobalState<NN, VV>, LocalState<NN, VV>> executeAssignments(
			LTSA<?, REFERENCE> transition,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		
		int newPCValue = this.repository.getProgramCounterValue(transition);
		Map<NN, VV> changedGlobalValues = new HashMap<NN, VV>();
		Map<NN, VV> changedLocalValues = new HashMap<NN, VV>();
		
		
		for(TBPAssignment<REFERENCE> a: transition.getAssignments()) {
			NN n = getVariableName(a.getIdf());
			assert(s.contains(n) || l.contains(n));
			VariableType t = getVariableTypeWithFallBack(n, s, l);
			switch (t) {
				case GLOBAL_VARIABLE:
					addAssignments(changedGlobalValues, n, a, s, l);
					break;
				case LOCAL_VARIABLE:
					addAssignments(changedLocalValues, n, a, s, l);
					break;
				default:
					HandleUnsupportedType(t);					
			}
		}
		
		LocalState<NN, VV> newLocalState = l.modify(newPCValue, changedLocalValues);
		
		GlobalState<NN, VV> newGlobalState = changedGlobalValues.isEmpty() ? s : s.modify(changedGlobalValues);
		
		return new Pair<GlobalState<NN, VV>, LocalState<NN, VV>>(newGlobalState, newLocalState);
	}
		
	private void addAssignments(
			Map<NN, VV> newValues,
			NN name,
			TBPAssignment<REFERENCE> a,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		TBPNode<REFERENCE> childNode = a.getChild();
		if (childNode instanceof TBPValue<?> ) {
			TBPValue<REFERENCE> childValue = (TBPValue<REFERENCE>)childNode;
			assert(!childValue.isMethodCall());
			newValues.put(name, getValueOfReference(childValue.getVarname(), s, l));
		} else {
			throw new RuntimeException("Bad value type in TBPAssignment.");
		}
	}
		
	private Collection<Pair<ObservingProvisionAutomaton<LocationReference>, List<NN>>> getProvisionAutomataSystem(int pc) {
		ThreadObservationSystem<NN> tos = repository.getThreadObservationSystem(pc);
		Collection<Pair<ObservingProvisionAutomaton<LocationReference>, List<NN>>> pas = null;
		if (tos != null) {
			pas = tos.getProvisionAutomata();
		}
		if (pas == null) {
			pas = Collections.emptySet();
		}
		return pas;
	}
	
	private Collection<Pair<Observer, List<NN>>> getObservers(int pc) {
		ThreadObservationSystem<NN> tos = repository.getThreadObservationSystem(pc);
		Collection<Pair<Observer, List<NN>>> os = null;
		if (tos != null) {
			os = tos.getObservers();
		} else {
			os = Collections.emptySet();
		}
		return os;
	}
	
	private LocalState<NN, VV> createIsolatedLocalState(
			int programCounter,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		// Try think of better way of doing this - the InvalidLocalState might cause some problems...
		return new InvalidLocalState(this, programCounter);
	}
	
	//
	// Evaluation of guards using three-state-logic...
	//
	private enum ThreeStateLogic{
		TRUE, FALSE, BOTH
	}
	
	private ThreeStateLogic combineAND(ThreeStateLogic a, ThreeStateLogic b) {
		return 
			(a == ThreeStateLogic.FALSE || b == ThreeStateLogic.FALSE)
				? ThreeStateLogic.FALSE
				: (a == ThreeStateLogic.BOTH || b == ThreeStateLogic.BOTH)
					? ThreeStateLogic.BOTH
					: ThreeStateLogic.TRUE;
	}
	private ThreeStateLogic combineOR(ThreeStateLogic a, ThreeStateLogic b) {
		return
			(a == ThreeStateLogic.TRUE || b == ThreeStateLogic.TRUE)
				? ThreeStateLogic.TRUE
				: (a == ThreeStateLogic.BOTH || b == ThreeStateLogic.BOTH)
					? ThreeStateLogic.BOTH
					: ThreeStateLogic.FALSE;
	}
	private ThreeStateLogic combineNOT(ThreeStateLogic a) {
		return
			a == ThreeStateLogic.FALSE
				? ThreeStateLogic.TRUE
				: a == ThreeStateLogic.TRUE
					? ThreeStateLogic.FALSE
					: ThreeStateLogic.BOTH;
	}
	
	private ThreeStateLogic evaluateGuard3S(ITBPCondition<REFERENCE> guard, GlobalState<NN, VV> s, LocalState<NN, VV> l) {
			if (guard instanceof TBPAnd<?>) {
				return
					combineAND(
							evaluateGuard3S(((TBPAnd<REFERENCE>) guard).getLeft(), s, l),
							evaluateGuard3S(((TBPAnd<REFERENCE>) guard).getRight(), s, l));
			} else if (guard instanceof TBPOr<?>) {
				return
					combineOR(
							evaluateGuard3S(((TBPOr<REFERENCE>) guard).getLeft(), s, l),
							evaluateGuard3S(((TBPOr<REFERENCE>) guard).getRight(), s, l));
			} else if (guard instanceof TBPNot<?>) {
				return 
					combineNOT(
							evaluateGuard3S(((TBPNot<REFERENCE>) guard).getOperand(), s, l));
			} else if (guard instanceof TBPEquals<?>) {
				return GuardEvaluateEquals(((TBPBinaryScalarCondition<REFERENCE>) guard).getLeft(), ((TBPBinaryScalarCondition<REFERENCE>) guard).getRight(), s, l);
			} else if (guard instanceof TBPNotEquals<?>) {
				return combineNOT(GuardEvaluateEquals(((TBPBinaryScalarCondition<REFERENCE>) guard).getLeft(), ((TBPBinaryScalarCondition<REFERENCE>) guard).getRight(), s, l));
			} else if (guard instanceof TBPInSet<?>) {
				return GuardEvaluateInSet(((TBPSetCondition<REFERENCE>) guard).getLeft(), ((TBPSetCondition<REFERENCE>) guard).getSet(), s, l);
			} else if (guard instanceof TBPNotInSet<?>) {
				return combineNOT(GuardEvaluateInSet(((TBPSetCondition<REFERENCE>) guard).getLeft(), ((TBPSetCondition<REFERENCE>) guard).getSet(), s, l));				
			} else if (guard instanceof TBPNondetCondition<?>) {
				return ThreeStateLogic.BOTH;
			} else {
				throw new UnsupportedOperationException("Unkown guard type.");
			}
	}
	
	private ThreeStateLogic GuardEvaluateEquals(REFERENCE left, REFERENCE right, GlobalState<NN, VV> s, LocalState<NN, VV> l) {
		return valueComparator.equals(getValueOfReference(left, s, l), getValueOfReference(right, s, l))
			? ThreeStateLogic.TRUE : ThreeStateLogic.FALSE;
	}
	
	private ThreeStateLogic GuardEvaluateInSet(REFERENCE v, Collection<REFERENCE> set, GlobalState<NN, VV> s, LocalState<NN, VV> l) {
		VV value = getValueOfReference(v, s, l);
		for(REFERENCE r : set) {
			if (valueComparator.equals(value, getValueOfReference(r, s, l))) {
				return ThreeStateLogic.TRUE;
			}
		}		 
		return ThreeStateLogic.FALSE;
	}
	
	
	///
	///
	///
	
	private VariableType getVariableTypeWithFallBack(
			NN n,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		
		VariableType t = getVariableType(n);
		if (t == VariableType.UNKNOWN) {
			if (s.contains(n)) {
				t = VariableType.GLOBAL_VARIABLE;
			} else if (l.contains(n)) {
				t = VariableType.LOCAL_VARIABLE;
			} else {
				t = VariableType.UNKNOWN;
			}
		}
		return t;
	}
	
	protected VV getValueOfReference(REFERENCE r, GlobalState<NN, VV> s, LocalState<NN, VV> l) {
		VV value = null;
		ReferenceType rt = getReferenceType(r);
		switch (rt) {
		case VARIABLE:
			NN n = getVariableName(r);
			value = getValueOfVariable(n, s, l);
			break;
		case CONSTANT:
			value = getConstant(r);
			break;
		default:
			HandleUnsupportedType(rt);
		}
		return value;
	}
	
	private Pair<GlobalState<NN, VV>, LocalState<NN, VV>> encodeAutomatonState(			
			List<NN> variables, boolean local,
			int stateNumber,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		
		List<VV> values = encodeStateNumber(stateNumber, variables);
		Map<NN, VV> globalStateModifications = new HashMap<NN, VV>(variables.size());
		Map<NN, VV> localStateModifications = new HashMap<NN, VV>(variables.size());
		
		ListIterator<NN> nameIterator = variables.listIterator();
		ListIterator<VV> valueIterator = values.listIterator();
		
		if (variables.size() != values.size()) {
			throw new RuntimeException("Number of variable values differs from number of variable names during encoding automaton state."); 
		}
		
		while(nameIterator.hasNext()) {
			NN n = nameIterator.next();
			VV v = valueIterator.next();
			VariableType vt = getVariableType(n);
			switch (vt) {
			case GLOBAL_VARIABLE:
				globalStateModifications.put(n, v);
				break;
			case LOCAL_VARIABLE:
				localStateModifications.put(n, v);
				break;
			default:
				HandleUnsupportedType(vt);
			}
		}
		
		return new Pair<GlobalState<NN,VV>, LocalState<NN,VV>>(
				globalStateModifications.isEmpty() ? s : s.modify(globalStateModifications),
				localStateModifications.isEmpty() ? l : l.modify(localStateModifications));
	}
	
	private int decodeAutomatonState(
			List<NN> variableNames,
			GlobalState<NN, VV> s,
			LocalState<NN, VV> l) {
		assert(variableNames != null);
		
		ArrayList<Pair<NN, VV>> values = new ArrayList<Pair<NN, VV>>(variableNames.size());
		
		for(NN n : variableNames) {
			values.add(new Pair<NN, VV>(n, getValueOfVariable(n, s, l)));
		}
		
		return decodeStateNumber(values);
	}
			
	private void HandleUnsupportedType(VariableType vt) {
		throw new RuntimeException("Unsupported variable type: " + vt.toString());

	}
	private void HandleUnsupportedType(ReferenceType rt) {
		throw new RuntimeException("Unsupported reference type: " + rt.toString());
	}
}
