/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import org.ow2.dsrg.fm.badger.ca.karpmiller.Transition;
import org.ow2.dsrg.fm.badger.ca.statespace.GlobalState;
import org.ow2.dsrg.fm.badger.ca.statespace.LocalState;

/** Immutable implementation of {@link Transition}.
 * 
 * <p>Contains {@link GlobalState global state} (as the result of transition),
 *    <i>abandoned</i> {@link LocalState local states}
 *    and <i>taken</i> {@link LocalState local states}.
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class ImmutableTransitionImpl<NAME, VAL> implements Transition<NAME, VAL> {
	private GlobalState<NAME, VAL> globalState;
	private Collection<LocalState<NAME, VAL>> abandonedLocalStates;
	private Collection<LocalState<NAME, VAL>> takenLocalStates;
	
	/** Creates immutable transition. 
	 * 
	 * @param abandonedLocalStates {@link Collection} of abandoned {@link LocalState}s
	 * @param globalState taken {@link GlobalState} 
	 * @param takenLocalStates {@link Collection} of taken {@link LocalState}s
	 */
	public ImmutableTransitionImpl(
			Collection<LocalState<NAME, VAL>> abandonedLocalStates,
			GlobalState<NAME, VAL> globalState,
			Collection<LocalState<NAME, VAL>> takenLocalStates) {
		
		this.globalState = globalState;
		if (abandonedLocalStates == null || abandonedLocalStates.isEmpty()) {
			this.abandonedLocalStates = Collections.emptyList();
		} else {
			this.abandonedLocalStates = Collections.unmodifiableList( 
				new ArrayList<LocalState<NAME, VAL>>(abandonedLocalStates));
		}
		if (takenLocalStates == null || takenLocalStates.isEmpty()) {
			this.takenLocalStates = Collections.emptyList();
		} else {
			this.takenLocalStates = Collections.unmodifiableList( 
					new ArrayList<LocalState<NAME, VAL>>(takenLocalStates));			
		}
	}	
	
	@Override
	public Collection<LocalState<NAME, VAL>> getAbandonedLocalStates() {
		return this.abandonedLocalStates;
	}
	@Override
	public GlobalState<NAME, VAL> getGlobalState() {
		return this.globalState;
	}
	@Override
	public Collection<LocalState<NAME, VAL>> getTakenLocalStates() {
		return this.takenLocalStates;
	}
}
