/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace.impl;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.ow2.dsrg.fm.badger.ca.statespace.GlobalState;
import org.ow2.dsrg.fm.badger.ca.statespace.LTSARepository;
import org.ow2.dsrg.fm.badger.ca.statespace.LocalState;
import org.ow2.dsrg.fm.badger.ca.util.EqualityComparator;
import org.ow2.dsrg.fm.badger.ca.util.Pair;
import org.ow2.dsrg.fm.tbplib.event.EventCode;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.reference.Constant;
import org.ow2.dsrg.fm.tbplib.reference.LocalVariable;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.reference.StateVariable;
import org.ow2.dsrg.fm.tbplib.reference.Variable;

/**
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class StateSpaceExplorerImpl
		extends AbstractStateSpaceExplorerImpl<Variable, Byte, Reference> {
	
	private EventTable eventTable;
	
	public StateSpaceExplorerImpl(
			LTSARepository<?, Reference, Variable, Byte> ltsaRepository,
			EventTable eventTable) {
		super(ltsaRepository, new EqualityComparator<Byte>() {
			@Override
			public boolean equals(Byte a, Byte b) {
				return a == b || (a != null && a.equals(b));
			}
		});
		this.eventTable = eventTable;
	}

	@Override
	protected Byte getConstant(Reference r) {
		assert(r instanceof Constant);
		return ((Constant)r).getCode();
	}

	@Override
	protected Variable getVariableName(Reference r) {
		assert(r instanceof Variable);
		return (Variable)r;
	}

	@Override
	protected AbstractStateSpaceExplorerImpl.ReferenceType getReferenceType(
			Reference r) {
		if (r instanceof Variable) {
			return ReferenceType.VARIABLE;
		} else if (r instanceof Constant) {
			return ReferenceType.CONSTANT;
		}
		return ReferenceType.UNKNOWN;
	}

	@Override
	protected AbstractStateSpaceExplorerImpl.VariableType getVariableType(
			Variable name) {
		if (name instanceof StateVariable) {
			return VariableType.GLOBAL_VARIABLE;
		} else if (name instanceof LocalVariable) {
			return VariableType.LOCAL_VARIABLE;
		}
		return VariableType.UNKNOWN;
	}
	

	@Override
	protected long getEventCode(Label<Reference> label,
			GlobalState<Variable, Byte> s, LocalState<Variable, Byte> l) {
		int methodCode = label.getMethodCode(eventTable);
		if (label.isMethodCall()) {
			byte [] params = new byte [label.getParameters().size()];
			int index = 0;
			for(Reference p: label.getParameters()) {
				params[index] = getValueOfReference(p, s, l);
				++index;
			}
			return eventTable.encodeMethodCall(methodCode, params);
		} else {
			byte returnValue = label.getReturnValue() != null
					? getValueOfReference(label.getReturnValue(), s, l)
					: EventCode.VOID_VALUE;
			return eventTable.encodeMethodReturn(methodCode, returnValue);
			
		}
	}	

	@Override
	public List<Byte> encodeStateNumber(int stateNumber,
			List<Variable> variableNames) {
		LinkedList<Byte> encoded = new LinkedList<Byte>();
		ListIterator<Variable> it = variableNames.listIterator(variableNames.size());
		while(it.hasPrevious()) {
			Variable current = it.previous();
			int baseSize = current.getType().getEnums().size();
			assert(baseSize <= Byte.MAX_VALUE);
			byte currentCode = (byte) (stateNumber % baseSize);
			stateNumber = (stateNumber - currentCode) / baseSize;
			encoded.addFirst(currentCode);
		}
		return encoded;
	}

	@Override
	protected int decodeStateNumber(List<Pair<Variable, Byte>> variables) {
		int currentValue = 0;
		for(Pair<Variable, Byte> valuePair: variables) {
			currentValue *= valuePair.getFirst().getType().getEnums().size();
			currentValue += valuePair.getSecond(); 
		}
		return currentValue;
	}
}