/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C)      2011  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace.impl;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.ow2.dsrg.fm.badger.ca.statespace.VariableValuation;

/**
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public abstract class VariableValuationImpl<NAME, VAL> implements VariableValuation<NAME, VAL> {
	
	protected Map<NAME, VAL> vars; // may be null befor finishInitialization called
	protected int hashCode;
	
	protected VariableValuationImpl(Map<NAME, VAL> valuation) {
		this.vars = null;
		this.addVariableValuesMapping(valuation);
		this.hashCode = 0;
	}

	@Override
	public VAL getValueOfVariable(NAME variableName) {
		return this.vars.get(variableName);
	}

	@Override
	public boolean contains(NAME v) {
		return this.vars.containsKey(v);
	}

	@Override
	public Map<NAME, VAL> getVariableValuation() {
		return Collections.unmodifiableMap(this.vars);
	}
	
	protected void addVariableValuesMapping(Map<NAME, VAL> variableValues) {
		if (variableValues != null && !variableValues.isEmpty()) {
			if (this.vars == null) {
				this.vars = new HashMap<NAME, VAL>(variableValues);
			} else {
				this.vars.putAll(variableValues);
			}
		}
	}
	
	protected void finishInitialization() {
		if (this.vars == null) {
			this.vars = Collections.emptyMap();
		} else {
			this.vars = Collections.unmodifiableMap(this.vars);
		}
		this.hashCode = this.vars.size();	// naive hash code based on maps
	}
	
	@Override
	public int hashCode() {
		return hashCode;
	}
		
	private boolean equalsVariableValuationImpl(VariableValuationImpl<?, ?> that) {
		return (	that != null
				&&	this.hashCode == that.hashCode
				&&	mapEquals(this.vars, that.vars));
	}
	private boolean equalsVariableValuation(VariableValuation<?, ?> that) {
		return that != null && mapEquals(this.vars, that.getVariableValuation());
	}
	private static boolean mapEquals(Map<?, ?> a, Map<?, ?> b) {
		if (a == b) {
			return true;
		} else if (a != null && b != null && a.size() == b.size()) {
			for(Map.Entry<?, ?> entry: a.entrySet()) {
				Object v = b.get(entry.getKey());
				if (v == null || !v.equals(entry.getValue())) {
					return false;
				}
			}
			return true;
		} else {
			return false;
		}
	}

	public boolean equals(Object o) {
		if (this == o) {
			return true;
		} else if (o instanceof VariableValuationImpl<?, ?>) {
			return equalsVariableValuationImpl((VariableValuationImpl<?, ?>)o);
		} else if (o instanceof VariableValuation<?, ?>) {
			return equalsVariableValuation((VariableValuation<?, ?>)o);
		} else {
			return false;
		}
	}
	
	protected StringBuilder appendAsString(StringBuilder sb, String separator, boolean startWithSeparator) {
		boolean first = !startWithSeparator;
		if (sb != null && this.vars != null) {
			for(Map.Entry<NAME, VAL> e: this.vars.entrySet()) {
				NAME n = e.getKey();
				VAL v = e.getValue();
				if (first) {
					first = false;
				} else {
					sb.append(separator);
				}
				sb.append(n == null ? "<null>" : n.toString())
				  .append("=")
				  .append(v == null ? "<null>" : v.toString());
			}
		}
		return sb;
	}
	
	@Override
	public String toString() {
		return this.appendAsString(new StringBuilder(), ", ", false).toString();
	}
	
}
