/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.statespace.impl.ltsa;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.ow2.dsrg.fm.badger.ca.provision.ThreadObservationSystem;
import org.ow2.dsrg.fm.badger.ca.statespace.LTSARepository;
import org.ow2.dsrg.fm.badger.ca.statespace.LocationReference;
import org.ow2.dsrg.fm.badger.ca.statespace.encoding.LocalStateEncoder;
import org.ow2.dsrg.fm.badger.ca.util.Pair;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPValue;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPAnd;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPInSet;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNondetCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNot;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotInSet;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPOr;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPSetCondition;
import org.ow2.dsrg.fm.tbplib.provision.CounterLimit;
import org.ow2.dsrg.fm.tbplib.provision.ObservingProvisionAutomaton;

/** Creates compact implementations of {@link LTSA} from {@link LTSAThread}.
 * 
 * <p>Also provides creation of isolated {@link LTSA} transitions
 *    using {@link #create(LocalStateEncoder, CounterLimit)} method.
 * </p>
 * <p>
 * In addition, provides repository of created {@link LTSA} transitions,
 * each {@link LTSA} transition is accessible by its
 * {@link CompactLTSAImpl#hashCode() hash code} (see {@link #get(int)}).
 * This fulfills requirement formulated within {@link LTSARepository}.
 * </p>
 * 
 * <p>
 * The creation of compact threads (and recursively all of its transitions)
 * is provided using {@link #create(LTSAThread, ThreadObservationSystem, LocalStateEncoder)} method.
 * Once all the LTSA-threads are and all the isolated
 * {@link LTSA} transitions are created, the {@link #stopFactory} method
 * should be invoked to switch this object into read-only mode
 * (no more transitions may be created then, the number of all LTSA-transitions is fixed,
 *  reading using {@link #get(int)} and {@link #getLocalStateEncoder(int)} optimized).
 * </p>
 * 
 * <p>
 * Note that compact implementation provides <i>program counter</i> identification
 * using {@link CompactLTSAImpl#hashCode() hashCode()} method.
 * The value returned from hashCode() method called on particular
 * compact LTSA transition is equal to <i>program counter</i>
 * value after execution the transition. 
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class CompactLTSAFactory<CODEREF_TYPE, REFERENCE, VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE>
		implements LTSARepository<CODEREF_TYPE, REFERENCE, VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> {
	
	private class StorageCell {
		public CompactLTSAImpl<CODEREF_TYPE, REFERENCE> ltsa;
		public CounterLimit limit;
		public LocalStateEncoder<VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> stateEncoder;
		public ThreadObservationSystem<VARIABLE_NAME_TYPE> observationSystem;
	}
	
	private CounterLimit unboundCounterLimit = new CounterLimit() {		
		@Override
		public int getUpperLimit() {
			return INFINITY;
		}
	};
	private boolean factoryWorking = true;
	private int nextId = 0;
	private LocalStateEncoder<VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> currentEncoder;
	private ThreadObservationSystem<VARIABLE_NAME_TYPE> currentObservationSystem;
	private List<StorageCell> storage = new LinkedList<StorageCell>(); 
	private Map<LTSA<CODEREF_TYPE, REFERENCE>, CompactLTSAImpl<CODEREF_TYPE, REFERENCE>> visitedMap = new HashMap<LTSA<CODEREF_TYPE,REFERENCE>, CompactLTSAImpl<CODEREF_TYPE, REFERENCE>>();
		
	/** Creates {@link CompactLTSAThreadImpl compact LTSA-thread} and associates
	 *   {@link LocalStateEncoder} <b>localStateEncoder</b> with it.
	 * 
	 * <p>In addition, the resulting LTSA-thread and all its LTSA-transitions are stored within this.
	 * </p>
	 * 
	 * @param thread
	 * @param localStateEncoder
	 * @return {@link LTSAThread}
	 */
	public LTSAThread<CODEREF_TYPE, REFERENCE> create(
			LTSAThread<CODEREF_TYPE, REFERENCE> thread,
			ThreadObservationSystem<VARIABLE_NAME_TYPE> observationSystem,
			LocalStateEncoder<VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> localStateEncoder) {
		if (!factoryWorking) {
			throw new UnsupportedOperationException("Can not create since the factory has been stop.");
		}
		int currentThreadId = getNextId();
		currentEncoder = localStateEncoder;
		currentObservationSystem = observationSystem;
		StorageCell s = createStorageCell(currentThreadId);
		s.limit = unboundCounterLimit;
		s.stateEncoder = localStateEncoder;
		s.observationSystem = observationSystem;
		
		CompactLTSAThreadImpl<CODEREF_TYPE, REFERENCE> result = new CompactLTSAThreadImpl<CODEREF_TYPE, REFERENCE>(currentThreadId, thread);
		s.ltsa = result;

		
		for(LTSA<CODEREF_TYPE, REFERENCE> ltsa : thread.getTransitions()) {
			s.ltsa.connectTransition(this.visitLTSA(ltsa));
		}
		
		s.ltsa.connectingFinished();
		
		return result;		
	}
	
	/** Creates isolated {@link LTSA} transition.
	 * 
	 * <p>Given {@link LocalStateEncoder} <b>localStateEncoder</b>
	 *    will be associated with resulting {@link LTSA} transition
	 *    as well as the {@link CounterLimit} <b>parallelLimit</b>
	 *    (if not null). 
	 *  
	 * @param localStateEncoder
	 * @param parallelLimit
	 * @return {@link LTSA} transition
	 */
	public LTSA<CODEREF_TYPE, REFERENCE> create(LocalStateEncoder<VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> localStateEncoder, CounterLimit parallelLimit) {
		int id = getNextId();
		CompactLTSAImpl<CODEREF_TYPE, REFERENCE> transition = new CompactLTSAImpl<CODEREF_TYPE, REFERENCE>(
				id,
				new LTSAImpl<CODEREF_TYPE, REFERENCE>()
				);
		StorageCell c = createStorageCell(id);
		c.stateEncoder = localStateEncoder;
		c.observationSystem = null;
		c.limit = parallelLimit;
		c.ltsa = transition;
		return transition;
	}
	
	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> get(int id) {
		return getStorageCell(id).ltsa;
	}
	
	@Override
	public LocalStateEncoder<VARIABLE_NAME_TYPE, VARIABLE_VALUE_TYPE> getLocalStateEncoder(int id) {
		return getStorageCell(id).stateEncoder;
	}
	
	@Override
	public CounterLimit getCounterLimit(int id) {
		return getStorageCell(id).limit;
	}
	
	@Override
	public ThreadObservationSystem<VARIABLE_NAME_TYPE> getThreadObservationSystem(int id) {
		return getStorageCell(id).observationSystem;
	}
	
	/** Prevents this from creating more objects.
	 * 
	 *  <p>
	 *  All consequent invocations of <code>create(...)</code>
	 *  will end up with {@link UnsupportedOperationException}.
	 *  On the other hand, benefits are:
	 *  <ul><li>{@link #size()} will never change</li>
	 *      <li>optimized access using trough getters
	 *          (get.* methods)</li>
	 *  </ul>
	 *  </p>
	 */
	public void stopFactory() {
		this.factoryWorking = false;
		this.storage = Collections.unmodifiableList(new ArrayList<StorageCell>(storage));
	}
	
	/** Has this factory been {@link #stopFactory() stopped}?
	 * 
	 * @return <code>true</code> if and only if <code>this.</code>{@link #stopFactory()}
	 *         has been called  
	 */
	public boolean isStopped() {
		return !this.factoryWorking;
	}
	
	/** Returns number of currently retained {@link LTSA LTSA-transitions}.
	 * 
	 * @return nuber of elements
	 */
	public int size()
	{
		return nextId;
	}
	
	@Override
	public int getInvalidStateProgramCounterValue() {
		if (factoryWorking) {
			throw new UnsupportedOperationException("Unable to determine invalid-state program counter value untill the factory is stopped.");
		}
		return size();
	}

	@Override
	public int getMaximalStorableProgramCounterValue() {
		if (factoryWorking) {
			throw new UnsupportedOperationException("Unable to determine invalid-state program counter value untill the factory is stopped.");
		}
		return size();
	}
	
	@Override
	public int getProgramCounterValue(LTSA<?, REFERENCE> ltsa) {
		assert(ltsa instanceof CompactLTSAThreadImpl<?, ?> || ltsa instanceof CompactLTSAImpl<?, ?>);
		return ltsa.hashCode(); 
	}
		
	private CompactLTSAImpl<CODEREF_TYPE, REFERENCE> visitLTSA(
			LTSA<CODEREF_TYPE, REFERENCE> transition) {
		assert(factoryWorking);
		CompactLTSAImpl<CODEREF_TYPE, REFERENCE> transformed = visitedMap.get(transition);
		if (transformed == null) {
			int id = getNextId();
			transformed = new CompactLTSAImpl<CODEREF_TYPE, REFERENCE>(id, transition);
			StorageCell s = createStorageCell(id);
			s.ltsa = transformed;
			s.stateEncoder = currentEncoder;
			s.observationSystem = currentObservationSystem;
			s.limit = unboundCounterLimit;
			visitedMap.put(transition, transformed);
			for(LTSA<CODEREF_TYPE, REFERENCE> succ: transition.getTransitions()) {
				transformed.connectTransition(visitLTSA(succ));
			}
			transformed.connectingFinished();
		}
		return transformed;
	}
	
	private StorageCell getStorageCell(int id) {
		assert(id >= 0 && storage.size() > id);
		return storage.get(id);
	}
	
	private StorageCell createStorageCell(int id) {
		assert(factoryWorking);
		StorageCell s = new StorageCell();
		if (storage.size() == id) {
			storage.add(s);
		} else if (storage.size() < id) {
			while(storage.size() != id) {
				storage.add(null);
			}
			storage.add(s);
		} else {
			StorageCell old = storage.get(id);
			if (old == null) {
				storage.set(id, s);
			} else {
				throw new IllegalStateException("Can not replace storage cell.");				
			}
		}
		return s;
	}
	private int getNextId() {
		assert(factoryWorking);
		return nextId++;
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("%s@%d:\n", this.getClass().getName(), this.hashCode()));
		
		boolean isFirst = false;
		for(StorageCell c : storage) {
			if (c != null && c.ltsa != null) {
				sb.append(String.format("  PC=%d\n", c.ltsa.hashCode()));
				
				sb.append("    guard: ");
				appendGuard(sb, c.ltsa.getGuard());
				sb.append("\n");
				
				sb.append("    labels: ");
				isFirst = true;
				for(Label<REFERENCE> l : c.ltsa.getLabels()) {
					if (isFirst) { isFirst = false; } else { sb.append("; "); }
					appendLabel(sb, l);
				}
				sb.append("\n");
				
				sb.append("    assignments: ");
				isFirst = true;
				for(TBPAssignment<REFERENCE> a: c.ltsa.getAssignments()) {
					if (isFirst) { isFirst = false; } else { sb.append("; "); }
					appendAssignment(sb, a);
				}
				sb.append("\n");
				
				sb.append("    connections: ");
				isFirst = true;
				for(LTSA<?, ?> ltsa: c.ltsa.getTransitions()) {
					if (isFirst) { isFirst = false; } else { sb.append("; "); }
					if (ltsa != null) {
						sb.append(String.format("PC=%d", ltsa.hashCode()));
					} else {
						sb.append("<null>");
					}
				}
				sb.append("\n")
				  .append("    limit: ").append(c.limit != null ? c.limit.toString() : "<null>").append("\n")
				  .append("    encoder: ").append(c.stateEncoder != null ? c.stateEncoder.toString() : "<null>").append("\n")
				  .append("    observation: ").append(c.observationSystem != null ? observationSystemToString(c.observationSystem) : "<null>").append("\n");

				
			}
		}
		
		return sb.toString();
	}
	
	
	// toString() auxiliary methods:
	private String observationSystemToString(ThreadObservationSystem<VARIABLE_NAME_TYPE> tos) {
		StringBuilder sb = new StringBuilder();
		Collection<Pair<ObservingProvisionAutomaton<LocationReference>, List<VARIABLE_NAME_TYPE>>> tObservers = tos.getProvisionAutomata();
		sb.append("[");
		for(Pair<ObservingProvisionAutomaton<LocationReference>, List<VARIABLE_NAME_TYPE>> tObs: tObservers) {
			if(sb.length() > 1) {
				sb.append(", ");
			}
			if (tObs == null || tObs.getFirst() == null) {
				sb.append("<null>");
			} else {
				sb.append(tObs.getFirst().getClass().getSimpleName())
				  .append("@")
				  .append(Integer.toHexString(tObs.getFirst().hashCode()));
			}
		}
		for(Pair<Observer, List<VARIABLE_NAME_TYPE>> tObs: tos.getObservers()) {
			if (sb.length() > 1) {
				sb.append(", ");
			}
			if (tObs == null || tObs.getFirst() == null) {
				sb.append("<null>");
			} else {
				sb.append(tObs.getFirst().getClass().getSimpleName())
				  .append("@")
				  .append(Integer.toHexString(tObs.getFirst().hashCode()));
			}
		}
		sb.append("]");
		return sb.toString();
	}	
	private void appendAssignment(StringBuilder sb, TBPAssignment<REFERENCE> a) {
		sb.append(a.getIdf().toString()).append(" = ");
		TBPNode<REFERENCE> valueNode = a.getChild();
		if (valueNode != null && valueNode instanceof TBPValue<?>) {
			TBPValue<REFERENCE> value = (TBPValue<REFERENCE>)valueNode;
			if (value.isMethodCall()) {
				sb.append(value.getMethodCall() != null ? value.getMethodCall().toString() : "<null method call>");
			} else {
				sb.append(value.getVarname());
			}
		} else {
			sb.append(valueNode != null ? valueNode.toString() : "<null value node>");
		}
	}
	private void appendGuard(StringBuilder sb, ITBPCondition<REFERENCE> guard) {
		if (guard == null) {
		} else if (guard instanceof TBPAnd<?>) {
			sb.append("(");
			appendGuard(sb, ((TBPAnd<REFERENCE>)guard).getLeft());
			sb.append(") && (");
			appendGuard(sb, ((TBPAnd<REFERENCE>)guard).getRight());
			sb.append(")");
		} else if (guard instanceof TBPOr<?>) {
			sb.append("(");
			appendGuard(sb, ((TBPOr<REFERENCE>)guard).getLeft());
			sb.append(") || (");
			appendGuard(sb, ((TBPOr<REFERENCE>)guard).getRight());
			sb.append(")");
		} else if (guard instanceof TBPNot<?>) {
			sb.append("!(");
			appendGuard(sb, ((TBPNot<REFERENCE>)guard).getOperand());
			sb.append(")");
		} else if (guard instanceof TBPEquals<?>) {
			sb.append(((TBPEquals<REFERENCE>)guard).getLeft().toString())
				.append(" == ")
				.append(((TBPEquals<REFERENCE>)guard).getRight().toString());
		} else if (guard instanceof TBPNotEquals<?>) {
			sb.append(((TBPNotEquals<REFERENCE>)guard).getLeft().toString())
				.append(" != ")
				.append(((TBPNotEquals<REFERENCE>)guard).getRight().toString());				
		} else if (guard instanceof TBPSetCondition<?>) {
			sb.append(((TBPSetCondition<REFERENCE>)guard).getLeft().toString());
			if (guard instanceof TBPInSet<?>) {
				sb.append(" in ");
			} else if (guard instanceof TBPNotInSet<?>) {
				sb.append(" not-in ");
			}
			boolean first = true;
			for(REFERENCE reference : ((TBPSetCondition<REFERENCE>)guard).getSet()) {
				if (first) {
					first = false;
				} else {
					sb.append(", ");
				}
				sb.append(reference.toString());
			}
		} else if (guard instanceof TBPNondetCondition<?>) {
			sb.append("*");
		} else {
			throw new UnsupportedOperationException("Unkown guard.");
		}
	}
	private void appendLabel(StringBuilder sb, Label<REFERENCE> label) {
		if (label != null) {
			sb.append(label.isMethodCall() ? "^" : "$");
			sb.append(label.getInterface().toString())
				.append(".")
				.append(label.getMethod().toString())
				.append("(");
			if (label.getParameters() != null && !label.getParameters().isEmpty()) {
				StringBuilder tmpSb = new StringBuilder();
				for (REFERENCE p : label.getParameters()) {
					if (tmpSb.length() > 0) {
						tmpSb.append(", ");
					}
					tmpSb.append(p.toString());
				}
				sb.append(tmpSb);
			}
			sb.append(")");
			if (label.getReturnValue() != null) {
				sb.append(": ").append(label.getReturnValue().toString());
			}
		}
	}
}
