/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.ca.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.dsrg.fm.badger.ca.provision.AutomataSymbolTransformer;
import org.ow2.dsrg.fm.badger.ca.provision.ThreadObservationSystem;
import org.ow2.dsrg.fm.badger.ca.statespace.LTSARepository;
import org.ow2.dsrg.fm.badger.ca.statespace.LocationReference;
import org.ow2.dsrg.fm.badger.ca.statespace.encoding.LocalStateEncoder;
import org.ow2.dsrg.fm.badger.ca.statespace.encoding.LocalStateEncoderFactory;
import org.ow2.dsrg.fm.badger.ca.statespace.impl.ltsa.CompactLTSAFactory;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.ltsa.visitor.CollectLTSALabelsVisitor;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPProvisionContainerNode;
import org.ow2.dsrg.fm.tbplib.node.TBPVariableDefinition;
import org.ow2.dsrg.fm.tbplib.node.visitor.TranslateProvisionToCallSequencesVisitor;
import org.ow2.dsrg.fm.tbplib.provision.CallSequence;
import org.ow2.dsrg.fm.tbplib.provision.CompactProvisionAutomaton;
import org.ow2.dsrg.fm.tbplib.provision.DynamicProvisionAutomaton;
import org.ow2.dsrg.fm.tbplib.provision.ObservingProvisionAutomaton;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.VariableDescriptor;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.VariableFactory;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.reference.Variable;
import org.ow2.dsrg.fm.tbplib.util.TransformationChain;


/** Provides functionality for building structures necessary for
 *  checking using <i>Counter Abstraction</i>.
 *  
 *  <p>Results of an individual method call on this object is 
 *     strongly dependent on previous calls.
 *     Thus there is a recommended call protocol which ensures
 *     that all call will behave as specified:
 *    <ul><li><i>Phase 1</i>:
 *            {@link #transformProvision(TBPProvisionContainerNode, EventTable)}
 *            or {@link #transformProvisionWithUnlimitedParallel(TBPProvisionContainerNode, EventTable)}
 *            as many times as necessary
 *            <ul><li>these calls create appropriate provision automata</li></ul>
 *          </li>
 *        <li><i>Phase 2</i>:
 *            {@link #transformThread(LTSAThread, EventTable, VariableFactory)}
 *            as many times as necessary
 *            <ul><li>these calls associates threads with provision automata created in <i>Phase 1</i></ili></ul>
 *          </li>
 *        <li><i>Phase 3</i>:
 *            {@link #getLTSARepository()} (eventually repeated)
 *            <ul><li>first call fixes the repository (which was being created during <i>Phase 1</i> and <i>Phase 2</i>)
 *                    and releases additional auxiliary data structures
 *                    (which may cause unspecified behavior of other methods)
 *        </li>
 *    </ul>
 *  </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class CAbsStructuresBuilder {
		
	private CompactLTSAFactory<TBPNode<Reference>, Reference, Variable, Byte> ltsaFactory;	
	private LocalStateEncoder<Variable, Byte> emptyThreadLocalStateEncoder;
	private LocalStateEncoderFactory<Variable, Byte> encoderFactory;
	private List<ObservingProvisionAutomaton<LocationReference>> provisionAutomata;
	private List<Observer> provisionObservers;
	private Map<Observer, Pair<Observer, List<Variable>>> observerPairCache;
	private CollectLTSALabelsVisitor<TBPNode<Reference>, Reference> labelsCollector;
	private static final int MAX_TYPE_SIZE = 100;
	
	public CAbsStructuresBuilder(LocalStateEncoderFactory<Variable, Byte> encoderFactory) {
		this.encoderFactory = encoderFactory;
		this.provisionAutomata = new LinkedList<ObservingProvisionAutomaton<LocationReference>>();
		this.provisionObservers = new LinkedList<Observer>();
		this.observerPairCache = new HashMap<Observer, Pair<Observer,List<Variable>>>();
	}
		
	private CompactLTSAFactory<TBPNode<Reference>, Reference, Variable, Byte> getLTSAFactory() {
		if (ltsaFactory == null) {
			ltsaFactory = new CompactLTSAFactory<TBPNode<Reference>, Reference, Variable, Byte>();
		}
		return ltsaFactory;
	}
	private LocalStateEncoder<Variable, Byte> getEmptyThreadEncoder() {
		if (emptyThreadLocalStateEncoder == null) {
			List<Variable> emptyList = Collections.emptyList();
			emptyThreadLocalStateEncoder = encoderFactory.create(emptyList);
		}
		return emptyThreadLocalStateEncoder;
	}
	private CollectLTSALabelsVisitor<TBPNode<Reference>, Reference> getLabelsCollector() {
		if (this.labelsCollector == null) {
			this.labelsCollector = new CollectLTSALabelsVisitor<TBPNode<Reference>, Reference>(); 
		}
		return this.labelsCollector; 
	}
		
	/** Returns {@link LTSARepository}.
	 * 
	 * <p>The method fetches result of building process.
	 *    After a call of this method, auxiliary internal structures
	 *    might be eventually released.
	 *    Since that, all consequent calls of other methods have unspecified behavior.
	 * </p>
	 * 
	 * @return {@link LTSARepository}
	 */
	public LTSARepository<TBPNode<Reference>, Reference, Variable, Byte> getLTSARepository () {
		CompactLTSAFactory<TBPNode<Reference>, Reference, Variable, Byte> tmpFactory =
			getLTSAFactory();
		this.emptyThreadLocalStateEncoder = null;
		this.encoderFactory = null;
		this.provisionAutomata = null;
		this.provisionObservers = null;
		this.observerPairCache = null;
		this.labelsCollector = null;
		if (tmpFactory != null && !tmpFactory.isStopped()) {
			tmpFactory.stopFactory();
		}
		return tmpFactory;
	}
	
	/** Creates {@link ObservingProvisionAutomaton provision automaton} that
	 *  supports unlimited parallel.
	 * 
	 * @param provisionContainer provision specification
	 * @param eventTable table of events (event encoding purposes)
	 * @return {@link ObservingProvisionAutomaton}
	 */
	public ObservingProvisionAutomaton<LocationReference> transformProvisionWithUnlimitedParallel(
			TBPProvisionContainerNode<Reference> provisionContainer,
			EventTable eventTable) {
		
        TranslateProvisionToCallSequencesVisitor<Reference> toCallSeqVisitor = new TranslateProvisionToCallSequencesVisitor<Reference>(eventTable);
        
    	Collection<CallSequence> intermediateProvision = provisionContainer.visit(toCallSeqVisitor);
    	
    	DynamicProvisionAutomaton<CallSequence> dpa = new DynamicProvisionAutomaton<CallSequence>(intermediateProvision, CallSequence.MERGING_FACTORY, CallSequence.COMPARATOR);
    	
    	CompactProvisionAutomaton<LocationReference> compactPA = CompactProvisionAutomaton.compactProvisionAutomaton(
    			dpa, 
    			new AutomataSymbolTransformer<CallSequence, Variable, Byte>(getLTSAFactory(), getEmptyThreadEncoder()),
    			eventTable);
    	
    	provisionAutomata.add(compactPA);
    	
    	return compactPA;
	}
	
	/** Creates {@link Observer provision automaton} that does not support
	 *  unlimited parallel.
	 * 
	 * @param provisionContainer provision specification
	 * @param eventTable table of events (event encoding purposes)
	 * @return {@link Observer}
	 */
	public Observer transformProvision(
			TBPProvisionContainerNode<Reference> provisionContainer,
			EventTable eventTable) {
		Observer o = TransformationChain.transformProvision(provisionContainer, eventTable);
		provisionObservers.add(o);
		return o;
	}
	
	
	private Collection<Integer> mapLabelsToMethodCodes(Collection<Label<Reference>> labels, EventTable eventTable) {
		Set<Integer> methodCodes = new HashSet<Integer>();
		for(Label<?> label : labels) {
			methodCodes.add(label.getMethodCode(eventTable));
		}
		return methodCodes;
	}
	
	private List<Variable> createStateEncodingGlobalVariables(int numberOfStates, VariableFactory<Reference> variableFactory) {
		List<Variable> encodingVariables = new LinkedList<Variable>();
		
		for(int index = 0; numberOfStates > 0; ++index) {
			int currentSize = numberOfStates % MAX_TYPE_SIZE;
			numberOfStates = (numberOfStates - currentSize) / MAX_TYPE_SIZE;
			VariableDescriptor<Reference> vDesc =
				variableFactory.createStateVariable(String.format("_state_%d_O_", index), currentSize);
			encodingVariables.add((Variable)vDesc.getIdentifier());
		}
		
		// NOTE: The values of variable that represent initial state
		//       are supposed to be initialized at verification time,
		//       when a thread is created.
		//       The best situation would be to ensure that initial
		//       state will have id 0.		
		return encodingVariables;
	}

	/* Creates variables for encoding of state number, adds these variables into <b>occupiedNames</b> list. */
	private List<Variable> createStateEncodingLocalVariables(int numberOfStates, List<Reference> occupiedNames, VariableFactory<Reference> variableFactory) {
		List<Variable> encodingVariables = new LinkedList<Variable>();
		
		for(int index = 0; numberOfStates > 0; ++index) {
			int currentSize = numberOfStates % MAX_TYPE_SIZE;
			numberOfStates = (numberOfStates - currentSize) / MAX_TYPE_SIZE;
			TBPVariableDefinition<Reference> varDef = variableFactory.createLocalVariable(
					String.format("_state_%d_PA_", index),
					currentSize,
					0,
					occupiedNames);
			occupiedNames.add(varDef.getName());
			encodingVariables.add((Variable)varDef.getName());
		}
		
		// NOTE: The values of variable that represent initial state
		//       are supposed to be initialized at verification time,
		//       when a thread is created.
		//       The best situation would be to ensure that initial
		//       state will have id 0.
		return encodingVariables;
	}
	
	private ThreadObservationSystem<Variable> createObservationSystem(
			LTSAThread<TBPNode<Reference>, Reference> thread,
			EventTable eventTable,
			VariableFactory<Reference> variableFactory) {
		
		final List<Pair<ObservingProvisionAutomaton<LocationReference>, List<Variable>>> provisionAutomataWithEncodings =
			new LinkedList<Pair<ObservingProvisionAutomaton<LocationReference>,List<Variable>>>();
		final List<Pair<Observer, List<Variable>>> provisionObserversWithEncodings =
			new LinkedList<Pair<Observer,List<Variable>>>();
		// prepare thread local variables
		List<Reference> localVariableNames = new LinkedList<Reference>();
		for(TBPVariableDefinition<Reference> localVariableDefinition: thread.getLocalVariables()) {
			localVariableNames.add(localVariableDefinition.getName());
		}
		
		
		Collection<Integer> methodCodes = mapLabelsToMethodCodes(
				thread.visit(this.getLabelsCollector()),
				eventTable);		
		
		for(ObservingProvisionAutomaton<LocationReference> opa: provisionAutomata) {
			boolean observes = false;
			for(Integer methodCode: methodCodes) {
				if (opa.observesMethod(methodCode)) {
					observes = true;
					break;
				}
			}
			
			if (observes) {
				provisionAutomataWithEncodings.add(
					new Pair<ObservingProvisionAutomaton<LocationReference>, List<Variable>>(
							opa, createStateEncodingLocalVariables(opa.getNumberOfState(), localVariableNames, variableFactory))
					);
			}
			
		}
		for(Observer o: provisionObservers) {
			boolean observes = false;
			for(Integer methodCode: methodCodes) {
				if (o.observesMethod(methodCode)) {
					observes = true;
					break;
				}
			}
			
			if (observes) {
				Pair<Observer, List<Variable>> withEncoding =
					observerPairCache.get(o);
				if (withEncoding == null) {
					withEncoding = new Pair<Observer, List<Variable>>(
							o,
							createStateEncodingGlobalVariables(o.getNumberOfState(), variableFactory));
					observerPairCache.put(o, withEncoding);
				}
				provisionObserversWithEncodings.add(withEncoding);
			}
		}
		
		return new ThreadObservationSystem<Variable>() {
			@Override
			public Collection<Pair<ObservingProvisionAutomaton<LocationReference>, List<Variable>>> getProvisionAutomata() {
				return provisionAutomataWithEncodings;
			}

			@Override
			public Collection<Pair<Observer, List<Variable>>> getObservers() {
				return provisionObserversWithEncodings;
			}
		};			
	}
	
	/** Transforms given <b>thread</b> and puts it into {@link LTSARepository}.
	 * 
	 * <p>In particular, the transformation attaches program counter evaluation
	 *    to all transitions of the <b>thread</b>.
	 * </p>
	 * 
	 * <p>{@link ThreadObservationSystem} responsible checking provisions associated
	 *    with the thread and {@link LocalStateEncoder} responsible of encoding of
	 *    thread states are created and associated with the thread and its transitions.
	 *    These associations are maintained within {@link LTSARepository}
	 *    (which is product of this builder, see {@link #getLTSARepository()}).
	 * </p>
	 * 
	 * @param thread {@link LTSAThread}
	 * @param eventTable {@link EventTable} for call/return events encoding purposes
	 * @param variableFactory {@link VariableFactory}
	 *                        helps creating variables that encode states of automata that check provisions
	 * @return transformed {@link LTSAThread}
	 */
	public LTSAThread<TBPNode<Reference>, Reference> transformThread(
			LTSAThread<TBPNode<Reference>, Reference> thread,
			EventTable eventTable,
			VariableFactory<Reference> variableFactory) {
		ThreadObservationSystem<Variable> tos = createObservationSystem(thread, eventTable, variableFactory);
		List<Variable> localVariables = new LinkedList<Variable>();
		for(TBPVariableDefinition<Reference> varDefinition:  thread.getLocalVariables()) {
			localVariables.add((Variable)varDefinition.getName());
		}
		for(Pair<?, List<Variable>> p: tos.getProvisionAutomata()) {
			localVariables.addAll(p.getSecond());
		}
		LTSAThread<TBPNode<Reference>, Reference> result = this.getLTSAFactory().create(thread, tos, this.encoderFactory.create(localVariables));
		return result;
	}
}
