/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.reference;

import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.architecture.Interface;
import org.ow2.dsrg.fm.tbplib.reference.EnumerationType;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

import java.util.Map;

/**
 * Represenation of an architecture that is specialized for translation into
 * Java. The TranslationArchitecture should be created just after symbol
 * collection and surely before binding (bindings are not retained) and
 * reference resolution (which is oblivious to the concrete architecture).
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class TranslationArchitecture extends Architecture implements TranslationReference {

    private int numThreads = 0;
    private int numStateVariableSets = 0;
    private int numAssignedInitialPC = 0;

    public TranslationArchitecture(String name) {
        super(name);
    }

    public TranslationArchitecture(Architecture arch) {
        super(arch.getName());

        // Copy types
        for (EnumerationType type : arch.getTypes().values()) {
            TranslationEnumerationType newType =
                    new TranslationEnumerationType(type);

            types.put(newType.getName(), newType);
        }

        // Copy components
        for (Component comp : arch.getComponents()) {
            addComponent(new TranslationComponent(comp, this));
        }

        // Copy bindings

        for (Component comp : arch.getComponents()) {
            Component tComp = getComponent(comp.getName());
            Map<String, Interface> requiredIfaces = comp.getRequiredInterfaces();
            for (Interface requiredIface:requiredIfaces.values()){
                if (requiredIface.isBound()){
                    Interface requiredTranslationIface = tComp.getRequiredInterface(requiredIface.getName());
                    Interface providedInterface = requiredIface.getBoundInterface();

                    Component provTComp = getComponent(providedInterface.getParent().getName());
                    Interface providedTInterface = provTComp.getProvidedInterface(providedInterface.getName());

                    requiredTranslationIface.bindTo(providedTInterface);
                }
            }
        }

    }

    public int assignInitialPC() {
        return numAssignedInitialPC++;
    }

    public int assignThreadId() {
        return numThreads++;
    }

    public int assignStateVariableSetId() {
        return numStateVariableSets++;
    }

    public int getNumAssignedInitialPC() {
        return numAssignedInitialPC;
    }

    public int getNumThreads() {
        return numThreads;
    }

    public int getNumStateVariableSets() {
        return numStateVariableSets;
    }

    @Override
    public boolean writeDefinition(CodeWriter output) {
        output.println("// Representation of the enumeration types");
        output.println("public static final class Types {");
        output.increaseIndent();
        for (EnumerationType type : types.values()) {
            ((TranslationEnumerationType) type).writeDefinition(output);
        }
        output.decreaseIndent();
        output.println('}');
        output.println();
        output.println("// Representation of the components");
        output.println("public static final class Components {");
        output.increaseIndent();
        for (Component comp : components) {
            ((TranslationComponent) comp).writeDefinition(output);
        }
        output.decreaseIndent();
        output.println('}');

        return output.checkError();
    }

    @Override
    public boolean writeReadValue(CodeWriter output) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public boolean writeAssignValue(CodeWriter output, TranslationReference value) {
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
