/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.reference;

import java.util.*;
import org.ow2.dsrg.fm.tbplib.architecture.*;
import org.ow2.dsrg.fm.tbplib.reference.*;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

/**
 * Represenation of a component that is specialized for translation into Java.
 * TranslationComponents do not contain any types. All types are stored only in
 * the architecture.
 * FIXME: Also in the tbp-lib, components should not contain types, however, the
 * types should probably be available in the TBPSpecification.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class TranslationComponent extends Component implements TranslationReference {

    public TranslationComponent(Component component, TranslationArchitecture arch) {
        super(component.getName());

        // NOTE: We ignore the types, they should be handled on the level
        // of architecture and will not be associated with a particular
        // component

        // Copy state variables
        Collection<StateVariable> stateVars = component.getStateVariables().values();

        if (stateVars != null && stateVars.size() > 0) {
            int stateVarSet = arch.assignStateVariableSetId();
            for (StateVariable var : stateVars) {
                EnumerationType newType = arch.getTypes().get(var.getType().getName());
                assert (newType != null);
                Constant newInit = newType.getEnums().get(var.getInitialValue().getName());
                assert (newInit != null);
                TranslationStateVariable newVar =
                        new TranslationStateVariable(newType, var.getName(),
                        newInit, this, stateVarSet, stateVariables.size());

                stateVariables.put(var.getName(), newVar);
            }
        }

        // Copy Interfaces
        for (Interface iface : component.getProvidedInterfaces().values()) {
            addProvidedInterface(copyInterface(iface, arch));
        }
        for (Interface iface : component.getRequiredInterfaces().values()) {
            addRequiredInterface(copyInterface(iface, arch));
        }

        // Copy Threads
        for (Autonomous thread : component.getThreads().values()) {
            arch.assignThreadId();
            TranslationAutonomous newThread =
                    new TranslationAutonomous(thread.getName());
            newThread.setInitialPC(arch.assignInitialPC());
            // Copy local variables
            copyLocals(thread, newThread, arch);
            addThread(newThread);
        }
    }

    private TranslationInterface copyInterface(Interface iface, TranslationArchitecture arch) {
        TranslationInterface newIface =
                new TranslationInterface(iface.getName(), this, iface.getDirection());

        // Copy Methods and their signatures, local variables, and parameters
        for (Method method : iface.getMethods().values()) {
            TranslationMethod newMethod = new TranslationMethod(newIface, method.getName());

            if (iface.getDirection() == Interface.Direction.PROVIDED) {
                newMethod.setInitialPC(arch.assignInitialPC());
            }

            // Prepare the new signature
            MethodSignature<Reference> signature = method.getMethodSignature();
            if (signature != null) {
                Reference returnType = signature.getReturnType();
                EnumerationType newReturnType =
                        returnType == null ? null : arch.getTypes().get(returnType.getName());

                LinkedHashMap<Reference, Reference> params = signature.getParams();
                List<Parameter> newParamNames = new ArrayList<Parameter>(params.size());
                LinkedHashMap<Parameter, EnumerationType> newParams = new LinkedHashMap<Parameter, EnumerationType>();

                for (Reference param : params.keySet()) {
                    EnumerationType newType = arch.getTypes().get(params.get(param).getName());
                    Parameter p = new TranslationParameter(method, newType,
                            param.getName(), newParams.size());

                    newParams.put(p, newType);
                    newParamNames.add(p);
                }
                MethodSignature<Reference> newSignature =
                        new MethodSignature<Reference>(newParams, newReturnType);
                newMethod.setMethodSignature(newSignature);
                newMethod.setParameters(newParamNames);
            }

            // Copy local variables
            copyLocals(method, newMethod, arch);

            // Add to the interface
            newIface.addMethod(newMethod);
        }

        return newIface;
    }

    private void copyLocals(Method methodFrom, Method methodTo,
            Architecture arch) {

        for (LocalVariable local : methodFrom.getLocals().values()) {
            EnumerationType newType = arch.getTypes().get(local.getType().getName());
            assert (newType != null);
            Constant newInit = newType.getEnums().get(local.getInitialValue().getName());
            assert (newInit != null);
            TranslationLocalVariable newLocal =
                    new TranslationLocalVariable(methodTo, newType,
                    local.getName(), newInit, methodTo.getLocals().size());

            methodTo.addLocalVariable(newLocal);
        }
    }

    @Override
    public boolean writeDefinition(CodeWriter output) {
        output.append("public static final class ").append(getName()).println(" {");
        output.increaseIndent();

        output.println("// Representation of the state variables");
        for (StateVariable var : getStateVariables().values()) {
            ((TranslationStateVariable) var).writeDefinition(output);
        }
        output.println();
        output.println("// Representation of the provided methods");
        for (Interface iface : getProvidedInterfaces().values()) {
            ((TranslationInterface) iface).writeDefinition(output);
        }
        output.println();
        output.println("// Representation of the autonomous threads");
        for (Autonomous thread : getThreads().values()) {
            ((TranslationAutonomous) thread).writeDefinition(output);
        }

        output.decreaseIndent();
        output.println('}');

        return output.checkError();
    }

    @Override
    public boolean writeReadValue(CodeWriter output) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public boolean writeAssignValue(CodeWriter output, TranslationReference value) {
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
