/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.simulation.state;

import java.util.Arrays;

/**
 * Mutable, copied on every transition.
 * The <code>prepareThreadExecution</code> method has to be called before
 * modifying any thread related data. Calling <code>finishThreadExecution</code>
 * is necessary after thread data modification to ensure up-to-date hashCode.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class State extends StateElement {

    public ThreadState[] threads;
    private VariableStore[] stateVariables;
    public int[] provisions;

    public State(ThreadState[] threads, VariableStore[] stateVariables, int[] provisions) {
        // TODO: Better hashcode computation (try to minimize collisions).
        super(2053 * Arrays.hashCode(threads) +
                5351 * Arrays.hashCode(stateVariables) +
                7591 * Arrays.hashCode(provisions));

        this.threads = threads;
        this.stateVariables = stateVariables;
        this.provisions = provisions;
    }

    private void recomputeHashCode() {
        precomputedHashCode = 2053 * Arrays.hashCode(threads) +
                5351 * Arrays.hashCode(stateVariables) +
                7591 * Arrays.hashCode(provisions);
    }

    public void prepareThreadExecution(int threadId) {
        threads = threads.clone();
        threads[threadId] = threads[threadId].clone();
    }

    public byte loadStateVariable(int varSetId, int varId) {
        return stateVariables[varSetId].load(varId);
    }

    public void assignStateVariable(int varSetId, int varId, byte value) {
        VariableStore modifiedVars = stateVariables[varSetId].assign(varId, value);

        if (modifiedVars == null) {
            return;
        }

        stateVariables = stateVariables.clone();
        stateVariables[varSetId] = modifiedVars;

    // NOTE: Hash is recomputed when the thread execution is finished;
    }

    public void finishThreadExecution(int threadId) {
        recomputeHashCode();
    }

    @Override
    public State clone() {
        return new State(threads, stateVariables, provisions);
    }

    public boolean differsInStateVars(State other) {
        if (stateVariables!=null) {
            if (other.stateVariables!=null) {
                int vsI = 0;
                if (stateVariables.length != other.stateVariables.length) return true;

                for (int i = 0; i<stateVariables.length;i++){
                    VariableStore vs = stateVariables[i];
                    VariableStore vsOther = other.stateVariables[i];
                    int maxindex = vs.getSize();
                    if (maxindex!=vsOther.getSize()) return true;
                    for (int j = 0; j<maxindex ; j++) {
                        if (vs.load(j)!=vsOther.load(j)) return true;
                    }
                }

                return false;
            } else {
                return true;
            }
        } else {
            return stateVariables!=other.stateVariables;
        }

    }


    public String toString(boolean includeStateVars) {
        StringBuilder bfr = new StringBuilder();
        boolean first = true;

        bfr.append('[');
        if (threads != null) {
            for (ThreadState ts : threads) {
                if (first) {
                    first = false;
                } else {
                    bfr.append(',');
                }

                bfr.append('<');
                bfr.append("pc=");
                bfr.append(ts == null ? "END" : ts.getProgramCounter());
                bfr.append('>');
            }
        }

        if (stateVariables!=null && includeStateVars) {
            for (VariableStore vs:stateVariables){
                bfr.append(",{");
                int maxindex = vs.getSize();
                first=true;
                for (int i=0;i<maxindex;i++){
                    if (!first) {
                        bfr.append(',');
                    }else {
                        first = false;
                    }
                    bfr.append(vs.load(i));

                }

                bfr.append('}');
            }
        }

        if (provisions != null) {
            for (int p : provisions) {
                bfr.append(',');
                bfr.append(p);
            }
        }
        bfr.append(']');




        return bfr.toString();
    }

    public boolean allThreadsAreInFinalState() {
        if (threads == null) {
            return true;
        }
        for (ThreadState ts:threads) {
            if (ts!=null) return false;
        }
        return true;
    }
    
}
