/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.badger.simulation.state;

/**
 * Mutable, copied on every transition (for the active thread).
 * The stackTop has to be cloned on every modification.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class ThreadState extends StateElement {

    private StackFrame stackTop;
    private int pc;

    public ThreadState(StackFrame stackTop, int pc) {
        // TODO: Better hashcode computation (try to minimize collisions).
        super(
                (stackTop == null ? 0 : stackTop.hashCode()) +
                pc * 6473
                );
        this.stackTop = stackTop;
        this.pc = pc;
    }

    public void pushStackFrame(ConstantVariableStore parameters, VariableStore variables,
            int returnAddress) {
        stackTop = new StackFrame(
                    parameters,
                    variables,
                    returnAddress,
                    stackTop);
        recomputeHashCode();
    }

    /**
     * Removes the topmost stack frame
     * @return
     */
    public void popStackFrame() {
        assert (stackTop != null);

        if (stackTop.nextFrame == null) {
            stackTop = null;
        } else {
            stackTop = stackTop.nextFrame;
        }
        recomputeHashCode();
    }

    public int getProgramCounter() {
        return pc;
    }

    public void setProgramCounter(int pc) {
        this.pc = pc;
        recomputeHashCode();
    }

    public int getReturnAddress() {
        return stackTop.returnAddress;
    }

    public ConstantVariableStore getParameters() {
        return stackTop.parameters;
    }

    // TODO: remove this method
    private VariableStore getLocalVariables() {
        return stackTop.variables;
    }
    
    /**
     * Assigns given value to local variable identified by given identifier.
     */
    public void assignLocalVariable(int varId, byte value) {
    	VariableStore modifiedVariables = stackTop.variables.assign(varId, value);
    	if (modifiedVariables == null) {
    		return;
    	}
    	
    	stackTop.variables = modifiedVariables;
    	stackTop.recomputeHashCode();
    	recomputeHashCode();
    }
    
    /**
     * Loads local variable identified by given identifier.
     */
    public byte loadLocalVariable(int varId) {
    	return stackTop.variables.load(varId);
    }
    
    // TODO: remove this method
    private void modifyLocalVariables(VariableStore variables) {
        stackTop = stackTop.clone();
        stackTop.variables = variables;
        stackTop.recomputeHashCode();
        recomputeHashCode();
    }

    public void setReturnValue(byte returnValue) {
        stackTop = stackTop.clone();
        stackTop.returnValue = returnValue;
        stackTop.recomputeHashCode();
        recomputeHashCode();
    }

    public byte getReturnValue() {
        return stackTop.returnValue;
    }

    private void recomputeHashCode() {
        // TODO: Better hashcode computation (try to minimize collisions).
        precomputedHashCode =
                (stackTop == null ? 0 : stackTop.hashCode()) +
                pc * 6473;
    }

    @Override
    public ThreadState clone() {
        return new ThreadState(stackTop, pc);
    }
}
