/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.simulation.transition;

import org.ow2.dsrg.fm.badger.simulation.TBPPropertyViolationException;
import org.ow2.dsrg.fm.badger.simulation.state.State;

/**
 * Represents the actual behavior (transition relation) defined by a component.
 * For optimization purposes, implementations of this interface will be
 * automatically generated from the TBP specifications. Their responsibilities
 * are:
 * <ol>
 * <li> Given a State and identification of a transition
 *      execute the transition by a given thread and return the modified State
 *      and returns an event encoding.
 * </li>
 * <li> Given a State and a given thread, return list of next transitions
 *      associated with identification of guards.
 * </li>
 * <li> Given a State and identification of a guard - evaluate the guard.</li>
 * </ol>
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public interface BehaviorInterpreter {

    /**
     * Execute the transition on the given state <b>in place</b> and return
     * a list of next transitions of the thread associated with identification
     * of guards.
     *
     * @param state The current state to be modified in place
     * @param threadId Identification of the executing thread
     * @param transition Identification of a transition to be executed
     * @return identification of an observable event or of a tau action
     * @throws TBPPropertyViolationException Thrown when a property is violated
     *      during transition execution.
     */
    long executeTransition(State state, int threadId, int transition) throws TBPPropertyViolationException;

    /**
     * Given a State and a given thread, return list of next transitions
     * associated with identification of guards.
     *
     * @param state The current state
     * @param threadId Identification of the executing thread
     * @return Array of the next transitions and their guards
     */
    Transition[] nextTransitions(State state, int threadId);

    /**
     * Given a State and identification of a guard - evaluate the guard
     *
     * @param state The current state upon which to evaluate
     * @param threadId Identification of the executing thread
     * @param guard Identification of a guard
     * @return Value of the given guard in the given state
     */
    boolean evaluateGuard(State state, int threadId, int guard);

    /**
     * Creates an the initial state of the simulation.
     *
     * @return State with initial states of all threads, state variables,
     *      and provisions.
     */
    State getInitialState();


    boolean isFinal(State s);
    public Long getAnEventExpectedByObserver(State s);
}
