/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.translation;

import org.ow2.dsrg.fm.badger.reference.*;
import org.ow2.dsrg.fm.badger.simulation.transition.Transition;
import org.ow2.dsrg.fm.tbplib.architecture.Autonomous;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNondetCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotInSet;
import org.ow2.dsrg.fm.tbplib.node.visitor.TBPVisitor;
import org.ow2.dsrg.fm.tbplib.reference.*;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

import java.util.*;

/**
 * Visitor for encoding of behavior into Java representation that implements
 * the BehaviorInterpreter interface.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class BehaviorEncodingVisitor implements TBPVisitor<Object, TranslationReference> {

    private String className = null;
    private int numAssignedTransitions = 0;
    private int numAssignedPC = 0;
    private int startPC = 0;
    private int endPC = 0;
    private GuardKeeper guardKeeper;
    private List<List<Transition>> transitions = new ArrayList<List<Transition>>();
    private TranslationArchitecture architecture;
    private Observer[] provisions;
    private CodeWriter output;
    private Set<Integer> methodEncoderDemand = new HashSet<Integer>();
    private Set<TranslationMethod> methodEncoders = new HashSet<TranslationMethod>();
    private TranslationMethod currentMethod = null;

    public BehaviorEncodingVisitor(Observer[] provisions, TranslationArchitecture architecture) {
        this(new CodeWriter(System.out, true), provisions, architecture);
    }

    public BehaviorEncodingVisitor(CodeWriter output, Observer[] provisions,
            TranslationArchitecture architecture) {
        this.provisions = provisions;
        this.architecture = architecture;
        this.output = output;
    }

    /**
     * Perform the whole transformation work from TBP to Java.
     *
     * @param tbp TBP Specification to be transformed. Typically this will be
     *      a result of a TBPComposition.
     * @return <code>true</code> on success, <code>false</code> on an I/O error.
     */
    public boolean transformTBP(TBPSpecification<TranslationReference> tbp,
            String className, String packageName) {

        this.className = className;
        guardKeeper = new GuardKeeper();

        writePreamble(className, packageName);
        tbp.visit(this);
        writeEpilogue();

        return !output.checkError();
    }



    private boolean writePreamble(String className, String packageName) {
        output.println("/*");
        output.println(" *   !!! THIS FILE WAS AUTOMATICALY GENERATED BY:");
        output.println(" *");
        output.println(" *   Badger - Threaded Behavior Protocols analysis toolchain");
        output.println(" *   Copyright (C) 2008-2010  DSRG, Charles University in Prague");
        output.println(" *                            http://dsrg.mff.cuni.cz/");
        output.println(" */");
        output.println();
        output.println("package " + packageName + ";");
        output.println();
        output.println("import java.util.*;");
        output.println("import org.ow2.dsrg.fm.badger.simulation.transition.*;");
        output.println("import org.ow2.dsrg.fm.badger.simulation.state.*;");
        output.println("import org.ow2.dsrg.fm.badger.simulation.*;");
        output.println("import org.ow2.dsrg.fm.tbplib.ltsa.*;");
        output.println("import org.ow2.dsrg.fm.tbplib.ltsa.Observer;");
        output.println("import org.ow2.dsrg.fm.tbplib.event.*;");
        output.println();
        output.println("public class " + className + " extends BehaviorInterpreterImpl {");
        output.increaseIndent();
        return !output.checkError();
    }

    private boolean writeEpilogue() {
        output.println();

        // Guard table
        guardKeeper.writeGuardEvaluation(output);

        output.println();
        output.println("@Override");
        output.println("public State getInitialState() {");
        output.increaseIndent();
        output.println("return initialState;");
        output.decreaseIndent();
        output.println("}");
        output.println();
        output.println("@Override");
        output.println("public Transition[] nextTransitions(State state, int threadId) {");
        output.increaseIndent();
        output.println("return transtionTable[state.threads[threadId].getProgramCounter()];");
        output.decreaseIndent();
        output.println("}");
        output.println();

        // Architectural definitions
        architecture.writeDefinition(output);

        output.println();


        // Initial state
        writeInitialState();

        // Transiton table
        writeTransitionTable();

        // Write method encode classes
        writeMethodEncoders();

        output.decreaseIndent();
        output.println("}");
        output.close();
        return !output.checkError();
    }

    @Override
    public Object visitParsedImperativeSequence(TBPImperativeSequence<TranslationReference> node) {
        int oldEndPC = endPC;
        int middlePC = assignNewPC();

        endPC = middlePC;
        node.getLeft().visit(this);
        startPC = middlePC;
        endPC = oldEndPC;
        node.getRight().visit(this);

        return null;
    }

    @Override
    public Object visitVariableDefinition(TBPVariableDefinition<TranslationReference> node) {
        return null;
    }

    @Override
    public Object visitParsedSwitch(TBPSwitch<TranslationReference> node) {
        int oldStartPC = startPC;
        boolean isNondeterministic = node.isNondeterministic();

        // Cases
        Set<TranslationReference> cases = new HashSet<TranslationReference>();

        Iterator<TBPNode<TranslationReference>> it = node.getChildren().iterator();

        for (TranslationReference c : node.getCases()) {
            cases.add(c);
            int bodyPC = assignNewPC();
            int transitionId = assignNewTransition();
            ITBPCondition<TranslationReference> cond =
            	isNondeterministic ? 
            		null : 
            		new TBPEquals<TranslationReference>(c,
            				node.getValue().getVarname());
            output.print("case ");
            output.print(transitionId);
            output.println(':');
            output.increaseIndent();
            writeNextPC(bodyPC);
            output.println("event = EventCode.TAU_ACTION;");
            output.println("break;");
            output.decreaseIndent();
            addTransition(oldStartPC, transitionId, cond);

            startPC = bodyPC;
            it.next().visit(this);
        }

        // Default branch
        TBPNode<TranslationReference> defaultBranch = node.getDefaultBranch();

        int transitionId = assignNewTransition();
        ITBPCondition<TranslationReference> cond =
        	isNondeterministic ? 
        		null : 
        		new TBPNotInSet<TranslationReference>(
        			node.getValue().getVarname(),
        			cases);
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();

        if (defaultBranch == null) {
            writeNextPC(endPC);
            output.println("event = EventCode.TAU_ACTION;");
            output.println("break;");
            output.decreaseIndent();
            addTransition(oldStartPC, transitionId, cond);
            return null;
        }

        // With default body
        int bodyPC = assignNewPC();
        writeNextPC(bodyPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();
        addTransition(oldStartPC, transitionId, cond);

        startPC = bodyPC;
        defaultBranch.visit(this);

        return null;
    }

    @Override
    public Object visitParsedIf(TBPIf<TranslationReference> node) {
        int oldStartPC = startPC;
        
        // Then branch
        int thenPC = assignNewPC();
        int transitionId = assignNewTransition();
        ITBPCondition<TranslationReference> cond = node.getCondition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writeNextPC(thenPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();
        boolean isDeterministic = ! (cond instanceof TBPNondetCondition);
        addTransition(startPC, transitionId,  isDeterministic ?  cond: null);

        startPC = thenPC;
        node.getChild(0).visit(this);

        // Else branch
        if (node.hasElse()) {
            int elsePC = assignNewPC();
            transitionId = assignNewTransition();
            output.print("case ");
            output.print(transitionId);
            output.println(':');
            output.increaseIndent();
            writeNextPC(elsePC);
            output.println("event = EventCode.TAU_ACTION;");
            output.println("break;");
            output.decreaseIndent();
            addTransition(oldStartPC, transitionId, isDeterministic ? cond.createNegation() : null);

            startPC = elsePC;
            node.getChild(1).visit(this);
        }
        return null;
    }

    @Override
    public Object visitParsedValue(TBPValue<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedReturn(TBPReturn<TranslationReference> node) {
        int transitionId = assignNewTransition();

        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        // TODO: Handle the return correctly
        TranslationReference retVal = node.getIdf();

        if (retVal != null) {
            writeMethodReturn(retVal);
            // Emit method return event
            writeMethodReturnEvent(retVal, currentMethod);
        } else {
            writeMethodVoidReturn();
            // Emit method return event
            writeMethodReturnVoidEvent(currentMethod);
        }
        output.decreaseIndent();

        addTransition(startPC, transitionId, null);

        return null;
    }

    @Override
    public Object visitParsedWhile(TBPWhile<TranslationReference> node) {
        int oldStartPC = startPC;
        int oldEndPC = endPC;

        // While branch
        int whilePC = assignNewPC();
        int transitionId = assignNewTransition();
        ITBPCondition<TranslationReference> cond = node.getCondition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writeNextPC(whilePC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();
        boolean isDeterministic = ! (cond instanceof TBPNondetCondition);
        addTransition(oldStartPC, transitionId, isDeterministic ? cond : null);

        startPC = whilePC;
        endPC = oldStartPC;
        node.getChild().visit(this);

        // Skip cycle branch
        transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writeNextPC(oldEndPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();
        addTransition(oldStartPC, transitionId,
                isDeterministic ? cond.createNegation() : null);

        return null;
    }

    @Override
    public Object visitParsedAssignment(TBPAssignment<TranslationReference> node) {
        TBPValue<TranslationReference> child = (TBPValue<TranslationReference>) node.getChild();

        TranslationMethod method = null;
        if (child.isMethodCall()) {       	
        	// call method
        	MethodCall<TranslationReference> methodCall = child.getMethodCall();
        	method = (TranslationMethod) methodCall.getMethod();
        	
        	boolean hasReturn = method.getMethodSignature().getReturnType() != null;     	
        	if (!hasReturn) throw new UnsupportedOperationException("Not supported yet.");
        	
        	int transitionId = assignNewTransition();
        	int afterCall = assignNewPC();
        	output.print("case ");
        	output.print(transitionId);
        	output.println(':');
        	output.increaseIndent();
        	writePushStack(methodCall, method, afterCall);
            writeNextPC(method.getInitialPC());
            // Emit the corresponding event
            writeMethodCallEvent(methodCall, method);
            output.decreaseIndent();

            addTransition(startPC, transitionId, null);
            startPC = afterCall;        	
        } 

        
        int transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();

        if (child.isMethodCall()) {
            output.println("returnValue = threadState.getReturnValue();");
            writePopStack();
            node.getIdf().writeAssignValue(output,
                new TranslationReturnValue((EnumerationType) method.getMethodSignature().getReturnType()));
        } else {
            node.getIdf().writeAssignValue(output, child.getVarname());
        }
        
        writeNextPC(endPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");

        output.decreaseIndent();
        addTransition(startPC, transitionId, null);
            
        return null;
    }

    @Override
    public Object visitParsedMethodDeclaration(
            TBPMethodDeclaration<TranslationReference> node) {
        currentMethod = (TranslationMethod) node.getMethod();
        startPC = currentMethod.getInitialPC();
        methodEncoders.add(currentMethod);

        if (currentMethod.getMethodSignature().getReturnType() == null) {
            // Implicit return transition is necessary only for void returning
            // methods
            endPC = assignNewPC();
            int transitionId = assignNewTransition();

            output.print("case ");
            output.print(transitionId);
            output.println(':');
            output.increaseIndent();
            writeMethodVoidReturn();

            // Emit method return event
            writeMethodReturnVoidEvent(currentMethod);
            output.decreaseIndent();

            addTransition(endPC, transitionId, null);
        }

        // Visit the body
        node.getChild().visit(this);

        currentMethod = null;

        return null;
    }

    @Override
    public Object visitParsedSync(
            TBPSync<TranslationReference> node) {

        // Can't think of a better way of getting the mutex type and constants :o(
        TranslationEnumerationType mutexType =
                (TranslationEnumerationType)architecture.getTypes().get(
                EnumerationType.MUTEX_TYPE.getName());
        TranslationConstant UNLOCKED = (TranslationConstant)mutexType.getEnums().get(
                EnumerationType.UNLOCKED.getName());
        TranslationConstant LOCKED = (TranslationConstant)mutexType.getEnums().get(
                EnumerationType.LOCKED.getName());
        ITBPCondition<TranslationReference> guard = new TBPEquals<TranslationReference>(
                node.getMutex(), UNLOCKED);

        int oldStartPC = startPC;
        int oldEndPC = startPC;
        int beforePC = assignNewPC();
        int afterPC = assignNewPC();

        // Before, atomicaly check and lock the mutex
        int transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        node.getMutex().writeAssignValue(output, LOCKED);
        writeNextPC(beforePC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();
        addTransition(oldStartPC, transitionId, guard);

        startPC = beforePC;
        endPC = afterPC;
        node.getChild(0).visit(this);

        // After, unlock the mutex
        transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        node.getMutex().writeAssignValue(output, UNLOCKED);
        writeNextPC(oldEndPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();
        addTransition(afterPC, transitionId, null);

        return null;
    }

    @Override
    public Object visitParsedWait(
            TBPWait<TranslationReference> node) {
        throw new NoSuchMethodError();
    }

    
    
    @Override
    public Object visitParsedImperativeNull(
            TBPImperativeNull<TranslationReference> node) {
        int transitionId = assignNewTransition();

        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writeNextPC(endPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();

        addTransition(startPC, transitionId, null);

        return null;
    }

    @Override
    public Object visitParsedEmit(
            TBPEmit<TranslationReference> node) {

        MethodCall<TranslationReference> methodCall = node.getMethodCall();
        TranslationMethod method = (TranslationMethod) methodCall.getMethod();
        boolean hasReturn = method.getMethodSignature().getReturnType() != null;

        // For a method with return value, a stack pop have to be emited
        // as well
        if (hasReturn) {
            int transitionId = assignNewTransition();
            int popPC = assignNewPC();
            output.print("case ");
            output.print(transitionId);
            output.println(':');
            output.increaseIndent();	
            writePopStack();
            writeNextPC(endPC);
            output.println("event = EventCode.TAU_ACTION;");
            output.println("break;");
            output.decreaseIndent();

            addTransition(popPC, transitionId, null);
            endPC = popPC;
        }

        int transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writePushStack(methodCall, method, endPC);
        writeNextPC(method.getInitialPC());

        // Emit the corresponding event
        writeMethodCallEvent(methodCall, method);
        output.decreaseIndent();

        addTransition(startPC, transitionId, null);

        return null;
    }

    @Override
    public Object visitParsedAlternative(
            TBPAlternative<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedParallel(
            TBPParallel<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedParallelOr(
            TBPParallelOr<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedSequence(
            TBPSequence<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedAccept(
            TBPAccept<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedNull(
            TBPProvisionNull<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitLimitedReentrancy(
            TBPLimitedReentrancy<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedRepetition(
            TBPRepetition<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitUnlimitedReentrancy(
            TBPUnlimitedReentrancy<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedProvisionContainerNode(
            TBPProvisionContainerNode<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedThreadContainerNode(
            TBPThreadContainerNode<TranslationReference> node) {
        TranslationAutonomous thread = (TranslationAutonomous) node.getName();
        startPC =
                thread.getInitialPC();
        endPC =
                assignNewPC();
        int transitionId = assignNewTransition();

        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        // Thread exit --> thread state == null
        output.println("state.threads[threadId] = null;");
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();

        addTransition(endPC, transitionId, null);

        // Visit the body
        node.getChild().visit(this);

        return null;
    }

    @Override
    public Object visitSpecification(
            TBPSpecification<TranslationReference> node) {
        output.println();
        /*output.println("private EventTable eventTable;");
        output.println("private Observer[] provisions;");*/
        output.println();
        output.println("public long executeTransition(State state, int threadId, int transition) throws TBPPropertyViolationException {");
        output.increaseIndent();
        output.println("state.prepareThreadExecution(threadId);");
        output.println("ThreadState threadState = state.threads[threadId];");
        output.println("long event;");
        output.println("byte returnValue;");
        output.println();
        output.println("switch(transition) {");
        output.increaseIndent();

        // Initial segment of PCs is allocated to the entrypoints of threads 
        // and reactions
        numAssignedPC =
                architecture.getNumAssignedInitialPC();

        for (TBPThreadContainerNode<TranslationReference> thread : node.getThreads()) {
            thread.visit(this);
        }

        for (TBPMethodDeclaration<TranslationReference> reaction : node.getReactions()) {
            reaction.visit(this);
        }

        output.println("default: throw new UnsupportedOperationException(\"Transition number out of bounds.\");");
        output.decreaseIndent();
        output.println("}");
        output.println("event = advanceProvisions(state, event);");
        output.println("state.finishThreadExecution(threadId);");
        output.println("return event;");
        output.decreaseIndent();
        output.println("}");
        output.println();

        return null;
    }

    private int assignNewPC() {
        return numAssignedPC++;
    }

    private int assignNewTransition() {
        return numAssignedTransitions++;
    }

    /**
     * Adds the specified transition to the given program location.
     *
     * @param pc Program location where the given transition should start.
     * @param transitionId Id of the transition as used in the executeTransition
     *      method.
     * @param guard Guard of the transition or <code>null</code> for no guard,
     *      i.e., an aways enabled transition.
     */
    private void addTransition(int pc, int transitionId,
            ITBPCondition<TranslationReference> guard) {
        List<Transition> state;
        if (pc >= transitions.size() || (state = transitions.get(pc)) == null) {
            state = new LinkedList<Transition>();
            while (transitions.size() <= pc) {
                transitions.add(null);
            }

            transitions.set(pc, state);
        }

        state.add(new Transition(transitionId, guardKeeper.indexGuard(guard)));
    }

    private void writeTransitionTable() {
        output.println("private static final Transition[][] transtionTable = new Transition[][]{");
        output.increaseIndent();
        for (List<Transition> state : transitions) {
            if (state == null || state.size() == 0) {
                output.println("null,");
            } else {
                output.print("new Transition[] {");
                boolean first = true;
                for (Transition t : state) {
                    if (first) {
                        first = false;
                    } else {
                        output.append(',');
                    }
                    output.print("new Transition(");
                    output.print(t.transitionId);
                    output.append(',');
                    output.print(t.guardId);
                    output.print(')');
                }

                output.println("},");
            }

        }
        output.println("null");
        output.decreaseIndent();
        output.println("};");
    }

    private void writeInitialState() {
        output.println("private static final State initialState = new State(");
        output.increaseIndent();

        if (architecture.getNumThreads() == 0) {
            output.println("null,");
        } else {
            output.print("new ThreadState[]{");
            boolean first = true;
            for (Component c : architecture.getComponents()) {
                for (Autonomous t : c.getThreads().values()) {
                    TranslationAutonomous thread = (TranslationAutonomous) t;

                    if (first) {
                        first = false;
                    } else {
                        output.print(',');
                    }

                    if (t.getLocals() == null || t.getLocals().isEmpty()) {
                        output.print("new ThreadState(null,");
                    } else {
                        output.print("new ThreadState(");
                        output.print("new StackFrame(null, new VariableStore(new byte[]{");

                        boolean f = true;
                        for (LocalVariable l : t.getLocals().values()) {
                            if (!f) {
                                output.print(", ");
                            } else {
                                f = false;
                            }
                            output.print(l.getInitialValue().getCode());
                        }

                        output.print("}), 0, null),");
                    }

                    output.print(thread.getInitialPC());
                    output.print(")");
                }
            }
            output.println("},");
        }

        if (architecture.getNumStateVariableSets() == 0) {
            output.println("null,");
        } else {
            output.print("new VariableStore[]{");
            boolean firstSet = true;

            for (Component c : architecture.getComponents()) {
                Collection<StateVariable> vars = c.getStateVariables().values();
                if (vars != null && vars.size() > 0) {
                    boolean first = true;

                    if (firstSet) {
                        firstSet = false;
                    } else {
                        output.print(',');
                    }
                    output.print("new VariableStore(new byte[]{");

                    for (StateVariable var : vars) {
                        if (first) {
                            first = false;
                        } else {
                            output.print(',');
                        }
                        TranslationConstant initVal = (TranslationConstant) var.getInitialValue();
                        initVal.writeReadValue(output);
                    }
                    output.print("})");
                }
            }
            output.println("},");
        }

        // Provisions
        if (provisions == null || provisions.length == 0) {
            output.println("null");
        } else {
            output.print("new int[]{");

            boolean first = true;
            for (Observer prov : provisions)  {
                if (first) {
                    first = false;
                } else {
                    output.print(',');
                }

                output.print(prov.initialState());
            }
            
            output.println("}");
        }

        output.decreaseIndent();
        output.println(");");
    }

    private void writeNextPC(int nextPC) {
        output.print("threadState.setProgramCounter(");
        output.print(nextPC);
        output.println(");");
    }

    private void writeMethodVoidReturn() {
        output.println("threadState.setProgramCounter(threadState.getReturnAddress());");
        output.println("threadState.popStackFrame();");
    }

    private void writeMethodReturn(TranslationReference retValue) {
        output.print("threadState.setReturnValue(");
        retValue.writeReadValue(output);
        output.println(");");

        output.println("threadState.setProgramCounter(threadState.getReturnAddress());");
    }

    private void writePopStack() {
        output.println("threadState.popStackFrame();");
    }

    private void writePushStack(MethodCall<TranslationReference> call, TranslationMethod method, int retAddress) {
        LinkedHashMap<String, LocalVariable> locals = method.getLocals();
        LinkedHashMap<String, Parameter> params = method.getParameters();

        output.println("threadState.pushStackFrame(");
        output.increaseIndent();
        // Parameters
        if (params == null || params.isEmpty()) {
            output.println("null,");
        } else {
            output.print("new VariableStore(new byte[]{");
            writeReferenceList(call.getParamDecl());
            output.println("}),");
        }
        // Local variables
        if (locals == null || locals.isEmpty()) {
            output.println("null,");
        } else {
            output.print("new VariableStore(new byte[]{");
            boolean first = true;
            for (LocalVariable var : locals.values()) {
                TranslationConstant initVal = (TranslationConstant) var.getInitialValue();
                if (first) {
                    first = false;
                } else {
                    output.print(", ");
                }
                initVal.writeReadValue(output);
            }
            output.println("}),");
        }
        output.print(retAddress);
        output.println(");");
        output.decreaseIndent();
    }

    private void writeMethodEncoders() {
        output.println();
        for (TranslationMethod method : methodEncoders) {
            writeMethodEncoder(method);
        }
        output.println();
        for (int numParams : methodEncoderDemand) {
            writeMethodEncoderClass(numParams);
        }
        // Contructor
        output.println();
        output.print("public ");
        output.print(className);
        output.println("(Observer[] provisions, EventTable eventTable) {");
        output.increaseIndent();
        output.println("super(provisions, eventTable);");        
        output.println();

        for (TranslationMethod m : methodEncoders) {
            // output.print("eventTable.addMethodEncoder(");
            // NOTE: The event table is left intact, only the corresponding
            // method code is extracted and used in the generated encoder.
            
            m.writeReadValue(output);
            output.print(".setMethodCode(eventTable.encodeMethod(");
            m.writeReadValue(output);
            output.println(".getMethodName()));");

        }
        output.decreaseIndent();
        output.println("}");
    }

    private void writeMethodEncoder(TranslationMethod method) {
        MethodSignature<Reference> sig = method.getMethodSignature();
        Collection<Reference> params = sig.getParams().keySet();
        output.print("private static final MethodEncoder");
        output.print(params.size());
        methodEncoderDemand.add(params.size());
        output.print(' ');
        method.writeReadValue(output);
        output.print(" = new MethodEncoder");
        output.print(params.size());
        output.print("(\"");
        // method.writeReadValue(output);
        output.print(method.getFullname());
        output.print("\", ");
        EnumerationType retType = (TranslationEnumerationType) sig.getReturnType();
        if (retType != null) {
            output.print(retType.getEnums().size());
        } else {
            output.print('0');
        }

        for (Reference paramRef : params) {
            TranslationParameter param = (TranslationParameter) paramRef;
            output.print(", ");
            output.print(param.getType().getEnums().size());
        }
        output.println(");");
    }

    private void writeMethodEncoderClass(int params) {
        output.println();
        output.print("private static class MethodEncoder");
        output.print(params);
        output.println(" extends EventTableImpl.MethodEncoderImpl {");
        output.increaseIndent();
        output.print("public MethodEncoder");
        output.print(params);
        output.print("(String methodName, int retValDomain");
        for (int i = 0; i < params; i++) {
            output.print(", int dom");
            output.print(i);
        }
        output.println(") {");
        output.increaseIndent();
        output.print("super(methodName, retValDomain");
        for (int i = 0; i < params; i++) {
            output.print(", dom");
            output.print(i);
        }
        output.println(");");
        output.decreaseIndent();
        output.println("}");
        output.println();
        output.print("public long encodeMethodCall");
        output.print(params);
        if (params > 0) {
            output.print("(int param0");
            for (int i = 1; i < params; i++) {
                output.print(", int param");
                output.print(i);
            }
            output.println(") {");
        } else {
            output.println("() {");
        }
        output.increaseIndent();

        output.println("int paramCode = 0;");

        for (int i = params - 1; i >= 0; i--) {
            output.print("assert (param");
            output.print(i);
            output.print(" < parameterDomains[");
            output.print(i);
            output.print("] && 0 <= param");
            output.print(i);
            output.println(");");

            output.print("paramCode += param");
            output.print(i);
            output.println(';');

            if (i != 0) {
                output.print("paramCode *= parameterDomains[");
                output.print(i - 1);
                output.println("];");
            }
        }

        output.println("return EventCode.encodeMethodCall(methodCode, paramCode);");
        output.decreaseIndent();
        output.println("}");
        output.decreaseIndent();
        output.println("}");
    }

    private void writeMethodCallEvent(MethodCall<TranslationReference> call, TranslationMethod method) {
        List<TranslationReference> params = call.getParamDecl();
        output.print("event = ");
        method.writeReadValue(output);
        output.print(".encodeMethodCall");
        output.print(params.size());
        output.print('(');
        writeReferenceList(call.getParamDecl());
        output.println(");");
        output.println("break;");
    }

    private void writeMethodReturnEvent(TranslationReference retValue, TranslationMethod method) {
        output.print("event = ");
        method.writeReadValue(output);
        output.print(".encodeMethodReturn(");
        retValue.writeReadValue(output);
        output.println(");");
        output.println("break;");
    }

    private void writeMethodReturnVoidEvent(TranslationMethod method) {
        output.print("event = ");
        method.writeReadValue(output);
        output.println(".encodeMethodReturn(EventCode.VOID_VALUE);");
        output.println("break;");
    }

    private void writeReferenceList(List<TranslationReference> list) {
        boolean first = true;
        for (TranslationReference ref : list) {
            if (first) {
                first = false;
            } else {
                output.print(", ");
            }
            ref.writeReadValue(output);
        }
    }
}
