package org.ow2.dsrg.fm.badger.translation;

import org.ow2.dsrg.fm.badger.reference.*;
import org.ow2.dsrg.fm.badger.simulation.transition.Transition;
import org.ow2.dsrg.fm.tbplib.architecture.Autonomous;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.ltsa.*;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.ltsa.visitor.LTSLabelsVisitorAdapter;
import org.ow2.dsrg.fm.tbplib.ltsa.visitor.LTSVisitor;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.reference.*;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

import java.util.*;

/**
 * Visitor for encoding of behavior into Java representation that implements
 * the BehaviorInterpreter interface.
 * Source representation is LTS representation.
 * 
 * Visits nodes of LTS representation (LTSVisitor) and labels of such
 * nodes. These labels are TBP nodes so this class implements TBPVisitor
 * (via LTSLabelsVisitorAdapter).
 * 
 * @author David Hauzar
 */
public class BehaviorFromLTSEncodingVisitor extends LTSLabelsVisitorAdapter<Object, TranslationReference> implements LTSVisitor<Object, TranslationReference> {
	private Map< LTS<?, TranslationReference>, TransitionPCAndID > visitedTransitions = new HashMap<LTS<?,TranslationReference>, TransitionPCAndID>();
	
	/** The guard used with actual transition. */
	private ITBPCondition<TranslationReference> guard = null;
	private String className = null;
    private int numAssignedTransitions = 0;
    private int numAssignedPC = 0;
    private int startPC = 0;
    private int endPC = 0;
    private GuardKeeper guardKeeper;
    private List<List<Transition>> transitions = new ArrayList<List<Transition>>();
    private TranslationArchitecture architecture;
    private Observer[] provisions;
    private CodeWriter output;
    private Set<Integer> methodEncoderDemand = new HashSet<Integer>();
    private Set<TranslationMethod> methodEncoders = new HashSet<TranslationMethod>();
    private TranslationMethod currentMethod = null;
    
    /** Stores program counter and transition id of a transition. */
    private static class TransitionPCAndID {
    	private final int transitionID;
    	private final int programCounter;
    	
    	public TransitionPCAndID(int transitionID, int programCounter) {
    		this.transitionID = transitionID;
    		this.programCounter = programCounter;
    	}
    }

    public BehaviorFromLTSEncodingVisitor(Observer[] provisions, TranslationArchitecture architecture) {
        this(new CodeWriter(System.out, true), provisions, architecture);
    }

    public BehaviorFromLTSEncodingVisitor(CodeWriter output, Observer[] provisions,
            TranslationArchitecture architecture) {
        this.provisions = provisions;
        this.architecture = architecture;
        this.output = output;
    }

    /**
     * Perform the whole transformation work from LTS to Java.
     *
     * @param ltsSpecification LTS Specification to be transformed.
     * @return <code>true</code> on success, <code>false</code> on an I/O error.
     */
    public boolean transformLTS(LTSSpecification<TranslationReference> ltsSpecification,
            String className, String packageName) {

        this.className = className;
        guardKeeper = new GuardKeeper();

        writePreamble(className, packageName);
        
        visitLTSSpecification(ltsSpecification);
        
        writeEpilogue();

        return !output.checkError();
    }
    /*
    private boolean writeAdvanceProvisions() {
        output.println("");

        output.println("private long advanceProvisions(State state, long event) throws TBPPropertyViolationException {");
        output.increaseIndent();
        output.println("if (!EventCode.isObservable(event)) return event;");
        output.println();
        output.println("int method = EventCode.decodeMethod(event);");
        output.println("boolean reallocated = false;");
        output.println("int[] provStates = state.provisions;");

        output.println("for (int i=0; i<provisions.length; i++){");
        output.increaseIndent();
        output.println("Observer prov = provisions[i];");
        output.println("if (!prov.observesMethod(method)) continue;");
        output.println("int nextState = prov.nextState(provStates[i], event);");
        output.println("// No change");
        output.println("if (nextState == provStates[i]) {");
        output.increaseIndent();
        output.println("continue;");
        output.decreaseIndent();
        output.println("} else if (nextState == DFA.NO_SUCH_TRANSITION) {");
        output.increaseIndent();
        output.println("// Signalize the bad activity error");
        output.println("throw new TBPBadActivityException(event);");
        output.decreaseIndent();
        output.println("}");

        output.println("if (!reallocated) {");
        output.increaseIndent();
        output.println("// Make a copy");
        output.println("provStates = Arrays.copyOf(provStates, provStates.length);");
        output.println("reallocated = true;");
        output.decreaseIndent();
        output.println("}");
        output.println("provStates[i] = nextState;");
        output.decreaseIndent();
        output.println("}");

        output.println("state.provisions = provStates;");
        output.println("return event;");
        output.decreaseIndent();
        output.println('}');
        return !output.checkError();
    }
      */
    private boolean writePreamble(String className, String packageName) {
        output.println("/*");
        output.println(" *   !!! THIS FILE WAS AUTOMATICALY GENERATED BY:");
        output.println(" *");
        output.println(" *   Badger - Threaded Behavior Protocols analysis toolchain");
        output.println(" *   Copyright (C) 2008-2010  DSRG, Charles University in Prague");
        output.println(" *                            http://dsrg.mff.cuni.cz/");
        output.println(" */");
        output.println();
        output.println("package " + packageName + ";");
        output.println();
        output.println("import java.util.*;");
        output.println("import org.ow2.dsrg.fm.badger.simulation.transition.*;");
        output.println("import org.ow2.dsrg.fm.badger.simulation.state.*;");
        output.println("import org.ow2.dsrg.fm.badger.simulation.*;");
        output.println("import org.ow2.dsrg.fm.tbplib.ltsa.*;");
        output.println("import org.ow2.dsrg.fm.tbplib.ltsa.Observer;");
        output.println("import org.ow2.dsrg.fm.tbplib.event.*;");
        output.println();
        output.println("public class " + className + " extends BehaviorInterpreterImpl {");
        output.increaseIndent();
        return !output.checkError();
    }

    private boolean writeEpilogue() {
        output.println();

        // Guard table
        guardKeeper.writeGuardEvaluation(output);

        output.println();
        output.println("@Override");
        output.println("public State getInitialState() {");
        output.increaseIndent();
        output.println("return initialState;");
        output.decreaseIndent();
        output.println("}");
        output.println();
        output.println("@Override");
        output.println("public Transition[] nextTransitions(State state, int threadId) {");
        output.increaseIndent();
        output.println("return transtionTable[state.threads[threadId].getProgramCounter()];");
        output.decreaseIndent();
        output.println("}");
        output.println();

        // Architectural definitions
        architecture.writeDefinition(output);

        output.println();

        // Write method encode classes

        // Initial state
        writeInitialState();

        // Transiton table
        writeTransitionTable();

        // Write method encode classes
        writeMethodEncoders();

        output.decreaseIndent();
        output.println("}");
        output.close();
        return !output.checkError();
    }

    @Override
    public Object visitVariableDefinition(TBPVariableDefinition<TranslationReference> node) {
        return -1;
    }

    @Override
    public Object visitParsedValue(TBPValue<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedReturn(TBPReturn<TranslationReference> node) {
        int transitionId = assignNewTransition();

        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        // TODO: Handle the return correctly
        TranslationReference retVal = node.getIdf();

        if (retVal != null) {
            writeMethodReturn(retVal);
            // Emit method return event
            writeMethodReturnEvent(retVal, currentMethod);
        } else {
            writeMethodVoidReturn();
            // Emit method return event
            writeMethodReturnVoidEvent(currentMethod);
        }
        output.decreaseIndent();

        //assert(guard == null);
        addTransition(startPC, transitionId, guard);

        return transitionId;
    }

    @Override
    public Object visitParsedAssignment(TBPAssignment<TranslationReference> node) {
        TBPValue<TranslationReference> child = (TBPValue<TranslationReference>) node.getChild();

        TranslationMethod method = null;
        if (child.isMethodCall()) {       	
        	// call method
        	MethodCall<TranslationReference> methodCall = child.getMethodCall();
        	method = (TranslationMethod) methodCall.getMethod();
        	
        	boolean hasReturn = method.getMethodSignature().getReturnType() != null;     	
        	if (!hasReturn) throw new UnsupportedOperationException("Not supported yet.");
        	
        	int transitionId = assignNewTransition();
        	int afterCall = assignNewPC();
        	output.print("case ");
        	output.print(transitionId);
        	output.println(':');
        	output.increaseIndent();
        	writePushStack(methodCall, method, afterCall);
            writeNextPC(method.getInitialPC());
            // Emit the corresponding event
            writeMethodCallEvent(methodCall, method);
            output.decreaseIndent();

            addTransition(startPC, transitionId, guard);
            guard = null;
            startPC = afterCall;        	
        } 

        
        int transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();

        if (child.isMethodCall()) {
            output.println("returnValue = threadState.getReturnValue();");
            writePopStack();
            node.getIdf().writeAssignValue(output,
                new TranslationReturnValue((EnumerationType) method.getMethodSignature().getReturnType()));
        } else {
            node.getIdf().writeAssignValue(output, child.getVarname());
        }
        
        writeNextPC(endPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");

        output.decreaseIndent();
        addTransition(startPC, transitionId, guard);
            
        return transitionId;
    }

    @Override
    public Object visitParsedMethodDeclaration(
            TBPMethodDeclaration<TranslationReference> node) {
        currentMethod = (TranslationMethod) node.getMethod();
        startPC = currentMethod.getInitialPC();
        methodEncoders.add(currentMethod);

        if (currentMethod.getMethodSignature().getReturnType() == null) {
            // Implicit return transition is necessary only for void returning
            // methods
            endPC = assignNewPC();
            int transitionId = assignNewTransition();

            output.print("case ");
            output.print(transitionId);
            output.println(':');
            output.increaseIndent();
            writeMethodVoidReturn();

            // Emit method return event
            writeMethodReturnVoidEvent(currentMethod);
            output.decreaseIndent();

            addTransition(endPC, transitionId, null);
        }

        return -1;
    }

    @Override
    public Object visitParsedImperativeNull(
            TBPImperativeNull<TranslationReference> node) {
        int transitionId = assignNewTransition();

        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writeNextPC(endPC);
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();

        addTransition(startPC, transitionId, guard);

        return transitionId;
    }

    @Override
    public Object visitParsedEmit(
            TBPEmit<TranslationReference> node) {

        MethodCall<TranslationReference> methodCall = node.getMethodCall();
        TranslationMethod method = (TranslationMethod) methodCall.getMethod();
        boolean hasReturn = method.getMethodSignature().getReturnType() != null;

        // For a method with return value, a stack pop have to be emited
        // as well
        if (hasReturn) {
            int transitionId = assignNewTransition();
            int popPC = assignNewPC();
            output.print("case ");
            output.print(transitionId);
            output.println(':');
            output.increaseIndent();	
            writePopStack();
            writeNextPC(endPC);
            output.println("event = EventCode.TAU_ACTION;");
            output.println("break;");
            output.decreaseIndent();

            addTransition(popPC, transitionId, null);
            endPC = popPC;
        }

        int transitionId = assignNewTransition();
        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        writePushStack(methodCall, method, endPC);
        writeNextPC(method.getInitialPC());

        // Emit the corresponding event
        writeMethodCallEvent(methodCall, method);
        output.decreaseIndent();

        addTransition(startPC, transitionId, guard);

        return transitionId;
    }

    @Override
    public Object visitParsedAlternative(
            TBPAlternative<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedParallel(
            TBPParallel<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedParallelOr(
            TBPParallelOr<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedSequence(
            TBPSequence<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedAccept(
            TBPAccept<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedNull(
            TBPProvisionNull<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitLimitedReentrancy(
            TBPLimitedReentrancy<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedRepetition(
            TBPRepetition<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitUnlimitedReentrancy(
            TBPUnlimitedReentrancy<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedProvisionContainerNode(
            TBPProvisionContainerNode<TranslationReference> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object visitParsedThreadContainerNode(
            TBPThreadContainerNode<TranslationReference> node) {
    	
        TranslationAutonomous thread = (TranslationAutonomous) node.getName();
        startPC =
                thread.getInitialPC();
        endPC =
                assignNewPC();
        int transitionId = assignNewTransition();

        output.print("case ");
        output.print(transitionId);
        output.println(':');
        output.increaseIndent();
        // Thread exit --> thread state == null
        output.println("state.threads[threadId] = null;");
        output.println("event = EventCode.TAU_ACTION;");
        output.println("break;");
        output.decreaseIndent();

        assert(guard == null);
        addTransition(endPC, transitionId, null);

        return transitionId;
    }
    
    public Object visitLTSSpecification(
            LTSSpecification<TranslationReference> ltsSpecification) {
        output.println();
        output.println("private EventTable eventTable;");
        output.println("private Observer[] provisions;");
        output.println();
        output.println("public long executeTransition(State state, int threadId, int transition) throws TBPPropertyViolationException {");
        output.increaseIndent();
        output.println("state.prepareThreadExecution(threadId);");
        output.println("ThreadState threadState = state.threads[threadId];");
        output.println("long event;");
        output.println("byte returnValue;");
        output.println("byte[] params = null;");
        output.println();
        output.println("switch(transition) {");
        output.increaseIndent();

        // Initial segment of PCs is allocated to the entrypoints of threads
        // and reactions
        numAssignedPC =
                architecture.getNumAssignedInitialPC();

        int tcounter = 0;
        for (LTSThread<TranslationReference> ltsThread : ltsSpecification.getThreads()) {
            output.println("// Thread "+tcounter++);
        	ltsThread.visit(this);
        }

        int mcounter = 0;
        for (LTSMethod<TranslationReference> ltsMethod : ltsSpecification.getMethods()) {
            output.println("// Method "+mcounter++);
        	ltsMethod.visit(this);
        }

        output.println("default: throw new UnsupportedOperationException(\"Transition number out of bounds.\");");
        output.decreaseIndent();
        output.println("}");
        output.println("event = advanceProvisions(state, event);");
        output.println("state.finishThreadExecution(threadId);");
        output.println("return event;");
        output.decreaseIndent();
        output.println("}");
        output.println();

        return null;
    }

    private int assignNewPC() {
        return numAssignedPC++;
    }

    private int assignNewTransition() {
        return numAssignedTransitions++;
    }

    /**
     * Adds the specified transition to the given program location.
     *
     * @param pc Program location where the given transition should start.
     * @param transitionId Id of the transition as used in the executeTransition
     *      method.
     * @param guard Guard of the transition or <code>null</code> for no guard,
     *      i.e., an aways enabled transition.
     */
    private void addTransition(int pc, int transitionId,
            ITBPCondition<TranslationReference> guard) {
        List<Transition> state;
        if (pc >= transitions.size() || (state = transitions.get(pc)) == null) {
            state = new LinkedList<Transition>();
            while (transitions.size() <= pc) {
                transitions.add(null);
            }

            transitions.set(pc, state);
        }

        state.add(new Transition(transitionId, guardKeeper.indexGuard(guard)));
    }

    private void writeTransitionTable() {
        output.println("private static final Transition[][] transtionTable = new Transition[][]{");
        output.increaseIndent();
        for (List<Transition> state : transitions) {
            if (state == null || state.size() == 0) {
                output.println("null,");
            } else {
                output.print("new Transition[] {");
                boolean first = true;
                for (Transition t : state) {
                    if (first) {
                        first = false;
                    } else {
                        output.append(',');
                    }
                    output.print("new Transition(");
                    output.print(t.transitionId);
                    output.append(',');
                    output.print(t.guardId);
                    output.print(')');
                }

                output.println("},");
            }

        }
        output.println("null");
        output.decreaseIndent();
        output.println("};");
    }

    private void writeInitialState() {
        output.println("private static final State initialState = new State(");
        output.increaseIndent();

        if (architecture.getNumThreads() == 0) {
            output.println("null,");
        } else {
            output.print("new ThreadState[]{");
            boolean first = true;
            for (Component c : architecture.getComponents()) {
                for (Autonomous t : c.getThreads().values()) {
                    TranslationAutonomous thread = (TranslationAutonomous) t;

                    if (first) {
                        first = false;
                    } else {
                        output.print(',');
                    }

                    if (t.getLocals() == null || t.getLocals().isEmpty()) {
                        output.print("new ThreadState(null,");
                    } else {
                        output.print("new ThreadState(");
                        output.print("new StackFrame(null, new VariableStore(new byte[]{");

                        boolean f = true;
                        for (LocalVariable l : t.getLocals().values()) {
                            if (!f) {
                                output.print(", ");
                            } else {
                                f = false;
                            }
                            output.print(l.getInitialValue().getCode());
                        }

                        output.print("}), 0, null),");
                    }

                    output.print(thread.getInitialPC());
                    output.print(")");
                }
            }
            output.println("},");
        }

        if (architecture.getNumStateVariableSets() == 0) {
            output.println("null,");
        } else {
            output.print("new VariableStore[]{");
            boolean firstSet = true;

            for (Component c : architecture.getComponents()) {
                Collection<StateVariable> vars = c.getStateVariables().values();
                if (vars != null && vars.size() > 0) {
                    boolean first = true;

                    if (firstSet) {
                        firstSet = false;
                    } else {
                        output.print(',');
                    }
                    output.print("new VariableStore(new byte[]{");

                    for (StateVariable var : vars) {
                        if (first) {
                            first = false;
                        } else {
                            output.print(',');
                        }
                        TranslationConstant initVal = (TranslationConstant) var.getInitialValue();
                        initVal.writeReadValue(output);
                    }
                    output.print("})");
                }
            }
            output.println("},");
        }

        // Provisions
        if (provisions == null || provisions.length == 0) {
            output.println("null");
        } else {
            output.print("new int[]{");

            boolean first = true;
            for (Observer prov : provisions)  {
                if (first) {
                    first = false;
                } else {
                    output.print(',');
                }

                output.print(prov.initialState());
            }
            
            output.println("}");
        }

        output.decreaseIndent();
        output.println(");");
    }

    private void writeNextPC(int nextPC) {
        output.print("threadState.setProgramCounter(");
        output.print(nextPC);
        output.println(");");
    }

    private void writeMethodVoidReturn() {
        output.println("threadState.setProgramCounter(threadState.getReturnAddress());");
        output.println("threadState.popStackFrame();");
    }

    private void writeMethodReturn(TranslationReference retValue) {
        output.print("threadState.setReturnValue(");
        retValue.writeReadValue(output);
        output.println(");");

        output.println("threadState.setProgramCounter(threadState.getReturnAddress());");
    }

    private void writePopStack() {
        output.println("threadState.popStackFrame();");
    }

    private void writePushStack(MethodCall<TranslationReference> call, TranslationMethod method, int retAddress) {
        LinkedHashMap<String, LocalVariable> locals = method.getLocals();
        LinkedHashMap<String, Parameter> params = method.getParameters();

        // Parameters
        output.print("params = ");
        if (params == null || params.isEmpty()) {
            output.println("null;");
        } else {
            output.print("new byte[]{");
            writeReferenceList(call.getParamDecl());
            output.println("};");
        }
        output.println("threadState.pushStackFrame(");
        output.increaseIndent();
        // Parameters
        if (params == null || params.isEmpty()) {
            output.println("null,");
        } else {
            output.println("new VariableStore(params),");
        }
        // Local variables
        if (locals == null || locals.isEmpty()) {
            output.println("null,");
        } else {
            output.print("new VariableStore(new byte[]{");
            boolean first = true;
            for (LocalVariable var : locals.values()) {
                TranslationConstant initVal = (TranslationConstant) var.getInitialValue();
                if (first) {
                    first = false;
                } else {
                    output.print(", ");
                }
                initVal.writeReadValue(output);
            }
            output.println("}),");
        }
        output.print(retAddress);
        output.println(");");
        output.decreaseIndent();
    }

    private void writeMethodEncoders() {
        output.println();
        for (TranslationMethod method : methodEncoders) {
            writeMethodEncoder(method);
        }
        output.println();
        for (int numParams : methodEncoderDemand) {
            writeMethodEncoderClass(numParams);
        }
        // Contructor
        output.println();
        output.print("public ");
        output.print(className);
        output.println("(Observer[] provisions, EventTable eventTable) {");
        output.increaseIndent();
        output.println("super(provisions, eventTable);");
        output.println();

        for (TranslationMethod m : methodEncoders) {
            // output.print("eventTable.addMethodEncoder(");
            // NOTE: The event table is left intact, only the corresponding
            // method code is extracted and used in the generated encoder.
            
            m.writeReadValue(output);
            output.print(".setMethodCode(eventTable.encodeMethod(");
            m.writeReadValue(output);
            output.println(".getMethodName()));");

        }
        output.decreaseIndent();
        output.println("}");
    }

    private void writeMethodEncoder(TranslationMethod method) {
        MethodSignature<Reference> sig = method.getMethodSignature();
        Collection<Reference> params = sig.getParams().keySet();
        output.print("private static final MethodEncoder");
        output.print(params.size());
        methodEncoderDemand.add(params.size());
        output.print(' ');
        method.writeReadValue(output);
        output.print(" = new MethodEncoder");
        output.print(params.size());
        output.print("(\"");
        // method.writeReadValue(output);
        output.print(method.getFullname());
        output.print("\", ");
        EnumerationType retType = (TranslationEnumerationType) sig.getReturnType();
        if (retType != null) {
            output.print(retType.getEnums().size());
        } else {
            output.print('0');
        }

        for (Reference paramRef : params) {
            TranslationParameter param = (TranslationParameter) paramRef;
            output.print(", ");
            output.print(param.getType().getEnums().size());
        }
        output.println(");");
    }

    private void writeMethodEncoderClass(int params) {
        output.println();
        output.print("private static class MethodEncoder");
        output.print(params);
        output.println(" extends EventTableImpl.MethodEncoderImpl {");
        output.increaseIndent();
        output.print("public MethodEncoder");
        output.print(params);
        output.print("(String methodName, int retValDomain");
        for (int i = 0; i < params; i++) {
            output.print(", int dom");
            output.print(i);
        }
        output.println(") {");
        output.increaseIndent();
        output.print("super(methodName, retValDomain");
        for (int i = 0; i < params; i++) {
            output.print(", dom");
            output.print(i);
        }
        output.println(");");
        output.decreaseIndent();
        output.println("}");
        output.println();
        output.print("public long encodeMethodCall");
        output.print(params);
        if (params > 0) {
            output.print("(int param0");
            for (int i = 1; i < params; i++) {
                output.print(", int param");
                output.print(i);
            }
            output.println(") {");
        } else {
            output.println("() {");
        }
        output.increaseIndent();

        output.println("int paramCode = 0;");

        for (int i = params - 1; i >= 0; i--) {
            output.print("assert (param");
            output.print(i);
            output.print(" < parameterDomains[");
            output.print(i);
            output.print("] && 0 <= param");
            output.print(i);
            output.println(");");

            output.print("paramCode += param");
            output.print(i);
            output.println(';');

            if (i != 0) {
                output.print("paramCode *= parameterDomains[");
                output.print(i - 1);
                output.println("];");
            }
        }

        output.println("return EventCode.encodeMethodCall(methodCode, paramCode);");
        output.decreaseIndent();
        output.println("}");
        output.decreaseIndent();
        output.println("}");
    }

    private void writeMethodCallEvent(MethodCall<TranslationReference> call, TranslationMethod method) {
        List<TranslationReference> params = call.getParamDecl();
        output.print("event = ");
        method.writeReadValue(output);
        output.println(".encodeMethodCall(params);");
        output.println("break;");
    }

    private void writeMethodReturnEvent(TranslationReference retValue, TranslationMethod method) {
        output.print("event = ");
        method.writeReadValue(output);
        output.print(".encodeMethodReturn(");
        retValue.writeReadValue(output);
        output.println(");");
        output.println("break;");
    }

    private void writeMethodReturnVoidEvent(TranslationMethod method) {
        output.print("event = ");
        method.writeReadValue(output);
        output.println(".encodeMethodReturn(EventCode.VOID_VALUE);");
        output.println("break;");
    }

    private void writeReferenceList(List<TranslationReference> list) {
        boolean first = true;
        for (TranslationReference ref : list) {
            if (first) {
                first = false;
            } else {
                output.print(", ");
            }
            ref.writeReadValue(output);
        }
    }

	@Override
	public <T extends TBPNode<TranslationReference> > Object visit(LTS<T, TranslationReference> lts) {
		if (visitedTransitions.containsKey(lts)) {
			// The code for this transition is already generated. Just
			// connect the transition to previous transition.
			addTransition(startPC, visitedTransitions.get(lts).transitionID, lts.getGuard());
			return null;
		}
				
		this.guard = lts.getGuard();

		int oldEndPC = endPC;
		int newPC= assignNewPC();
		if (lts.getTransitionsCount() > 0) {
			endPC = newPC;
		}
		
		if (lts.getTransitionsCount() >= 1) {
			// if transitions connected to this transitions are visited,
			// set program counter to go to such visited transitions
			// this assumes: pc!=null => all connected transitions all visited
			TransitionPCAndID pcAndID = visitedTransitions.get( lts.getTransitions().get(0) );
			if (pcAndID != null) endPC = pcAndID.programCounter;
		}
		
		// Generate the code for this transition
		visitedTransitions.put(lts, 
				new TransitionPCAndID( (Integer) lts.getLabel().visit(this), 
						startPC ));
		
		startPC = newPC;
		endPC = oldEndPC;
			
		// Visit connected transitions
		int tempStartPC = startPC;
		for (LTS<? extends TBPNode<TranslationReference>, TranslationReference> transition: lts.getTransitions()) {
			startPC = tempStartPC;
			transition.visit( this );
		}
		
		return null;
	
	}
	
	@Override
	public Object visitMethodDeclaration(LTSMethod<TranslationReference> ltsMethod) {
		return visitLTSNodeWithoutGeneratingPC(ltsMethod);
	}
	
	@Override
	public Object visitThread(LTSThread<TranslationReference> ltsThread) {
		return visitLTSNodeWithoutGeneratingPC(ltsThread);
	}

	/**
	 * Visits LTS node without generating new program counter.
	 * 
	 * @param lts lts node to visit.
	 * @return null
	 */
	private <T extends TBPNode<TranslationReference> > Object visitLTSNodeWithoutGeneratingPC(
			LTS<T, TranslationReference> lts) {
		this.guard = null;		
		// Generate the code for this transition
		lts.getLabel().visit(this);
		
			
		// Visit connected transitions
		int tempStartPC = startPC;
		for (LTS<? extends TBPNode<TranslationReference>, TranslationReference> transition: lts.getTransitions()) {
			startPC = tempStartPC;
    		transition.visit( this );
		}
		
		return null;
	}
}
