/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.badger.translation;

import org.ow2.dsrg.fm.badger.reference.TranslationReference;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNondetCondition;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

import java.util.*;

/**
 * This class is a container for guards to be used in the TBP to Java
 * translation. It assigns unique guards unique indices and merges identical 
 * guards under common index. It can also pretty-print the resulting Java
 * representation of the previously indexed guards
 * (@see BehaviorInterpreter.evaluateGuard(State,int,int)).
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class GuardKeeper {

    private Map<ITBPCondition<TranslationReference>, Integer> guardsToIndices =
            new HashMap<ITBPCondition<TranslationReference>, Integer>();

    private List<ITBPCondition<TranslationReference>> guards =
            new ArrayList<ITBPCondition<TranslationReference>>();

    public GuardKeeper() {
        // An always satisfied guard
        guards.add(null);
    }

    /**
     * Assigns an integer index to the given guard. In case an equal guard has
     * already been indexed, the same index is returned.
     *
     * @param guard Guard to be indexed or <code>null</code> for no guard =
     *      an always satisfied guard.
     * @return Unique integer index for a new guard or an index previously
     *      assigned to an equal guard.
     */
    public int indexGuard(ITBPCondition<TranslationReference> guard) {
        
        if (guard == null || guard instanceof TBPNondetCondition){
            return 0;
        }

        Integer index = guardsToIndices.get(guard);

        if (index == null) {
            index = guards.size();
            guardsToIndices.put(guard, index);
            guards.add(guard);
        }
        
        return index;
    }

    /**
     * Writes out the Java representation of the (so-far) indexed guards.
     * (@see BehaviorInterpreter.evaluateGuard(State,int,int)).
     *
     * @param output Stream to which to output the representation. At least
     *      a preamble is expected to be written prior to this call.
     * @return <code>true</code> on success.
     */
    public boolean writeGuardEvaluation(CodeWriter output) {
        // Preamble
        output.println("@Override");
        output.println("public boolean evaluateGuard(State state, int threadId, int guard) {");
        output.increaseIndent();
        output.println("ThreadState threadState = state.threads[threadId];");
        output.println();
        output.println("switch (guard) {");
        output.increaseIndent();

        ConditionEncodingVisitor cev = new ConditionEncodingVisitor(output);

        // Cases
        Iterator<ITBPCondition<TranslationReference>> it = guards.iterator();
        for (int i = 0; it.hasNext(); i++) {
            ITBPCondition<TranslationReference> guard = it.next();
            output.print("case ");output.print(i);output.println(':');
            output.increaseIndent();
            if (guard == null)  {
                // An always satisfied guard
                output.println("return true;");
            } else {
                output.print("return ");
                guard.visit(cev);
                output.println(";");
            }

            output.decreaseIndent();
        }
        output.println("default:");
        output.increaseIndent();
        output.println("throw new UnsupportedOperationException(\"Guard number out of bounds.\");");
        output.decreaseIndent();

        // Epilogue
        output.decreaseIndent();
        output.println("}");
        output.decreaseIndent();
        output.println("}");

        return !output.checkError();
    }
}
