/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.translation;

import org.ow2.dsrg.fm.tbplib.ltsa.*;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNondetCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotInSet;
import org.ow2.dsrg.fm.tbplib.node.visitor.TBPVisitor;
import org.ow2.dsrg.fm.tbplib.reference.Reference;

import java.util.*;

/**
 * Visitor for encoding of behavior into LTS automaton.
 *
 * @author David Hauzar (modified by Ondrej Cerny &lt;ondrej.cerny@email.cz&gt)
 */
public class LTSEncodingVisitor<REFERENCE extends Reference> implements TBPVisitor<LTS<? extends TBPNode<REFERENCE>, REFERENCE>, REFERENCE> {
	/** Transitions to that actual transition should be connected. */
	private Vector< LTS<? extends TBPNode<REFERENCE>, REFERENCE>> previousTransitions = new Vector<LTS<? extends TBPNode<REFERENCE>,REFERENCE>>();
	/** The guard that enables to enter actual transition. */
	private ITBPCondition<REFERENCE> guard = null;
	
    //private Architecture architecture;


    private REFERENCE UNLOCKED;
    private REFERENCE LOCKED;


    /**
     * Creates the instance of LTSEncodingVisitor.
     */
    public LTSEncodingVisitor(REFERENCE mutexLockedConstant, REFERENCE mutexUnlockedConstant) {
    	UNLOCKED = mutexUnlockedConstant;
    	LOCKED = mutexLockedConstant;
     }

    /**
     * Perform the whole transformation work from TBP to LTS.
     *
     * @param tbp TBP Specification to be transformed. Typically this will be
     *      a result of a TBPComposition.
     * @return transformed TBP Specificication.
     */
    public LTSSpecification<REFERENCE> transformTBP(TBPSpecification<REFERENCE> tbp) {    	
        return visitSpecificationInternal(tbp);
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedImperativeSequence(TBPImperativeSequence<REFERENCE> node) {
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> retValue = node.getLeft().visit(this);
        node.getRight().visit(this);
        guard = null;

        return retValue;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitVariableDefinition(TBPVariableDefinition<REFERENCE> node) {
        return null;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedSwitch(TBPSwitch<REFERENCE> node) {
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> entryTransition = LTSImpl.createInstance(new TBPImperativeNull<REFERENCE>(), guard);
    	connectTransitionToPreviousTransitions(entryTransition);
    	    	
    	// Cases
        Set<REFERENCE> casesForDefault = new HashSet<REFERENCE>();
        Vector< LTS<? extends TBPNode<REFERENCE>, REFERENCE>> tempPreviousTransitions = new Vector<LTS<? extends TBPNode<REFERENCE>,REFERENCE>>();


        List<REFERENCE> cases = node.getCases();
        Iterator<REFERENCE> cIt = (cases==null)?null:cases.iterator();
        TBPNode<REFERENCE> defBranch = null;
        for (TBPNode<REFERENCE> branch: node.getChildren()){
            ITBPCondition<REFERENCE> cond = null;
            if (cIt!=null){
                if (cIt.hasNext()){
                    REFERENCE c = cIt.next();
                    casesForDefault.add(c);
                    cond = new TBPEquals<REFERENCE>(c, node.getValue().getVarname());
                } else {
                    assert defBranch == null;
                    defBranch = branch;
                    //break;
                    continue;
                }
            }

            setNewPreviousTransitionAndGuard(entryTransition, cond );

            branch.visit(this);
            
            tempPreviousTransitions.addAll(previousTransitions);
        }
    	
    	// Default branch - only for deterministic cases
    	if (cIt!=null){

            ITBPCondition<REFERENCE> cond =
                node.isNondeterministic() ?
                    null :
                    new TBPNotInSet<REFERENCE>(node.getValue().getVarname(),casesForDefault);

            TBPNode<REFERENCE> defaultBranch = node.getDefaultBranch();
            if (defaultBranch == null) {
                // create new null statement 
                LTS<? extends TBPNode<REFERENCE>, REFERENCE> defaultTransition = LTSImpl.createInstance(new TBPImperativeNull<REFERENCE>(), cond);
                entryTransition.connectTransition(defaultTransition);
                tempPreviousTransitions.add(defaultTransition);
            } else {
	            // With default body
	            setNewPreviousTransitionAndGuard(entryTransition, cond);
                setNewPreviousTransitionAndGuard(entryTransition, cond);            	
	            defaultBranch.visit(this);
	            tempPreviousTransitions.addAll(previousTransitions);
            }
        }
        guard = null;
        previousTransitions.clear();
        previousTransitions.addAll(tempPreviousTransitions);
        
        return entryTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedIf(TBPIf<REFERENCE> node) {
    	TBPImperativeNull<REFERENCE> nullNode = new TBPImperativeNull<REFERENCE>();
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> entryTransition = LTSImpl.createInstance(nullNode, guard);
    	connectTransitionToPreviousTransitions(entryTransition);
    	
    	// Then branch
    	ITBPCondition<REFERENCE> cond = node.getCondition();
        boolean isDeterministic = !(cond instanceof TBPNondetCondition);
    	setNewPreviousTransitionAndGuard(entryTransition, isDeterministic ? cond : null);
    	node.getChild(0).visit(this);

        ITBPCondition<REFERENCE> negCond = isDeterministic ? cond.createNegation() : null;

    	// Else branch
    	if (node.hasElse()) {
    		@SuppressWarnings("unchecked")    		
			Vector< LTS<? extends TBPNode<REFERENCE>, REFERENCE>> tempPreviousTransitions = (Vector< LTS<? extends TBPNode<REFERENCE>, REFERENCE>>) previousTransitions.clone();
    		setNewPreviousTransitionAndGuard(entryTransition, negCond);
    		node.getChild(1).visit(this);    		
    		previousTransitions.addAll(tempPreviousTransitions);
    	} else {
            //Vector< LTS<?, REFERENCE>> tempPreviousTransitions = (Vector< LTS<?, REFERENCE>>) previousTransitions.clone();                		            
            LTS<? extends TBPNode<REFERENCE>, REFERENCE> defaultTransition = LTSImpl.createInstance(new TBPImperativeNull<REFERENCE>(), negCond);
            entryTransition.connectTransition(defaultTransition);
            previousTransitions.add(defaultTransition);

        }
    	
    	return entryTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedValue(TBPValue<REFERENCE> node) {
    	// TODO: what is the problem?
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedReturn(TBPReturn<REFERENCE> node) {
        
        for(REFERENCE mutex:lockedMutexes){
            LTS<? extends TBPNode<REFERENCE>, REFERENCE> unlockTransition = getUnlockTransition(mutex);
            connectTransitionToPreviousTransitions(unlockTransition);
            setNewPreviousTransitionAndGuard(unlockTransition, null);
        }


    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> ltsTransition = LTSImpl.createInstance(node, guard);
    	connectTransitionToPreviousTransitions(ltsTransition);
    	setNewPreviousTransitionAndGuard(ltsTransition, null);
    	return ltsTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedWhile(TBPWhile<REFERENCE> node) {
    	TBPImperativeNull<REFERENCE> nullNode = new TBPImperativeNull<REFERENCE>();
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> ltsTransition = LTSImpl.createInstance(nullNode, guard);
    	connectTransitionToPreviousTransitions(ltsTransition);
    	    	
    	// Entry transition
    	//nullNode = new TBPImperativeNull<REFERENCE>();
    	//LTS<? extends TBPNode<REFERENCE>, REFERENCE> entryTransition = LTSImpl.createInstance(nullNode, null);
    	//ltsTransition.addTransition(entryTransition);

        // While branch
        ITBPCondition<REFERENCE> cond = node.getCondition();
        boolean isDeterministic = !(cond instanceof TBPNondetCondition);

        setNewPreviousTransitionAndGuard(ltsTransition, isDeterministic ? cond : null);
        // visit
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> entryTransition = (LTS<? extends TBPNode<REFERENCE>, REFERENCE>) node.getChild().visit(this);
        // go back to entry transition
        connectTransitionToPreviousTransitions(entryTransition);

        // Skip cycle branch
        // endTransition must be there: we must end with guard set to null
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> endTransition = LTSImpl.createInstance(nullNode, isDeterministic ? cond.createNegation() : null);
        connectTransitionToPreviousTransitions(endTransition);
        ltsTransition.connectTransition(endTransition);
        setNewPreviousTransitionAndGuard(endTransition, null);
        

        return ltsTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedAssignment(TBPAssignment<REFERENCE> node) {
    	//System.out.println("Assignment");
    	
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> ltsTransition = LTSImpl.createInstance(node, guard);
    	connectTransitionToPreviousTransitions(ltsTransition);
    	setNewPreviousTransitionAndGuard(ltsTransition, null);
    	return ltsTransition;
    }

	private void connectTransitionToPreviousTransitions(
			LTS<? extends TBPNode<REFERENCE>, REFERENCE> ltsTransition) {
		for ( LTS<?, REFERENCE> transition: previousTransitions ) {
    		transition.connectTransition(ltsTransition);
    	}
	}

    /**
     * Sets new previous transition and guard.
     * @param ltsTransition new previous transition.
     * @param guard the guard for entering next transition from previous
     * 	null if the transition has no associated condition.
     * 	transition.
     */
	private void setNewPreviousTransitionAndGuard(
			LTS<?, REFERENCE> ltsTransition,
			ITBPCondition<REFERENCE> guard) {
		this.guard = guard;
    	previousTransitions = new Vector<LTS<? extends TBPNode<REFERENCE>, REFERENCE>>();
    	previousTransitions.add(ltsTransition);
	}

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedMethodDeclaration(
            TBPMethodDeclaration<REFERENCE> node) {
    	
    	LTS<?, REFERENCE> ltsTransition = LTSImpl.createInstance(node, null);
    	setNewPreviousTransitionAndGuard(ltsTransition, null);
    	
    	node.getChild().visit(this);
    	
    	return ltsTransition;
    }



    private Stack<REFERENCE> lockedMutexes = new Stack<REFERENCE>();

    private LTS<? extends TBPNode<REFERENCE>, REFERENCE> getLockTransition(REFERENCE mutex){
        ITBPCondition<REFERENCE> lockGuard = new TBPEquals<REFERENCE>(mutex, UNLOCKED);
        TBPAssignment<REFERENCE> lockLabel = new TBPAssignment<REFERENCE>(mutex, new TBPValue<REFERENCE>(LOCKED));
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> lockTransition = LTSImpl.createInstance(lockLabel, lockGuard);
    	return lockTransition;
    }

    private LTS<? extends TBPNode<REFERENCE>, REFERENCE> getUnlockTransition(REFERENCE mutex){
        TBPAssignment<REFERENCE> unlockLabel = new TBPAssignment<REFERENCE>(mutex, new TBPValue<REFERENCE>(UNLOCKED));
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> unlockTransition = LTSImpl.createInstance(unlockLabel, guard);
        return unlockTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedSync(TBPSync<REFERENCE> node) {
    	
    	// Entry transition with guard
        TBPImperativeNull<REFERENCE> nullNode = new TBPImperativeNull<REFERENCE>();
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> entryTransition = LTSImpl.createInstance(nullNode, guard);
    	connectTransitionToPreviousTransitions(entryTransition);


        // Before, atomically check and lock the mutex
        REFERENCE mutex =node.getMutex();
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> lockTransition = getLockTransition(mutex);
        entryTransition.connectTransition(lockTransition);
    	setNewPreviousTransitionAndGuard(lockTransition, null);


    	// Visit
        lockedMutexes.push(mutex);
        node.getChild(0).visit(this);
        lockedMutexes.pop();

        // After, unlock the mutex
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> unlockTransition = getUnlockTransition(mutex);
    	connectTransitionToPreviousTransitions(unlockTransition);
    	setNewPreviousTransitionAndGuard(unlockTransition, null);

        return lockTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedWait(TBPWait<REFERENCE> node) {
        
        REFERENCE mutex =node.getMutex();        
        if (!lockedMutexes.contains(mutex)) {
            throw new Error("Attempt to wait out of critical section of mutex "+mutex.getName());
        }
        
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> unlockTransition = getUnlockTransition(mutex);
    	connectTransitionToPreviousTransitions(unlockTransition);
    	
        LTS<? extends TBPNode<REFERENCE>, REFERENCE> lockTransition = getLockTransition(mutex);
        unlockTransition.connectTransition(lockTransition);
    	setNewPreviousTransitionAndGuard(lockTransition, null);
        return unlockTransition;

    }
    	
    
    
    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedImperativeNull(
            TBPImperativeNull<REFERENCE> node) {
    	// TODO: maybe null transitions could be skipped
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> ltsTransition = LTSImpl.createInstance(node, guard);
    	connectTransitionToPreviousTransitions(ltsTransition);
    	setNewPreviousTransitionAndGuard(ltsTransition, null);
    	return ltsTransition;    
    }

    @Override
    public  LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedEmit(
            TBPEmit<REFERENCE> node) {
    	
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> ltsTransition = LTSImpl.createInstance(node, guard);
    	connectTransitionToPreviousTransitions(ltsTransition);
    	setNewPreviousTransitionAndGuard(ltsTransition, null);
    	return ltsTransition;
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedAlternative(
            TBPAlternative<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedParallel(
            TBPParallel<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedParallelOr(
            TBPParallelOr<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedSequence(
            TBPSequence<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedAccept(
            TBPAccept<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedNull(
            TBPProvisionNull<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitLimitedReentrancy(
            TBPLimitedReentrancy<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedRepetition(
            TBPRepetition<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitUnlimitedReentrancy(
            TBPUnlimitedReentrancy<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedProvisionContainerNode(
            TBPProvisionContainerNode<REFERENCE> node) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitParsedThreadContainerNode(
            TBPThreadContainerNode<REFERENCE> node) {
    	LTS<?, REFERENCE> ltsTransition = LTSImpl.createInstance(node, null);
    	setNewPreviousTransitionAndGuard(ltsTransition, null);

        // Visit the body
        node.getChild().visit(this);

        return ltsTransition;
    }

    /**
     * @deprecated Use {@link #transformTBP(TBPSpecification)} instead.
     * <p><b>DO NOT USE THIS METHOD !!!</b></p>
     */
    @Override
    public LTS<? extends TBPNode<REFERENCE>, REFERENCE> visitSpecification(
            TBPSpecification<REFERENCE> node) {
    	
    	// Useless implementation - not recommended to use this method...
    	LTS<? extends TBPNode<REFERENCE>, REFERENCE> result = LTSImpl.createInstance(new TBPImperativeNull<REFERENCE>(), null);
    	
        for (TBPThreadContainerNode<REFERENCE> thread : node.getThreads()) {
            result.connectTransition( thread.visit(this) );
        }

        for (TBPMethodDeclaration<REFERENCE> reaction : node.getReactions()) {
        	result.connectTransition( reaction.visit(this) );
        }

        return result;
    }
    
	private LTSSpecification<REFERENCE> visitSpecificationInternal(
            TBPSpecification<REFERENCE> node) {
    	        
    	LTSSpecification<REFERENCE> ltsSpecitifation = new LTSSpecification<REFERENCE>();

    	
        for (TBPThreadContainerNode<REFERENCE> thread : node.getThreads()) {
            @SuppressWarnings("unchecked")        	
        	LTSThread<REFERENCE> temp = (LTSThread<REFERENCE>) thread.visit(this);
            ltsSpecitifation.addThread( temp );
        }

        for (TBPMethodDeclaration<REFERENCE> reaction : node.getReactions()) {
            @SuppressWarnings("unchecked")        	
        	LTSMethod<REFERENCE> temp = (LTSMethod<REFERENCE>) reaction.visit(this);
        	ltsSpecitifation.addMethod( temp );
        }
        return ltsSpecitifation;
    }
    
}