/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.traversal;

import org.ow2.dsrg.fm.badger.BadgerLog;
import org.ow2.dsrg.fm.badger.simulation.TBPBadActivityException;
import org.ow2.dsrg.fm.badger.simulation.TBPNoActivityException;
import org.ow2.dsrg.fm.badger.simulation.TBPPropertyViolationException;
import org.ow2.dsrg.fm.badger.simulation.state.State;
import org.ow2.dsrg.fm.badger.simulation.transition.BehaviorInterpreter;
import org.ow2.dsrg.fm.badger.simulation.transition.Transition;
import org.ow2.dsrg.fm.tbplib.event.EventCode;
import org.ow2.dsrg.fm.tbplib.event.EventTable;

import java.io.PrintStream;
import java.util.Stack;

/**
 * Testing traversal of the behavior interpreter.
 *
 * FIXME: Use bit caching to conserve memory.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class Traversal {

    private BehaviorInterpreter interpreter;
    private EventTable eventTable;
    private static final boolean silent = true;

    public Traversal(BehaviorInterpreter interpreter, EventTable eventTable) {
        this.interpreter = interpreter;
        this.eventTable = eventTable;
    }

    public void traverse() throws TBPPropertyViolationException{
        State current = interpreter.getInitialState();
        TraversalInfo currentInfo = new TraversalInfo(current);
        Stack<TraversalInfo> dfsStack = new Stack<TraversalInfo>();
        StateCache stateCache = new BitStateCache(10000000);
        long visited = 0;
        long reportStats = 10;

        dfsStack.push(currentInfo);
        stateCache.tryStoreState(current);
        visited++;
        if (!silent) BadgerLog.logger.finest("Initial state #" + current.hashCode()+ " " + current);

        boolean thereWasEnabledTransition = false;

        while (true) {

            if (!currentInfo.nextChoice()) {
                // No more choices, let's backtrack...

                if (!silent) BadgerLog.logger.finest("Backtrack.");

                if (!thereWasEnabledTransition) {
                    if (!interpreter.isFinal(current)){
                        boolean allThreadsInFinalState = current.allThreadsAreInFinalState();
                        TBPNoActivityException ex = new TBPNoActivityException(allThreadsInFinalState);
                        Long eventId = interpreter.getAnEventExpectedByObserver(current);
                        if (eventId!=null){
                            String eventName = eventTable.decodeEventName(eventId);
                            ex.setExpectedEventName(eventName);
                        }
                        ex.setStatesVisited(visited);
                        printStack(dfsStack, ex.getCallStackForWriting());
                        throw ex;
                    }
                }

                thereWasEnabledTransition = true;
                dfsStack.pop();

                if (dfsStack.empty()) {
                    // No more states on the stack...we are done.
                    break;
                }
                currentInfo = dfsStack.peek();
                current = currentInfo.state;
                continue;
            }

            // Check the transition guard
            if (!interpreter.evaluateGuard(current, currentInfo.currentThread,
                    currentInfo.choices[currentInfo.currentBranch].guardId)) {
                // Skip the disabled branch
                if (!silent) BadgerLog.logger.finest("Disabled branch.");
                continue;
            }

            thereWasEnabledTransition = true;

            // Execute the transition
            long event = 0;
            current = current.clone();

            try {
                event = interpreter.executeTransition(current,
                        currentInfo.currentThread,
                        currentInfo.choices[currentInfo.currentBranch].transitionId);

                currentInfo.lastEvent = event;
            } catch (TBPBadActivityException ex) {
                currentInfo.lastEvent = ex.event;
                ex.setEventString(eventTable.decodeEventName(ex.event));
                ex.setCurrentThread(currentInfo.currentThread);
                ex.setStatesVisited(visited);
                printStack(dfsStack, ex.getCallStackForWriting());
                throw ex;


            } catch (TBPPropertyViolationException ex) {
                ex.setStatesVisited(visited);
                printStack(dfsStack, ex.getCallStackForWriting());
                throw ex;
            }

            if (!silent) BadgerLog.logger.finest("Event: " + eventTable.decodeEventName(event));

            if (event == EventCode.IGNORE_BRANCH) {
                // Skip the branch
                if (!silent) BadgerLog.logger.finest("Ignoring branch.");
                continue;
            }

            // Try to add to the state cache and check for collisions
            if (!stateCache.tryStoreState(current)) {
                // Already visited, skip it
                if (!silent) BadgerLog.logger.finest("Cache hit #" + current.hashCode() + ".");
                current = currentInfo.state;
                continue;
            }

            visited++;
            if (visited >= reportStats) {
                BadgerLog.logger.fine("States visited so far: " + visited);
                if (reportStats < 100000) {
                    reportStats *= 10;
                } else {
                    reportStats += 100000;
                }
            }
            if (!silent) BadgerLog.logger.finest("New state #" + current.hashCode()+ " " + current);

            // Push new state onto the dfsStack,
            currentInfo = new TraversalInfo(current);
            dfsStack.push(currentInfo);
            thereWasEnabledTransition = false;
        }
        BadgerLog.logger.info("-----------------------------");
        BadgerLog.logger.info("THE SYSTEM IS SAFE!");
        BadgerLog.logger.fine("Total visited states: " + visited);
    }

    private void printStack(Stack<TraversalInfo> dfsStack, PrintStream out) {
        State initial = interpreter.getInitialState();
        int[] outTableCols = new int[initial.threads.length + 2];
        String[] outTableFormats = new String[initial.threads.length + 2];
        String[][] outTable = new String[dfsStack.size()][initial.threads.length + 2];

        // Detecting column widths
        int id = 0;
        State lastState = null;
        for (TraversalInfo info : dfsStack) {
            outTable[id][0] = Integer.toString(id) + ":";
            outTableCols[0] = Math.max(outTableCols[0], outTable[id][0].length());
            boolean differsFromLast = true;
            if (lastState!=null) {
                differsFromLast = info.state.differsInStateVars(lastState);
            }
            outTable[id][1] = info.state.toString(differsFromLast);
            outTableCols[1] = Math.max(outTableCols[1], outTable[id][1].length());

            for (int t = 0; t < info.state.threads.length; t++) {
                if (t == info.currentThread) {
                    outTable[id][t + 2] = eventTable.decodeEventName(info.lastEvent);
                } else {
                    outTable[id][t + 2] = "|";
                }
                outTableCols[t + 2] = Math.max(outTableCols[t + 2], outTable[id][t + 2].length());
            }
            id++;
            lastState = info.state;
        }
        outTableCols[1] = Math.max(outTableCols[1], ("State").length());
        for (int t = 0; t < initial.threads.length; t++) {
            outTableCols[t + 2] = Math.max(outTableCols[t + 2], ("Thread #" + (t+1)).length());
        }

        outTableFormats[0] = "%" + outTableCols[0] + "s ";
        for (int i = 1; i < outTableFormats.length; i++) {
            outTableFormats[i] = "%-" + (outTableCols[i]+1) + "." +
                    (outTableCols[i]) + "s";
        }

        out.println("ERROR TRACE:");
        out.printf(outTableFormats[0], "");
        out.printf(outTableFormats[1], "State");
        for (int t = 0; t < initial.threads.length; t++) {
            out.printf(outTableFormats[t+2], "Thread #" + (t+1));
        }
        out.println();
        out.printf(outTableFormats[0], "");
        for (int t = 1; t < outTableCols.length; t++) {
            for (int i = 0; i < outTableCols[t]; i++) {
                out.print('-');
            }
            out.print(' ');
        }

        // Print the trace
        for (id = 0; id < outTable.length; id++) {
            out.println();
            for (int t = 0; t < outTable[id].length; t++) {
                out.printf(outTableFormats[t], outTable[id][t]);
            }
        }
        out.println();
    }

    private class TraversalInfo {

        public State state;
        public int currentThread;
        public int currentBranch;
        public Transition[] choices;
        public long lastEvent;

        public TraversalInfo(State state) {
            this.lastEvent = -1;
            this.state = state;
            this.currentThread = -1;
            this.currentBranch = -1;
            this.choices = null;
        }

        public boolean nextChoice() {
            while (choices == null) {
                currentThread++;
                currentBranch = -1;
                // No threads --> no transitions
                if (state.threads == null)
                    return false;

                // No more thread --> no transitions
                if (currentThread == state.threads.length)
                    return false;

                if (state.threads[currentThread] == null)
                    continue;

                choices = interpreter.nextTransitions(state, currentThread);
            }

            currentBranch++;

            if (currentBranch == choices.length) {
                choices = null;
                return nextChoice();
            }

            return true;
        }
    }
}
