/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.util;

import org.ow2.dsrg.fm.badger.reference.TranslationArchitecture;
import org.ow2.dsrg.fm.badger.reference.TranslationConstant;
import org.ow2.dsrg.fm.badger.reference.TranslationEnumerationType;
import org.ow2.dsrg.fm.badger.reference.TranslationReference;
import org.ow2.dsrg.fm.badger.simulation.transition.BehaviorInterpreter;
import org.ow2.dsrg.fm.badger.translation.BehaviorEncodingVisitor;
import org.ow2.dsrg.fm.badger.translation.BehaviorFromLTSEncodingVisitor;
import org.ow2.dsrg.fm.badger.translation.LTSEncodingVisitor;
import org.ow2.dsrg.fm.tbplib.TBPLibLog;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.reference.EnumerationType;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

import javax.tools.JavaCompiler;
import javax.tools.ToolProvider;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;
import java.util.logging.Level;

/**
 * Set of utility static methods for performing task necessary for checking
 * composition correctness.
 * FIXME: Use standard logger.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class CheckingChain {

    private static String WORKING_DIR = "output/";
    private static String TEST_WORKING_DIR = "output_test/";
    private static String GENERATION_DIR = "org/ow2/dsrg/fm/badger/generated/";
    private static String GENERATION_PACKAGE = "org.ow2.dsrg.fm.badger.generated";
    private static String GENERATION_CLASS = "Generated";
    private static int numberGenerated = 0;
    
    
    /**
     * Transforms the given TBP specification into LTS specification. 
     * Typically the input  specification is a result of TBP composition 
     * and binding.
     *
     * @param tbp The TBP specification for which a Java behavior interpreter
     *      is to be created..
     * @param arch Architectural information to be used during the construction.
     *      This information is used to make the output more human-friendly.
     * @return LTS specification.
     */
    public static LTSSpecification<TranslationReference> transformToLTS(
            TBPSpecification<TranslationReference> tbp,
            TranslationArchitecture arch) {
    	
    	
        // Can't think of a better way of getting the mutex type and constants :o(
        TranslationEnumerationType mutexType =
                (TranslationEnumerationType)arch.getTypes().get(
                EnumerationType.MUTEX_TYPE.getName());

        TranslationReference lockedConst = (TranslationConstant)mutexType.getEnums().get(
                EnumerationType.LOCKED.getName());
        TranslationReference unlockedConst = (TranslationConstant)mutexType.getEnums().get(
                EnumerationType.UNLOCKED.getName());

        // Create visitor and transform TBPSpecification to LTSSpecification
        LTSEncodingVisitor<TranslationReference> visitor = new LTSEncodingVisitor<TranslationReference>(lockedConst, unlockedConst);

        return visitor.transformTBP(tbp);
    }
    
    /**
     * Transforms the given LTS specification into Java.
     *
     * @param lts The LTS specification for which a Java behavior interpreter
     *      is to be created.
     * @param provisions A list of provisions to be embedded in the resulting
     *      behavior (they are not encoded into Java, but passed to
     *      the generated behavior during construction).
     * @param arch Architectural information to be used during the construction.
     *      This information is used to make the output more human-friendly.
     * @param table EventTable to be used during generation. Note that,
     *      Some MethodEncoders in this table will be replaced by new generated
     *      encoders (for the sake of traversal speed).
     * @return Loaded and initialized instance of the generated behavior.
     */
    public static BehaviorInterpreter transformFromLTSToJava(LTSSpecification<TranslationReference> lts, List<Observer> provisions,
            TranslationArchitecture arch, EventTable table) {
    	try {
            String className = GENERATION_CLASS + ++numberGenerated;
            String fileName = TEST_WORKING_DIR + GENERATION_DIR + className + ".java";
            Observer[] provs = provisions.toArray(new Observer[provisions.size()]);
            ByteArrayOutputStream bfr = new ByteArrayOutputStream();

            new File(TEST_WORKING_DIR + GENERATION_DIR).mkdirs();

            TBPLibLog.logger.info("Creating file: "+fileName);
            BehaviorFromLTSEncodingVisitor visitor =
                    new BehaviorFromLTSEncodingVisitor(new CodeWriter(fileName), provs, arch);

            visitor.transformLTS(lts, className, GENERATION_PACKAGE);

            JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();

            if (compiler == null) {
                TBPLibLog.logger.severe("No Java compiler found.");
                return null;
            }
            int result = compiler.run(null, bfr, bfr, fileName);
            
            if (result != 0) {
                TBPLibLog.logger.severe("Compilation error:\n" + bfr.toString());
                return null;
            }


            URLClassLoader cl = new URLClassLoader(new URL[]{
                new URL("file:" + TEST_WORKING_DIR)});
            Class c = cl.loadClass(
                    GENERATION_PACKAGE + "." + className);

            return (BehaviorInterpreter)c.getConstructor(
                    Observer[].class, EventTable.class).newInstance(
                    provs, table);

        } catch (NoSuchMethodException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (SecurityException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (IllegalArgumentException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (InvocationTargetException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (InstantiationException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (IllegalAccessException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Illegal access.", ex);
            return null;
        } catch (MalformedURLException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Malformed URL to class.", ex);
            return null;
        } catch (ClassNotFoundException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Class could not be loaded.", ex);
            return null;
        } catch (FileNotFoundException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Temporary file could not be created.", ex);
            return null;
        }
    }

    /**
     * Transforms the given TBP specification into Java. Typically the input
     * specification is a result of TBP composition and binding.
     *
     * @param tbp The TBP specification for which a Java behavior interpreter
     *      is to be created.
     * @param provisions A list of provisions to be embedded in the resulting
     *      behavior (they are not encoded into Java, but passed to
     *      the generated behavior during construction).
     * @param arch Architectural information to be used during the construction.
     *      This information is used to make the output more human-friendly.
     * @param table EventTable to be used during generation. Note that,
     *      Some MethodEncoders in this table will be replaced by new generated
     *      encoders (for the sake of traversal speed).
     * @return Loaded and initialized instance of the generated behavior.
     */
    public static BehaviorInterpreter transformToJava(
            TBPSpecification<TranslationReference> tbp, List<Observer> provisions,
            TranslationArchitecture arch, EventTable table) {
        try {
            String className = GENERATION_CLASS + ++numberGenerated;
            String fileName = WORKING_DIR + GENERATION_DIR + className + ".java";
            Observer[] provs = provisions.toArray(new Observer[provisions.size()]);
            ByteArrayOutputStream bfr = new ByteArrayOutputStream();

            new File(WORKING_DIR + GENERATION_DIR).mkdirs();

            BehaviorEncodingVisitor visitor =
                    new BehaviorEncodingVisitor(new CodeWriter(fileName), provs, arch);

            visitor.transformTBP(tbp, className, GENERATION_PACKAGE);

            JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();

            if (compiler == null) {
                TBPLibLog.logger.severe("No Java compiler found.");
                return null;
            }
            int result = compiler.run(null, bfr, bfr, fileName);
            
            if (result != 0) {
                TBPLibLog.logger.severe("Compilation error:\n" + bfr.toString());
                return null;
            }


            URLClassLoader cl = new URLClassLoader(new URL[]{
                new URL("file:" + WORKING_DIR)});
            Class c = cl.loadClass(
                    GENERATION_PACKAGE + "." + className);

            return (BehaviorInterpreter)c.getConstructor(
                    Observer[].class, EventTable.class).newInstance(
                    provs, table);

        } catch (NoSuchMethodException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (SecurityException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (IllegalArgumentException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (InvocationTargetException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (InstantiationException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Instantiation error.", ex);
            return null;
        } catch (IllegalAccessException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Illegal access.", ex);
            return null;
        } catch (MalformedURLException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Malformed URL to class.", ex);
            return null;
        } catch (ClassNotFoundException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Class could not be loaded.", ex);
            return null;
        } catch (FileNotFoundException ex) {
            TBPLibLog.logger.log(Level.SEVERE, "Temporary file could not be created.", ex);
            return null;
        }
    }
}
