/*
 *  $Id$
 * 
 *   Badger - Threaded Behavior Protocols analysis toolchain
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.badger.translation;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.ow2.dsrg.fm.badger.BadgerLog;
import org.ow2.dsrg.fm.badger.CheckArch;
import org.ow2.dsrg.fm.badger.reference.TranslationArchitecture;
import org.ow2.dsrg.fm.badger.reference.TranslationReference;
import org.ow2.dsrg.fm.badger.simulation.TBPBadActivityException;
import org.ow2.dsrg.fm.badger.simulation.TBPPropertyViolationException;
import org.ow2.dsrg.fm.badger.simulation.transition.BehaviorInterpreter;
import org.ow2.dsrg.fm.badger.traversal.Traversal;
import org.ow2.dsrg.fm.badger.util.CheckingChain;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.util.TransformationChain;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import org.ow2.dsrg.fm.badger.simulation.TBPNoActivityException;

/**
 * Tests encoding on trivial examples.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
@RunWith(Parameterized.class)
public class EncodingArchLTSTest {

    private static String testPrefix = "test/input/";

    @Parameterized.Parameters
    public static Collection<Object[]> testInputs() {
        return Arrays.asList(new Object[][]{
                    {"simple/01_ok.arch", null},
                    {"simple/log.arch", TBPBadActivityException.class},
                    {"simple/plog.arch", null},
                    {"session_manager/sm.arch", null},
                    {"session_manager/sm_ba.arch", TBPBadActivityException.class},
                    {"session_manager/sm_na.arch", TBPNoActivityException.class}           
                });
    }
    ;
    public String file;
    public Class exClass;

    public EncodingArchLTSTest(String file, Class exClass) {
        this.file = file;
        this.exClass = exClass;
    }

    private BehaviorInterpreter createInterpreterFromFile(String filename, EventTable table) throws Exception {

        CheckArch.ArchitectureReader ar = new CheckArch.ArchitectureReader(filename);

        List<TBPSpecification<String>> tbps = ar.getTbps();

        Architecture arch =ar.getArch();

        TranslationArchitecture tArch = new TranslationArchitecture(arch);

        BadgerLog.logger.info("Implicit binding");
        TransformationChain.performImplicitBinding(tArch);

        BadgerLog.logger.info("Resolving symbols");
        List<TBPSpecification<Reference>> ref_tbps =
                TransformationChain.resolveArchitecture(tbps, tArch);
        TransformationChain.fillEventTable(arch, table);

        BadgerLog.logger.info("Transformation of provisions");
        List<Observer> provisions = TransformationChain.transformProvisions(ref_tbps, table);

        BadgerLog.logger.info("Creating composition of components");
        TBPSpecification tbpComp =
                TransformationChain.performComposition(ref_tbps, null);
        TransformationChain.replaceArbitraryValuesInImperativePartsByAlternatives(tbpComp);

        BadgerLog.logger.info("Transformation to LTS represenation");
        LTSSpecification<TranslationReference> ltsSpecification =
                CheckingChain.transformToLTS(tbpComp, tArch);

        BadgerLog.logger.info("Transformation to Java");
        BehaviorInterpreter interpreter =
            CheckingChain.transformFromLTSToJava(ltsSpecification, provisions, tArch, table);
        return interpreter;

    }




    private boolean checkException(Exception ex) throws Exception{
            if (ex.getClass().equals(exClass)) {
                System.out.println("An exception " + exClass.getCanonicalName() + " thrown as expected.");
                return true;
            } else {
                System.out.println("An unexpected exception " + ex.getClass().getCanonicalName() + " thrown.");
                throw ex;
            }

    }

    @Test
    public void testEncodingAndTraversal() throws Exception {
        try {
            EventTable table = new EventTableImpl();

            BehaviorInterpreter interpreter = createInterpreterFromFile(
                    testPrefix + file, table);
            Assert.assertNotNull(interpreter);

            Traversal tr = new Traversal(interpreter, table);

            tr.traverse();
        } catch (TBPPropertyViolationException ex) {
            ex.printError(System.out);
            if (checkException(ex)) return;
        } catch (Exception ex) {
            if (checkException(ex)) return;
        }
        if (exClass != null) {
            Assert.fail("An exception " + exClass.getCanonicalName() + " expected.");
        }
    }
}