grammar TBP;

options {
	language=Java;
	k=5;
	output=AST;
}

tokens { 
    TYPEDEF; 
    VARDEF;
    PROVISION;
    REACTION;
    THREAD;
    ANNOTATION;
    ANN_PROVISION;
    ANN_STATEMENT;
    EVENT;

    BLOCK;

	METHOD_LIST;
	METHOD_LIST_ELEMENT;
    METHOD_CALL;
    METHOD_DECL;
    PARLIST;
    PARLIST_DECL;
    VALUE;
    ASSIGNMENT;
    RETURN;
    SWITCH_ND;
}

@header {
	package org.ow2.dsrg.fm.tbplib.parser;
import org.ow2.dsrg.fm.tbplib.util.ErrorCollector;
}

@lexer::header {
package org.ow2.dsrg.fm.tbplib.parser;
import org.ow2.dsrg.fm.tbplib.util.ErrorCollector;
}


@members {

    private ErrorCollector errorReporter = null;
    public void setErrorReporter(ErrorCollector errorReporter) {
        this.errorReporter = errorReporter;
    }
    public void emitErrorMessage(String msg) {
	if (errorReporter!=null){
	    errorReporter.addError(msg);
	} else {
	    super.emitErrorMessage(msg);	
	}
    }
}


@lexer::members {

    private ErrorCollector errorReporter = null;
    public void setErrorReporter(ErrorCollector errorReporter) {
        this.errorReporter = errorReporter;
    }
    public void emitErrorMessage(String msg) {
	if (errorReporter!=null){
	    errorReporter.addError(msg);
	} else {
	    super.emitErrorMessage(msg);	
	}
    }
 /* public void emitErrorMessage(String s){
 	super.emitErrorMessage(s);
 	throw new RuntimeException(s);
 }*/
}



// tokens
DIGITS  :   '0'..'9'+;
IDF 	:   ('a'..'z'|'A'..'Z')('a'..'z'|'A'..'Z'|'0'..'9'|'_')*;

// ignore whitespace and comments
WS      :   (' '|'\n'|'\r'|'\t')+ {skip();}; 
COMMENT :   '/*'  .* '*/' {skip();}
		|	'#' .* '\n' {skip();}
	;

//start symbol
tbp	: 'component' IDF '{'
		'types' '{' (type (';' type)*)? (';')? '}'
		'vars'  '{' (var (';' var)*)? (';')?  '}'
		'provisions' '{' provision* '}'
		'reactions'  '{' reaction*  '}'
		'threads'    '{' thread* '}'
	 '}' -> 
        ^('component' IDF ^('types' type*) 		
                          ^('vars'  var*)
                          ^('provisions' provision* )
                          ^('reactions' reaction* )
                          ^('threads'  thread*)
         ); 
	
type	:	IDF '=' '{' IDF (',' IDF)* '}' -> ^(TYPEDEF IDF IDF+);

var_decl	:	IDF IDF '=' IDF -> ^(VARDEF IDF IDF IDF);
mutex_decl	:	'Mutex' IDF     -> ^(VARDEF IDF);

var	:	var_decl
	|	mutex_decl
	;

provision   :	IDF?  p_protocol  'for' '{' method_list '}'
            ->  ^(PROVISION p_protocol method_list IDF?);

reaction    :	annotation? method_decl '{' (var_decl ';')* imperative '}'
            ->  ^(REACTION method_decl 'vars' var_decl* ^(BLOCK imperative) annotation?)
            ;

thread	    :	IDF? '{' (var_decl ';')* imperative '}' 
            ->  ^(THREAD 'vars' var_decl* ^(BLOCK imperative) IDF?);

annotation  :	'@' IDF ( '(' (IDF '=' IDF)+ ')' )?
            ->  ^(ANNOTATION IDF IDF*);


// Provisions 
method_list : (method_list_element (',' method_list_element)*)?
	-> ^(METHOD_LIST method_list_element*)
	;
	
method_list_element	:
	'*' -> ^(METHOD_LIST_ELEMENT '*')
	| IDF '.' '*' -> ^(METHOD_LIST_ELEMENT IDF '*')
	| IDF '.' IDF -> ^(METHOD_LIST_ELEMENT IDF IDF)
	;

p_protocol  :  p_alt ;

p_alt  		:  p_seq  ( '+'^  p_seq)* ;
p_seq       :  p_par  ( ';'^  p_par)* ;
p_par       :  p_opar ( '|'^  p_opar)* ;
p_opar	    :  p_rep  ( '||'^ p_rep)*;

p_rep       :  ap_term (ap_op^)*; 

ap_op	    : '*'  
            | '|*' 
            | '|'^ DIGITS
            | '||'^ DIGITS
            ;          

ap_term     : annotation p_term -> ^(ANN_PROVISION annotation p_term)
            | p_term;

p_term      : '{' p_protocol '}' -> p_protocol
            | '(' p_protocol ')' -> p_protocol
            | p_event            
            ;   

p_event     : method_call ( ':' IDF )? -> ^(EVENT method_call IDF?)
            | 'NULL'                       -> ^(EVENT)
            ;

// Imperative 
imperative	: imperative2 (';')? -> ^(imperative2);

imperative2  : ar_stmt (';'^ ar_stmt)*;

ar_stmt     : annotation r_stmt -> ^(ANN_STATEMENT annotation r_stmt)
            | r_stmt;
            
r_stmt      : r_while
            | r_switch
            | r_if
            | r_synchronized
            | r_event
            | r_wait
            ;

r_switch    : 'switch' '(' IDF ')' '{' r_switchbody  
					('default' ':' '{' imperative '}')?  
					'}'    -> ^('switch' ^(VALUE IDF) r_switchbody ^('default' imperative)? )
	        | 'switch' '('  '?'  ')' '{' r_nd_switchbody '}'    -> ^(SWITCH_ND r_nd_switchbody)
    	    ;

r_switchbody: r_case+;
r_case 	    : 'case' IDF ':' '{' imperative '}' -> ^('case' IDF imperative)
			| 'case' IDF ':' ar_stmt -> ^('case' IDF ar_stmt)
			;

r_nd_switchbody: r_nd_case+;
r_nd_case   : 'case' ':' '{' imperative '}'     -> ^('case' imperative)
			| 'case' ':' ar_stmt     -> ^('case' ar_stmt)
			;

r_while     : 'while' '(' cond ')' block
        ->  ^('while' cond block)
    ;

r_if        : 'if' '(' cond ')' block ('else' block)?
        ->  ^('if' cond block block?)
            ;

r_synchronized: 'sync' '(' IDF ')' block
        ->  ^( 'sync' IDF block);

r_wait      : 'wait' '(' IDF ')' -> ^( 'wait' IDF);

r_event     : method_call -> method_call
            | r_assign        
            | r_return        
            | 'NULL' -> ^('NULL')
	    ;
r_assign    : IDF '=' value  -> ^(ASSIGNMENT IDF value);

value       : IDF             -> ^(VALUE IDF)
    	    | method_call -> ^(VALUE method_call)
    	    ;

r_return    : 'return' IDF?   -> ^(RETURN IDF?);

cond 	    : condor ('&&' condor)* 
            ;

condor      : condelem ('||' condelem)* 
            ;


condelem    :IDF '==' IDF  
	    |IDF '!=' IDF  
            | '!' condelem
            | '(' cond ')'
            | '?'
            ;


block       : '{' imperative '}' -> ^(BLOCK imperative);

method_name : IDF '.' IDF -> IDF IDF;

method_decl : method_name '(' parlist_decl ')' ( ':' IDF )?
        ->  ^(METHOD_DECL method_name parlist_decl IDF?);

method_call : method_name '(' parlist ')'
        ->  ^(METHOD_CALL method_name parlist);

parlist_decl: (IDF IDF (',' IDF IDF)* )?
        ->  ^(PARLIST_DECL IDF*);

param : IDF 
      | '?'
      ;

parlist	    : (param  (',' param)* )?
        ->  ^(PARLIST param*);

