/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib;

/**
 * Represents pattern for method name.
 * 
 * Used to represent patterns in for section of provision.
 * 
 * @author David Hauzar
 *
 */
public abstract class TBPMethodNamePattern {
	/**
	 * Returns true if given method from given interface
	 * can be matched by this pattern.
	 * 
	 * @param interfaceName the interface of the method.
	 * @param methodName the name of the method.
	 * @return true if given method from given interface
	 * 	can be matched by this pattern.
	 */
	public abstract boolean matchMethod(String interfaceName, String methodName);
	
	/**
	 * Factory method. Creates pattern that matches methods.
	 * @return the pattern that matches methods.
	 */
	public static TBPMethodNamePattern allMethodsMatcher() {
		return new TBPAllMethodNamePattern();
	}
	
	/**
	 * Factory method. Creates pattern that matches all methods of given
	 * provided interface.
	 * @param ifaceName the name of the interface to that methods that
	 * 	are matched by this pattern belong.
	 * 	should be provided interface
	 * @return the pattern that matches all provided methods of given 
	 * interface.
	 */
	public static TBPMethodNamePattern allMethodsInGivenProvidedIfaceMatcher(String ifaceName) {
		return new TBPAllProvidedMethodsInGivenInterfaceNamePattern(ifaceName);
	}
	
	/**
	 * Factory method. Creates pattern that matches given concrete provided
	 * method.
	 * 
	 * @param ifaceName the name of the interface to that matched method
	 * 	belongs. Should be provided interface.
	 * @param methodName the name of the method that is matched by 
	 * 	returned pattern.
	 * @return the pattern that matches method given by ifaceName 
	 * 	and methodName.
	*/
	public static TBPMethodNamePattern concreteProvidedMethodMatcher(String ifaceName, String methodName) {
		return new TBPConcreteProvidedMethodNamePattern(ifaceName, methodName);
	}
	
	private static class TBPAllMethodNamePattern extends TBPMethodNamePattern {
		@Override
		public boolean matchMethod(String interfaceName, String methodName) {
			return true;
		}
	}
	
	private static class TBPAllProvidedMethodsInGivenInterfaceNamePattern extends TBPMethodNamePattern {
		private final String interfaceName;
		
		public TBPAllProvidedMethodsInGivenInterfaceNamePattern(String interfaceName) {
			// TODO: if the interface is not provided, show warning
			this.interfaceName = interfaceName;
		}
		
		@Override
		public boolean matchMethod(String interfaceName, String methodName) {
			if (this.interfaceName.equals(interfaceName)) return true;
			else return false;
		}
	}
	
	private static class TBPConcreteProvidedMethodNamePattern extends TBPAllProvidedMethodsInGivenInterfaceNamePattern {
		private final String methodName;
		
		public TBPConcreteProvidedMethodNamePattern(String interfaceName, String methodName) {
			super(interfaceName);
			this.methodName = methodName;
			
			// TODO: if the method is not provided method, show warning
			//if (!MethodsInfo.isMethodProvided(interfaceName, methodName)) show warning
		}
		
		@Override
		public boolean matchMethod(String interfaceName, String methodName) {
			if (super.matchMethod(interfaceName, methodName) && this.methodName.equals(methodName)) return true;
			else return false;
		}
	}
}
