/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.architecture;

import org.ow2.dsrg.fm.tbplib.TBPResolvingException;
import org.ow2.dsrg.fm.tbplib.reference.EnumerationType;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.reference.StateVariable;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Representation of a component in an architectural information.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class Component extends Container<Reference> implements Reference {

    protected Architecture architecture;
    protected LinkedHashMap<String, EnumerationType> types = new LinkedHashMap<String, EnumerationType>();
    protected LinkedHashMap<String, StateVariable> stateVariables = new LinkedHashMap<String, StateVariable>();
    protected LinkedHashMap<String, Autonomous> threads = new LinkedHashMap<String, Autonomous>();
    protected LinkedHashMap<String, Interface> providedIfaces = new LinkedHashMap<String, Interface>();
    protected LinkedHashMap<String, Interface> requiredIfaces = new LinkedHashMap<String, Interface>();

    protected Component(String name) {
        super(name);
    }

    public Component(String name, List<EnumerationType> types,
            List<StateVariable> stateVariables) {
        super(name);
        setTypes(types);
        setStateVariables(stateVariables);
    }

    public Component(String name, List<EnumerationType> types) {
        super(name);
        setTypes(types);
    }

    public void setStateVariables(List<StateVariable> stateVariables) {
        switchSymbols(this.stateVariables.values(), stateVariables);
        setSymbolsIn(this.stateVariables, stateVariables);
    }

    private void setTypes(List<EnumerationType> types) {
        for (EnumerationType type : types) {
            // Should not be set at this time
            if (architecture != null) {
                type = architecture.addType(type);
            }
            this.types.put(type.getName(), type);
            addSymbol(type);
        }
    }

    public void addThread(Autonomous thread) {
        String threadName = thread.getName();
        if (threads.containsKey(threadName)) {
            throw new TBPResolvingException("Multiple threads share the same name: " + threadName);
        }
        threads.put(threadName, thread);
    }

    public void addProvidedInterface(Interface iface) {
        assert (iface.getDirection() == Interface.Direction.PROVIDED);
        providedIfaces.put(iface.getName(), iface);
    }

    public void addRequiredInterface(Interface iface) {
        assert (iface.getDirection() == Interface.Direction.REQUIRED);
        requiredIfaces.put(iface.getName(), iface);
    }

    public LinkedHashMap<String, Interface> getProvidedInterfaces() {
        return providedIfaces;
    }

    public LinkedHashMap<String, Interface> getRequiredInterfaces() {
        return requiredIfaces;
    }

    public Interface getProvidedInterface(String name) {
        return providedIfaces.get(name);
    }

    public Interface getRequiredInterface(String name) {
        return requiredIfaces.get(name);
    }


    public LinkedHashMap<String, Autonomous> getThreads() {
        return threads;
    }

    public void addStateVariable(StateVariable stateVariable) {
        stateVariables.put(stateVariable.getName(), stateVariable);
        addSymbol(stateVariable);
    }

    @Override
    public Architecture getParent() {
        return architecture;
    }

    void setArchitecture(Architecture architecture) {
        this.architecture = architecture;
        
        for(Map.Entry<String, EnumerationType> entry : types.entrySet()) {
            // NOTE: Conflicts are checked in addType
            entry.setValue(architecture.addType(entry.getValue()));
        }
    }

    public LinkedHashMap<String, EnumerationType> getTypes() {
        return types;
    }

    public LinkedHashMap<String, StateVariable> getStateVariables() {
        return stateVariables;
    }
}
