/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.event;

/**
 * A set of helper static methods for encoding method and event identifications
 * into <code>long</code>.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class EventCode {

    /** Classification of the event types. */
    public static enum EventType {
        METHOD_CALL,
        METHOD_RETURN,
        SPECIAL
    }

    /** Classification of the event types. */
    public static enum SpecialEvent {
        TAU_ACTION,
        BAD_ACTIVITY,
        NO_ACTIVITY,
        IGNORE_BRANCH
    }
    /** Place holder for the void parameterisation or return value */
    public static final byte VOID_VALUE = -1;
    /** Bit offset of the event type code in the <code>long</code> event
     * representation. */
    private static final int EVENT_TYPE_SHIFT = 62;
    /** Bit offset of the method code in the <code>long</code> event
     * representation. */
    private static final int METHOD_SHIFT = 32;
    /** Event type code for method call. */
    private static final int METHOD_CALL_CODE = 0;
    /** Event type code for return from method. */
    private static final int METHOD_RETURN_CODE = 1;
    /** Event type code for special event (e.g., tau action and bad activity)*/
    private static final int SPECIAL_CODE = 2;
    /** Bit mask of the event type code in the <code>long</code> event
     * representation. */
    private static final long EVENT_TYPE_MASK = 0xc000000000000000l;
    /** Bit mask of the method code in the <code>long</code> event
     * representation. */
    private static final long METHOD_MASK = 0x3fffffff00000000l;
    /** Bit mask of the parameters encoding in the <code>long</code> event
     * representation. */
    private static final long PARAMETERS_MASK = 0x00000000ffffffffl;
    /** Bit mask of the return value encoding in the <code>long</code> event
     * representation. */
    private static final long RETURN_VALUE_MASK = 0x00000000ffffffffl;
    /** Represenattion of the tau event, i.e., an internal event. */
    public static final long TAU_ACTION =
            (((long) SPECIAL_CODE) << EVENT_TYPE_SHIFT) | 0l;
    /** Representation of the bad activity erroneous event. */
    public static final long BAD_ACTIVITY =
            (((long) SPECIAL_CODE) << EVENT_TYPE_SHIFT) | 1l;
    /** Representation of the no activity erroneous event. */
    public static final long NO_ACTIVITY =
            (((long) SPECIAL_CODE) << EVENT_TYPE_SHIFT) | 2l;
    /** This event represents a branch that should not be further traversed. */
    public static final long IGNORE_BRANCH =
            (((long) SPECIAL_CODE) << EVENT_TYPE_SHIFT) | 3l;

    /**
     * Is the given event a method related event - an observable event?
     *
     * @param event Integer encoding of anevent to be decoded.
     * @return <code>true</code> if the given event corresponds to a method call
     *      or method return, i.e., it might be observed by a provision.
     */
    public static boolean isObservable(long event) {
        int eventType = (int)((event & EVENT_TYPE_MASK) >>> EVENT_TYPE_SHIFT);

        return eventType == METHOD_CALL_CODE ||
                eventType == METHOD_RETURN_CODE;
    }

    /**
     * Extracts code of the method corresponding to the given event code.
     * Validity of the event is not checked.
     *
     * @param event An event code to be analyzed.
     * @return The <code>int</code> code of a method.
     */
    public static int decodeMethod(long event) {
        return (int) ((event & METHOD_MASK) >>> METHOD_SHIFT);
    }

    /**
     * Extracts type of the given event.
     *
     * @param event An event code to be analyzed.
     * @return Type of the given event.
     */
    public static EventType decodeEventType(long event) {
        switch ((int) ((event & EVENT_TYPE_MASK) >>> EVENT_TYPE_SHIFT)) {
            case METHOD_CALL_CODE:
                return EventType.METHOD_CALL;
            case METHOD_RETURN_CODE:
                return EventType.METHOD_RETURN;
            case SPECIAL_CODE:
                return EventType.SPECIAL;
        }
        throw new IllegalArgumentException("Unsupported event type");
    }

    /**
     * Extracts type of the given special event.
     *
     * @param event A special event
     * @return Type of the given special event.
     */
    public static SpecialEvent decodeSpecialEvent(long event) {
        switch ((int) (event & PARAMETERS_MASK)) {
            case 0:
                return SpecialEvent.TAU_ACTION;
            case 1:
                return SpecialEvent.BAD_ACTIVITY;
            case 2:
                return SpecialEvent.NO_ACTIVITY;
            case 3:
                return SpecialEvent.IGNORE_BRANCH;
        }
        throw new IllegalArgumentException("Unsupported special event type");
    }

    /**
     * Extracts code of the parameters corresponding to the given event code.
     * Validity of the event is not checked.
     *
     * @param event An event code to be analyzed.
     * @return The <code>int</code> encoding of the parameters.
     */
    public static int decodeParameters(long event) {
        return (int) (event & PARAMETERS_MASK);
    }

    /**
     * Extracts code of the return value corresponding to the given event code.
     * Validity of the event is not checked.
     *
     * @param event An event code to be analyzed.
     * @return The <code>int</code> encoding of the return value.
     */
    public static int decodeReturnValue(long event) {
        return (int) (event & RETURN_VALUE_MASK);
    }

    /**
     * Composes event code representing a method call of the given method with
     * the given encoding of parameters (possibly a <code>VOID_VALUE</code>).
     *
     * @param method Code of the method
     * @param parameters Encoding of the method parameters or <code>VOID_VALUE</code>.
     * @return The <code>long</code> encoding of the method call event.
     */
    public static long encodeMethodCall(int method, int parameters) {
        return (((long) METHOD_CALL_CODE) << EVENT_TYPE_SHIFT) |
                ((((long) method) << METHOD_SHIFT) & METHOD_MASK)|
                (((long) parameters) & PARAMETERS_MASK);
    }

    /**
     * Composes event code representing a return from the given method with
     * the given encoding of return value (possibly a <code>VOID_VALUE</code>).
     *
     * @param method Code of the method
     * @param retValue Encoding of the return value or <code>VOID_VALUE</code>.
     * @return The <code>long</code> encoding of the return-from-a-method event.
     */
    public static long encodeMethodReturn(int method, int retValue) {
        return (((long) METHOD_RETURN_CODE) << EVENT_TYPE_SHIFT) |
                ((((long) method) << METHOD_SHIFT) & METHOD_MASK) |
                (((long) retValue) & RETURN_VALUE_MASK);
    }
}
