/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.event;

/**
 * Interface of an EventTable as suitable for the TBP checker. Events are
 * encoded as <code>int</code>s. Mapping to string is provided.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public interface EventTable {

    /**
     * Encoding of events corresponding to a particular method.
     * Implementations of this interfaces will be generated to allow efficient
     * computation of parameter encoding (bypassing the array allocation for
     * varargs).
     */
    static public interface MethodEncoder {

        /**
         * Assigns method code to the method encoder.
         *
         * @param methodCode Encoding of the represented method
         */
        public void setMethodCode(int methodCode);

        /**
         * Getter.
         * @return The assigned (fully qualified) method name.
         */
        public String getMethodName();

        /**
         * Computes integer encoding of the method call with the given parameter
         * values.
         *
         * @param parameters Values of parameters to be used for computation.
         * @return Integer identification of the corresponding event including
         *      the offset.
         */
        public long encodeMethodCall(byte... parameters);

        /**
         * Computes integer encoding of the method return with the given
         * return value.
         *
         * @param retVal Return value to be encoded in the event or
         *      <code>VOID_VALUE</code> for methods without return value.
         * @return Integer identification of the corresponding event including
         *      the offset.
         */
        public long encodeMethodReturn(byte retVal);

        /**
         * Reconstructs the textual close-to-human-readable representation.
         *
         * @param event Integer identification of the event (offset included)
         * @return Textual representation of the given event
         */
        public String decodeEventName(long event);

        /**
         * Computes the number of method call events assciated with the
         * method represented by this encoder.
         *
         * @return Number of events that correspond to a call of this method
         *      with different parameterisations. If the method has
         *      no parameters, EventCode.VOID_VALUE is returned.
         */
        public int getNumberOfCallEvents();

        /**
         * Computes the number of method return events assciated with the
         * method represented by this encoder.
         *
         * @return Number of events that correspond to a return from this method
         *      with different return values. If the method has return type
         *      EventCode.VOID_VALUE is returned.
         */
        public int getNumberOfReturnEvents();
    }

    /**
     * Assigns the method encoder to the given method name. The encoder is
     * notified about the assigned method code via the <code>setMethodCode</code>
     * method.
     *
     * @param encoder <code>MethodEncoder</code> to be used when encoding events
     *      corresponding to the given method.
     * @return The newly assigned integer method identification.
     */
    public int addMethodEncoder(MethodEncoder encoder);

    /**
     * Does the EventTable already contain the given method?
     * 
     * @param methodName Textual method identification containing
     *      fully qualified name (including interface and possible renaming)
     * @return <code>true</code> if the method is already present.
     */
    public boolean containsMethod(String methodName);

    /**
     * Computes integer encoding for the given method
     *
     * @param methodName Textual method identification containing
     *      fully qualified name (including interface and possible renaming)
     * @return Integer identification of the method
     */
    public int encodeMethod(String methodName);

    /**
     * Computes integer encoding of the method call with the given parameter
     * values.
     *
     * @param method Integer identification of the method.
     * @param parameters Values of parameters to be used for computation.
     * @return Integer identification of the corresponding event including
     *      the offset.
     * @return Integer identification of the corresponding event including
     *      the offset.
     */
    public long encodeMethodCall(int method, byte... parameters);

    /**
     * Computes integer encoding of the method return with the given
     * return value.
     *
     * @param method Integer identification of the method.
     * @param retVal Return value to be encoded in the event or
     *      <code>VOID_VALUE</code> for methods without return value.
     * @return Integer identification of the corresponding event including
     *      the offset.
     */
    public long encodeMethodReturn(int method, byte retVal);

    /**
     * Reconstructs the textual close-to-human-readable representation.
     *
     * @param method Integer identification of the method.
     * @return Textual representation of the given method
     */
    public String decodeMethodName(int method);

    /**
     * Reconstructs the textual close-to-human-readable representation.
     *
     * @param event Integer identification of the event (offset included)
     * @return Textual representation of the given event.
     */
    public String decodeEventName(long event);

    /**
     * Retrieves the size of the event table, i.e., the number of indexed
     * methods.
     *
     * @return The number of indexed methods. Any method index between 0 and
     *      this number - 1 is a valid index of a method.
     */
    public int getNumberOfMethods();

    /**
     * Finds the method encoder responsible for encoding the given method.
     * 
     * @param method Index of a method for which the encoder is to be found.
     * @return MethodEncoder corresponding to the given method.
     */
    public MethodEncoder getEncoder(int method);
}
