/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2008-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.event;

import java.util.*;

/**
 * Straightforward implementation of the EventTable interface.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class EventTableImpl implements EventTable {

    protected Map<String, Integer> methodCodes = new HashMap<String, Integer>();
    protected List<MethodEncoder> methodInfo = new ArrayList<MethodEncoder>();

    @Override
    public int addMethodEncoder(MethodEncoder encoder) {
        String methodName = encoder.getMethodName();

        if (methodCodes.containsKey(methodName)) {
            throw new IllegalArgumentException("Duplicate method encoder: "+methodName);
        }

        int code = methodInfo.size();
        methodInfo.add(encoder);
        methodCodes.put(methodName, code);
        encoder.setMethodCode(code);

        return code;
    }

    @Override
    public boolean containsMethod(String methodName) {
        return methodCodes.containsKey(methodName);
    }

    @Override
    public int encodeMethod(String methodName) {
        Integer code = methodCodes.get(methodName);
        if (code == null) {
            throw new IllegalArgumentException("Method not stored in the EventTable.");
        }

        return code;
    }

    @Override
    public long encodeMethodCall(int method, byte... parameters) {
        return methodInfo.get(method).encodeMethodCall(parameters);
    }

    @Override
    public long encodeMethodReturn(int method, byte retVal) {
        return methodInfo.get(method).encodeMethodReturn(retVal);
    }

    @Override
    public String decodeMethodName(int method) {
        return methodInfo.get(method).getMethodName();
    }

    @Override
    public String decodeEventName(long event) {
        EventCode.EventType type = EventCode.decodeEventType(event);

        if (type.equals(EventCode.EventType.SPECIAL)) {
            switch (EventCode.decodeSpecialEvent(event)) {
                case TAU_ACTION:
                    return "<Tau>";
                case BAD_ACTIVITY:
                    return "<BadActivity>";
                case NO_ACTIVITY:
                    return "<NoActivity>";
                case IGNORE_BRANCH:
                    return "<IgnoreBranch>";
                default:
                    throw new IllegalArgumentException("Unsupported special event type.");
            }
        }

        int method = EventCode.decodeMethod(event);
        return methodInfo.get(method).decodeEventName(event);
    }

    @Override
    public int getNumberOfMethods() {
        return methodInfo.size();
    }

    @Override
    public MethodEncoder getEncoder(int method) {
        return methodInfo.get(method);
    }

    static public class MethodEncoderImpl implements MethodEncoder {

        protected int methodCode;
        protected String methodName;
        protected int[] parameterDomains;
        protected int retValDomain;

        public MethodEncoderImpl(String methodName,
                int retValDomain, int... parameterDomains) {
            this.methodName = methodName;
            this.retValDomain = retValDomain;
            this.parameterDomains = parameterDomains;
        }

        @Override
        public void setMethodCode(int methodCode) {
            this.methodCode = methodCode;
        }

        @Override
        public String getMethodName() {
            return methodName;
        }

        @Override
        public long encodeMethodCall(byte... parameters) {
            assert (parameters == null || parameters.length == parameterDomains.length);

            if (parameters == null || parameters.length == 0) {
                return EventCode.encodeMethodCall(methodCode, EventCode.VOID_VALUE);
            }

            int paramCode = 0;

            for (int i = parameters.length - 1; i >= 0; i--) {
                assert (parameters[i] < parameterDomains[i] && 0 <= parameters[i]);

                paramCode += parameters[i];
                paramCode *= (i == 0 ? 1 : parameterDomains[i - 1]);
            }

            return EventCode.encodeMethodCall(methodCode, paramCode);
        }

        @Override
        public long encodeMethodReturn(byte retVal) {
            assert ((retVal < retValDomain && 0 <= retVal) ||
                    (retValDomain == 0 && retVal == EventCode.VOID_VALUE));

            return EventCode.encodeMethodReturn(methodCode, retVal);
        }

        @Override
        public String decodeEventName(long event) {
            StringBuilder bfr = new StringBuilder(methodName);
            EventCode.EventType eventType = EventCode.decodeEventType(event);
            int params = EventCode.decodeParameters(event);

            // Handle different event types
            switch (eventType) {
                case METHOD_CALL:
                    boolean first = true;
                    bfr.append("^(");
                    for (int dom : parameterDomains) {
                        if (first) {
                            first = false;
                        } else {
                            bfr.append(',');
                        }
                        bfr.append(params % dom);
                        params = params / dom;
                    }
                    bfr.append(")");
                    break;
                case METHOD_RETURN:
                    if (params == EventCode.VOID_VALUE) {
                        bfr.append("$");
                    } else {
                        bfr.append("$:");
                        bfr.append(params);
                    }
                    break;
                default:
                    throw new IllegalArgumentException("Unexpected event type.");
            }

            return bfr.toString();
        }

        @Override
        public int getNumberOfCallEvents() {
            if(parameterDomains == null || parameterDomains.length == 0)
                return EventCode.VOID_VALUE;

            int events = 1;

            for (int dom : parameterDomains) {
                events *= dom;
            }
            return events;
        }

        @Override
        public int getNumberOfReturnEvents() {
            if (retValDomain == 0)
                return EventCode.VOID_VALUE;

            return retValDomain;
        }
    }
}
