/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa;

import java.io.OutputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPValue;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPAnd;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPInSet;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNondetCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNot;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotInSet;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPOr;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPSetCondition;
import org.ow2.dsrg.fm.tbplib.util.DOTPrintStream;
import org.ow2.dsrg.fm.tbplib.util.DOTPrintStreamWithSubgraphs;
import org.ow2.dsrg.fm.tbplib.util.DOTPrintStreamWithSubgraphs.NodeShape;

/**
 * Helper class with static methods to print out graphs of provisions, reactions 
 * and threads.
 * 
 * @author caitt3am
 *
 */
public class AutomatonToDot {
	
	
	/** Creates labels for nodes and edges.
	 */
	public static interface LabelFactory <REFERENCE> {
		/** Creates label for edge.
		 * 
		 * @param sourceState source state forming source node of the edge
		 * @param event event
		 * @param targetState target state forming target node of the edge
		 * @return label
		 */
		public String createEdgeLabel(Automaton a, int sourceState, long event, int targetState);
		/** Creates label for node.
		 * 
		 * @param state state forming the node.
		 * @return label
		 */
		public String createNodeLabel(Automaton a, int state);
		/** Creates label for {@link LTS} transition.
		 * 
		 * @param transition {@link LTS} transition.
		 * @return label
		 */
		public String createTransitionLabel(LTS<? extends TBPNode<REFERENCE>,REFERENCE> transition);
		/** Creates label for {@link LTSA} transition.
		 * 
		 * @param transition {@link LTSA} transition.
		 * @return label
		 */
		public String createTransitionLabel(LTSA<?,REFERENCE> transition);		
		/** Converts REFERENCE to string.
		 * 
		 * <p>Used for printing names of entities (threads, components etc.).</p>
		 * 
		 * @param reference REFERENCE
		 * @return label
		 */		
		public String referenceToString(REFERENCE reference);
	}
	public static abstract class AbstractLabelFactory<REFERENCE> implements LabelFactory<REFERENCE> {
		@Override
		public String createEdgeLabel(Automaton a, int sourceState, long event, int targetState) {
			return Long.valueOf(event).toString();
		}
		@Override
		public String createNodeLabel(Automaton a, int state) {
			return "";
		}
		@Override
		public String createTransitionLabel(
				LTS<? extends TBPNode<REFERENCE>, REFERENCE> transition) {
			String label = null;
			String guard = null;
			String assign = null;
			StringBuilder result = new StringBuilder();
			
			if (transition != null) {
				if (transition.getLabel() != null) {
					label = ltsLabelToString(transition.getLabel());
					assign = ltsAssignmentToString(transition.getLabel());
				}
				if (transition.getGuard() != null) {
					guard = guardToString(transition.getGuard());
				}
				if (label != null) {
					result.append("[").append(label).append("]");
				}
				if (guard != null) {
					if (result.length() > 0) {
						result.append("\\n");
					}
					result.append("?(").append(guard).append(")");
				}
				if (assign != null) {
					if (result.length() > 0) {
						result.append("\\n");
					}
					result.append(assign);
				}
			}
			return result.toString();			
		}
		@Override
		public String createTransitionLabel( LTSA<?, REFERENCE> transition) {
			String label = null;
			String guard = null;
			String assign = null;
			StringBuilder result = new StringBuilder();
			
			if (transition != null) {
				if (transition.getLabels() != null && !transition.getLabels().isEmpty()) {
					for(Label<REFERENCE> l : transition.getLabels()) {
						if (result.length() > 0) {
							result.append(", ");
						}
						result.append(labelToString(l));
					}
					label = result.toString();
					result.delete(0, result.length());
				}
				if (transition.getGuard() != null) {
					guard = guardToString(transition.getGuard());
				}
				if (transition.getAssignments() != null && !transition.getAssignments().isEmpty()) {
					for(TBPAssignment<REFERENCE> a: transition.getAssignments()) {
						if (result.length() > 0) {
							result.append(", ");
						}
						result.append(assignmentToString(a));
					}
					assign = result.toString();
					result.delete(0, result.length());
				}
				
				result.append("#").append(transition != null ? transition.hashCode() : -1).append("\\n");
				if (label != null) {
					result.append("[").append(label).append("]");
				}
				if (guard != null) {
					if (result.length() > 0) {
						result.append("\\n");
					}
					result.append("?(").append(guard).append(")");
				}
				if (assign != null) {
					if (result.length() > 0) {
						result.append("\\n");
					}
					result.append(assign);
				}
			}
			return result.toString();			
		}
		@Override
		public String referenceToString(REFERENCE reference) {
			return reference != null ? reference.toString() : "";
		}
		
		/** Creates <i>guard</i> part of LTS(A) transition label. 
		 * @param guard
		 * @return Condition or <code>null</code>.
		 */
		protected abstract String guardToString(ITBPCondition<REFERENCE> guard);
		/** Creates <i>assignment</i> part of LTSA transition label. 
		 * @param assignment
		 * @return String representation of assignment or <code>null</code>.
		 */
		protected abstract String assignmentToString(TBPAssignment<REFERENCE> assignment);
		/** Creates <i>label</i> part of LTSA transition label. 
		 * @param label
		 * @return String representation of label or <code>null</code>.
		 */
		protected abstract String labelToString(Label<REFERENCE> label);		
		/** Creates <i>label (i.e. method call)</i> part of LTS transition label. 
		 * @param label
		 * @return String representation of LTS Label or <code>null</code>.
		 */
		protected abstract String ltsLabelToString(TBPNode<REFERENCE> label);
		/** Creates <i>assignment</i> part of LTS transition label. 
		 * @param assignment
		 * @return String representation of assignment or <code>null</code>.
		 */
		protected abstract String ltsAssignmentToString(TBPNode<REFERENCE> assignment);
		
	}
	
	/** Provides base for label factory for printing LTSA.
	 * 
	 */
	public static abstract class LTSALabelFactoryBase<REFERENCE> extends AbstractLabelFactory<REFERENCE> {
		@Override
		protected String guardToString(ITBPCondition<REFERENCE> guard) {
			StringBuilder sb = new StringBuilder();
			if (guard instanceof TBPAnd<?>) {
				sb.append("(").append(guardToString(((TBPAnd<REFERENCE>)guard).getLeft())).append(") && (").append(guardToString(((TBPAnd<REFERENCE>)guard).getRight())).append(")");
			} else if (guard instanceof TBPOr<?>) {
				sb.append("(").append(guardToString(((TBPOr<REFERENCE>)guard).getLeft())).append(") || (").append(guardToString(((TBPOr<REFERENCE>)guard).getRight())).append(")");
			} else if (guard instanceof TBPNot<?>) {
				sb.append("!(").append(guardToString(((TBPNot<REFERENCE>)guard).getOperand())).append(")");
			} else if (guard instanceof TBPEquals<?>) {
				sb.append(referenceToString(((TBPEquals<REFERENCE>)guard).getLeft())).append(" == ").append(referenceToString(((TBPEquals<REFERENCE>)guard).getRight()));
			} else if (guard instanceof TBPNotEquals<?>) {
				sb.append(referenceToString(((TBPNotEquals<REFERENCE>)guard).getLeft())).append(" != ").append(referenceToString(((TBPNotEquals<REFERENCE>)guard).getRight()));				
			} else if (guard instanceof TBPSetCondition<?>) {
				sb.append(referenceToString(((TBPSetCondition<REFERENCE>)guard).getLeft()));
				if (guard instanceof TBPInSet<?>) {
					sb.append(" in ");
				} else if (guard instanceof TBPNotInSet<?>) {
					sb.append(" not-in ");
				}
				boolean first = true;
				for(REFERENCE reference : ((TBPSetCondition<REFERENCE>)guard).getSet()) {
					if (first) {
						first = false;
					} else {
						sb.append(", ");
					}
					sb.append(referenceToString(reference));
				}
			} else if (guard instanceof TBPNondetCondition<?>) {
				sb.append("*");
			} else {
				throw new UnsupportedOperationException("Unkown gurad.");
			}
			return sb.toString();
		}
		@Override
		protected String assignmentToString(TBPAssignment<REFERENCE> assignment) {
			StringBuilder sb = new StringBuilder();
			if (assignment != null) {
				sb.append(referenceToString(assignment.getIdf())).append(" := ");
				if (assignment.getChild() instanceof TBPValue<?>) {
					TBPValue<REFERENCE> v = (TBPValue<REFERENCE>)assignment.getChild();
					if (v.isMethodCall()) {
						sb.append("?").append(v.getMethodCall().getFullname());
					} else {
						sb.append(referenceToString(v.getVarname()));
					}
				} else {
					sb.append("<error: unable to describe value>");
				}
			}
			return sb.toString();
		}
		@Override
		protected String labelToString(Label<REFERENCE> label) {
			StringBuilder sb = new StringBuilder();
			if (label != null) {
				sb.append(label.isMethodCall() ? "^" : "$");
				sb.append(referenceToString(label.getInterface()))
					.append(".")
					.append(referenceToString(label.getMethod()))
					.append("(");
				if (label.getParameters() != null && !label.getParameters().isEmpty()) {
					StringBuilder tmpSb = new StringBuilder();
					for (REFERENCE p : label.getParameters()) {
						if (tmpSb.length() > 0) {
							tmpSb.append(", ");
						}
						tmpSb.append(referenceToString(p));
					}
					sb.append(tmpSb);
				}
				sb.append(")");
				if (label.getReturnValue() != null) {
					sb.append(": ").append(referenceToString(label.getReturnValue()));
				}
			}
			return sb.toString();
		}		
	}
	
	/** Creates default labels.
	 * 
	 * <ul>
	 * 	<li>Empty label for nodes</li>
	 *  <li>Human readable event names for edges</li>
	 *  <li>Poor support for LTS printing</li>
	 *  <li>Good support for LTSA printing</li>
	 * </ul>
	 */
	public static class DefaultLabelFactory<REFERENCE> extends LTSALabelFactoryBase<REFERENCE> {
		private EventTable table;
		public DefaultLabelFactory(EventTable table) {
			this.table = table;
		}
		@Override
		public String createEdgeLabel(Automaton a, int sourceState, long event, int targetState) {
			return table.decodeEventName(event);
		}
		@Override
		public String createNodeLabel(Automaton a, int state) {
			return "";
		}
		@Override
		protected String ltsLabelToString(TBPNode<REFERENCE> label) {
			return label.toString();
		}
		@Override
		protected String ltsAssignmentToString(TBPNode<REFERENCE> assignment) {
			return null;
		}
	}
	
	//
	//
	// DFA
	//
	//
	
	public static void printDFAsToDot(Map<String, DFA> dfas, EventTable et, OutputStream stream) {
		printDFAsToDot(dfas, new DefaultLabelFactory<Object>(et), stream);
	}
	
	public static void printDFAsToDot(Map<String, DFA> dfas, LabelFactory<?> lf, OutputStream stream) {
		DOTPrintStreamWithSubgraphs out = new DOTPrintStreamWithSubgraphs(stream);
		for(String name : dfas.keySet()) {
			printDFAtoDOTasSubgraph(dfas.get(name), name, lf, out);
		}
		out.DOTclose();
	}
		
    public static void printDFAtoDot(DFA dfa, EventTable et, OutputStream stream) {
    	printDFAtoDot(dfa, new DefaultLabelFactory<Object>(et), stream);
    }
	
    public static void printDFAtoDot(DFA dfa, LabelFactory<?> lf, OutputStream stream) {
        DOTPrintStream out = new DOTPrintStream(stream);
        printDFAtoDOTPrintStream(dfa, lf, out);
        out.DOTclose();
    }
    
    //
    // private methods for DFA
    //
    
    private static void printDFAtoDOTasSubgraph(DFA dfa, String name, LabelFactory<?> lf, DOTPrintStreamWithSubgraphs out)
    {
		out.openSubGraph(name);
		out.nextComponent();
		printDFAtoDOTPrintStream(dfa, lf, out);
		out.closeSubGraph();    	
    }
    
    private static void printDFAtoDOTPrintStream(DFA dfa, LabelFactory<?> lf, DOTPrintStream out) {
        for (int state = 0; state < dfa.getNumberOfState(); state++) {
            printDFANode(out, dfa, state, lf);
        }    	
    }

    private static void printDFANode(DOTPrintStream out, DFA a, int state,
            LabelFactory<?> lf) {

        if (state == a.initialState()) {
            out.peripheries = 2;
        } else {
            out.peripheries = 1;
        }

        out.filled = a.isFinal(state);

        out.node(state, lf.createNodeLabel(a, state));

        for (Map.Entry<Long, Integer> entry : a.getEdges(state).entrySet()) {
            out.edge(state, entry.getValue(), lf.createEdgeLabel(a, state, entry.getKey(), entry.getValue()));
        }

        out.peripheries = 1;
        out.filled = false;
    }
    
    
    //
    //
    // NFA
    //
    //

    public static void printNFAtoDot(NFA nfa, EventTable et, OutputStream stream) {
    	printNFAtoDot(nfa, new DefaultLabelFactory<Object>(et), stream);
    }


    public static void printNFAtoDot(NFA nfa, LabelFactory<?> lf, OutputStream stream) {
        DOTPrintStream out = new DOTPrintStream(stream);

        for (int state = 0; state < nfa.getNumberOfState(); state++) {
            printNFANode(out, nfa, state, lf);
        }

        out.DOTclose();
    }
    
    //
    // Private methods for NFA
    //
    
    private static void printNFANode(DOTPrintStream out, NFA a, int state,
            LabelFactory<?> lf) {

        if (state == a.initialState()) {
            out.peripheries = 2;
        } else {
            out.peripheries = 1;
        }

        out.filled = a.isFinal(state);

        out.node(state, lf.createNodeLabel(a, state));

        for (Map.Entry<Long, List<Integer>> entry : a.getEdges(state).entrySet()) {
            for (int target : entry.getValue()) {
                out.edge(state, target, lf.createEdgeLabel(a, state, entry.getKey(), target));
            }
        }

        out.peripheries = 1;
        out.filled = false;
    }
    
    //
    //
    // LTS
    //
    //
    
    /** Prints threads to clustered DOT graph.
     * 
     * <p>Each thread resides in separate cluster labeled with thread name.</p>
     * 
     * @param <REFERENCE>
     * @param specification {@link LTSASpecification}
     * @param lf {@link LabelFactory}
     * @param stream {@link OutputStream}
     */
    public static <REFERENCE> void printLTSSpecificationToDot(LTSSpecification<REFERENCE> specification, LabelFactory<REFERENCE> lf, OutputStream stream) {
    	DOTPrintStreamWithSubgraphs out = new DOTPrintStreamWithSubgraphs(stream);
    	printLTSSpecificationToDOTPrintStream(specification, lf, out);
    	out.DOTclose();
    }
    
    /** Prints {@link LTSThread} to single DOT graph.
     * 
     * @param <REFERENCE>
     * @param lts {@link LTS}
     * @param lf {@link LabelFactory}
     * @param stream {@link OutputStream}
     */
    public static <REFERENCE> void printLTSToDot(LTS<? extends TBPNode<REFERENCE>, REFERENCE> lts, LabelFactory<REFERENCE> lf, OutputStream stream) {
    	DOTPrintStreamWithSubgraphs out = new DOTPrintStreamWithSubgraphs(stream);
    	printLTStoDOTPrintStream(lts, lf, out);
    	out.DOTclose();
    }
    
    //
    // Auxiliary classes for LTS and LTSA
    //
    
    private static class LTSNode {
    	private static int lastId = 0;
    	private int id;
    	private Collection<?> transitions;
    	
    	private static synchronized int getNextId() {
    		return ++lastId;
    	}    	
    	
    	public LTSNode(Collection<?> transitions) {
    		this.id = getNextId();
    		this.transitions = transitions;
    		if (transitions == null) {
    			throw new NullPointerException();
    		}
    	}
    	public int getId() {
    		return id;
    	}
    	public int getSize() {
    		return transitions.size();
    	}
    	@Override
    	public boolean equals(Object o) {
    		boolean result = false;
    		if (o instanceof LTSNode) {
    			LTSNode that = (LTSNode)o;
    			if (that.transitions.size() == this.transitions.size()) {
    				result = true;
    				for(Object t : this.transitions) {
    					if (!that.transitions.contains(t)) {
    						result = false;
    					}
    					break;
    				}
    			}
    		}
    		return result;
    	}
    }
    
    public static class LTSNodeSet {
    	private Map<Integer, Collection<LTSNode>> base = new HashMap<Integer, Collection<LTSNode>>();
    	public LTSNode get(LTSNode node) {
    		LTSNode result = null;
    		Collection<LTSNode> nodes = base.get(node.getSize());
    		if (nodes != null) {
    			for(LTSNode n : nodes) {
    				if (node.equals(n)) {
    					result = n;
    					break;
    				}
    			}
    		}
    		return result;
    	}
    	public void insert(LTSNode node) {
    		if (get(node) == null) {
    			Collection<LTSNode> nodes = base.get(node.getSize());
    			if (nodes == null) {
    				nodes = new LinkedList<LTSNode>();
    				base.put(node.getSize(), nodes);
    			}
				nodes.add(node);
    		}
    	}
    }
    
    //
    // Private methods for LTS
    //
    
    private static <REFERENCE> void printLTSSpecificationToDOTPrintStream(LTSSpecification<REFERENCE> specification, LabelFactory<REFERENCE> lf, DOTPrintStreamWithSubgraphs out) {
    	int nonameCounter = 0;
    	
    	// threads
    	out.openSubGraph("Threads");
    	for(LTSThread<REFERENCE> t : specification.getThreads()) {
    		String name = t != null && t.getLabel() != null && t.getLabel().getName() != null ? lf.referenceToString(t.getLabel().getName()) : null;
    		if (name == null || name.isEmpty()) {
    			name = "(unnamed-thread-" + (nonameCounter++) + ")";
    		}
    		out.openSubGraph(name);
    		printLTStoDOTPrintStream(t, lf, out);
    		out.closeSubGraph();
    	}
    	out.closeSubGraph();
    	
    	// methods
    	nonameCounter = 0;
    	out.openSubGraph("Reactions");
    	for(LTSMethod<REFERENCE> m : specification.getMethods()) {
    		String name = m != null && m.getLabel() != null && m.getLabel().getFullname() != null ? m.getLabel().getFullname() : null;
    		if (name == null || name.isEmpty()) {
    			name = "(unnamed-reaction-" + (nonameCounter++) + ")";    			
    		}
    		out.openSubGraph(name);
    		printLTStoDOTPrintStream(m, lf, out);
    		out.closeSubGraph();
    	}
    	out.closeSubGraph();
    }
    
    
    private static <REFERENCE> void printLTStoDOTPrintStream(LTS<? extends TBPNode<REFERENCE>, REFERENCE> lts, LabelFactory<REFERENCE> lf, DOTPrintStreamWithSubgraphs out) {
    	LTSNodeSet visited = new LTSNodeSet();
    	NodeShape oldShape = out.getCurrentNodeShape();
    	out.changeNodeShape(NodeShape.POINT);
    	
    	LinkedList<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> transitions = new LinkedList<LTS<? extends TBPNode<REFERENCE>, REFERENCE>>();
    	transitions.add(lts);
        LTSNode startNode = new LTSNode(transitions);
        visited.insert(startNode);
        printLTSNode(startNode.getId(), out);
        printLTSTransitionRecursively(startNode, lts, visited, lf, out);
    	out.changeNodeShape(oldShape);
        
    }
    
    private static <REFERENCE> void printLTSTransitionRecursively(
    		LTSNode node,
    		LTS<? extends TBPNode<REFERENCE>, REFERENCE> lts,
    		LTSNodeSet visited,
    		LabelFactory<REFERENCE> lf,
    		DOTPrintStream out) {
    	
    	LTSNode endNode = new LTSNode(lts.getTransitions());
    	LTSNode oldNode = visited.get(endNode);
    	if (oldNode != null) {
    		endNode = oldNode;
    	}
    	printLTSTransition(node.getId(), lts, lf, endNode.getId(), out);
    	
    	if (oldNode == null) {	// not visited yet
    		printLTSNode(endNode.getId(), out);
    		visited.insert(endNode);
        	for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> l : lts.getTransitions()) {
        		printLTSTransitionRecursively(endNode, l, visited, lf, out);
        	}    		
    	} // else stop recursion
    }
    
    private static void printLTSNode(int key, DOTPrintStream out) {
    	out.node(key, "");
    }
    
    private static <REFERENCE> void printLTSTransition(int from, LTS<? extends TBPNode<REFERENCE>, REFERENCE> lts, LabelFactory<REFERENCE> lf, int to, DOTPrintStream out) {
    	out.edge(from, to, lf.createTransitionLabel(lts));
    }
    

    
    //
    //
    // LTSA
    //
    //
    
    /** Prints {@link LTSASpecification} to clustered DOT graph.
     * 
     * @param <REFERENCE>
     * @param specification {@link LTSASpecification}
     * @param lf {@link LabelFactory}
     * @param stream {@link OutputStream}
     */
    public static <REFERENCE> void printLTSASpecificationToDot(LTSASpecification<?, REFERENCE> specification, LabelFactory<REFERENCE> lf, OutputStream stream) {
    	DOTPrintStreamWithSubgraphs out = new DOTPrintStreamWithSubgraphs(stream);
    	printLTSASpecificationToDOTPrintStream(specification, lf, out);
    	out.DOTclose();
    }
    
    /** Prints {@link LTSA} to single DOT graph.
     * 
     * @param <REFERENCE>
     * @param ltsa {@link LTSA}
     * @param lf {@link LabelFactory}
     * @param stream {@link OutputStream}
     */
    public static <REFERENCE> void printLTSAToDot(LTSA<?, REFERENCE> ltsa, LabelFactory<REFERENCE> lf, OutputStream stream) {
    	DOTPrintStreamWithSubgraphs out = new DOTPrintStreamWithSubgraphs(stream);
    	printLTSAtoDOTPrintStream(ltsa, lf, out);
    	out.DOTclose();
    }
    
    //
    // Private methods for LTSA
    //
    
    private static <REFERENCE> void printLTSASpecificationToDOTPrintStream(LTSASpecification<?, REFERENCE> specification, LabelFactory<REFERENCE> lf, DOTPrintStreamWithSubgraphs out) {
    	int nonameCounter = 0;
    	
    	// threads
    	out.openSubGraph("Threads");
    	for(LTSAThread<?, REFERENCE> t : specification.getThreads()) {
    		String name = t != null && t.getName() != null ? lf.referenceToString(t.getName()) : null;
    		if (name == null || name.isEmpty()) {
    			name = "(unnamed-thread-" + (nonameCounter++) +")";
    		}
    		out.openSubGraph(name + " (*" + (t.getReplicationFactor() == LTSAThread.INFINITY ? "" : t.getReplicationFactor()) + ")");
    		printLTSAtoDOTPrintStream(t, lf, out);
    		out.closeSubGraph();
    	}
    	out.closeSubGraph();
    	
    	// methods
    	out.openSubGraph("Reactions");
    	for(LTSAMethod<?, REFERENCE> m : specification.getMethods()) {
    		String name = m != null && m.getName() != null ? lf.referenceToString(m.getName()) : null;
    		if (name == null || name.isEmpty()) {
    			name = "(unnamed-reaction-" + (nonameCounter++) +")";    			
    		}
    		out.openSubGraph(name);
    		printLTSAtoDOTPrintStream(m, lf, out);
    		out.closeSubGraph();
    	}
    	out.closeSubGraph();
    	
    }
        
    private static <REFERENCE> void printLTSAtoDOTPrintStream(LTSA<?, REFERENCE> ltsa, LabelFactory<REFERENCE> lf, DOTPrintStreamWithSubgraphs out) {
    	LTSNodeSet visited = new LTSNodeSet();
    	NodeShape oldShape = out.getCurrentNodeShape();
    	out.changeNodeShape(NodeShape.POINT);
    	
    	LinkedList<LTSA<?, REFERENCE>> transitions = new LinkedList<LTSA<?, REFERENCE>>();
    	transitions.add(ltsa);
        LTSNode startNode = new LTSNode(transitions);
        visited.insert(startNode);
        printLTSNode(startNode.getId(), out);
        printLTSATransitionRecursively(startNode, ltsa, visited, lf, out);
    	out.changeNodeShape(oldShape);
        
    }
    
    private static <REFERENCE> void printLTSATransitionRecursively(
    		LTSNode node,
    		LTSA<?, REFERENCE> ltsa,
    		LTSNodeSet visited,
    		LabelFactory<REFERENCE> lf,
    		DOTPrintStream out) {
    	
    	LTSNode endNode = new LTSNode(ltsa.getTransitions());
    	LTSNode oldNode = visited.get(endNode);
    	if (oldNode != null) {
    		endNode = oldNode;
    	}
    	printLTSATransition(node.getId(), ltsa, lf, endNode.getId(), out);
    	
    	if (oldNode == null) {	// not visited yet
    		printLTSNode(endNode.getId(), out);
    		visited.insert(endNode);
        	for(LTSA<?, REFERENCE> l : ltsa.getTransitions()) {
        		printLTSATransitionRecursively(endNode, l, visited, lf, out);
        	}    		
    	} // else stop recursion
    }
        
    private static <REFERENCE> void printLTSATransition(int from, LTSA<?, REFERENCE> ltsa, LabelFactory<REFERENCE> lf, int to, DOTPrintStream out) {
    	out.edge(from, to, lf.createTransitionLabel(ltsa));
    }
}