/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.*;
import org.ow2.dsrg.fm.tbplib.event.*;

/**
 * A compacted representation of a DFA. The class is aware of the eventIdx
 * encodings and uses the information (separation into a method and 
 * the parameterisation) to compact the transition tables.
 * 
 * This representation is immutable and optimised for size and state traversal
 * speed.
 *
 * TODO: More sensible encoding might be method calls and method returns
 * to be separated (these are typically not accepted at the same state).
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class CompactDFA implements Observer {

    private boolean[] methodFilter;
    private boolean[] finalStates;
    private int initialState;
    private int[] methodOffsets;
    private int[] methodReturnEventsOffset;
    private int[] methodCallEvents;
    private int[] methodReturnEvents;
    /**
     * For (every state) X (every method) X (every call / return item) -> index
     * of the next state or NO_SUCH_TRANSITION.
     */
    private int[][][] transitionTable;

    private CompactDFA() {
    }

    @Override
    public int nextState(int stateId, long event) {
        // This method should be called only when the given eventIdx is observed
        // by this DFA.
        int method = methodOffsets[EventCode.decodeMethod(event)];
        int[] transitions = transitionTable[stateId][method];

        if (transitions == null) {
            return NO_SUCH_TRANSITION;
        }

        EventCode.EventType type = EventCode.decodeEventType(event);

        if (type == EventCode.EventType.METHOD_CALL) {
            int params = EventCode.decodeParameters(event);
            return transitions[params == EventCode.VOID_VALUE ? 0 : params];

        } else if (type == EventCode.EventType.METHOD_RETURN) {
            int returns = EventCode.decodeReturnValue(event);
            return transitions[(returns == EventCode.VOID_VALUE ? 0 : returns) +
                    methodReturnEventsOffset[method]];

        } else {
            throw new IllegalArgumentException(
                    "Only method calls and method returns might be observed.");
        }
    }

    @Override
    public int initialState() {
        return initialState;
    }

    @Override
    public boolean isFinal(int stateId) {
        return finalStates[stateId];
    }

    @Override
    public boolean observesEvent(long event) {
        if (EventCode.isObservable(event)) {
            return methodFilter[EventCode.decodeMethod(event)];
        }
        return false;
    }

    @Override
    public boolean observesMethod(int method) {
        return methodFilter[method];
    }

    @Override
    public int getNumberOfState() {
        return transitionTable.length;
    }

    @Override
    public Map<Long, Integer> getEdges(int state) {
        // NOTE: This implementation is very inefficient and should be used only
        // for debuging and printing purposes.
        
        Map<Long, Integer> edges = new HashMap<Long, Integer>();

        for (int method = 0; method < methodFilter.length; method++) {
            if (!methodFilter[method])
                continue;

            int mIdx = methodOffsets[method];
            int[] nextStates = transitionTable[state][mIdx];
            int callEvents = methodCallEvents[mIdx];
            int returnEvents = methodReturnEvents[mIdx];
            int returnEventsOffset = methodReturnEventsOffset[mIdx];

            if (nextStates == null)
                continue;

            int events = returnEventsOffset +
                    (returnEvents == EventCode.VOID_VALUE ? 1 : returnEvents);

            for (int eventIdx = 0; eventIdx < events; eventIdx++) {
                int target = nextStates[eventIdx];

                if (target == NO_SUCH_TRANSITION)
                    continue;
                
                // Construct the corresponding event code
                long eventCode;
                if (callEvents == EventCode.VOID_VALUE && eventIdx == 0) {
                    eventCode = EventCode.encodeMethodCall(method, EventCode.VOID_VALUE);
                } else if (eventIdx < returnEventsOffset) {
                    eventCode = EventCode.encodeMethodCall(method, eventIdx);
                } else if (returnEvents == EventCode.VOID_VALUE) {
                    eventCode = EventCode.encodeMethodReturn(method, EventCode.VOID_VALUE);
                } else {
                    eventCode = EventCode.encodeMethodReturn(method, eventIdx - returnEventsOffset);
                }

                // Add the corresponding edge
                edges.put(eventCode, target);
            }
        }

        return edges;
    }

    /**
     * Constructs a compact representation of the given DFA. To obtain the most
     * compact representation, the input DFA should be minimal and observe only
     * methods from the given set of methods (filter). In particular, such
     * DFA can be created by calling DynamicDFA.makeMinimal().
     *
     * @param sourceDFA DFA to be compacted.
     * @param eventTable EventTable to be used during compaction for construction
     *      of a compressed transitionTable.
     * @param filter Methods to be observed by the resulting Observer.
     * @return Compacted representation of the sourceDFA observing methods from
     *      filter.
     */
    static public CompactDFA makeCompact(DFA sourceDFA, EventTable eventTable,
            Set<Integer> filter) {

        CompactDFA dfa = new CompactDFA();
        int states = sourceDFA.getNumberOfState();
        int allMethods = eventTable.getNumberOfMethods();
        int acceptedMethods = filter.size();

        dfa.methodFilter = new boolean[allMethods];
        dfa.methodOffsets = new int[allMethods];
        dfa.methodCallEvents = new int[acceptedMethods];
        dfa.methodReturnEvents = new int[acceptedMethods];
        dfa.methodReturnEventsOffset = new int[acceptedMethods];

        // Create method filter
        int mIdx = 0;
        for (int method : filter) {
            dfa.methodFilter[method] = true;
            dfa.methodOffsets[method] = mIdx;

            dfa.methodCallEvents[mIdx] =
                    eventTable.getEncoder(method).getNumberOfCallEvents();
            dfa.methodReturnEvents[mIdx] =
                    eventTable.getEncoder(method).getNumberOfReturnEvents();
            dfa.methodReturnEventsOffset[mIdx] =
                    dfa.methodCallEvents[mIdx] == EventCode.VOID_VALUE ? 1 : dfa.methodCallEvents[mIdx];

            mIdx++;
        }

        // Create transition table
        dfa.transitionTable = new int[states][][];
        dfa.finalStates = new boolean[states];
        for (int state = 0; state < states; state++) {
            // Mark final states
            if (sourceDFA.isFinal(state)) {
                dfa.finalStates[state] = true;
            }

            int[][] transitions = new int[acceptedMethods][];
            dfa.transitionTable[state] = transitions;

            for (int method : filter) {
                mIdx = dfa.methodOffsets[method];
                int callEvents = dfa.methodCallEvents[mIdx];
                int returnEvents = dfa.methodReturnEvents[mIdx];
                int returnEventsOffset = dfa.methodReturnEventsOffset[mIdx];
                int[] nextStates = transitions[mIdx];

                int events = returnEventsOffset +
                        (returnEvents == EventCode.VOID_VALUE ? 1 : returnEvents);

                for (int eventIdx = 0; eventIdx < events; eventIdx++) {
                    long eventCode;

                    // Construct the corresponding event code
                    if (callEvents == EventCode.VOID_VALUE && eventIdx == 0) {
                        eventCode = EventCode.encodeMethodCall(method, EventCode.VOID_VALUE);
                    } else if (eventIdx < returnEventsOffset) {
                        eventCode = EventCode.encodeMethodCall(method, eventIdx);
                    } else if (returnEvents == EventCode.VOID_VALUE) {
                        eventCode = EventCode.encodeMethodReturn(method, EventCode.VOID_VALUE);
                    } else {
                        eventCode = EventCode.encodeMethodReturn(method, eventIdx - returnEventsOffset);
                    }

                    // Add the corresponding edge
                    int target = sourceDFA.nextState(state, eventCode);

                    if (target != DFA.NO_SUCH_TRANSITION) {
                        if (nextStates == null) {
                            // If no eventIdx was previously present, we create the
                            // array and set the prefix transitions to
                            // NO_SUCH_TRANSITION
                            nextStates = new int[events];
                            transitions[mIdx] = nextStates;
                            if (eventIdx > 0) {
                                Arrays.fill(nextStates, 0, eventIdx, DFA.NO_SUCH_TRANSITION);
                            }
                        }
                        nextStates[eventIdx] = target;
                    } else {
                        if (nextStates != null) {
                            nextStates[eventIdx] = target;
                        }
                    }
                }
            }
        }

        // Set initial state
        dfa.initialState = sourceDFA.initialState();

        return dfa;
    }
}
