/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.*;

/**
 * Representation of DFA, which can be incrementally constructed and modified.
 * All method may result in runtime exceptions when inconsistent
 * parameters (e.g., an identifier of an unexistent state) are passed to them.
 *
 * @author caitt3am
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class DynamicDFA implements DFA {

    private List<DFAState> states;
    private DFAState initialState;
    private Set<DFAState> finalStates;

    /**
     * Constructs an empty automaton. The initialState has to be set before
     * making the automaton accesible.
     */
    public DynamicDFA() {
        this.states = new ArrayList<DFAState>();
        this.initialState = null;
        this.finalStates = new HashSet<DFAState>();
    }

    /**
     * Constructs an empty automaton. The initialState has to be set before
     * making the automaton accesible.
     */
    private DynamicDFA(int states, int finalStates) {
        this.states = new ArrayList<DFAState>(states);
        this.initialState = null;
        this.finalStates = new HashSet<DFAState>(finalStates);
    }

    @Override
    public int nextState(int stateId, long event) {
        Integer target = states.get(stateId).edges.get(event);

        if (target == null)
            return NO_SUCH_TRANSITION;

        return target;
    }

    @Override
    public int initialState() {
        return initialState.getId();
    }

    @Override
    public boolean isFinal(int stateId) {
        return finalStates.contains(states.get(stateId));
    }

    @Override
    public int getNumberOfState() {
        return states.size();
    }

    @Override
    public Map<Long, Integer> getEdges(int state) {
        return states.get(state).edges;
    }

    static protected class DFAState extends FAState<Integer> {

        public DFAState(int id) {
            super(id);
        }

        /**
         * Add edge into list of outgoing edges.
         *
         * @param event Event code labelling the new edge.
         * @param state Target state of the new edge
         */
        public void addEdge(long event, int state) {
            assert (!edges.containsKey(event));

            edges.put(event, state);
        }

        /**
         * Add edge into list of outgoing edges.
         *
         * @param newEdges Map of edges to be all added into this state.
         */
        public void addAllEdges(Map<Long, Integer> newEdges) {
            for (Map.Entry<Long, Integer> entry : newEdges.entrySet()) {
                addEdge(entry.getKey(), entry.getValue());
            }
        }

        public Map<Long, Integer> getEdges() {
            return edges;
        }

        @Override
        public String toString() {
            StringBuilder bld = new StringBuilder("E(");
            boolean first = true;

            for (Map.Entry<Long, Integer> entry : edges.entrySet()) {
                if (first) {
                    first = false;
                } else {
                    bld.append(',');
                }

                bld.append(entry.getKey());
                bld.append("-> ");
                bld.append(entry.getValue());
            }
            bld.append(')');
            return bld.toString();
        }

        private void offsetState(int offset) {
            offsetId(offset);
            for (Map.Entry<Long, Integer> entry : edges.entrySet()) {
                entry.setValue(entry.getValue() + offset);
            }
        }
    }

    /**
     * Removes all states from this automaton. An initial state should be
     * set before using the automaton any further.
     */
    public void clear() {
        initialState = null;
        states.clear();
        finalStates.clear();
    }

    /**
     * Creates new states and inserts it into the automaton.
     *
     * @return Index of the newly allocated state.
     */
    public int allocateState() {
        DFAState state = new DFAState(states.size());
        states.add(state);
        return state.getId();
    }

    public void addEdge(int src, long event, int dst) {
        DFAState state = states.get(src);

        state.addEdge(event, dst);
    }

    public boolean hasInitial() {
        return initialState != null;
    }

    public void markFinal(int state) {
        finalStates.add(states.get(state));
    }

    public void unmarkFinal(int state) {
        finalStates.remove(states.get(state));
    }

    public void setInitial(int state) {
        initialState = states.get(state);
    }

    /**
     * The automaton should not contain unreachable states. In particular, this
     * holds for the result of determinisation of DynamicNFA.
     *
     * @return The corresponding minimal DynamicDFA
     */
    public DynamicDFA makeMinimal() {
        DynamicDFA dfa = new DynamicDFA();
        Map<DFAState, Integer> stateToClass = new HashMap<DFAState, Integer>();
        Map<Integer, Set<DFAState>> classToState = new HashMap<Integer, Set<DFAState>>();

        // Do the hard minimalisation work by creating equivalence classes
        createPartitions(states, finalStates, classToState, stateToClass);

        // Fill in the new automaton
        Set<Integer> classes = classToState.keySet();

        Map<Integer, Integer> classesToRepresentants =
                new HashMap<Integer, Integer>();

        // Alocate new states
        for (Integer i : classes) {
            classesToRepresentants.put(i, dfa.allocateState());
        }

        // Add edges
        for (int c : classes) {
            final int newSource = classesToRepresentants.get(c);
            DFAState oldSource = classToState.get(c).iterator().next(); // Pick one, it should not matter which

            for (Map.Entry<Long, Integer> entry : oldSource.getEdges().entrySet()) {
                final int classTarget = stateToClass.get(states.get(entry.getValue()));
                final Integer newTarget = classesToRepresentants.get(classTarget);
                if (newTarget != null) {
                    dfa.addEdge(newSource, entry.getKey(), newTarget);
                }
            }
        }

        // Set initial state
        final int classInit = stateToClass.get(initialState);
        final int newInit = classesToRepresentants.get(classInit);

        dfa.setInitial(newInit);

        // Mark final states
        for (DFAState f : finalStates) {
            Integer classFinal = stateToClass.get(f);
            Integer newFinal = classesToRepresentants.get(classFinal);

            dfa.markFinal(newFinal);
        }
        return dfa;
    }

    /**
     * Returns mapping from state to mapping from symbol to target
     * class (determined from stateToClass mapping)
     */
    private static Map<DFAState, Map<Long, Integer>> getStatesNeigborhood(
            List<DFAState> states, Map<DFAState, Integer> stateToClass,
            DFAState deadEndState) {

        Map<DFAState, Map<Long, Integer>> env = new HashMap<DFAState, Map<Long, Integer>>();
        int deadEndClass = stateToClass.get(deadEndState);

        for (DFAState s : states) {
            Map<Long, Integer> myEnv = new HashMap<Long, Integer>();
            for (Map.Entry<Long, Integer> entry : s.getEdges().entrySet()) {
                final Integer targetClass = stateToClass.get(states.get(entry.getValue()));

                // Ignore transitions comming to the dead end class
                if (targetClass != deadEndClass) {
                    myEnv.put(entry.getKey(), targetClass);
                }
            }
            env.put(s, myEnv);
        }

        return env;
    }

    private static void createPartitions(
            List<DFAState> allStates, Collection<DFAState> finalStates,
            Map<Integer, Set<DFAState>> classToState, Map<DFAState, Integer> stateToClass) {

        // Just to be sure
        classToState.clear();
        stateToClass.clear();

        List<Integer> classes = new LinkedList<Integer>();
        DFAState deadEndState = new DFAState(NO_SUCH_TRANSITION);

        // Add an artificial state representing a dead end. Its class (e.g.,
        // all states from which no final state is ever reachable) will be
        // removed later.
        allStates.add(deadEndState);

        // Initialize two equivalence classes: 1 - final states, 0 - others
        for (DFAState s : allStates) {
            stateToClass.put(s, 0);
        }
        for (DFAState s : finalStates) {
            stateToClass.put(s, 1);
        }
        classToState.put(0, new HashSet<DFAState>());
        classToState.put(1, new HashSet<DFAState>());

        for (DFAState s : allStates) {
            classToState.get(stateToClass.get(s)).add(s);
        }

        classes.add(0);
        classes.add(1);
        int lastClass = 2;

        // While the set of classes is being updated
        boolean hasChanged = true;
        while (hasChanged) {

            hasChanged = false;
            Map<DFAState, Map<Long, Integer>> env = 
                    getStatesNeigborhood(allStates, stateToClass, deadEndState);

            for (ListIterator<Integer> it = classes.listIterator(); it.hasNext();) {
                int c = it.next();
                Set<DFAState> classStates = classToState.get(c);

                // Gather unique transition patterns (subclasses) in this class
                Set<Map<Long, Integer>> transitionPattern = new HashSet<Map<Long, Integer>>();
                for (DFAState s : classStates) {
                    transitionPattern.add(env.get(s));
                }
                int numSubClasses = transitionPattern.size();

                // The equivalence class is consistent, no class split necessary
                if (numSubClasses == 1) {
                    continue;
                }

                // Split class c into numSubClasses new classes
                // update classes, classToState, stateToClass
                it.remove();
                classToState.remove(c);

                // Preparing for new classes
                for (int i = 0; i < numSubClasses; i++) {
                    int newClass = lastClass + i;
                    it.add(newClass);
                    classToState.put(newClass, new HashSet<DFAState>());
                }

                List<Map<Long, Integer>> transList =
                        new ArrayList<Map<Long, Integer>>(transitionPattern);

                for (DFAState s : classStates) {
                    int clsIdx = transList.indexOf(env.get(s));
                    assert clsIdx != -1;
                    int newClass = lastClass + clsIdx;
                    stateToClass.put(s, newClass);
                    classToState.get(newClass).add(s);
                }
                lastClass += numSubClasses;
                hasChanged = true;
            }
        }

        // Remove the artificial state representing a dead end along with its
        // equivalence class
        allStates.remove(allStates.size()-1);
        classToState.remove(stateToClass.get(deadEndState));
    }
}
