/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.*;

/**
 * Representation of NFA, which can be incrementally constructed and modified.
 * When composing different automatons, their alphabets are expected
 * to be equal. All method may result in runtime exceptions when inconsistent
 * parameters (e.g., an identifier of an unexistent state) are passed to them.
 *
 * @author caitt3am
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class DynamicNFA implements NFA {

    private List<NFAState> states;
    private NFAState initialState;
    private Set<NFAState> finalStates;

    /**
     * Constructs an empty automaton. The initialState has to be set before
     * making the automaton accesible.
     */
    public DynamicNFA() {
        this.states = new ArrayList<NFAState>();
        this.initialState = null;
        this.finalStates = new HashSet<NFAState>();
    }

    /**
     * Constructs an empty automaton. The initialState has to be set before
     * making the automaton accesible.
     */
    private DynamicNFA(int states, int finalStates) {
        this.states = new ArrayList<NFAState>(states);
        this.initialState = null;
        this.finalStates = new HashSet<NFAState>(finalStates);
    }

    @Override
    public List<Integer> nextStates(int stateId, long event) {
        return states.get(stateId).getEdges().get(event);
    }

    @Override
    public int initialState() {
        return initialState.getId();
    }

    @Override
    public boolean isFinal(int stateId) {
        return finalStates.contains(states.get(stateId));
    }

    @Override
    public int getNumberOfState() {
        return states.size();
    }

    @Override
    public Map<Long, List<Integer>> getEdges(int state) {
        return states.get(state).edges;
    }

    /**
     * Removes all states from this automaton. An initial state should be
     * set before using the automaton any further.
     */
    public void clear() {
        initialState = null;
        states.clear();
        finalStates.clear();
    }

    static protected class NFAState extends FAState<List<Integer>> {

        public NFAState(int id) {
            super(id);
        }

        /**
         * Retrieves the list of neighbours for the given event. If necessary,
         * new list is created and inserted into the edges map.
         *
         * @param event Event code of the neighborhood in question.
         */
        private List<Integer> assureNeighbourList(long event) {
            List<Integer> l = edges.get(event);

            if (l == null) {
                l = new ArrayList<Integer>();
                edges.put(event, l);
            }

            return l;
        }

        /**
         * Add edge into list of outgoing edges.
         *
         * @param event Event code labelling the new edge.
         * @param state Target state of the new edge
         */
        public void addEdge(long event, int state) {
            assureNeighbourList(event).add(state);
        }

        /**
         * Add edge into list of outgoing edges.
         *
         * @param event Event code labelling the new edges.
         * @param states Target states of the new edges
         */
        public void addEdges(long event, List<Integer> states) {
            List<Integer> l = assureNeighbourList(event);

            l.addAll(states);
        }

        /**
         * Add edge into list of outgoing edges.
         *
         * @param newEdges Map of edges to be all added into this state.
         */
        public void addAllEdges(Map<Long, List<Integer>> newEdges) {
            for (Map.Entry<Long, List<Integer>> entry : newEdges.entrySet()) {
                addEdges(entry.getKey(), entry.getValue());
            }
        }

        public Map<Long, List<Integer>> getEdges() {
            return edges;
        }

        @Override
        public String toString() {
            StringBuilder bld = new StringBuilder("E(");
            boolean first = true;

            for (Map.Entry<Long, List<Integer>> entry : edges.entrySet()) {
                if (first) {
                    first = false;
                } else {
                    bld.append(',');
                }

                boolean f = true;

                bld.append(entry.getKey());
                bld.append("->{");
                for (int state : entry.getValue()) {
                    if (f) {
                        f = false;
                    } else {
                        bld.append(',');
                    }

                    bld.append(state);
                }
                bld.append('}');
            }
            bld.append(')');
            return bld.toString();
        }

        private void offsetState(int offset) {
            offsetId(offset);
            for (List<Integer> l : edges.values()) {
                for (int i = 0; i < l.size(); i++) {
                    l.set(i, l.get(i) + offset);
                }
            }
        }
    }

    /**
     * Imports the given set of states. This operation alters the given states
     * by offseting their edge target indexes and id.
     */
    private void importStates(List<NFAState> newStates) {
        int offset = states.size();
        for (NFAState state : newStates) {
            state.offsetState(offset);
            states.add(state);
        }
    }

    /**
     * Creates a new state and adds it to this automaton.
     */
    private NFAState createState() {
        NFAState state = new NFAState(states.size());
        states.add(state);
        return state;
    }

    private void set(DynamicNFA a) {
        this.initialState = a.initialState;
        this.finalStates = a.finalStates;
        this.states = a.states;
    }

    /**
     * Consumes the given automaton!
     */
    public DynamicNFA addAlternative(DynamicNFA aut) {
        assert Collections.disjoint(states, aut.states);

        importStates(aut.states);

        initialState.addAllEdges(aut.initialState.getEdges());
        finalStates.addAll(aut.finalStates);

        if (aut.finalStates.contains(aut.initialState)) {
            finalStates.add(initialState);
        }

        aut.clear();
        return this;
    }

    private static DynamicNFA createParallel(DynamicNFA a, DynamicNFA b) {        
        final int sz1 = a.states.size();
        final int sz2 = b.states.size();
        System.out.println("Parallel: size1:"+sz1+" size2:"+sz2);
        DynamicNFA result = new DynamicNFA();
        NFAState temp[][] = new NFAState[sz1][sz2];

        // Generate states of the cartesian product
        for (int i = 0; i < sz1; i++) {
            for (int j = 0; j < sz2; j++) {
                temp[i][j] = result.createState();
            }
        }

        // Generate edges of the cartesian product
        for (int i = 0; i < sz1; i++) {
            for (int j = 0; j < sz2; j++) {
                NFAState s = temp[i][j];

                // First coordinate
                NFAState c1 = a.states.get(i);
                for (Map.Entry<Long, List<Integer>> entry : c1.getEdges().entrySet()) {
                    for (int idx : entry.getValue()) {
                        s.addEdge(entry.getKey(), temp[idx][j].getId());
                    }
                }

                // Second coordinate
                NFAState c2 = b.states.get(j);
                for (Map.Entry<Long, List<Integer>> entry : c2.getEdges().entrySet()) {
                    for (int idx : entry.getValue()) {
                        s.addEdge(entry.getKey(), temp[i][idx].getId());
                    }
                }
            }
        }

        // Set initial state
        result.initialState = temp[a.initialState.getId()][b.initialState.getId()];

        // Generate final states
        for (NFAState s1 : a.finalStates) {
            for (NFAState s2 : b.finalStates) {
                result.finalStates.add(temp[s1.getId()][s2.getId()]);
            }
        }

        return result;
    }

    public DynamicNFA addParallel(DynamicNFA aut) {
        set(createParallel(this, aut));
        return this;
    }

    /**
     * Consumes the given automaton!
     */
    public DynamicNFA addParallelOr(DynamicNFA aut) {
        DynamicNFA par = createParallel(this, aut);
        return addAlternative(aut).addAlternative(par);
    }

    public DynamicNFA addLimitedReentrancy(int k) {
        // "A |k" is equivalent to A*|A*|A*|...|A* for k occurances of A*
        assert k >= 1;
        if (k == 1) {
            return this;
        }

        addRepetition();

        DynamicNFA tmp = this;
        for (int i = 0; i < k - 1; i++) {
            tmp = createParallel(this, tmp);
        }
        set(tmp);

        return this;
    }

    public DynamicNFA addRepetition() {
        NFAState newStart = createState();
        finalStates.add(newStart);
        for (NFAState s : finalStates) {
            s.addAllEdges(initialState.getEdges());
        }
        initialState = newStart;
        return this;
    }

    /**
     * Creates new states and inserts it into the automaton.
     *
     * @return Index of the newly allocated state.
     */
    public int allocateState() {
        return createState().getId();
    }

    public void addEdge(int src, long event, int dst) {
        NFAState state = states.get(src);

        state.addEdge(event, dst);
    }

    public boolean hasInitial() {
        return initialState != null;
    }

    public void markFinal(int state) {
        finalStates.add(states.get(state));
    }

    public void unmarkFinal(int state) {
        finalStates.remove(states.get(state));
    }

    public void setInitial(int state) {
        initialState = states.get(state);
    }

    /**
     * Consumes the given automaton!
     */
    public DynamicNFA addSequence(DynamicNFA aut) {
        assert Collections.disjoint(states, aut.states);

        importStates(aut.states);

        for (NFAState s : finalStates) {
            s.addAllEdges(aut.initialState.getEdges());
        }

        if (aut.finalStates.contains(aut.initialState)) {
            finalStates.addAll(aut.finalStates);
        } else {
            finalStates = aut.finalStates;
        }

        return this;
    }

    public static DynamicNFA createNull() {
        DynamicNFA result = new DynamicNFA(1, 1);

        result.initialState = result.createState();
        result.finalStates.add(result.initialState);

        return result;
    }

    public static DynamicNFA createSimple(long symbol) {
        DynamicNFA result = new DynamicNFA(2, 1);

        result.initialState = result.createState();
        NFAState fin = result.createState();
        result.initialState.addEdge(symbol, fin.getId());
        result.finalStates.add(fin);

        return result;
    }

    public DynamicDFA determinize() {

        DynamicDFA dfa = new DynamicDFA();

        List<Integer> detStates = new ArrayList<Integer>();
        Queue<Integer> queue = new ArrayDeque<Integer>();
        Map<Set<NFAState>, Integer> setToState =
                new HashMap<Set<NFAState>, Integer>();
        Map<Integer, Set<NFAState>> stateToSet =
                new HashMap<Integer, Set<NFAState>>();

        int detStart = dfa.allocateState();
        dfa.setInitial(detStart);
        if (finalStates.contains(initialState)) {
            dfa.markFinal(detStart);
        }

        Set<NFAState> startSet = Collections.singleton(initialState);

        queue.add(detStart);
        stateToSet.put(detStart, startSet);
        setToState.put(startSet, detStart);

        Integer state;
        Collection<NFAState> actSet;
        while ((state = queue.poll()) != null) {
            detStates.add(state);
            actSet = stateToSet.get(state);

            // Get accesible sets
            EdgeMultiMap m = new EdgeMultiMap();
            for (NFAState s : actSet) {
                for (Map.Entry<Long, List<Integer>> entry : s.getEdges().entrySet()) {
                    for (int dst : entry.getValue()) {
                        m.put(entry.getKey(), states.get(dst));
                    }
                }
            }

            // Each entry results in one edge in new graph
            for (Map.Entry<Long, Set<NFAState>> entry : m.entrySet()) {
                long symbol = entry.getKey();
                Set<NFAState> setTarget = entry.getValue();

                // Get deterministic state for target
                Integer target = setToState.get(setTarget);
                if (target == null) { // We have to create one

                    target = dfa.allocateState();

                    if (!Collections.disjoint(setTarget, finalStates)) {
                        dfa.markFinal(target);
                    }

                    queue.add(target);
                    setToState.put(setTarget, target);
                    stateToSet.put(target, setTarget);
                }

                dfa.addEdge(state, symbol, target);
            }

        }

        return dfa;
    }

    private static class EdgeMultiMap extends HashMap<Long, Set<NFAState>> {

        public void put(Long key, NFAState value) {

            if (containsKey(key)) {
                get(key).add(value);
            } else {
                Set<NFAState> newBucket = new HashSet<NFAState>();
                newBucket.add(value);
                put(key, newBucket);
            }
        }
    }
}
