/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.visitor.LTSAVisitor;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;

/**
 * Represents transition of Labeled Transition System with Assignments.
 * 
 * <p>
 * A transition is supposed to be composed of following units:
 * <table>
 * 	<tr>
 * 		<td>Labels</td>
 * 		<td>Sequence of {@link Label}s</td>
 * 		<td>Represents observable events (method calls and returns) triggered during execution of this transition.</td>
 * 	</tr>
 * 	<tr>
 * 		<td>Guard</td>
 * 		<td>{@link ITBPCondition}&lt;REFERENCE&gt;</td>
 * 		<td>Represents a condition under which this transition is enabled.</td>
 * 	</tr>
 * 	<tr>
 * 		<td>Assignments</td>
 * 		<td>Collection of {@link TBPAssignment}&lt;REFERENCE&gt;s</td>
 * 		<td>Represents set of variable assignments that performed during execution of this transition.</td>
 * 	</tr>
 * 	<tr>
 * 		<td>Code-Reference</td>
 * 		<td>Specified by 2nd template parameter</td>
 * 		<td>Reference to the original code forming this transition (typically a TBPNode).</td>
 * 	</tr>
 * </table>
 * </p>
 *
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 *
 */
public interface LTSA <CODEREF_TYPE, REFERENCE> {

	/**
	 * Returns condition under which this transition is enabled.
	 * 
	 * <p>
	 * Note that <code>null</code> represents condition which is always true.
	 * </p>
	 * 
	 * @return {@link ITBPCondition}
	 */
	ITBPCondition< REFERENCE > getGuard();
	
	/**
	 * Returns list of observable events that
	 * are triggered during execution of this transition. 
	 * 
	 * <p>
	 * Note that the returned list may be read-only.
	 * </p>
	 * 
	 * @return {@link List} of {@link Label}s
	 */
	List< Label<REFERENCE> > getLabels();
	
	/**
	 * Returns list of assignments that are performed during
	 * execution of this transition.
	 * 
	 * <p>
	 * Note that the returned list may be read-only.
	 * </p>
	 * 
	 * @return {@link List} of {@link TBPAssignment}s
	 */
	List< TBPAssignment< REFERENCE > > getAssignments();
	
	/**
	 * Returns reference to forming code.
	 * 
	 * @return CODEREF
	 */
	CODEREF_TYPE getCodeReference();

	/** Returns list of transitions connected to this one.
	 * 
	 * @return {@link List} of {@link LTSA}
	 */
	List< LTSA< CODEREF_TYPE, REFERENCE> > getTransitions();	

	/** Connects successive transition to this one.
	 * 
	 * @param transition Transition to be connected.
	 */
	void connectTransition(LTSA< CODEREF_TYPE, REFERENCE> transition);
	
	<RESULT> RESULT visit(LTSAVisitor<RESULT, CODEREF_TYPE, REFERENCE> visitor);
}
