/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.visitor.LTSAVisitor;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;

/** Implementation of general {@link LTSA} transition.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class LTSAImpl<CODEREF_TYPE, REFERENCE> implements LTSA<CODEREF_TYPE, REFERENCE> {
	
	private List<Label<REFERENCE>> labels;
	private ITBPCondition<REFERENCE> guard;
	private List<TBPAssignment<REFERENCE>> assignments;
	private CODEREF_TYPE codeReference;
	private List<LTSA<CODEREF_TYPE, REFERENCE>> connectedTransitions;
	
	public LTSAImpl() {
		this(null, null, null, null, null);
	}
	
	public LTSAImpl(
			Collection<Label<REFERENCE>> labels,
			ITBPCondition<REFERENCE> guard,
			Collection<TBPAssignment<REFERENCE>> assignments) {
		this(labels, guard, assignments, null, null);
	}
	
	public LTSAImpl(
			Collection<Label<REFERENCE>> labels,
			ITBPCondition<REFERENCE> guard,
			Collection<TBPAssignment<REFERENCE>> assignments,
			CODEREF_TYPE codeReference) {
		this(labels, guard, assignments, codeReference, null);
	}
		
	protected LTSAImpl(
			Collection<Label<REFERENCE>> labels,
			ITBPCondition<REFERENCE> guard,
			Collection<TBPAssignment<REFERENCE>> assignments,
			CODEREF_TYPE codeReference,
			List<LTSA<CODEREF_TYPE, REFERENCE>> connectedTransitions) {
		
		// initialize data structures 
		this.labels = new LinkedList<Label<REFERENCE>>();
		this.guard = null;
		this.assignments = new LinkedList<TBPAssignment<REFERENCE>>();
		this.codeReference = null;
		this.connectedTransitions = new LinkedList<LTSA<CODEREF_TYPE,REFERENCE>>();		
		
		// labels
		if (labels != null) {
			this.labels.addAll(labels);
		}
		// guard
		this.guard = guard;
		// assignments
		if (assignments != null) {
			this.assignments.addAll(assignments);
		}
		// code reference
		this.codeReference = codeReference;
		// transitions
		if (connectedTransitions != null) {
			this.connectedTransitions.addAll(connectedTransitions);			
		}
		// fix
		this.labels = Collections.unmodifiableList(this.labels);
		this.assignments = Collections.unmodifiableList(this.assignments);
	}

	@Override
	public ITBPCondition<REFERENCE> getGuard() {
		return guard;
	}

	@Override
	public List<Label<REFERENCE>> getLabels() {
		return this.labels;
	}

	@Override
	public List<TBPAssignment<REFERENCE>> getAssignments() {
		return this.assignments;
	}

	@Override
	public CODEREF_TYPE getCodeReference() {
		return codeReference;
	}

	@Override
	public List<LTSA<CODEREF_TYPE, REFERENCE>> getTransitions() {
		return Collections.unmodifiableList(this.connectedTransitions);
	}

	@Override
	public void connectTransition(
			LTSA<CODEREF_TYPE, REFERENCE> transition) {
		if (transition == null) {
			throw new NullPointerException("Can not connect null transition.");
		}
		this.connectedTransitions.add(transition);
	}

	@Override
	public <RESULT> RESULT visit(
			LTSAVisitor<RESULT, CODEREF_TYPE, REFERENCE> visitor) {
		return visitor.visit(this);
	}
}
