/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.TBPRecursiveCallException;
import org.ow2.dsrg.fm.tbplib.ltsa.visitor.FindMethodCallsVisitor;
import org.ow2.dsrg.fm.tbplib.ltsa.visitor.InlineMethodCallsVisitor;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.VariableFactory;
import org.ow2.dsrg.fm.tbplib.reference.Reference;

/** Provides encapsulated LTSA in-linement.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class LTSAInlineHepler<REFERENCE extends Reference> {

	/** Transforms {@link LTSASpecification} into in-lined one.  
	 * 
	 * <p>The resulting {@link LTSASpecification} contains no method invocation statements.</p>
	 *
	 * @param <REFERENCE>
	 * @param specification original {@link LTSASpecification}
	 * @param callLocator {@link FindMethodCallsVisitor call locator} (locates method invocation target) 
	 * @param variableFactory {@link VariableFactory variable factory} (provides creation of neccessary local variables)
	 * @return transformed {@link LTSASpecification}
	 */
	public static <REFERENCE extends Reference> LTSASpecification<TBPNode<REFERENCE>, REFERENCE> inline(
			LTSASpecification<TBPNode<REFERENCE>, REFERENCE> specification,
			FindMethodCallsVisitor<REFERENCE> callLocator,
			VariableFactory<REFERENCE> variableFactory) {
		
		return (new LTSAInlineHepler<REFERENCE>(specification, callLocator, variableFactory)).inline();
	}
	
	/* Target of inlining */
	private LTSASpecification<TBPNode<REFERENCE>, REFERENCE> subject;
	/* Mapping of original methods to in-lined methods. */
	private Map<LTSAMethod<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>>
		originalToInlined;
	/* Set of methods that are currently being processed. */
	private Set<LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> inProc;
	/* Provides locating of call targets (callees). */
	private FindMethodCallsVisitor<REFERENCE> callLocator;
	/* Provides step of in-linement (single code fragment - method/thread). */
	private InlineMethodCallsVisitor<TBPNode<REFERENCE>, REFERENCE> inliner;
		
	/** Creates instance of this class.
	 * 
	 * <p>Used from {@link #inline(LTSASpecification, FindMethodCallsVisitor, VariableFactory)} method.
	 * </p>
	 * 
	 * @param subject {@link LTSASpecification} that will be processed
	 * @param callLocator {@link FindMethodCallsVisitor}
	 * @param variableFactory {@link VariableFactory}
	 */
	protected LTSAInlineHepler(
			LTSASpecification<TBPNode<REFERENCE>, REFERENCE> subject,
			FindMethodCallsVisitor<REFERENCE> callLocator,
			VariableFactory<REFERENCE> variableFactory) {
		super();
		this.callLocator = callLocator;
		this.inliner = new InlineMethodCallsVisitor<TBPNode<REFERENCE>, REFERENCE>(variableFactory);
		this.inProc = new HashSet<LTSAMethod<TBPNode<REFERENCE>,REFERENCE>>();
		this.subject = subject;
	}

	/** Performs in-linement.
	 * 
	 * <p>
	 * Traverses threads and methods and for each such object performs
	 * in-line. In particular, the process consists of following steps:
	 * <ul><li>locate calls and their targets (using {@link #callLocator})</li>
	 *     <li>in-line called methods to caller (using {@link #inliner})</li>
	 * </ul>
	 * </p>
	 * <p>
	 * During the whole process of transformation the
	 * {@link LTSASpecification}, a mapping of original method to processed
	 * methods ({@link #originalToInlined}) is kept in order not to process
	 * single method more than once.
	 * </p>
	 * <p>
	 * In addition a set of methods currently being processed
	 * (@link {@link #inProc}) is kept as well
	 * in order to detect recursive calls.
	 * Recursion is reported using {@link TBPRecursiveCallException}.
	 * </p>
	 * 
	 * @return in-lined version of original {@link LTSASpecification} 
	 */
	protected LTSASpecification<TBPNode<REFERENCE>, REFERENCE> inline() {
		assert(subject != null);

		LTSASpecification<TBPNode<REFERENCE>, REFERENCE> inlinedSpecification = new LTSASpecification<TBPNode<REFERENCE>, REFERENCE>();

		List<LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> originalMethods = subject.getMethods();
		List<LTSAThread<TBPNode<REFERENCE>, REFERENCE>> originalThreads = subject.getThreads();
		
		this.originalToInlined = new HashMap<LTSAMethod<TBPNode<REFERENCE>,REFERENCE>, LTSAMethod<TBPNode<REFERENCE>,REFERENCE>>(originalMethods.size());
		
		this.inProc.clear();
		
		for(LTSAThread<TBPNode<REFERENCE>, REFERENCE> t: originalThreads) {
			inlinedSpecification.addThread(inlineIntoThreadRecursively(t));
		}
				
		for(LTSAMethod<TBPNode<REFERENCE>, REFERENCE> m : originalMethods) {
			inlinedSpecification.addMethod(inlineIntoMethodRecursively(m));
		}
		
		this.inProc.clear();
		
		return inlinedSpecification;
	}
	
	
	private LTSAMethod<TBPNode<REFERENCE>, REFERENCE> inlineIntoMethodRecursively(
			LTSAMethod<TBPNode<REFERENCE>, REFERENCE> method) {
		LTSAMethod<TBPNode<REFERENCE>, REFERENCE> result = this.originalToInlined.get(method);
		if (result == null) { // if not visited yet
			// check recursion...
			if (this.inProc.contains(method)) {
				throw new TBPRecursiveCallException();
			}
			
			this.inProc.add(method);
			
			callLocator.reset();
			Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> 
				callMap = method.visit(this.callLocator);
			Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>>
				inlinedCallMap = new HashMap<LTSA<TBPNode<REFERENCE>,REFERENCE>, LTSAMethod<TBPNode<REFERENCE>,REFERENCE>>();
			
			for(Map.Entry<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> entry: callMap.entrySet()) {
				inlinedCallMap.put(entry.getKey(), this.inlineIntoMethodRecursively(entry.getValue()));
			}
			
			inliner.setCallMap(inlinedCallMap);
			result = inliner.inlineToMethod(method);
			
			this.originalToInlined.put(method, result);			
		}
		return result;
	}
	private LTSAThread<TBPNode<REFERENCE>, REFERENCE> inlineIntoThreadRecursively(LTSAThread<TBPNode<REFERENCE>, REFERENCE> thread) {
		callLocator.reset();
		Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> 
			callMap = thread.visit(this.callLocator);
		Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>>
			inlinedCallMap = new HashMap<LTSA<TBPNode<REFERENCE>,REFERENCE>, LTSAMethod<TBPNode<REFERENCE>,REFERENCE>>();
			
		
		for(Map.Entry<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> entry: callMap.entrySet()) {
			inlinedCallMap.put(entry.getKey(), inlineIntoMethodRecursively(entry.getValue()));
		}
		
		inliner.setCallMap(inlinedCallMap);
		return inliner.inlineToThread(thread);
	}
}
