/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.visitor.LTSAVisitor;
import org.ow2.dsrg.fm.tbplib.node.TBPMethodDeclaration;
import org.ow2.dsrg.fm.tbplib.node.TBPVariableDefinition;
import org.ow2.dsrg.fm.tbplib.reference.MethodSignature;

/** Represents a method (reaction).
 * 
 * <p>
 * Usual implementation of {@link LTSAMethod}.
 * </p> 
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class LTSAMethodImpl<CODEREF_TYPE, REFERENCE> extends LTSAAutonomousBase<CODEREF_TYPE, REFERENCE> implements
		LTSAMethod<CODEREF_TYPE, REFERENCE> {
	
	private REFERENCE interfaceName;
	private MethodSignature<REFERENCE> signature; 
	private List<LTSA<CODEREF_TYPE, REFERENCE>> returnStatements;

	/** Creates instance of <i>this</i> by fileds.
	 * 
	 * @param codeReference reference to the original code forming the method
	 * @param interfaceName interface 
	 * @param name name of the method
	 * @param localVariables local variables used within the method 
	 * @param signature signature of the method
	 * @param returnStatements LTSA transition that represent <code>return</code> statements 
	 */
	public LTSAMethodImpl(CODEREF_TYPE codeReference, REFERENCE interfaceName, REFERENCE name, List<TBPVariableDefinition<REFERENCE>> localVariables, MethodSignature<REFERENCE> signature, List<LTSA<CODEREF_TYPE, REFERENCE>> returnStatements) {
		super(codeReference, name, localVariables);
		this.interfaceName = interfaceName;
		this.signature = signature;
		if (returnStatements != null) {
			this.returnStatements = new ArrayList<LTSA<CODEREF_TYPE,REFERENCE>>(returnStatements.size());
			this.returnStatements.addAll(returnStatements);
		} else {
			this.returnStatements = Collections.emptyList();
		}
		this.returnStatements = Collections.unmodifiableList(this.returnStatements);		
	}
	/** Creates instance of <i>this</i> from {@link TBPMethodDeclaration}. 
	 * 
	 * @param codeReference reference to the original code forming the method
	 * @param methodDeclaration {@link TBPMethodDeclaration}
	 * @param returnStatements LTSA transition that represent <code>return</code> statements 
	 */
	public LTSAMethodImpl(CODEREF_TYPE codeReference, TBPMethodDeclaration<REFERENCE> methodDeclaration, List<LTSA<CODEREF_TYPE, REFERENCE>> returnStatements) {
		this(
				codeReference,
				methodDeclaration.getInterface(),
				methodDeclaration.getMethod(),
				methodDeclaration.getLocalVariables(),
				methodDeclaration.getSignature(),
				returnStatements
			);
	}
	

	@Override
	public REFERENCE getInterfaceName() {
		return interfaceName;
	}

	@Override
	public MethodSignature<REFERENCE> getSignature() {
		return signature;
	}

	@Override
	public List<LTSA<CODEREF_TYPE, REFERENCE>> getReturnTransitions() {
		return returnStatements;
	}
	
	@Override
	public <RESULT> RESULT visit(LTSAVisitor<RESULT, CODEREF_TYPE, REFERENCE> visitor) {
		return visitor.visitMethod(this);
	}

}
