/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa;

import java.util.Collections;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.event.EventTable;

/** Label for Labeled Transition System with Assignments ({@link LTSA}). 
 * 
 * <p>Represents observable event that is triggered on LTSA transition execution.
 *    The event may be divided into two major types; <i>method call</i> and <i>return from method</i>.
 *    An instance of this class is supposed to contain all information
 *    necessary for creating appropriate encoding of event at simulation-time
 *    (note that actual values of method parameters and return value are rarely
 *     known before simulation).
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class Label<REFERENCE> { 
	protected REFERENCE method;
	protected REFERENCE iface;
	protected REFERENCE returnValue;
	protected boolean returnValueSpecified; 
	protected List<REFERENCE> parameters;
	protected String fullMethodName;
	
	/** Default constructor.
	 */
	protected Label() {
		returnValueSpecified = false;
	}
	
	/** Returns method name.
	 * 
	 * @return REFERENCE
	 */
	public REFERENCE getMethod() {
		return method;
	}
	
	/** Returns interface name.
	 * 
	 * @return REFERENCE
	 */
	public REFERENCE getInterface() {
		return iface;
	}
	
	/** Returns identifier of variable with return value
	 *  (null if this is method call or return from method with no return value).
	 * 
	 * @return REFERENCE or <code>null</code>
	 */
	public REFERENCE getReturnValue() {
		return returnValue;
	}
	
	/** Returns list of identifiers of parameters
	 *  (null if this is return)
	 * 
	 * @return REFERENCE or <code>null</code>
	 */
	public List<REFERENCE> getParameters() {
		return parameters;
	}
	
	/** Returns <code>true</code> iff this is call of a method.
	 * 
	 * @return boolean
	 */
	public boolean isMethodCall() {
		return (parameters != null);
	}
	
	/** Returns <code>true</code> iff this is return from a method.
	 * 
	 * @return boolean
	 */
	public boolean isReturnFromMethod() {
		return !isMethodCall();
	}
	
	/** Returns <code>true</code> iff the return value has been specified.
	 * 
	 * @return boolean
	 */
	public boolean isReturnValueSpecified() {
		return returnValueSpecified;
	}
	
	/** Returns full method name.
	 * 
	 * <p>Intended for encoding method using {@link EventTable}.</p>
	 * 
	 * @return full name of method
	 */
	public String getFullMethodName() {
		return fullMethodName;
	}
	
	/** Returns method code according to given {@link EventTable}.
	 * 
	 * @param eventTable
	 * @return code of method
	 */
	public int getMethodCode(EventTable eventTable) {
		return eventTable.encodeMethod(getFullMethodName());
	}

	/** Creates {@link Label} representing method call.
	 * 
	 * @param <REF> REFERENCE
	 * @param iface Interface
	 * @param method Method
	 * @param parameters Parameters
	 * @return {@link Label}
	 */
	public static <REF> Label<REF> createCallLabel(REF iface, REF method, List<REF> parameters, String fullName) {
		Label<REF> result = new Label<REF>();
		setEssentialFields(result, iface, method, fullName);
		result.parameters = parameters;
		if (result.parameters == null) {
			result .parameters = Collections.emptyList(); 
		}
		result.returnValue = null;
		return result;
	}
	
	/** Creates {@link Label} representing return from a method (return value unspecified).
     *
	 * @param <REF> REFERENCE
	 * @param iface Interface
	 * @param method Method
	 * @return {@link Label}
	 */
	public static <REF> Label<REF> createReturnLabel(REF iface, REF method, String fullName) {
		Label<REF> result = new Label<REF>();
		setEssentialFields(result, iface, method, fullName);
		result.parameters = null;
		return result;
	}
	
	/** Creates {@link Label} representing return from a method with specified return value.
	 * 
	 * @param <REF> REFERENCE
	 * @param iface Interface
	 * @param method Method
	 * @param returnValue Return value
	 * @return {@link Label} 
	 */
	public static <REF> Label<REF> createReturnLabel(REF iface, REF method, REF returnValue, String fullName) {
		Label<REF> result = new Label<REF>();
		setEssentialFields(result, iface, method, fullName);
		result.parameters = null;
		result.returnValue = returnValue;
		result.returnValueSpecified = true;
		return result;
	}
	
	private static <REF> void setEssentialFields(Label<REF> label, REF iface, REF method, String fullName) {
		label.iface = iface;
		label.method = method;
		label.fullMethodName = fullName;
	}
}
