/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa.visitor;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethod;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;

/** Collects method call/return labels from LTSA.
 *
 * <p>This is <b>not</b> thread safe.</p> 
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class CollectLTSALabelsVisitor<C, REFERENCE> implements
		LTSAVisitor<Collection<Label<REFERENCE>>, C, REFERENCE> {
	
	private Set<LTSA<?, REFERENCE>> visited;
	private Collection<Label<REFERENCE>> collectedLabels;
	boolean inSpecification, asAncestor;

	public CollectLTSALabelsVisitor() {
		visited = new HashSet<LTSA<?,REFERENCE>>();
		collectedLabels = null;
		inSpecification = false;
		asAncestor = false;
	}
		
	private Collection<Label<REFERENCE>> getCollectedLabels() {
		if (collectedLabels == null) {
			collectedLabels = new LinkedList<Label<REFERENCE>>();
		}
		return collectedLabels;
	}
	
	private void resetCollectedLabels() {
		this.collectedLabels = null;
	}

	@Override
	public Collection<Label<REFERENCE>> visit(LTSA<C, REFERENCE> transition) {
			
		boolean previsousAsAncestor = this.asAncestor;
		this.asAncestor = true;
		
		Collection<Label<REFERENCE>> collected = getCollectedLabels();
		
		if (!visited.contains(transition)) {
			for(Label<REFERENCE> label : transition.getLabels()) {
				collected.add(label);
			}
			visited.add(transition);
			for(LTSA<C, REFERENCE> connectedTransition : transition.getTransitions()) {
				connectedTransition.visit(this);				
			}			
		}
		
		this.asAncestor = previsousAsAncestor;
		if (!this.asAncestor && !this.inSpecification) {
			resetCollectedLabels();
		}		
		if (!this.asAncestor) {
			visited.clear();
		}		
		return collected;
	}

	@Override
	public Collection<Label<REFERENCE>> visitMethod(
			LTSAMethod<C, REFERENCE> method) {
		return this.visit(method);
	}

	@Override
	public Collection<Label<REFERENCE>> visitThread(
			LTSAThread<C, REFERENCE> thread) {
		return this.visit(thread);
	}

	@Override
	public Collection<Label<REFERENCE>> visitSpecification(
			LTSASpecification<C, REFERENCE> specification) {
		
		this.inSpecification = true;
				
		for(LTSAMethod<C, REFERENCE> method: specification.getMethods()) {
			this.visitMethod(method);
		}
		
		for(LTSAThread<C, REFERENCE> thread: specification.getThreads()) {
			this.visitThread(thread);
		}
		
		Collection<Label<REFERENCE>> collected = getCollectedLabels();
		resetCollectedLabels();
		
		this.inSpecification = false;
		return collected;
	}	
}
