/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa.visitor;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethod;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;

/** Searches method calls within visited {@link LTSA} transitions
 *  and locates targets of the calls. 
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 *
 */
public abstract class FindMethodCallsVisitor<REFERENCE> implements
		LTSAVisitor<Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>>, TBPNode<REFERENCE>, REFERENCE> {
	
	private Set<LTSA<TBPNode<REFERENCE>, REFERENCE>> visited;
	private Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> result;

	public FindMethodCallsVisitor() {
		super();
		
		this.visited = new HashSet<LTSA<TBPNode<REFERENCE>,REFERENCE>>();
		this.reset();
	}
	
	/** Resolves called method.
	 * 
	 * @param iface name of the called interface
	 * @param method name of the called method
	 * @param parameters parameters of the call
	 * @return {@link LTSAMethod}
	 */
	protected abstract LTSAMethod<TBPNode<REFERENCE>, REFERENCE> resolveCalledMethod(REFERENCE iface, REFERENCE method, List<REFERENCE> parameters);
	
	/** Resets this visitor for further utilization.
	 * 
	 * <p>Clears set of visited transitions (which is maintained in order
	 *    to avoid visiting single transition more than once) and
	 *    creates new (empty) object for result.</p>
	 */
	public void reset() {
		this.visited.clear();
		this.result = new HashMap<LTSA<TBPNode<REFERENCE>,REFERENCE>, LTSAMethod<TBPNode<REFERENCE>,REFERENCE>>();
	}

	@Override
	public Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> visit(LTSA<TBPNode<REFERENCE>, REFERENCE> transition) {
		if (!visited.contains(transition)) {
			for(Label<REFERENCE> call : getMethodCalls(transition)) {
				result.put(transition, 
						resolveCalledMethod(call.getInterface(), call.getMethod(), call.getParameters()));
			}
			
			visited.add(transition);
			
			for(LTSA<TBPNode<REFERENCE>, REFERENCE> t:  transition.getTransitions()) {
				t.visit(this);
			}
		}
		return result;
	}

	@Override
	public Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> visitMethod(LTSAMethod<TBPNode<REFERENCE>, REFERENCE> method) {
		if (!visited.contains(method)) {
			visited.add(method);
			for(LTSA<TBPNode<REFERENCE>, REFERENCE> t: method.getTransitions()) {
				t.visit(this);
			}
		}
		return result;
	}

	@Override
	public Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> visitThread(LTSAThread<TBPNode<REFERENCE>, REFERENCE> thread) {
		if (!visited.contains(thread)) {
			visited.add(thread);
			for(LTSA<TBPNode<REFERENCE>, REFERENCE> t: thread.getTransitions()) {
				t.visit(this);
			}
		}
		return result;
	}

	@Override
	public Map<LTSA<TBPNode<REFERENCE>, REFERENCE>, LTSAMethod<TBPNode<REFERENCE>, REFERENCE>> visitSpecification(
			LTSASpecification<TBPNode<REFERENCE>, REFERENCE> specification) {
		
		for(LTSAMethod<TBPNode<REFERENCE>, REFERENCE> method: specification.getMethods()) {
			method.visit(this);
		}		
		for(LTSAThread<TBPNode<REFERENCE>, REFERENCE> thread: specification.getThreads()) {
			thread.visit(this);
		}
		return result;
	}
	
	private List<Label<REFERENCE>> getMethodCalls(LTSA<TBPNode<REFERENCE>, REFERENCE> transition) {
		List<Label<REFERENCE>> callLabels = new LinkedList<Label<REFERENCE>>();
		for(Label<REFERENCE> label: transition.getLabels()) {
			if (label.isMethodCall()) {
				callLabels.add(label);
			}
		}
		return callLabels;
	}
}
