/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa.visitor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethod;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethodImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThreadImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPValue;
import org.ow2.dsrg.fm.tbplib.node.TBPVariableDefinition;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.LTSAManipulationHelper;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.VariableFactory;

/** In-lines method calls into caller.
 * 
 * <p>This visitor <b>does not support standard visitor methods</b>.
 *    Instead, just {@link #inlineToMethod(LTSAMethod)} and {@link #inlineToThread(LTSAThread)}
 *    should be used (from caller's perspective).
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class InlineMethodCallsVisitor<CODEREF_TYPE, REFERENCE> implements
		LTSAVisitor<LTSA<CODEREF_TYPE, REFERENCE>, CODEREF_TYPE, REFERENCE> {
	
	/** Map of visited nodes. */
	private Map<LTSA<CODEREF_TYPE, REFERENCE>, LTSA<CODEREF_TYPE, REFERENCE>>
		visitedMap;
	/** Local variables. */
	private List<TBPVariableDefinition<REFERENCE>>
		localVariables;
	private Collection<REFERENCE>
		locallyUsedNames;
	private Map<LTSA<CODEREF_TYPE, REFERENCE>, LTSAMethod<CODEREF_TYPE, REFERENCE>>
		callMap;
	private VariableFactory<REFERENCE> variableFactory;
	private RenameLocalVariablesVisitor<CODEREF_TYPE, REFERENCE> renameMethodBodyVisitor;
	
	private Map<REFERENCE, List<REFERENCE>> typesMap;
	
	private REFERENCE currentAutonomous;
	
	/** Creates this visitor.
	 * 
	 * @param variableFactory {@link VariableFactory} for creating variables
	 */
	public InlineMethodCallsVisitor(VariableFactory<REFERENCE> variableFactory) {
		this(variableFactory, new HashMap<LTSA<CODEREF_TYPE, REFERENCE>, LTSAMethod<CODEREF_TYPE, REFERENCE>>());
	}
	
	/** Creates this visitor.
	 * 
	 * @param variableFactory {@link VariableFactory} for creating variables
	 * @param callMap mapping of call statements to called methods
	 */
	public InlineMethodCallsVisitor(VariableFactory<REFERENCE> variableFactory, Map<LTSA<CODEREF_TYPE, REFERENCE>, LTSAMethod<CODEREF_TYPE, REFERENCE>> callMap) {
		super();
		
		this.renameMethodBodyVisitor = new RenameLocalVariablesVisitor<CODEREF_TYPE, REFERENCE>();
		this.typesMap = new HashMap<REFERENCE, List<REFERENCE>>();			
		this.visitedMap = new HashMap<LTSA<CODEREF_TYPE,REFERENCE>, LTSA<CODEREF_TYPE,REFERENCE>>();					
		this.variableFactory = variableFactory;
		this.setCallMap(callMap);
		this.localVariables = null;
		this.locallyUsedNames = null;
	}
	
	/** Sets call map.
	 * 
	 * @param callMap mapping of call statements to called methods
	 */
	public void setCallMap(Map<LTSA<CODEREF_TYPE, REFERENCE>, LTSAMethod<CODEREF_TYPE, REFERENCE>> callMap) {
		this.callMap = Collections.unmodifiableMap(callMap);
	}
				
	/** Visits ordinary {@link LTSA} node.
	 * <p><b>DO NOT USE THIS METHOD.</b></p>
	 */
	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visit(
			LTSA<CODEREF_TYPE, REFERENCE> transition) {
		if (visitedMap.containsKey(transition)) {
			return visitedMap.get(transition);
		}
		LTSA<CODEREF_TYPE, REFERENCE> result = null;
		List<LTSA<CODEREF_TYPE, REFERENCE>> endTransitions = new LinkedList<LTSA<CODEREF_TYPE,REFERENCE>>();
		
		if (callMap.containsKey(transition)) {
			
			//  in-line the call
			result = inlineMethodCall(transition, callMap.get(transition), endTransitions);
			 
		} else {
			// create pure copy
			result = new LTSAImpl<CODEREF_TYPE, REFERENCE>(
					transition.getLabels(),
					transition.getGuard(),
					transition.getAssignments(),
					transition.getCodeReference());
			endTransitions.add(result);
		}

		visitedMap.put(transition, result);		
		
		for(LTSA<CODEREF_TYPE, REFERENCE> connected : transition.getTransitions()) {
			LTSAManipulationHelper.connectAll(endTransitions, connected.visit(this));
		}
		
		return result;
	}
	
	/** Replaces method calls of given {@link LTSAMethod} with bodies of called methods. 
	 * 
	 * @param method {@link LTSAMethod}
	 * @return new {@link LTSAMethod} with replaced calls
	 */
	public LTSAMethod<CODEREF_TYPE, REFERENCE> inlineToMethod(LTSAMethod<CODEREF_TYPE, REFERENCE> method) {
		currentAutonomous = method.getName();
		LinkedList<LTSA<CODEREF_TYPE, REFERENCE>> connections = new LinkedList<LTSA<CODEREF_TYPE,REFERENCE>>();
		LinkedList<LTSA<CODEREF_TYPE, REFERENCE>> returnStatements = new LinkedList<LTSA<CODEREF_TYPE,REFERENCE>>();
		this.reset(method.getSignature().getParams().keySet(), new LinkedList<TBPVariableDefinition<REFERENCE>>(method.getLocalVariables()));
		for(LTSA<CODEREF_TYPE, REFERENCE> t: method.getTransitions()) {
			connections.add(this.visit(t));
		}
		for(LTSA<CODEREF_TYPE, REFERENCE> t: method.getReturnTransitions()) {
			returnStatements.add(visitedMap.get(t));
		}
		
		// create new method
		LTSAMethod<CODEREF_TYPE, REFERENCE> newMethod = new LTSAMethodImpl<CODEREF_TYPE, REFERENCE>
			(method.getCodeReference(), method.getInterfaceName(), method.getName(),  this.localVariables, method.getSignature(), returnStatements);  		
		for(LTSA<CODEREF_TYPE, REFERENCE> t: connections) {
			newMethod.connectTransition(t);
		}
		return newMethod;		
	}
	
	/** Replaces method calls of given {@link LTSAThread} with bodies of called methods. 
	 * 
	 * @param thread {@link LTSAThread}
	 * @return new {@link LTSAThread} with replaced calls
	 */
	public LTSAThread<CODEREF_TYPE, REFERENCE> inlineToThread(LTSAThread<CODEREF_TYPE, REFERENCE> thread) {
		currentAutonomous = thread.getName();
		LinkedList<LTSA<CODEREF_TYPE, REFERENCE>> connections = new LinkedList<LTSA<CODEREF_TYPE,REFERENCE>>();
		this.reset(null, new LinkedList<TBPVariableDefinition<REFERENCE>>(thread.getLocalVariables()));
		for(LTSA<CODEREF_TYPE, REFERENCE> t: thread.getTransitions()) {
			connections.add(this.visit(t));
		}
		// create new thread
		LTSAThread<CODEREF_TYPE, REFERENCE> newThread = new LTSAThreadImpl<CODEREF_TYPE, REFERENCE>(thread.getCodeReference(), thread.getName(), this.localVariables, thread.getReplicationFactor());
		for(LTSA<CODEREF_TYPE, REFERENCE> t: connections) {
			newThread.connectTransition(t);
		}
		return newThread;		
	}
	
	/** <b>NOT SUPPORTED</b> - throws {@link UnsupportedOperationException}.
	 */
	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visitSpecification(
			LTSASpecification<CODEREF_TYPE, REFERENCE> specification) {
		throw new UnsupportedOperationException();		
	}
	/** <b>NOT SUPPORTED</b> - throws {@link UnsupportedOperationException}.
	 */	
	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visitMethod(
			LTSAMethod<CODEREF_TYPE, REFERENCE> method) {
		throw new UnsupportedOperationException();		
	}
	/** <b>NOT SUPPORTED</b> - throws {@link UnsupportedOperationException}.
	 */	
	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visitThread(
			LTSAThread<CODEREF_TYPE, REFERENCE> thread) {
			throw new UnsupportedOperationException();			
	}		
	
	//
	// Internal methods
	//
	
	/** In-lines method into caller. 
	 * 
	 * @param call Caller transition
	 * @param method Called method (supposed to be replaced by the result)
	 * @param returnPoints <i>[out]</i> {@link List} of returning {@link LTSA} transitions 
	 * @return First {@link LTSA} transition 
	 */
	private LTSA<CODEREF_TYPE, REFERENCE> inlineMethodCall(LTSA<CODEREF_TYPE, REFERENCE> call, LTSAMethod<CODEREF_TYPE, REFERENCE> method, List<LTSA<CODEREF_TYPE, REFERENCE>> returnPoints) {
		assert(call != null && method != null && returnPoints != null);
		// extract return value target
		REFERENCE returnValueTarget = null;
		for(TBPAssignment<REFERENCE> assignment : call.getAssignments()) {
			TBPNode<REFERENCE> assChild = assignment.getChild();
			if (assChild instanceof TBPValue<?>) {
				TBPValue<REFERENCE> value = (TBPValue<REFERENCE>)assChild;
				if (value.isMethodCall() && value.getMethodCall().getInterface().equals(method.getInterfaceName()) && value.getMethodCall().getMethod().equals(method.getName())) {
					if (returnValueTarget == null) {
						returnValueTarget = assignment.getIdf();
					} else {
						throw new UnsupportedOperationException("Multiple calls of the same method within single transition not supported.");
					}
				}
			} else {
				throw new UnsupportedOperationException("Unsupported assignment.");
			}
		}
				
		// create variable renaming map
		Map<REFERENCE, REFERENCE> renaming = createVariableRenaming(method);
		
		// extract method body with renamed local variables
		this.renameMethodBodyVisitor.reset(renaming, returnValueTarget);
		LTSA<CODEREF_TYPE, REFERENCE> methodStart = method.visit(this.renameMethodBodyVisitor);
		returnPoints.addAll(this.renameMethodBodyVisitor.getReturnStatements());		
				
		// extract parameters
		List<REFERENCE> parameters = null;
		Label<REFERENCE> theCallLabel = null;
		for(Label<REFERENCE> label: call.getLabels()) {
			// NOTE that this cycle asserts that there are just single call label within the call.getLabels()
			if (label.isMethodCall()) {
				if (theCallLabel == null) {
					parameters = label.getParameters();
					theCallLabel = label;
				} else {
					throw new UnsupportedOperationException("Multiple calls of the same method within single transition not supported.");
				}
				if (!label.getInterface().equals(method.getInterfaceName()) || !label.getMethod().equals(method.getName())) {
					throw new UnsupportedOperationException("Method call mismatch.");
				}
				
			}			
		}
		if (theCallLabel == null) {
			throw new UnsupportedOperationException("Inlining of method call without call label is not supported.");
		}		
		if (parameters == null) {
			parameters = Collections.emptyList();
		}
		List<REFERENCE> formalParams = new ArrayList<REFERENCE>(method.getSignature().getParams().entrySet().size());
		for(Entry<REFERENCE, REFERENCE> parameter: method.getSignature().getParams().entrySet()) {
			REFERENCE renamed = renaming.get(parameter.getKey());
			if (renamed == null) {
				formalParams.add(parameter.getKey());
			} else {
				formalParams.add(renamed);
			}
		}
		if (formalParams.size() != parameters.size()) {
			// NOTE: this may be either mistake (mismatching parameters of call and method)
			//       or call with default parameters (which is hardly supported by TBP laguage)
			throw new UnsupportedOperationException("Inlining of method with default parameters not supported yet.");
		}
		
		// create method entry transition		
		LinkedList<TBPAssignment<REFERENCE>> initialAssignments = new LinkedList<TBPAssignment<REFERENCE>>();
		for(TBPAssignment<REFERENCE> assignment : methodStart.getAssignments()) {
			if (!formalParams.contains(assignment.getIdf())) {
				initialAssignments.add(assignment);
			}
		}
		Iterator<REFERENCE> formalParamsIt = formalParams.iterator();
		Iterator<REFERENCE> paramsIt = parameters.iterator();
		while(formalParamsIt.hasNext() && paramsIt.hasNext()) {
			initialAssignments.add(new TBPAssignment<REFERENCE>(formalParamsIt.next(), new TBPValue<REFERENCE>(paramsIt.next())));
		}
		assert(formalParamsIt.hasNext() == paramsIt.hasNext());
		
		ITBPCondition<REFERENCE> firstGuard = call.getGuard();
		ITBPCondition<REFERENCE> secondGuard = methodStart.getGuard();
		if (secondGuard == null) {
			secondGuard = firstGuard;
			firstGuard = null;
		}
				
		LTSA<CODEREF_TYPE, REFERENCE> methodEntryTransition = new LTSAImpl<CODEREF_TYPE, REFERENCE>(
				Collections.singletonList(theCallLabel),
				secondGuard,
				initialAssignments,
				methodStart.getCodeReference());
		for(LTSA<CODEREF_TYPE, REFERENCE> successor: methodStart.getTransitions()) {
			methodEntryTransition.connectTransition(successor);
		}
		
		if (firstGuard != null) {
			LTSA<CODEREF_TYPE, REFERENCE> tmpTransition = new LTSAImpl<CODEREF_TYPE, REFERENCE>(
				null,
				firstGuard,
				null,
				null);
			tmpTransition.connectTransition(methodEntryTransition);
			methodEntryTransition = tmpTransition;
		}
		
		return methodEntryTransition;
	}
	
	private void reset(Collection<REFERENCE> paramNames, List<TBPVariableDefinition<REFERENCE>> localVariables) {
		this.visitedMap.clear();
		this.typesMap.clear();
		this.localVariables = localVariables != null ? localVariables : new LinkedList<TBPVariableDefinition<REFERENCE>>();
		this.locallyUsedNames = new LinkedList<REFERENCE>();
		if (paramNames != null) {
			for(REFERENCE r : paramNames) {
				locallyUsedNames.add(r);
			}
		}
		for (TBPVariableDefinition<REFERENCE> d: this.localVariables) {
			locallyUsedNames.add(d.getName());
		}
	}
	
	private void addLocalVariable(TBPVariableDefinition<REFERENCE> variableDefinition) {
		localVariables.add(variableDefinition);
		locallyUsedNames.add(variableDefinition.getName());
	}
	
	
	private Map<REFERENCE, REFERENCE> createVariableRenaming(LTSAMethod<CODEREF_TYPE, REFERENCE> method) {
		Map<REFERENCE, Integer> typeUsageMap = new HashMap<REFERENCE, Integer>();
		
		Map<REFERENCE, REFERENCE> nameToName = new HashMap<REFERENCE, REFERENCE>();
		
		// handle parameters
		for (Entry<REFERENCE, REFERENCE> param : method.getSignature().getParams().entrySet()) {
			// is the parameter name contained within local variables?
			REFERENCE type = param.getValue();
			REFERENCE name = param.getKey();
			for(TBPVariableDefinition<REFERENCE> local : method.getLocalVariables()) {
				if (local.getName().equals(name)) {
					assert(local.getTypename().equals(type));
					name = null; type = null;
					break;
				}
			}
			if (type != null) {
				int nameIndex = adjustTypeArityMapping(typeUsageMap, type, 1);
				REFERENCE newName = getVariableName(type, name, nameIndex);
				nameToName.put(name, newName);
			}
		}
		
		// handle local variables
		for(TBPVariableDefinition<REFERENCE> local : method.getLocalVariables()) {
			REFERENCE name = local.getName();
			REFERENCE type = local.getTypename();
			int nameIndex = adjustTypeArityMapping(typeUsageMap, type, 1);
			REFERENCE newName = getVariableName(type, name, nameIndex);
			nameToName.put(name, newName);
		}
		
		return nameToName;
	}
	
	private int adjustTypeArityMapping(Map<REFERENCE, Integer> mapping, REFERENCE type, int arity) {
		int result = 0;
		if (mapping.containsKey(type)) {
			Integer currentAriry = mapping.remove(type);
			mapping.put(type, currentAriry.intValue() + arity);
			result = currentAriry.intValue();
		} else {
			mapping.put(type, arity);
		}
		return result;
	}
	
	private REFERENCE getVariableName(REFERENCE type, REFERENCE suggestedName, int index) {
		List<REFERENCE> variableNames = this.typesMap.get(type);
		if (variableNames == null) {
			variableNames = new LinkedList<REFERENCE>();
			this.typesMap.put(type, variableNames);
		}
		
		assert(variableNames.size() + 1 > index);
		REFERENCE result = null;
		if (variableNames.size() > index) {
			result = variableNames.get(index);
		} else {
			TBPVariableDefinition<REFERENCE> newVariable = 
				this.variableFactory.createLocalVariable(suggestedName, type, this.currentAutonomous , this.locallyUsedNames);
			addLocalVariable(newVariable);
			result = newVariable.getName();
			variableNames.add(result);
		}
		return result;
	}
}
