/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa.visitor;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.ow2.dsrg.fm.tbplib.ltsa.LTS;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethod;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethodImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThreadImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSMethod;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSThread;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPAtomic;
import org.ow2.dsrg.fm.tbplib.node.TBPEmit;
import org.ow2.dsrg.fm.tbplib.node.TBPImperativeSequence;
import org.ow2.dsrg.fm.tbplib.node.TBPMethodDeclaration;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPReturn;
import org.ow2.dsrg.fm.tbplib.node.TBPValue;
import org.ow2.dsrg.fm.tbplib.reference.MethodCall;

/** Provides conversion from {@link LTS} to {@link LTSA}.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class LTSToLTSAConversionVisitor<REFERENCE>
		implements LTSVisitor<LTSA<TBPNode<REFERENCE>, REFERENCE>, REFERENCE> {
	
	private Map<LTS<? extends TBPNode<REFERENCE>, REFERENCE>, LTSA<TBPNode<REFERENCE>, REFERENCE>>
		convertedDictionary = new HashMap<LTS<? extends TBPNode<REFERENCE>,REFERENCE>, LTSA<TBPNode<REFERENCE>,REFERENCE>>();
	
	private List<LTSA<TBPNode<REFERENCE>, REFERENCE>> methodReturnStatements = null;
	
	private TBPMethodDeclaration<REFERENCE> methodDeclaration = null;
	
	@Override
	public <T extends TBPNode<REFERENCE>> LTSA<TBPNode<REFERENCE>, REFERENCE> visit(
			LTS<T, REFERENCE> lts) {
		if (convertedDictionary.containsKey(lts)) {
			return convertedDictionary.get(lts);
		}
		// create the LTSA
		List<Label<REFERENCE>> labels = new LinkedList<Label<REFERENCE>>();
		List<TBPAssignment<REFERENCE>> assignments = new LinkedList<TBPAssignment<REFERENCE>>();
		boolean isReturn = collectSymbols(lts.getLabel(), labels, assignments);
		// FIXME: think of a better way of recognition of (silent) return from void method 		
		if (methodDeclaration != null && !isReturn && lts.getTransitions().isEmpty()) {
			labels.add(Label.createReturnLabel(methodDeclaration.getInterface(), methodDeclaration.getMethod(), methodDeclaration.getFullname()));
			isReturn = true;
		}
		LTSA<TBPNode<REFERENCE>, REFERENCE> result =
			new LTSAImpl<TBPNode<REFERENCE>, REFERENCE>(
					labels,
					lts.getGuard(),
					assignments,
					lts.getLabel());
		
		if (isReturn && methodReturnStatements != null) {
			methodReturnStatements.add(result);
		}
		
		// store intermediate result to dictionary before recursion
		convertedDictionary.put(lts, result);
		
		// connect transitions
		for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> connectedTransition: lts.getTransitions()) {
			result.connectTransition(connectedTransition.visit(this));
		}
					
		return result;
	}
	
	@Override
	public LTSA<TBPNode<REFERENCE>, REFERENCE> visitThread(
			LTSThread<REFERENCE> ltsThread) {
		return visitThreadInternal(ltsThread);
	}

	@Override
	public LTSA<TBPNode<REFERENCE>, REFERENCE> visitMethodDeclaration(
			LTSMethod<REFERENCE> ltsMethod) {
		return visitMethodDeclaration(ltsMethod);		
	}
	
	/**
	 * Converts {@link LTSSpecification} to {@link LTSASpecification}.
	 * 
	 * @param specification {@link LTSSpecification}
	 * @return {@link LTSASpecification}
	 */
	public LTSASpecification<TBPNode<REFERENCE>, REFERENCE> convertToLTSASpecification(LTSSpecification<REFERENCE> specification) {
		LTSASpecification<TBPNode<REFERENCE>, REFERENCE> result = new LTSASpecification<TBPNode<REFERENCE>, REFERENCE>();
		for(LTSMethod<REFERENCE> method : specification.getMethods()) {
			result.addMethod(this.visitMethodDeclarationInternal(method));			
		}
		for(LTSThread<REFERENCE> thread : specification.getThreads()) {
			result.addThread(this.visitThreadInternal(thread));
		}
		return result;
	}
	
	//
	// Internal methods
	//
	private LTSAThread<TBPNode<REFERENCE>, REFERENCE> visitThreadInternal(LTSThread<REFERENCE> thread) {
		List<LTSA<TBPNode<REFERENCE>, REFERENCE>> tmpMethodReturns = methodReturnStatements;
		methodReturnStatements = null;
		LTSAThread<TBPNode<REFERENCE>, REFERENCE> ltsaThread =
			new LTSAThreadImpl<TBPNode<REFERENCE>, REFERENCE>(thread.getLabel(), thread.getLabel(), 1);
		for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> lts : thread.getTransitions()) {
			ltsaThread.connectTransition(lts.visit(this));
		}
		methodReturnStatements = tmpMethodReturns;
		return ltsaThread;
	}
	public LTSAMethod<TBPNode<REFERENCE>, REFERENCE> visitMethodDeclarationInternal(
			LTSMethod<REFERENCE> ltsMethod) {
		if (methodReturnStatements == null) {
			methodReturnStatements = new LinkedList<LTSA<TBPNode<REFERENCE>,REFERENCE>>();
		} else {
			methodReturnStatements.clear();
		}
		methodDeclaration = ltsMethod.getLabel();		
		
		LinkedList<LTSA<TBPNode<REFERENCE>, REFERENCE>> transitions = new LinkedList<LTSA<TBPNode<REFERENCE>,REFERENCE>>();
		for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> lts : ltsMethod.getTransitions()) {
			transitions.add(lts.visit(this));
		}
		
		
		LTSAMethod<TBPNode<REFERENCE>, REFERENCE> ltsaMethod =
			new LTSAMethodImpl<TBPNode<REFERENCE>, REFERENCE>(ltsMethod.getLabel(), ltsMethod.getLabel(), methodReturnStatements);
		
		for(LTSA<TBPNode<REFERENCE>, REFERENCE> transition : transitions) {
			ltsaMethod.connectTransition(transition);
		}
		
		methodDeclaration = null;
		return ltsaMethod;		
	}
	private TBPValue<REFERENCE> getAssignmentValue(TBPAssignment<REFERENCE> assignment) {
		return (TBPValue<REFERENCE>)assignment.getChild();
	}
	private Label<REFERENCE> createCallLabelFromMethodCall(MethodCall<REFERENCE> methodCall) {
		return Label.createCallLabel(methodCall.getInterface(), methodCall.getMethod(), methodCall.getParamDecl(), methodCall.getFullname());
	}
	private Label<REFERENCE> createReturnLabelFromMethodCall(MethodCall<REFERENCE> methodCall) {
		return Label.createReturnLabel(methodCall.getInterface(), methodCall.getMethod(), methodCall.getFullname());
	}
	// Collects symbols (assignments and labels) from TBPNode (used as Label of LTS).
	// NOTE that the order of labels MUST the same as it would occur when the appropriate code is executed.
	// returns true if and only if node is a return statement
	private boolean collectSymbols(TBPNode<REFERENCE> node, List<Label<REFERENCE>> labels, List<TBPAssignment<REFERENCE>> assignments) {
		boolean isReturnStatement = false;
		assert(labels != null && assignments != null);
		if (node instanceof TBPReturn<?>) {
			// return statement
			assert(methodDeclaration != null);
			isReturnStatement = true;
			labels.add(Label.createReturnLabel(methodDeclaration.getInterface(), methodDeclaration.getMethod(), ((TBPReturn<REFERENCE>)node).getIdf(), methodDeclaration.getFullname()));			
		} else if (node instanceof TBPAssignment<?>) {
			// assignment 
			TBPAssignment<REFERENCE> assignment = (TBPAssignment<REFERENCE>) node;
			TBPValue<REFERENCE> value = getAssignmentValue(assignment);
			assignments.add(assignment);
			if (value.isMethodCall()) {
				labels.add(createCallLabelFromMethodCall(value.getMethodCall()));
				labels.add(createReturnLabelFromMethodCall(value.getMethodCall()));
			}
		} else if (node instanceof TBPEmit<?>) {
			// emit
			labels.add(createCallLabelFromMethodCall( ((TBPEmit<REFERENCE>) node).getMethodCall() ));
			labels.add(createReturnLabelFromMethodCall( ((TBPEmit<REFERENCE>) node).getMethodCall() )); 
		} else if (node instanceof TBPAtomic<?>) {
			// atomic - currently just sequence of plain assignments is supported
			LinkedList<TBPNode<REFERENCE>> stack = new LinkedList<TBPNode<REFERENCE>>(node.getChildren());
			while (!stack.isEmpty()) {
				TBPNode<REFERENCE> stmt = stack.pop();
				if (stmt instanceof TBPAssignment<?>) {
					assert(stmt.getChildCount() == 1);
					if (((TBPValue<REFERENCE>)((TBPAssignment<REFERENCE>) stmt).getChild()).isMethodCall()) {
						throw new UnsupportedOperationException("Method call not supported within atomic section (conversion to LTSA).");
					} else {
						assignments.add((TBPAssignment<REFERENCE>) stmt);						
					}
				} else if (stmt instanceof TBPImperativeSequence<?>) {
					stack.addAll(stmt.getChildren());
				} else {
					throw new UnsupportedOperationException("Unsuported statement within atomic section (conversion to LTSA).");
				}
			}
		}
		return isReturnStatement;
	}
}
