/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.ltsa.visitor;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAMethod;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPValue;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPAnd;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPInSet;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNondetCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNot;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotEquals;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPNotInSet;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPOr;
import org.ow2.dsrg.fm.tbplib.reference.MethodCall;

/** Renames local variables within body of a method.
 * 
 * <p>Creates copy of each LTSA transition with adjusted
 *    variable names. The names are adjusted accordingly to
 *    <b>renamingMap</b>, which is supposed to be given in
 *    {@link #RenameLocalVariablesVisitor(Map, Object)} constructor
 *    or {@link #reset(Map, Object)} method.
 * </p>
 * 
 * <p>Also keeps list of adjusted return transitions ({@link LTSAMethod#getReturnTransitions()}),
 *    which is afterwards accessible via the {@link #getReturnStatements()}.
 *    Note that the return transition is enriched with assignment to target variable (if any).
 * </p>
 * 
 * <p>Note that this visitor is not supposed to be used for
 *    {@link LTSASpecification} nor {@link LTSAThread},
 *    thus the {@link #visitSpecification(LTSASpecification)}
 *    and {@link #visitThread(LTSAThread)} methods throw
 *    {@link UnsupportedOperationException}.
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class RenameLocalVariablesVisitor<CODEREF_TYPE, REFERENCE> implements
		LTSAVisitor<LTSA<CODEREF_TYPE, REFERENCE>, CODEREF_TYPE, REFERENCE> {
	
	private Map<REFERENCE, REFERENCE> renamingMap;
	private Map<LTSA<CODEREF_TYPE, REFERENCE>, LTSA<CODEREF_TYPE, REFERENCE>> visitedTransitions;
	private List<LTSA<CODEREF_TYPE, REFERENCE>> returnStatements;
	private List<LTSA<CODEREF_TYPE, REFERENCE>> originalReturnStatements;
	private REFERENCE returnValueTarget;
	
	public RenameLocalVariablesVisitor() {
		this.renamingMap = null;
		this.visitedTransitions = new HashMap<LTSA<CODEREF_TYPE,REFERENCE>, LTSA<CODEREF_TYPE,REFERENCE>>();
		this.returnStatements = new LinkedList<LTSA<CODEREF_TYPE,REFERENCE>>();
		this.originalReturnStatements = null;
		this.returnValueTarget = null;
	}
	
	/** Creates and initializes this with given <b>renamingMap</p>
	 * 
	 * @param renamingMap renaming map
	 */
	public RenameLocalVariablesVisitor(Map<REFERENCE, REFERENCE> renamingMap, REFERENCE returnValueTarget) {
		this();
		reset(renamingMap, returnValueTarget);
	}
	
	/** Resets the visitor to be used for another renaming.
	 * 
	 * @param renamingMap new renaming map
	 */
	public void reset(Map<REFERENCE, REFERENCE> renamingMap, REFERENCE returnValueTarget) {
		this.returnStatements.clear();
		this.visitedTransitions.clear();
		this.renamingMap = renamingMap;
		this.returnValueTarget = returnValueTarget;
	}
	
	/** Returns list of return statements.
	 * 
	 * <p>The list of returns of a method is available after appropriate
	 *    call of {@link #visitMethod(LTSAMethod)} and is lost on
	 *    following {@link #reset(Map, Object)} or {@link #visitMethod(LTSAMethod)}
	 *    call. 
	 * </p>
	 * 
	 * @return {@link List} of {@link LTSA} transitions representing return from method.
	 */
	public List<LTSA<CODEREF_TYPE, REFERENCE>> getReturnStatements() {
		return returnStatements;
	}
	

	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visit(
			LTSA<CODEREF_TYPE, REFERENCE> transition) {
		if (this.visitedTransitions.containsKey(transition)) {
			return this.visitedTransitions.get(transition);
		}
		
		boolean isReturn = this.originalReturnStatements.contains(transition);
		
		// transform labels
		List<Label<REFERENCE>> originalLabels = transition.getLabels();
		List<Label<REFERENCE>> labels = new ArrayList<Label<REFERENCE>>(originalLabels.size());
		for(Label<REFERENCE> originalLabel : originalLabels) {
			Label<REFERENCE> label = null;
			if (originalLabel.isMethodCall()) {
				label = Label.createCallLabel(originalLabel.getInterface(), originalLabel.getMethod(), renameListOfReferences(originalLabel.getParameters()), originalLabel.getFullMethodName());
			} else if (originalLabel.isReturnValueSpecified()) {
				label = Label.createReturnLabel(originalLabel.getInterface(), originalLabel.getMethod(), renameReference(originalLabel.getReturnValue()), originalLabel.getFullMethodName());
			} else {
				label = Label.createReturnLabel(originalLabel.getInterface(), originalLabel.getMethod(), originalLabel.getFullMethodName());
			}
			assert(label != null);
			labels.add(label);
		}
		
		// transform guard
		ITBPCondition<REFERENCE> guard = renameGuard(transition.getGuard());
		
		// transform assignments
		List<TBPAssignment<REFERENCE>> originalAssignments = transition.getAssignments();
		List<TBPAssignment<REFERENCE>> assignments = new ArrayList<TBPAssignment<REFERENCE>>(originalAssignments.size());
		for(TBPAssignment<REFERENCE> originalAssignment : originalAssignments) {
			TBPAssignment<REFERENCE> assignment = new TBPAssignment<REFERENCE>(renameReference(originalAssignment.getIdf()), renameTBPValue(originalAssignment.getChild()));
			assignments.add(assignment);
		}
		
		LTSA<CODEREF_TYPE, REFERENCE> result, successor;
		REFERENCE returnValue = null;
		if (isReturn && this.returnValueTarget != null) {
			assert(labels.size() > 0);
			Label<REFERENCE> returnLabel = labels.get(labels.size() - 1);
			assert(returnLabel != null && returnLabel.isReturnFromMethod() && returnLabel.isReturnValueSpecified());
			returnValue = returnLabel.getReturnValue();
		}
		// assign return value to the target
		if (returnValue != null) {				
				List<TBPAssignment<REFERENCE>> finalAssignments =
					Collections.singletonList(new TBPAssignment<REFERENCE>(this.returnValueTarget, new TBPValue<REFERENCE>(returnValue)));
				if (assignments.isEmpty()) {
					result = new LTSAImpl<CODEREF_TYPE, REFERENCE>(labels, guard, finalAssignments, transition.getCodeReference());
					successor = result;				
				} else {
					// any assignment within return statement - split this transition into two
					Label<REFERENCE> returnLabel = labels.remove(labels.size() - 1);
					result = new LTSAImpl<CODEREF_TYPE, REFERENCE>(labels, guard, assignments, null);
					successor = new LTSAImpl<CODEREF_TYPE, REFERENCE>(Collections.singletonList(returnLabel), null, finalAssignments, transition.getCodeReference());
					result.connectTransition(successor);
				}
		} else {
			result = new LTSAImpl<CODEREF_TYPE, REFERENCE>(labels, guard, assignments, transition.getCodeReference());
			successor = result;
		}
		
		if (isReturn) {
			returnStatements.add(successor);
		}
		
		this.visitedTransitions.put(transition, result);
		
		for(LTSA<CODEREF_TYPE, REFERENCE> connected : transition.getTransitions()) {
			successor.connectTransition(connected.visit(this));
		}
		return result;
	}
	
	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visitMethod(
			LTSAMethod<CODEREF_TYPE, REFERENCE> method) {
		this.returnStatements.clear();
		this.originalReturnStatements = method.getReturnTransitions();
		
		return this.visit(method);
	}	

	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visitThread(
			LTSAThread<CODEREF_TYPE, REFERENCE> thread) {
		throw new UnsupportedOperationException();
	}

	@Override
	public LTSA<CODEREF_TYPE, REFERENCE> visitSpecification(
			LTSASpecification<CODEREF_TYPE, REFERENCE> specification) {
		throw new UnsupportedOperationException();		
	}
	
	//
	// Internal methods
	//
	private List<REFERENCE> renameListOfReferences(List<REFERENCE> refences) {
		List<REFERENCE> result = new ArrayList<REFERENCE>(refences.size());
		for(REFERENCE ref : refences) {
			result.add(renameReference(ref));
		}
		return result;
	}
	private Set<REFERENCE> renameSetOfReferences(Set<REFERENCE> references) {
		Set<REFERENCE> result = new HashSet<REFERENCE>(references.size());
		for(REFERENCE ref : references) {
			result.add(renameReference(ref));
		}
		return result;
	}
	
	private REFERENCE renameReference(REFERENCE reference) {
		if (renamingMap.containsKey(reference)) {
			return renamingMap.get(reference);
		} else {
			return reference;
		}
	}
	
	private ITBPCondition<REFERENCE> renameGuard(ITBPCondition<REFERENCE> guard) {
		ITBPCondition<REFERENCE> result = null;
		if (guard instanceof TBPEquals<?>) {
			result = new TBPEquals<REFERENCE>(renameReference(((TBPEquals<REFERENCE>) guard).getLeft()), renameReference(((TBPEquals<REFERENCE>) guard).getRight()));
		} else if (guard instanceof TBPNotEquals<?>) {
			result = new TBPNotEquals<REFERENCE>(renameReference(((TBPNotEquals<REFERENCE>) guard).getLeft()), renameReference(((TBPNotEquals<REFERENCE>) guard).getRight()));
		} else if (guard instanceof TBPInSet<?>) {
			result = new TBPInSet<REFERENCE>(renameReference(((TBPInSet<REFERENCE>) guard).getLeft()), renameSetOfReferences(((TBPInSet<REFERENCE>) guard).getSet()));
		} else if (guard instanceof TBPNotInSet<?>) {
			result = new TBPNotInSet<REFERENCE>(renameReference(((TBPNotInSet<REFERENCE>) guard).getLeft()), renameSetOfReferences(((TBPNotInSet<REFERENCE>) guard).getSet()));
		} else if (guard instanceof TBPAnd<?>) {
			result = new TBPAnd<REFERENCE>(renameGuard(((TBPAnd<REFERENCE>) guard).getLeft()), ((TBPAnd<REFERENCE>) guard).getRight());
		} else if (guard instanceof TBPOr<?>) {
			result = new TBPAnd<REFERENCE>(renameGuard(((TBPOr<REFERENCE>) guard).getLeft()), ((TBPOr<REFERENCE>) guard).getRight());			
		} else if (guard instanceof TBPNot<?>) {
			result = new TBPNot<REFERENCE>(renameGuard(((TBPNot<REFERENCE>)guard).getOperand()));
		} else if (guard instanceof TBPNondetCondition<?>) {
			result = new TBPNondetCondition<REFERENCE>();
		}
		return result;
	}
	
	private TBPValue<REFERENCE> renameTBPValue(TBPNode<REFERENCE> node) {
		TBPValue<REFERENCE> result = null;
		if (node instanceof TBPValue<?>) {
			TBPValue<REFERENCE> value = (TBPValue<REFERENCE>)node;
			if (value.isMethodCall()) {
				// method call
				MethodCall<REFERENCE> mc = new MethodCall<REFERENCE>(value.getMethodCall().getInterface(), value.getMethodCall().getMethod(), renameListOfReferences(value.getMethodCall().getParamDecl()));
				result = new TBPValue<REFERENCE>(mc);
			} else {
				// value
				result = new TBPValue<REFERENCE>(renameReference(value.getVarname()));
			}
		} else {
			throw new ClassCastException("Value expected.");
		}
		assert(result != null);
		return result;
	}
}
