/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.node;

import java.util.*;

/**
 * Condition is either '?' (represented by null in both right and left) or a comparison.
 * @author caitt3am
 *
 */
public class TBPCondition<REFERENCE> implements Cloneable {

	private REFERENCE right;
	private REFERENCE left;
	private boolean negate = false;
	private Set<REFERENCE> inSet;
    private int precomputedHashCode;

	public TBPCondition(REFERENCE left, REFERENCE right) {
		this(left, right, false);
	}

	public TBPCondition(REFERENCE left, REFERENCE right, boolean negate) {
		this.left = left;
		this.right = right;
        this.negate = negate;
        recomputeHashCode();
	}

	public TBPCondition(REFERENCE left, Set<REFERENCE> inSet) {
		this(left, inSet, false);
	}

	public TBPCondition(REFERENCE left, Set<REFERENCE> inSet, boolean negate) {
		this.left = left;
		this.inSet = inSet;
        this.negate = negate;
        recomputeHashCode();
	}
	
	/**
	 * Constructs '?' condition that is always true (even negated)
	 */
	public TBPCondition() {
		left = right = null;
        recomputeHashCode();
	}
	
	public REFERENCE getLeft() {
		return left;
	}
	
	public REFERENCE getRight() {
		return right;
	}

    public Set<REFERENCE> getInSet() {
        return inSet;
    }

	public void setRight(REFERENCE right) {
        assert(inSet == null);
		this.right = right;
        recomputeHashCode();
	}

	public void setLeft(REFERENCE left) {
        assert(inSet == null);
		this.left = left;
        recomputeHashCode();
	}

	public boolean isNondeterministic() {
		return left == null && right == null;
	}

	public boolean isDeterministic(){
		return !isNondeterministic();
	}

    public boolean isSetCondition() {
        return inSet != null;
    }

	public TBPCondition<REFERENCE> createNegation() {
		TBPCondition<REFERENCE> result = clone();
		result.negate = !negate;
		return result;
	}

    public boolean isNegated() {
        return negate;
    }

	@Override
	public TBPCondition<REFERENCE> clone() {
        if (inSet != null) {
    		return new TBPCondition<REFERENCE>(left, inSet, negate);
        } else if (isNondeterministic()){
    		return new TBPCondition<REFERENCE>();
        } else {
            return new TBPCondition<REFERENCE>(left, right, negate);
        }
	}

	@Override
	public String toString() {
		if(isNondeterministic()) {
			return "?";
		} else if (inSet == null) {
			if(negate)
				return left.toString() + " != " + right.toString();
			else
				return left.toString() + " == " + right.toString();
		} else {
			if(negate)
				return left.toString() + " notin " + inSetToString();
			else
				return left.toString() + " in " + inSetToString();
        }
    }

    private String inSetToString() {
        StringBuilder bfr = new StringBuilder();

        bfr.append('[');
        Iterator<REFERENCE> it = inSet.iterator();
        if (it.hasNext()) {
            REFERENCE ref = it.next();
            bfr.append(ref.toString());
        }
        while (it.hasNext()){
            REFERENCE ref = it.next();
            bfr.append(", ");
            bfr.append(ref.toString());
        }
        bfr.append(']');

        return bfr.toString();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof TBPCondition) {
            TBPCondition cond = (TBPCondition) obj;

            return
                    negate == cond.negate &&
                    ((left == null && cond.left == null) || left.equals(cond.left)) &&
                    ((right == null && cond.right == null) || right.equals(cond.right)) &&
                    ((inSet == null && cond.inSet == null) || inSet.equals(cond.inSet));
        }
        return false;
    }

    @Override
    public int hashCode() {
        return precomputedHashCode;
    }

    private void recomputeHashCode() {
        int result;
        int prime = 7823;

        result = inSet == null ? 0 : inSet.hashCode();
        result *= prime;
        result += left == null ? 0 : left.hashCode();
        result *= prime;
        result += right == null ? 0 : right.hashCode();
        result *= prime;
        result += negate ? 0 : 1;

        precomputedHashCode = result;
    }
}
