/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.node;

import org.ow2.dsrg.fm.tbplib.node.visitor.ImperativeResolvingVisitor;
import org.ow2.dsrg.fm.tbplib.node.visitor.ProvisionResolvingVisitor;
import org.ow2.dsrg.fm.tbplib.node.visitor.TBPDefaultVisitor;
import org.ow2.dsrg.fm.tbplib.node.visitor.TBPVisitor;
import org.ow2.dsrg.fm.tbplib.reference.ReferenceTransformer;

import java.util.ArrayList;
import java.util.List;

/**
 * Container for a TBPSpecification, i.e., a set of provisions, reactions, and
 * threads. Definitionas will be kept aside in the architecture structure.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class TBPSpecification<REFERENCE> {

    public void setComponentInstance(REFERENCE componentInstance) {
        this.componentInstance = componentInstance;
    }

    protected REFERENCE componentInstance;
    protected String componentType;
    protected List<TBPVariableDefinition<REFERENCE>> variables;
    protected List<TBPProvisionContainerNode<REFERENCE>> provisions;
    protected List<TBPMethodDeclaration<REFERENCE>> reactions;
    protected List<TBPThreadContainerNode<REFERENCE>> threads;

    public TBPSpecification(
            REFERENCE componentInstance,
            List<TBPVariableDefinition<REFERENCE>> variables,
            List<TBPProvisionContainerNode<REFERENCE>> provisions,
            List<TBPMethodDeclaration<REFERENCE>> reactions,
            List<TBPThreadContainerNode<REFERENCE>> threads) {

        this.componentInstance = componentInstance;
        this.variables = variables;
        this.provisions = provisions;
        this.reactions = reactions;
        this.threads = threads;
    }

    public List<TBPVariableDefinition<REFERENCE>> getVariables() {
        return variables;
    }

    public List<TBPProvisionContainerNode<REFERENCE>> getProvisions() {
        return provisions;
    }

    public List<TBPMethodDeclaration<REFERENCE>> getReactions() {
        return reactions;
    }

    public List<TBPThreadContainerNode<REFERENCE>> getThreads() {
        return threads;
    }

    public REFERENCE getComponent() {
        return componentInstance;
    }

    /**
     * All occurrences of unlimited reentrancy in all provisions are replaced
     * by limited reentrancy. This metod changes provisions in place. It does not
     * make copies.
     *
     * @param limit parameter of limited reentrancy
     */
    public void limitReentrancy(final int limit){
    	TBPVisitor<Object, REFERENCE> replaceReentrancy = new TBPDefaultVisitor<Object, REFERENCE>(){
    		@Override
    		public Object visitUnlimitedReentrancy(TBPUnlimitedReentrancy<REFERENCE> node) {

    			TBPProvisionNode<REFERENCE> child = (TBPProvisionNode<REFERENCE>) node.getChild();
    			TBPNode<REFERENCE> parent = node.getParent();
    			int idx = parent.getChildren().indexOf(node);

    			node.setChild(new TBPProvisionNull<REFERENCE>());
    			TBPNode<REFERENCE> replacement = new TBPLimitedReentrancy<REFERENCE>(child, limit);
    			parent.setChild(idx, replacement);

    			return null;
    		}

    	};

    	for(TBPProvisionNode<REFERENCE> provision: provisions){
			for(TBPNode<REFERENCE> node : provision){
				((TBPProvisionNode<REFERENCE>)node).visit(replaceReentrancy);
			}
		}
    }

    /**
     * Resolves references in this TBP specification using the given reference
     * transformer.
     *
     * @param transformer Reference transformer to be called back during
     *      the resolving proces.
     * @return The resolved TBPSpecification.
     */
	public <TO> TBPSpecification<TO> resolve(
            ReferenceTransformer<REFERENCE, TO> transformer) {

		// Resolving provsion part
		List<TBPProvisionContainerNode<TO>> resolved_provisions = 
                new ArrayList<TBPProvisionContainerNode<TO>>(provisions.size());
		TBPVisitor<TBPProvisionNode<TO>, REFERENCE> pvisitor =
			new ProvisionResolvingVisitor<REFERENCE, TO>(transformer);

		for (TBPProvisionContainerNode<REFERENCE> pnode : provisions) {
			resolved_provisions.add((TBPProvisionContainerNode<TO>)pnode.visit(pvisitor));
		}

		// Resolving threads
		List<TBPThreadContainerNode<TO>> resolved_threads = 
                new ArrayList<TBPThreadContainerNode<TO>>(threads.size());
		TBPVisitor<TBPImperativeNode<TO>, REFERENCE> ivisitor =
			new ImperativeResolvingVisitor<REFERENCE, TO>(transformer);

		for (TBPThreadContainerNode<REFERENCE> imperativeNode : threads) {
			resolved_threads.add(
					(TBPThreadContainerNode<TO>)imperativeNode.visit(ivisitor));
		}

        // Resolving reactions
		List<TBPMethodDeclaration<TO>> resolved_reactions = 
                new ArrayList<TBPMethodDeclaration<TO>>(reactions.size());
		for (TBPImperativeNode<REFERENCE> imperativeNode : reactions) {
			resolved_reactions.add((TBPMethodDeclaration<TO>) imperativeNode.visit(ivisitor));
		}

        // Resolving state variables
		List<TBPVariableDefinition<TO>> resolved_variables =
                new ArrayList<TBPVariableDefinition<TO>>(variables.size());
		for (TBPVariableDefinition<REFERENCE> variable : variables) {
			resolved_variables.add((TBPVariableDefinition<TO>) variable.visit(ivisitor));
		}

        return new TBPSpecification<TO>(
                transformer.resolveComponent(componentInstance),
                resolved_variables,
                resolved_provisions,
                resolved_reactions,
                resolved_threads);
    }

	public <RESULT> RESULT visit(TBPVisitor<RESULT, REFERENCE> visitor) {
		return visitor.visitSpecification(this);
	}
}
