/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.node.visitor;

import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.reference.MethodCall;
import org.ow2.dsrg.fm.tbplib.reference.MethodSignature;
import org.ow2.dsrg.fm.tbplib.reference.ReferenceTransformer;

import java.util.ArrayList;
import java.util.List;

/**
 * Translates trees of parsed imperative nodes into resolved imperative nodes.
 * 
 * @author caitt3am
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class ImperativeResolvingVisitor<FROM, TO> extends TBPCheckingVisitor<TBPImperativeNode<TO>, FROM> {

    private ReferenceTransformer<FROM, TO> referenceTransformer;

    /**
     * Creates visitor that transforms parsing nodes to resolve nodes in threads
     * and reaction sections
     *
     * @param referenceTransformer The reference transformer to be used for
     *      reference resolving.
     */
    public ImperativeResolvingVisitor(ReferenceTransformer<FROM, TO> referenceTransformer) {
        this.referenceTransformer = referenceTransformer;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedSwitch(TBPSwitch<FROM> node) {
        TBPImperativeNode<TO> result;

        List<TBPImperativeNode<TO>> b = new ArrayList<TBPImperativeNode<TO>>();
        for (TBPNode<FROM> ch : node.getChildren()) {
            b.add(ch.visit(this));
        }

        if (node.isNondeterministic()) {
            result = new TBPSwitch<TO>(b);
        } else {
            TBPValue<TO> rval = (TBPValue<TO>) node.getValue().visit(this);
            List<TO> cases = resolveList(node.getCases());
            result = new TBPSwitch<TO>(rval, b, cases);
        }
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedIf(TBPIf<FROM> node) {
        TBPImperativeNode<TO> result;

        ITBPCondition<TO> rcond = resolveCondition(node.getCondition());
        TBPImperativeNode<TO> rifb = node.getChild(0).visit(this);

        if (node.hasElse()) {
            TBPImperativeNode relseb = node.getChild(1).visit(this);
            result = new TBPIf<TO>(rcond, rifb, relseb);
        } else {
            result = new TBPIf<TO>(rcond, rifb, null);
        }
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedWhile(TBPWhile<FROM> node) {
        TBPImperativeNode<TO> result;

        result = new TBPWhile<TO>(
                resolveCondition(node.getCondition()),
                node.getChild().visit(this));
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedImperativeSequence(
            TBPImperativeSequence<FROM> node) {
        TBPImperativeNode<TO> result;

        result = new TBPImperativeSequence<TO>(
                node.getLeft().visit(this),
                node.getRight().visit(this));
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedAssignment(
            TBPAssignment<FROM> node) {

        TBPImperativeNode<TO> result;
        TBPValue<TO> rval = (TBPValue<TO>) node.getChild().visit(this);
        TO ref = referenceTransformer.resolveIdentifier(node.getIdf());

        result = new TBPAssignment<TO>(ref, rval);
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedEmit(TBPEmit<FROM> node) {
        TBPImperativeNode<TO> result;

        result = new TBPEmit<TO>(
                referenceTransformer.resolveMethodCall(node.getMethodCall(), true));

        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedImperativeNull(
            TBPImperativeNull<FROM> node) {
        TBPImperativeNode<TO> result = new TBPImperativeNull<TO>();
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedMethodDeclaration(
            TBPMethodDeclaration<FROM> node) {

        TO iface = referenceTransformer.resolveInterface(node.getInterface());
        TO method = referenceTransformer.resolveMethod(node.getMethod(), iface);
        MethodSignature ms =
                referenceTransformer.resolveMethodSignature(node.getSignature(), method);
        
        referenceTransformer.enterMethod(method);

        // Resolve local variables.
        List<TBPVariableDefinition<TO>> resolved_variables = getResolvedVariables(node.getLocalVariables());

        TBPImperativeNode<TO> result = new TBPMethodDeclaration(
                iface,
                method,
                ms, node.getChild().visit(this),
                resolved_variables);
        referenceTransformer.leaveMethod();

        result.setAnnotation(node.getAnnotation());
        return result;
    }

    /**
     * Resolves given variables and returns them.
     * @param variables variables to resolve
     * @return the list of resolved variables.
     */
	private List<TBPVariableDefinition<TO>> getResolvedVariables(
			List<TBPVariableDefinition<FROM>> variables) {
		// Resolving variables
		List<TBPVariableDefinition<TO>> resolved_variables =
                new ArrayList<TBPVariableDefinition<TO>>(variables.size());
		for (TBPVariableDefinition<FROM> variable : variables) {
			resolved_variables.add((TBPVariableDefinition<TO>) variable.visit(this));
		}
		return resolved_variables;
	}

    @Override
    public TBPImperativeNode<TO> visitParsedReturn(TBPReturn<FROM> node) {
        TBPImperativeNode<TO> result;
        FROM retVal = node.getIdf();
        if (retVal!=null){
        final TO ridf = referenceTransformer.resolveIdentifier(retVal);
            // NOTE: We do not check type here.
            result = new TBPReturn<TO>(ridf);
        }else {
            result = new TBPReturn<TO>();
        }


        result.setAnnotation(node.getAnnotation());

        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedSync(TBPSync<FROM> node) {
        TBPImperativeNode<TO> result;
        TO ref = referenceTransformer.resolveIdentifier(node.getMutex());

        result = new TBPSync<TO>(ref, node.getChild().visit(this));
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedWait(TBPWait<FROM> node) {
        TBPImperativeNode<TO> result;
        TO ref = referenceTransformer.resolveIdentifier(node.getMutex());
        result = new TBPWait<TO>(ref);
        result.setAnnotation(node.getAnnotation());
        return result;
    }
    
    
    @Override
    public TBPImperativeNode<TO> visitParsedValue(TBPValue<FROM> node) {
        TBPImperativeNode<TO> result;
        MethodCall<FROM> mc = node.getMethodCall();
        FROM varname = node.getVarname();

        assert mc != null || varname != null;

        if (mc == null) {
            TO ref = referenceTransformer.resolveIdentifier(varname);
            result = new TBPValue<TO>(ref);
        } else {
            MethodCall<TO> rmc = referenceTransformer.resolveMethodCall(mc, true);
            result = new TBPValue<TO>(rmc);
        }
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitParsedThreadContainerNode(
            TBPThreadContainerNode<FROM> node) {

        TO thread = referenceTransformer.resolveThread(node.getName());

        referenceTransformer.enterThread(thread);
        
        // Resolve local variables.
        List<TBPVariableDefinition<TO>> resolved_variables = getResolvedVariables(node.getLocalVariables());
        
        TBPImperativeNode<TO> result = new TBPThreadContainerNode<TO>(
                thread,
                node.getChild().visit(this),
                resolved_variables);
        referenceTransformer.leaveThread();

        result.setAnnotation(node.getAnnotation());
        return result;
    }

    @Override
    public TBPImperativeNode<TO> visitVariableDefinition(TBPVariableDefinition<FROM> node) {
        TBPVariableDefinition<TO> result;

        if (node.isMutex()) {
            result = new TBPVariableDefinition<TO>(
                    referenceTransformer.resolveIdentifier(node.getName()));
        } else {
            TO type = referenceTransformer.resolveType(node.getTypename());
            result = new TBPVariableDefinition<TO>(
                    type,
                    referenceTransformer.resolveVariable(node.getName()),
                    referenceTransformer.resolveConstant(node.getInitialValue(), type)
                    );
        }
        result.setAnnotation(node.getAnnotation());
        return result;
    }

    /**
     * Helper function for resolving a list of references.
     * 
     * @param source List of references of the FROM type.
     * @return List of references of the TO type.
     */
    private List<TO> resolveList(List<FROM> source) {
        List<TO> dest = new ArrayList<TO>(source.size());

        for (FROM element : source) {
            dest.add(referenceTransformer.resolveIdentifier(element));
        }
        return dest;
    }

    /**
     * Helper function for resolving a condition. This is not a method from
     * the TBPVisitor interface.

     * @param source Condition with references in the FROM type.
     * @return Condition with references in the TO type.
     */
    private ITBPCondition<TO> resolveCondition(ITBPCondition<FROM> source) {

        ConditionResolvingVisitor<FROM, TO> conditionVisitor = new ConditionResolvingVisitor<FROM, TO>(referenceTransformer);
        return source.visit(conditionVisitor);
        
        /*if (source.isDeterministic()) {
            TO lref = referenceTransformer.resolveIdentifier(source.getLeft());
            TO rref = referenceTransformer.resolveIdentifier(source.getRight());
            return new TBPCondition<TO>(lref, rref);
        } else {
            return new TBPCondition<TO>();
        }
        */
    }
}