/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.node.visitor;

import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.architecture.Method;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.reference.Constant;
import org.ow2.dsrg.fm.tbplib.reference.EnumerationType;
import org.ow2.dsrg.fm.tbplib.reference.MethodSignature;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * Prints parts of resolved hierarchy into given stream (System.out is default).
 * Use {@link #printComponentSpecification(TBPSpecification<Reference>, Component)}
 * to print the whole component.
 * 
 * @author caitt3am
 *
 */
public class PrintComponentVisitor<REFERENCE> implements TBPVisitor<Object, REFERENCE> {

    private CodeWriter o;
    private Component component;
    private static final String opening = " {";
    private static final String closing = "}";

    public PrintComponentVisitor() {
        o = new CodeWriter(System.out, true);
    }

    public PrintComponentVisitor(CodeWriter writer) {
        o = writer;
    }

    private void open() {
        o.println(opening);
        o.increaseIndent();
    }

    private void close() {
        o.decreaseIndent();
        o.println();
        o.print(closing);
    }

    private void appendSection(String name, Collection<? extends TBPNode<REFERENCE>> container) {
        o.append(name);
        open();
        for (TBPNode<REFERENCE> node : container) {
            node.visit(this);
        }
        o.decreaseIndent();
        o.println(closing);
    }

    private void appendType(EnumerationType t) {
        o.append(t.getName());
        o.append(" = ");
        o.append(opening);
        Iterator<Constant> it = t.getEnums().values().iterator();
        if (it.hasNext()) {
            o.append(it.next().getName());
        }
        while (it.hasNext()) {
            o.append(", ");
            o.append(it.next().getName());
        }
        o.print(closing);
        o.append(';');
    }

    private void appendCondition(ITBPCondition<REFERENCE> c) {
        o.append(c.toString());
    }

    public void printComponentSpecification(TBPSpecification<REFERENCE> tbp,
            Component component) {
        this.component = component;
        tbp.visit(this);
    }

    private void appendBinary(TBPProvisionBinaryNode<REFERENCE> node, String middle, boolean newline) {
        o.append('(');
        node.getLeft().visit(this);
        o.append(')');
        if (newline) {
            o.println();
            o.println(middle);
        } else {
            o.append(middle);
        }
        o.append('(');
        node.getRight().visit(this);
        o.append(')');
    }

    private void appendUnary(TBPProvisionUnaryNode<REFERENCE> node,
            String postfix) {
        o.append('(');
        node.getChild().visit(this);
        o.append(')').append(postfix);
    }

    private void appendBlock(TBPNode<REFERENCE> block) {
        open();
        block.visit(this);
        close();
    }

    @Override
    public Object visitParsedImperativeSequence(TBPImperativeSequence<REFERENCE> node) {
        TBPImperativeNode<REFERENCE> left =
                (TBPImperativeNode<REFERENCE>) node.getLeft();
        TBPImperativeNode<REFERENCE> right =
                (TBPImperativeNode<REFERENCE>) node.getRight();

        left.visit(this);
        o.println(";");
        right.visit(this);

        return null;
    }

    @Override
    public Object visitVariableDefinition(TBPVariableDefinition<REFERENCE> node) {
        if (node.isMutex()) {
            o.append("mutex ");
            o.append(node.getName().toString());
        } else {
            o.append(node.getTypename().toString()).append(' ');
            o.append(node.getName().toString()).append(" = ");
            o.append(node.getInitialValue().toString());
        }
        o.println(";");
        return null;
    }

    @Override
    public Object visitParsedSwitch(TBPSwitch<REFERENCE> node) {
        List<TBPNode<REFERENCE>> branches = node.getChildren();
        List<REFERENCE> cases = node.getCases();
        TBPImperativeNode<REFERENCE> db = node.getDefaultBranch();

        if (node.isNondeterministic()) {
            o.println("switch(?) {");
            for (TBPNode<REFERENCE> b : branches) {
                o.println("case: ");
                o.increaseIndent();

                b.visit(this);

                o.decreaseIndent();
                o.println();
            }
            o.print('}');

            return null;
        }

        o.append("switch(");
        node.getValue().visit(this);
        o.println(") {");
        for (int i = 0; i < cases.size(); i++) {
            String c = cases.get(i).toString();
            TBPImperativeNode<REFERENCE> b =
                    (TBPImperativeNode<REFERENCE>) branches.get(i);
            o.append("case ");
            o.append(c);
            o.println(": ");
            o.increaseIndent();

            b.visit(this);

            o.decreaseIndent();
            o.println();
        }
        if (db != null) {
            o.append("default: ");
            o.increaseIndent();

            db.visit(this);

            o.decreaseIndent();
            o.println();
        }
        o.print('}');

        return null;
    }

    @Override
    public Object visitParsedIf(TBPIf<REFERENCE> node) {
        o.append("if(");
        appendCondition(node.getCondition());
        o.append(')');
        open();
        node.getChild(0).visit(this);
        close();
        if (node.hasElse()) {
            o.append(" else ");
            open();
            node.getChild(1).visit(this);
            close();
        }
        return null;
    }

    @Override
    public Object visitParsedValue(TBPValue<REFERENCE> node) {
        if (node.isMethodCall()) {
            o.append('!');
            o.append(node.getMethodCall().toString());
        } else {
            o.print(node.getVarname().toString());
        }
        return null;
    }

    @Override
    public Object visitParsedReturn(TBPReturn<REFERENCE> node) {
        o.append("return ");
        o.append(node.getIdf().toString());
        return null;
    }

    @Override
    public Object visitParsedWhile(TBPWhile<REFERENCE> node) {
        o.append("while(");
        appendCondition(node.getCondition());
        o.append(")");
        appendBlock(node.getChild());
        return null;
    }

    @Override
    public Object visitParsedAssignment(TBPAssignment<REFERENCE> node) {
        o.append(node.getIdf().toString());
        o.append(" = ");
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedMethodDeclaration(TBPMethodDeclaration<REFERENCE> node) {
        MethodSignature<REFERENCE> signature = node.getSignature();
        o.print(node.getFullname());
        o.print(signature);
        appendBlock(node.getChild());
        o.println();
        return null;
    }

    @Override
    public Object visitParsedSync(TBPSync<REFERENCE> node) {
        o.append("sync(").append(node.getMutex().toString()).append(')');
        appendBlock(node.getChild());
        return null;
    }

    @Override
    public Object visitParsedWait(TBPWait<REFERENCE> node) {
        o.append("wait(").append(node.getMutex().toString()).append(')');        
        return null;
    }    
    
    @Override
    public Object visitParsedImperativeNull(TBPImperativeNull<REFERENCE> node) {
        o.append("NULL");
        return null;
    }

    @Override
    public Object visitParsedEmit(TBPEmit<REFERENCE> node) {
        o.append(node.getMethodCall().toString());
        return null;
    }

    @Override
    public Object visitParsedAlternative(TBPAlternative<REFERENCE> node) {
        appendBinary(node, "+", true);
        return null;
    }

    @Override
    public Object visitParsedParallel(TBPParallel<REFERENCE> node) {
        appendBinary(node, "|", true);
        return null;
    }

    @Override
    public Object visitParsedParallelOr(TBPParallelOr<REFERENCE> node) {
        appendBinary(node, "||", true);
        return null;
    }

    @Override
    public Object visitParsedSequence(TBPSequence<REFERENCE> node) {
        appendBinary(node, ";", false);
        return null;
    }

    @Override
    public Object visitParsedAccept(TBPAccept<REFERENCE> node) {
        //o.append('?');
        o.append(node.getMethodCall().toString());

        if (node.getReturnValue() != null) {
            o.append(':');
            o.append(node.getReturnValue().toString());
        }
        
        return null;
    }

    @Override
    public Object visitParsedNull(TBPProvisionNull<REFERENCE> node) {
        o.append("NULL");
        return null;
    }

    @Override
    public Object visitLimitedReentrancy(TBPLimitedReentrancy<REFERENCE> node) {
        appendUnary(node, "|");
        o.append(Integer.toString(node.getLimit()));
        return null;
    }

    @Override
    public Object visitParsedRepetition(TBPRepetition<REFERENCE> node) {
        appendUnary(node, "*");
        return null;
    }

    @Override
    public Object visitUnlimitedReentrancy(TBPUnlimitedReentrancy<REFERENCE> node) {
        appendUnary(node, "*");
        return null;
    }

    @Override
    public Object visitParsedProvisionContainerNode(TBPProvisionContainerNode<REFERENCE> node) {
        o.append(node.getName());
        appendBlock(node.getChild());

        o.append(" for {");
        boolean first = true;
        Set<REFERENCE> filter = node.getFilter();
        for (REFERENCE r: filter){
           if (!first) {
               o.append(", ");
           }
           first = false;
           if (r instanceof Method) {
                o.append(((Method)r).getFullname());
           } else {
                o.append(r.toString());
           }
        }
        o.append("}");

        o.println();
        return null;
    }

    @Override
    public Object visitParsedThreadContainerNode(TBPThreadContainerNode<REFERENCE> node) {
        o.append(node.getName().toString());
        appendBlock(node.getChild());
        o.println();
        return null;
    }

    @Override
    public Object visitSpecification(TBPSpecification<REFERENCE> node) {
        o.print("component ");
        o.print(node.getComponent().toString());
        open();

        o.append("types");
        open();
        for (EnumerationType type : component.getTypes().values()) {
            appendType(type);
            o.println();
        }
        o.decreaseIndent();
        o.println(closing);

        o.append("vars");
        open();
        for (TBPVariableDefinition<REFERENCE> v : node.getVariables()) {
            v.visit(this);
        }
        o.decreaseIndent();
        o.println(closing);

        appendSection("provisions", node.getProvisions());
        appendSection("reactions", node.getReactions());
        appendSection("threads", node.getThreads());

        close();
        o.println();
        return null;
    }
}
