/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.node.visitor;

import org.ow2.dsrg.fm.tbplib.TBPResolvingException;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.reference.MethodCall;
import org.ow2.dsrg.fm.tbplib.reference.ReferenceTransformer;

import java.util.HashSet;
import java.util.Set;

/**
 * Translates trees of parsed provision nodes into resolved provision nodes.
 * For provisions without a filter, the default filter (containing exactly
 * methods mentioned in the provision) will be assigned.
 * 
 * @author caitt3am
 */
public class ProvisionResolvingVisitor<FROM, TO> extends TBPCheckingVisitor<TBPProvisionNode<TO>, FROM> {

    private ReferenceTransformer<FROM, TO> transformer;
    private Set<TO> methods = null;

    public ProvisionResolvingVisitor(ReferenceTransformer<FROM, TO> transformer) {
        this.transformer = transformer;
    }

    @Override
    public TBPProvisionNode<TO> visitParsedAlternative(TBPAlternative<FROM> node) {
        return new TBPAlternative<TO>(
                node.getLeft().visit(this),
                node.getRight().visit(this));
    }

    @Override
    public TBPProvisionNode<TO> visitParsedParallel(TBPParallel<FROM> node) {
        return new TBPParallel<TO>(
                node.getLeft().visit(this),
                node.getRight().visit(this));
    }

    @Override
    public TBPProvisionNode<TO> visitParsedParallelOr(TBPParallelOr<FROM> node) {
        return new TBPParallelOr<TO>(
                node.getLeft().visit(this),
                node.getRight().visit(this));
    }

    @Override
    public TBPProvisionNode<TO> visitParsedSequence(TBPSequence<FROM> node) {
        return new TBPSequence<TO>(
                node.getLeft().visit(this),
                node.getRight().visit(this));
    }

    @Override
    public TBPProvisionNode<TO> visitParsedRepetition(TBPRepetition<FROM> node) {
        return new TBPRepetition<TO>(node.getChild().visit(this));
    }

    @Override
    public TBPProvisionNode<TO> visitLimitedReentrancy(TBPLimitedReentrancy<FROM> node) {
        return new TBPLimitedReentrancy<TO>(
                node.getChild().visit(this),
                node.getLimit());
    }

    @Override
    public TBPProvisionNode<TO> visitUnlimitedReentrancy(
            TBPUnlimitedReentrancy<FROM> node) {
        // NOTE: No check for unlimited reentrancy
        return new TBPUnlimitedReentrancy<TO>(
                node.getChild().visit(this));
    }

    @Override
    public TBPProvisionNode<TO> visitParsedProvisionContainerNode(
            TBPProvisionContainerNode<FROM> node) {


        Set<FROM> parsedFilter = node.getFilter();
        Set<TO> resolvedFilter = new HashSet<TO>();
        for (FROM item:parsedFilter){
            transformer.resolveFilterItem(item, resolvedFilter);

        }

        methods = resolvedFilter;

        TBPProvisionContainerNode<TO> provision = new TBPProvisionContainerNode<TO>(
                node.getChild().visit(this),
                node.getName());
        

        provision.setFilter(resolvedFilter);


        return provision;
    }

    @Override
    public TBPProvisionNode<TO> visitParsedNull(TBPProvisionNull<FROM> node) {
        return new TBPProvisionNull<TO>();
    }

    @Override
    public TBPProvisionNode<TO> visitParsedAccept(TBPAccept<FROM> node) {

        // NOTE: This resolving process is much simpler then the previous one,
        // no alternatives are made for unspecified parameters. This should be
        // performed by a separate visitor or during the automata construction.


        MethodCall<TO> mc = transformer.resolveMethodCall(node.getMethodCall(), false);

        TO m = mc.getMethod();

        if (!methods.contains(m)) {
            throw new TBPResolvingException("Method " + node.getFullname() + " used in a provision is not specified in for section of the provision.");
        }
        /*if (methods != null) {
            methods.add(mc.getMethod());
        } */
        
        return new TBPAccept<TO>(mc,
                node.getReturnValue() == null ? null : transformer.resolveIdentifier(node.getReturnValue()));
    }
}
