/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.node.visitor;

import org.ow2.dsrg.fm.tbplib.TBPResolvingException;
import org.ow2.dsrg.fm.tbplib.architecture.*;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.reference.*;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;

/**
 * Visitor collecting all symbol definitions. The architecture passed on creation
 * will be filled with symbols (Reference) during traversal.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class SymbolCollectingVisitor extends TBPCheckingVisitor<Object, String> {

    protected Architecture architecture;
    protected Component component;
    protected Method method;
    protected Autonomous thread;

    public SymbolCollectingVisitor(Architecture architecture) {
        this.architecture = architecture;
    }

    public void setComponent(Component component) {
        this.component = component;
    }

    @Override
    public Object visitParsedImperativeSequence(TBPImperativeSequence<String> node) {
        node.getLeft().visit(this);
        node.getRight().visit(this);
        return null;
    }

    @Override
    public Object visitVariableDefinition(TBPVariableDefinition<String> node) {
        if (method != null) {
            EnumerationType type = findType(node.getTypename());
            Constant init = type.getEnums().get(node.getInitialValue());
            if (init == null)
                throw new TBPResolvingException(
                        "Initial value '" + node.getInitialValue() +
                        "' is not element of the type of the local variable: " +
                        node.getName());
            method.addLocalVariable(
                    new LocalVariable(method, type,
                    node.getName(), init));
            return null;
        } else if (thread != null) {
            EnumerationType type = findType(node.getTypename());
            Constant init = type.getEnums().get(node.getInitialValue());
            if (init == null)
                throw new TBPResolvingException(
                        "Initial value '" + node.getInitialValue() +
                        "' is not element of the type of the local variable: " +
                        node.getName());
            thread.addLocalVariable(
                    new LocalVariable(thread, type,
                    node.getName(), init));
            return null;
        } else {
            EnumerationType type;
            Constant init;
            if (node.isMutex()) {
                type = EnumerationType.MUTEX_TYPE;
                init = EnumerationType.UNLOCKED;
            } else {
                type = findType(node.getTypename());
                init = type.getEnums().get(node.getInitialValue());
            }
            if (init == null)
                throw new TBPResolvingException(
                        "Initial value '" + node.getInitialValue() +
                        "' is not element of the type of the state variable: " +
                        node.getName());
            component.addStateVariable(
                    new StateVariable(type,
                    node.getName(), init, component));
            return null;
        }
    }

    @Override
    public Object visitParsedSwitch(TBPSwitch<String> node) {
        for (TBPNode<String> child : node.getChildren()) {
            child.visit(this);
        }
        if (node.isDeterministic()) {
            node.getValue().visit(this);
        }
        return null;
    }

    @Override
    public Object visitParsedIf(TBPIf<String> node) {
        node.getChild(0).visit(this);
        if (node.hasElse()) {
            node.getChild(1).visit(this);
        }
        return null;
    }

    @Override
    public Object visitParsedValue(TBPValue<String> node) {
        if (node.isMethodCall()) {
            // Collect this call
            prepareMethod(node.getMethodCall());
        }
        return null;
    }

    @Override
    public Object visitParsedReturn(TBPReturn<String> node) {
        return null;
    }

    @Override
    public Object visitParsedWhile(TBPWhile<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedAssignment(TBPAssignment<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedMethodDeclaration(TBPMethodDeclaration<String> node) {

        // Assure existing iface symbol.
        Interface iface = component.getProvidedInterfaces().get(node.getInterface());
        if (iface == null) {
            iface = new Interface(node.getInterface(), component, Interface.Direction.PROVIDED);
            component.addProvidedInterface(iface);
        }

        // Create new method - check conflicts
        if (iface.containsSymbol(node.getMethod()))
            throw new TBPResolvingException("Symbol conflict: " + node.getMethod());

        method = new Method(iface, node.getMethod());
        iface.addMethod(method);

        // Prepare the new signature
        MethodSignature<String> oldSignature = node.getSignature();
        EnumerationType returnType = findType(oldSignature.getReturnType());

        LinkedHashMap<String, String> oldParams = oldSignature.getParams();
        List<Parameter> newParamList = new ArrayList<Parameter>(oldParams.size());
        LinkedHashMap<Parameter, EnumerationType> newParams = new LinkedHashMap<Parameter, EnumerationType>();

        for (String param : oldParams.keySet()) {
            EnumerationType type = findType(oldParams.get(param));
            Parameter p = new Parameter(method, type, param);

            newParamList.add(p);
            newParams.put(p, type);
        }
        MethodSignature<Reference> newSignature =
                new MethodSignature<Reference>(newParams, returnType);
        method.setMethodSignature(newSignature);
        method.setParameters(newParamList);
        
        // Visit local variables
        List<TBPVariableDefinition<String>> localVars = node.getLocalVariables();
        for (TBPVariableDefinition<String> localVar: localVars) {
        	localVar.visit(this);
        }

        // Visit the implementation
        node.getChild().visit(this);
        method = null;
        return null;
    }

    @Override
    public Object visitParsedSync(TBPSync<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedWait(TBPWait<String> node) {        
        return null;
    }

    
    @Override
    public Object visitParsedImperativeNull(TBPImperativeNull<String> node) {
        return null;
    }

    @Override
    public Object visitParsedEmit(TBPEmit<String> node) {
        // Collect this call
        prepareMethod(node.getMethodCall());

        return null;
    }

    @Override
    public Object visitParsedProvisionContainerNode(TBPProvisionContainerNode<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitLimitedReentrancy(TBPLimitedReentrancy<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedAccept(TBPAccept<String> node) {
        // NOTE: Existence is checked but no method is added --> Using method
        // with no reaction is considered to be an error.
        Interface iface = component.getProvidedInterfaces().get(node.getInterface());
        if (iface == null) {
            throw new TBPResolvingException(
                    "Undefined provided interface used in provisions: " +
                    node.getInterface());
        }
        if (!iface.containsSymbol(node.getMethod()))
            throw new TBPResolvingException(
                    "Undefined method of a provided interface used in provisions: " +
                    node.getFullname());

        return null;
    }

    @Override
    public Object visitParsedAlternative(TBPAlternative<String> node) {
        node.getLeft().visit(this);
        node.getRight().visit(this);
        return null;
    }

    @Override
    public Object visitParsedNull(TBPProvisionNull<String> node) {
        return null;
    }

    @Override
    public Object visitParsedParallel(TBPParallel<String> node) {
        node.getLeft().visit(this);
        node.getRight().visit(this);
        return null;
    }

    @Override
    public Object visitParsedParallelOr(TBPParallelOr<String> node) {
        node.getLeft().visit(this);
        node.getRight().visit(this);
        return null;
    }

    @Override
    public Object visitParsedRepetition(TBPRepetition<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedSequence(TBPSequence<String> node) {
        node.getLeft().visit(this);
        node.getRight().visit(this);
        return null;
    }

    @Override
    public Object visitUnlimitedReentrancy(TBPUnlimitedReentrancy<String> node) {
        node.getChild().visit(this);
        return null;
    }

    @Override
    public Object visitParsedThreadContainerNode(TBPThreadContainerNode<String> node) {
        thread = new Autonomous(node.getName());
        component.addThread(thread);
        
        // Visit local variables
        List<TBPVariableDefinition<String>> localVars = node.getLocalVariables();
        for (TBPVariableDefinition<String> localVar: localVars) {
        	localVar.visit(this);
        }
        
        node.getChild().visit(this);
        thread = null;
        return null;
    }

    @Override
    public Object visitSpecification(TBPSpecification<String> node) {
        for(TBPMethodDeclaration<String> m : node.getReactions()) {
            m.visit(this);
        }
        for(TBPThreadContainerNode<String> t : node.getThreads()) {
            t.visit(this);
        }
        for(TBPVariableDefinition<String> vd : node.getVariables()) {
            vd.visit(this);
        }
        for(TBPProvisionContainerNode<String> p : node.getProvisions()) {
            p.visit(this);
        }
        return null;
    }

    private EnumerationType findType(String typeName) {
        if (typeName == null)
            return null;

        EnumerationType type = architecture.getTypes().get(typeName);

        if (type == null)
            throw new TBPResolvingException("Undefined type: " + typeName);

        return type;
    }

    private void prepareMethod(MethodCall<String> mc) {

        // Assure existing iface symbol.
        Interface iface = component.getRequiredInterfaces().get(mc.getInterface());
        if (iface == null) {
            iface = new Interface(mc.getInterface(), component, Interface.Direction.REQUIRED);
            component.addRequiredInterface(iface);
        }

        // Assure existing method
        if (!iface.containsSymbol(mc.getMethod())) {
            Method m = new Method(iface, mc.getMethod());
            iface.addMethod(m);
        }
    }
}
