/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.node.visitor;

import java.util.HashSet;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.ltsa.DynamicDFA;
import org.ow2.dsrg.fm.tbplib.ltsa.DynamicNFA;
import org.ow2.dsrg.fm.tbplib.node.TBPAccept;
import org.ow2.dsrg.fm.tbplib.node.TBPAlternative;
import org.ow2.dsrg.fm.tbplib.node.TBPLimitedReentrancy;
import org.ow2.dsrg.fm.tbplib.node.TBPParallel;
import org.ow2.dsrg.fm.tbplib.node.TBPParallelOr;
import org.ow2.dsrg.fm.tbplib.node.TBPProvisionContainerNode;
import org.ow2.dsrg.fm.tbplib.node.TBPProvisionNode;
import org.ow2.dsrg.fm.tbplib.node.TBPProvisionNull;
import org.ow2.dsrg.fm.tbplib.node.TBPRepetition;
import org.ow2.dsrg.fm.tbplib.node.TBPSequence;
import org.ow2.dsrg.fm.tbplib.node.TBPUnlimitedReentrancy;
import org.ow2.dsrg.fm.tbplib.provision.CallSequence;
import org.ow2.dsrg.fm.tbplib.provision.CounterLimit;
import org.ow2.dsrg.fm.tbplib.reference.Reference;

/** Translates provision expression to set of {@link CallSequence}s. 
 * 
 * <p>This visitor assumes that there is no <i>parallel</i> operator
 *    in a subtree of any <i>regular</i> operator in the visited tree.
 *    Nodes of types {@link TBPAlternative}, {@link TBPSequence} and {@link TBPRepetition} represent <i>regular</i> operators.
 *    Nodes of types {@link TBPParallel}, {@link TBPParallelOr}, {@link TBPLimitedReentrancy} and {@link TBPUnlimitedReentrancy} represent <i>parallel</i> operators.
 * </p>
 * 
 * <p>The upper parallel limit of a {@link CallSequence}
 *    (see {@link CallSequence#getUpperLimit() getUpperLimit()}, {@link CallSequence#setUpperLimit(int) setUpperLimit()})
 *    is computed as a product of upper parallel bounds of parallel operators on the path from call sequence
 *    (expression composed just from <i>regular</i> operators).
 *    Upper parallel bound of 
 *    <ul><li>{@link TBPParallel}, {@link TBPParallelOr} is 1</li>
 *        <li>{@link TBPLimitedReentrancy} is {@link TBPLimitedReentrancy#getLimit()}</li>
 *        <li>{@link TBPUnlimitedReentrancy} is {@link CallSequence#INFINITY infinity}</li>
 *    </ul>
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 *
 */
public class TranslateProvisionToCallSequencesVisitor<REFERENCE extends Reference>
	extends TBPCheckingVisitor<Set<CallSequence>, REFERENCE> {
	
	/** Event table for DFA symbols. */
	private EventTable eventTable;
	
	public TranslateProvisionToCallSequencesVisitor(EventTable eventTable) {
		this.eventTable = eventTable;
	}
	
    @Override
    public Set<CallSequence> visitParsedParallel(TBPParallel<REFERENCE> node) {
        Set<CallSequence> left = node.getLeft().visit(this);
        Set<CallSequence> right = node.getRight().visit(this);
        left.addAll(right);
        return left;
    }

    @Override
    public Set<CallSequence> visitParsedParallelOr(TBPParallelOr<REFERENCE> node) {
    	Set<CallSequence> left = node.getLeft().visit(this);
    	Set<CallSequence> right = node.getRight().visit(this);
    	left.addAll(right);
    	return left;
    }

    @Override
    public Set<CallSequence> visitLimitedReentrancy(TBPLimitedReentrancy<REFERENCE> node) {
    	//NOTE There's a possibility, that we will need to check, whether unary parallel operations contains just regular expressions
    	//     In such case we would have to check this here...
    	Set<CallSequence> child = node.getChild().visit(this);
    	for(CallSequence c : child) {
    		c.multiplyUpperLimitWith(node.getLimit());
    	}
    	return child;
    }
    
    @Override
    public Set<CallSequence> visitUnlimitedReentrancy(TBPUnlimitedReentrancy<REFERENCE> node) {
    	//NOTE There's a possibility, that we will need to check, whether unary parallel operations contains just regular expressions
    	//     In such case we would have to check this here...
    	Set<CallSequence> child = node.getChild().visit(this);
    	for(CallSequence c : child) {
    		c.setUpperLimit(CounterLimit.INFINITY);
    	}
    	return child;
    }

    @Override
    public Set<CallSequence> visitParsedProvisionContainerNode(TBPProvisionContainerNode<REFERENCE> node) {
    	return node.getChild().visit(this);
    }
    
    //
    // Following nodes are passed to TranslateRegularProvisionVisitor
    //
    
    /**
     * Calls {@link #visitRegularNode(TBPProvisionNode)} directly.
     */
    @Override
    public Set<CallSequence> visitParsedRepetition(TBPRepetition<REFERENCE> node) {
    	return visitRegularNode(node);
    }
    
    /**
     * Calls {@link #visitRegularNode(TBPProvisionNode)} directly.
     */    
    @Override
    public Set<CallSequence> visitParsedAlternative(TBPAlternative<REFERENCE> node) {
    	return visitRegularNode(node);
    }
    
    /**
     * Calls {@link #visitRegularNode(TBPProvisionNode)} directly.
     */    
    @Override
    public Set<CallSequence> visitParsedSequence(TBPSequence<REFERENCE> node) {
    	return visitRegularNode(node);
    }
    
    /**
     * Calls {@link #visitRegularNode(TBPProvisionNode)} directly.
     */
    @Override
    public Set<CallSequence> visitParsedAccept(TBPAccept<REFERENCE> node) {
    	return visitRegularNode(node);
    }
    
    /**
     * Calls {@link #visitRegularNode(TBPProvisionNode)} directly.
     */
    @Override
    public Set<CallSequence> visitParsedNull(TBPProvisionNull<REFERENCE> node) {
    	return visitRegularNode(node);
    }

    /** Creates {@link CallSequence} from expression (sub)tree.
     * 
     * Passes the job of creation of NFA from expression tree to {@link TranslateRegularProvisionVisitor}.
     * Then the result NFA is determinized using {@link DynamicNFA#determinize()}.
     * The deterministic FA is used to create {@link CallSequence}, with initial parallel limits (equal to 1).
     * 
     * @param node provision expression sub-tree
     * @return DFA of the (regular) sub-expression packed within a {@link TBPProvisionContainerNode}
     */
	protected Set<CallSequence> visitRegularNode(TBPProvisionNode<REFERENCE> node) {
		TranslateRegularProvisionVisitor<REFERENCE> dfaVisitor = new TranslateRegularProvisionVisitor<REFERENCE>(eventTable);
		DynamicNFA nfa = node.visit(dfaVisitor);
		DynamicDFA dfa = nfa.determinize();
		nfa = null;
		dfa = dfa.makeMinimal();
		Set<CallSequence> result = new HashSet<CallSequence>();
		ReconstructProvisionExpressionStringVisitor<REFERENCE> nameVisitor =  new ReconstructProvisionExpressionStringVisitor<REFERENCE>(); 		
		result.add(new CallSequence(node.visit(nameVisitor).toString(), dfa));
		return result;
	}
}
