/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.node.visitor;

import org.ow2.dsrg.fm.tbplib.architecture.Method;
import org.ow2.dsrg.fm.tbplib.event.EventCode;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.ltsa.DynamicNFA;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.reference.*;

import java.util.LinkedHashMap;
import java.util.List;

/**
 * Compiles provision into a nondeterministic automaton. The calls featuring
 * wildcards are rewritten using corresponding atomic events.
 *
 * @author caitt3am
 */
public class TranslateProvisionVisitor<REFERENCE extends Reference> extends TBPCheckingVisitor<DynamicNFA, REFERENCE> {

    private EventTable eventTable;

    public TranslateProvisionVisitor(EventTable eventTable) {
        this.eventTable = eventTable;
    }

    @Override
    public DynamicNFA visitParsedAlternative(TBPAlternative<REFERENCE> node) {
        return node.getLeft().visit(this).addAlternative(node.getRight().visit(this));
    }

    @Override
    public DynamicNFA visitParsedParallel(TBPParallel<REFERENCE> node) {
        return node.getLeft().visit(this).addParallel(node.getRight().visit(this));
    }

    @Override
    public DynamicNFA visitParsedParallelOr(TBPParallelOr<REFERENCE> node) {
        return node.getLeft().visit(this).addParallelOr(node.getRight().visit(this));
    }

    @Override
    public DynamicNFA visitParsedSequence(TBPSequence<REFERENCE> node) {
        return node.getLeft().visit(this).addSequence(node.getRight().visit(this));
    }

    @Override
    public DynamicNFA visitParsedAccept(TBPAccept<REFERENCE> node) {
        MethodCall<REFERENCE> mc = node.getMethodCall();
        MethodSignature<Reference> sig = ((Method) mc.getMethod()).getMethodSignature();
        LinkedHashMap<Reference,Reference> paramsDecl = sig.getParams();
        List<REFERENCE> params = mc.getParamDecl();
        int methodCode = eventTable.encodeMethod(mc.getFullname());

        DynamicNFA result = new DynamicNFA();

        int startState = result.allocateState();
        int middleState = result.allocateState();
        int endState = result.allocateState();

        result.setInitial(startState);
        result.markFinal(endState);

        // Create call edges
        if (paramsDecl.size() == 0) {
            // Method without parameters
            long code = eventTable.encodeMethodCall(methodCode);
            result.addEdge(startState, code, middleState);
        } else {
            // Method with parameters
            int i = 0;
            int[] doms = new int[paramsDecl.size()];
            byte[] codes = new byte[paramsDecl.size()];

            for (Reference type : sig.getParams().values()) {
                doms[i++] = ((EnumerationType) type).getEnums().size();
            }

            i = 0;
            for (REFERENCE val : params) {
                if (val == null || val instanceof ArbitraryValue) {
                    codes[i] = 0;
                } else {
                    doms[i] = 0;
                    codes[i] = ((Constant) val).getCode();
                }
                i++;
            }

            do {
                // Add edge
                long code = eventTable.encodeMethodCall(methodCode, codes);
                boolean carry = true;
                result.addEdge(startState, code, middleState);

                // Increment values + carry
                for (i = 0; i < codes.length; i++) {
                    // Increment
                    if (doms[i] == 0) // Skip any constant value
                    {
                        continue;
                    }

                    if (codes[i] + 1 < doms[i]) {
                        codes[i]++;
                        carry = false;
                        break;
                    } else {
                        // Carry
                        codes[i] = 0;
                    }
                }
                if (carry) break;
            } while (true);
        }

        // Create return edges
        if (node.getReturnValue() != null || sig.getReturnType() == null) {
            // Constant value or VOID
            long code = eventTable.encodeMethodReturn(methodCode,
                    node.getReturnValue() == null ? EventCode.VOID_VALUE : ((Constant) node.getReturnValue()).getCode());
            result.addEdge(middleState, code, endState);
        } else {
            // Wildcard
            for (byte i = 0; i < ((EnumerationType) sig.getReturnType()).getEnums().size(); i++) {
                long code = eventTable.encodeMethodReturn(methodCode, i);
                result.addEdge(middleState, code, endState);
            }
        }

        return result;
    }

    @Override
    public DynamicNFA visitParsedNull(TBPProvisionNull<REFERENCE> node) {
        return DynamicNFA.createNull();
    }

    @Override
    public DynamicNFA visitLimitedReentrancy(TBPLimitedReentrancy<REFERENCE> node) {
        return node.getChild().visit(this).addLimitedReentrancy(node.getLimit());
    }

    @Override
    public DynamicNFA visitParsedRepetition(TBPRepetition<REFERENCE> node) {
        return node.getChild().visit(this).addRepetition();
    }

    @Override
    public DynamicNFA visitUnlimitedReentrancy(TBPUnlimitedReentrancy<REFERENCE> node) {
        throw new UnsupportedOperationException("UnlimitedReentrancy cannot be translated into DFA.");
    }

    @Override
    public DynamicNFA visitParsedProvisionContainerNode(TBPProvisionContainerNode<REFERENCE> node) {
        return node.getChild().visit(this);
    }
}