/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.parser;

import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.ow2.dsrg.fm.tbplib.TBPParsingException;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.node.ParserResult;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.util.ErrorCollector;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * 
 * Helper class that facilitates parsing TBP files and 
 * hides interaction with ANTLR.
 * 
 * @author caitt3am
 *
 */
public class ParserFacade {

    /**
     * Parses a single TBP specification from the file represented by the given
     * filename. A corresponding component is added to the given architecture.
     *
     * @param filename Filename of a file to be opened, read, and parsed.
     * @param architecture Architecture to which a component representing the
     *      parsed specification is to be added.
     * @return The newly created TBP specification
     * @throws java.io.IOException On IO errors with the input stream.
     * @throws org.antlr.runtime.RecognitionException On any ANTLR parsing
     *      problems.
     */
    public static ParserResult parseFile(String filename) throws IOException, RecognitionException {
    	//MethodsInfo.changeTBPSpecification(filename);
        return parseFile(new FileInputStream(filename));
    }

    /**
     * Parses a single TBP specification from the given file.
     * A corresponding component is added to the given architecture.
     *
     * @param file File to be opened, read, and parsed.
     * @param architecture Architecture to which a component representing the
     *      parsed specification is to be added.
     * @return The newly created TBP specification
     * @throws java.io.IOException On IO errors with the input stream.
     * @throws org.antlr.runtime.RecognitionException On any ANTLR parsing
     *      problems.
     */
    public static ParserResult parseFile(File file) throws IOException, RecognitionException {
        return parseFile(new FileInputStream(file));
    }

    /**
     * Parses a single TBP specification from the given input stream.
     * A corresponding component is added to the given architecture.
     *
     * @param stream Input stream to be read and parsed.
     * @param architecture Architecture to which a component representing the
     *      parsed specification is to be added.
     * @return The newly created TBP specification
     * @throws java.io.IOException On IO errors with the input stream.
     * @throws org.antlr.runtime.RecognitionException On any ANTLR parsing
     *      problems.
     */
    public static ParserResult parseFile(InputStream stream)
            throws IOException, RecognitionException {

        ANTLRInputStream input = new ANTLRInputStream(stream);
        TBPLexer lexer = new TBPLexer(input);
        ErrorCollector ec = new ErrorCollector();
        lexer.setErrorReporter(ec);
        CommonTokenStream tokens = new CommonTokenStream(lexer);
        TBPParser parser = new TBPParser(tokens);
        parser.setErrorReporter(ec);
        TBPParser.tbp_return ret = parser.tbp();

        if (!ec.noErrors()){
            throw new TBPParsingException(ec.reportErrors());
        }

        CommonTree t = (CommonTree) ret.getTree();
        
        CommonTreeNodeStream nodes = new CommonTreeNodeStream(t);

        TBPTree evaluator = new TBPTree(nodes);
        evaluator.setErrorReporter(ec);
        ParserResult pr = evaluator.component();
        
        if (!ec.noErrors()){
            throw new TBPParsingException(ec.reportErrors());
        }
        return pr;

    }
}
