/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.provision;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.ltsa.DynamicDFA;

/** Represents pure regular expression from provision.
 * 
 * <p>Retains DFA of the regular expression and upper parallel limit
 *    (i.e. maximal number of parallel executions of the regular expression).
 * </p>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class CallSequence implements DFAHandler, CounterLimit {
	private int upperParallelLimit;
	private DynamicDFA dfa;
	protected final Set<CallSequence> coveredSequences; 
	private String name;
	/** Compares {@link CallSequence} by it's upperParallelLimit.
	 * 
	 * <p>Note: this comparer imposes orderings that are inconsistent with equals.</p>
	 */	
	public static final Comparator<CallSequence> COMPARATOR = new Comparator<CallSequence>() {
		@Override
		public int compare(CallSequence o1, CallSequence o2) {
			if (o1.getUpperLimit() == o2.getUpperLimit()) {
				return 0;
			}
			if (o2.getUpperLimit() == INFINITY || o1.getUpperLimit() < o2.getUpperLimit()) {
				return -1;
			}
			if (o1.getUpperLimit() == INFINITY || o1.getUpperLimit() > o2.getUpperLimit()) {
				return 1;
			}			
			throw new java.lang.IllegalArgumentException("Unable to compare.");
		}
	};
	/** {@link MergingFactory} for {@link CallSequence}.
	 */
	public static final MergingFactory<CallSequence> MERGING_FACTORY = new MergingFactory<CallSequence>() {
		@Override
		public CallSequence createCombined(Iterable<CallSequence> source) {
			HashSet<CallSequence> coveredSequencesSet = new HashSet<CallSequence>();
			CallSequence aCovered = null;
			StringBuilder nameBuilder = new StringBuilder();
			int limit = 0, numberOfCovered = 0;
			for(CallSequence c: source) {
				++numberOfCovered;
				aCovered = c;
				coveredSequencesSet.add(c);
				if (nameBuilder.length() > 0) {
					nameBuilder.append(" OR ");
				}
				nameBuilder.append(c.getName());
				if (limit != INFINITY) {
					if (c.getUpperLimit() == INFINITY) {
						limit = INFINITY;
					} else {
						limit += c.getUpperLimit();
					}
				}
			}
			return numberOfCovered <= 1 ? aCovered : new CallSequence(nameBuilder.toString(), coveredSequencesSet, limit);
		}		
	};
	
	/** Creates {@link CallSequence} covering given set of sequences.
	 * 
	 * <p>Utilized by {@link CallSequence#MERGING_FACTORY}.</p
	 * 	
	 * @param name name
	 * @param coveredSequences covered sequences
	 * @param upperParallelLimit upper parallel limit
	 */
	protected CallSequence(String name, Set<CallSequence> coveredSequences, int upperParallelLimit) {
		this.coveredSequences = Collections.unmodifiableSet(coveredSequences);
		this.upperParallelLimit = 0;
		this.name = name != null ? name : "";
		this.dfa = null;
	}
	
	/** Creates new {@link CallSequence}.
	 * 
	 * @param name Name of the call sequence.
	 * @param dfa DFA of the call sequence.
	 * @param upperParallelLimit Upper parallel limit (maximal safe number of parallel executions).
	 */
	public CallSequence(String name, DynamicDFA dfa, int upperParallelLimit) {
		assert(upperParallelLimit == INFINITY || upperParallelLimit > 0);
		this.name = name == null ? "" : name;
		this.dfa = dfa;
		this.upperParallelLimit = upperParallelLimit;
		this.coveredSequences = Collections.emptySet();
	}
	
	/** Creates new {@link CallSequence} with upper parallel limit equal to 1.
	 * 
	 * @param name Name of the call sequence.
	 * @param dfa DFA of the call sequence.
	 */
	public CallSequence(String name, DynamicDFA dfa) {
		this(name, dfa, 1);
	}
	
	/** Returns upper parallel limit of this.
	 * 
	 * @return upper parallel limit of this.
	 */
	public int getUpperLimit() {
		return upperParallelLimit;
	}
	/** Multiplies upper parallel limit with given value.
	 * 
	 * <p>This operation is aware of {@link CallSequence#INFINITY infinity}.
	 * 
	 * @param value Multiplier.
	 */
	public void multiplyUpperLimitWith(int value) {
		assert(value == INFINITY || value > 0);
		if (this.upperParallelLimit != INFINITY) {
			if (value == INFINITY) {
				this.upperParallelLimit = INFINITY;
			} else {
				this.upperParallelLimit *= value;
			}
		}
	}
	/** Sets upper parallel limit.
	 * 
	 * @param newValue New value of the upper parallel limit.
	 */
	public void setUpperLimit(int newValue) {
		assert(newValue == INFINITY || newValue > 0);
		this.upperParallelLimit = newValue;
	}
	/** Returns DFA.
	 * 
	 * @return DFA
	 */
	public DynamicDFA getDFA() {
		return dfa;
	}
	/** Returns name of this.
	 * 
	 * @return name.
	 */
	public String getName() {
		return name;
	}
	
	@Override
	public String toString() {
		return "(" + getName() + " |<= " + getUpperLimit() + ")@" + this.hashCode();
	}	
}
