/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.provision;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.ow2.dsrg.fm.tbplib.event.EventCode;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTable.MethodEncoder;

/**
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 *
 */
public class CompactProvisionAutomaton<SYMBOL extends CounterLimit> implements ObservingProvisionAutomaton<SYMBOL> {
	
	private static final int CALL = 0;
	private static final int RETURN = 1;
	
	private int initialState;
	private CounterLimit [] symbols;
	private boolean [] finalFlags;
	private boolean [] methodFilter;
    private int[] methodOffsets;
    

    // state X method X call/return X call/return code -> state 
	private Transition<SYMBOL> [][][][] transitions;
	private EventTable eventTable;
		
	protected CompactProvisionAutomaton() {
	}
	
	@Override
	public int initialState() {
		return this.initialState;
	}

	@Override
	public boolean isFinal(int stateId) {
		return this.finalFlags[stateId];
	}

	@Override
	public int getNumberOfState() {
		return this.finalFlags.length;
	}

	@Override
	public Transition<SYMBOL> getTransition(int state, long event) {
		Transition<SYMBOL>[][][] stateTransitions = transitions[state];
		if (stateTransitions == null) {
			return null;
		}
		
		int methodIdx = methodOffsets[EventCode.decodeMethod(event)];
		Transition<SYMBOL>[][] methodTransitions = stateTransitions[methodIdx];
		if (methodTransitions == null) {
			return null;
		}
		
		Transition<SYMBOL>[] eventTransitions;
		int eventOffset;
		switch(EventCode.decodeEventType(event)) {
		case METHOD_CALL:
			eventTransitions = methodTransitions[CALL];
			eventOffset = getEventOffsetFromEventCode(EventCode.decodeParameters(event));
			break;
		case METHOD_RETURN:
			eventTransitions = methodTransitions[RETURN];
			eventOffset = getEventOffsetFromEventCode(EventCode.decodeReturnValue(event));
			break;
		default:
			eventTransitions = null;
			eventOffset = -1;
		}
		if (eventTransitions == null) {
			return null;
		}
		
		return eventTransitions.length > 0 ? eventTransitions[eventOffset] : null;
	}

	@Override
	public Map<Long, Transition<SYMBOL>> getTransitions(int state) {
		Transition<SYMBOL>[][][] stateTransitions = transitions[state];
		if (stateTransitions != null) {
			HashMap<Long, Transition<SYMBOL>> transitionsMap = new HashMap<Long, Transition<SYMBOL>>();
			for(int methodOffset = 0; methodOffset < stateTransitions.length; ++methodOffset) {
				if (stateTransitions[methodOffset] != null) {
					int methodCode = getMethodCodeFromOffset(methodOffset);
					assert(methodOffsets[methodCode] == methodOffset);
					Transition<SYMBOL>[] eventTransitions = stateTransitions[methodOffset][CALL];
					if (eventTransitions != null) {
						for(int paramCode = 0; paramCode < eventTransitions.length; ++paramCode) {
							Transition<SYMBOL> transition = eventTransitions[paramCode];
							if (transition != null) {
								transitionsMap.put(
										EventCode.encodeMethodCall(methodCode, getEventCodeFromEventOffset(paramCode)),
										transition);
							}
						}
					}
					eventTransitions = stateTransitions[methodOffset][RETURN];
					if (eventTransitions != null) {
						for(int returnCode = 0; returnCode < eventTransitions.length; ++returnCode) {
							Transition<SYMBOL> transition = eventTransitions[returnCode];
							if (transition != null) {
								transitionsMap.put(
										EventCode.encodeMethodReturn(methodCode, getEventCodeFromEventOffset(returnCode)),
										transition);
							}
						}
					}
				}
			}
			return transitionsMap;
		} else {
			return Collections.emptyMap();
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public SYMBOL getStateSymbol(int state) {
		return (SYMBOL)symbols[state];
	}
	
	@Override
	public boolean observesEvent(long event) {
		if (EventCode.isObservable(event)) {
			return observesMethod(EventCode.decodeMethod(event));
		}
		return false;
	}

	@Override
	public boolean observesMethod(int method) {
		return methodFilter[method];
	}

	// DFA
	@Override
	public int nextState(int stateId, long event) {
		return this.getTransition(stateId, event).getTargetState();
	}

	// DFA
	@Override
	public Map<Long, Integer> getEdges(int state) {
		Map<Long, Integer> edges = new HashMap<Long, Integer>();
		Map<Long, Transition<SYMBOL>> transitions = this.getTransitions(state);
		if (transitions != null) {
			for(Map.Entry<Long, Transition<SYMBOL>> m : transitions.entrySet()) {
				edges.put(m.getKey(), m.getValue().getTargetState());
			}
		}
		return edges;
	}
	
		
	@SuppressWarnings("unchecked")
	public static <ORIGINAL_SYMBOL extends CounterLimit, SYMBOL extends CounterLimit> CompactProvisionAutomaton<SYMBOL> compactProvisionAutomaton(
			ProvisionAutomaton<ORIGINAL_SYMBOL> a,
			SymbolTransformer<ORIGINAL_SYMBOL, SYMBOL> t,
			EventTable e) {
		CompactProvisionAutomaton<SYMBOL> cpa = new CompactProvisionAutomaton<SYMBOL>();
		int numberOfStates = a.getNumberOfState();
		
		cpa.eventTable = e;
		cpa.initialState = a.initialState();
		cpa.symbols = new CounterLimit[numberOfStates];
		cpa.finalFlags = new boolean[numberOfStates];
		cpa.methodFilter = new boolean[cpa.eventTable.getNumberOfMethods()];
		cpa.methodOffsets = new int[cpa.eventTable.getNumberOfMethods()];
		cpa.transitions = (Transition<SYMBOL>[][][][])new Transition<?>[numberOfStates][][][];
		
		Map<Long, Transition<ORIGINAL_SYMBOL>> [] transitionMaps =
			(Map<Long, Transition<ORIGINAL_SYMBOL>>[])new Map<?, ?>[numberOfStates];
		
		int observedMethods = 0;		
		for(int stateId = 0; stateId < numberOfStates; ++stateId) {
			transitionMaps [stateId] = a.getTransitions(stateId) ;
			for(Map.Entry<Long, Transition<ORIGINAL_SYMBOL>> tEntry : transitionMaps[stateId].entrySet()) {

				int methodCode = EventCode.decodeMethod(tEntry.getKey());
				// register method
				if (!cpa.methodFilter[methodCode]) {
					cpa.methodFilter[methodCode] = true;	
					cpa.methodOffsets[methodCode] = observedMethods++;
				}
			}
		}
		
		for(int stateId = 0; stateId < numberOfStates; ++stateId) {
			cpa.finalFlags[stateId] = a.isFinal(stateId);
			cpa.symbols[stateId] = t.transform(a.getStateSymbol(stateId));
			
			if (transitionMaps[stateId] != null && !transitionMaps[stateId].isEmpty()) {
				cpa.transitions[stateId] = (Transition<SYMBOL>[][][])new Transition<?>[observedMethods][][];
				for(Map.Entry<Long, Transition<ORIGINAL_SYMBOL>> tEntry : transitionMaps[stateId].entrySet()) {
					long eventCode = tEntry.getKey();
					int methodCode = EventCode.decodeMethod(eventCode);
					int methodIdx = cpa.methodOffsets[methodCode];
					MethodEncoder menc = cpa.eventTable.getEncoder(methodCode);
					
					if (cpa.transitions[stateId][methodIdx] == null) {
						cpa.transitions[stateId][methodIdx] = (Transition<SYMBOL>[][])new Transition<?>[2][];
					}
					
					switch (EventCode.decodeEventType(eventCode)) {
					case METHOD_CALL:
						if (cpa.transitions[stateId][methodIdx][CALL] == null) {
							int numberOfMethodCalls = menc.getNumberOfCallEvents();
							cpa.transitions[stateId][methodIdx][CALL] = (Transition<SYMBOL>[])new Transition<?>[(numberOfMethodCalls <= 0 ? 0 : numberOfMethodCalls) + 1]; // reserve extra space for VOID
						}
						int paramCode = EventCode.decodeParameters(eventCode);
						int paramOffset = getEventOffsetFromEventCode(paramCode);
						cpa.transitions[stateId][methodIdx][CALL][paramOffset] = transformTransition(tEntry.getValue(), t);  
						
						break;
					case METHOD_RETURN:
						if (cpa.transitions[stateId][methodIdx][RETURN] == null) {
							int numberOfReturns = menc.getNumberOfReturnEvents();
							cpa.transitions[stateId][methodIdx][RETURN] = (Transition<SYMBOL>[])new Transition<?>[(numberOfReturns <= 0 ? 0 : numberOfReturns) + 1]; // reserve extra space for VOID
						}
						int returnCode = EventCode.decodeReturnValue(eventCode);
						int returnOffset = getEventOffsetFromEventCode(returnCode);
						cpa.transitions[stateId][methodIdx][RETURN][returnOffset] = transformTransition(tEntry.getValue(), t);
						
						break;
					default:
						throw new IllegalArgumentException("Only method call or return from method can be observed by provision automaton.");
					}
				}
			} 
		}
		
		return cpa;
	}
	
	private static int getEventOffsetFromEventCode(int code) {
		if (code == EventCode.VOID_VALUE) {
			return 0;
		} else {
			return code + 1;
		}
	}
	
	private static int getEventCodeFromEventOffset(int offset) {
		if (offset == 0) {
			return EventCode.VOID_VALUE;
		} else {
			return offset - 1;
		}
	}
	
	private static <SYMBOL, ORIGINAL_SYMBOL> Transition<SYMBOL> transformTransition(
			Transition<ORIGINAL_SYMBOL> transition,
			SymbolTransformer<ORIGINAL_SYMBOL, SYMBOL> transformer)
	{
		Collection<Transition.OutputSymbol<ORIGINAL_SYMBOL>> origSymbols = transition.getOutputSymbols();
		ArrayList<Transition.OutputSymbol<SYMBOL>> symbols = new ArrayList<Transition.OutputSymbol<SYMBOL>>(origSymbols.size());
		for(Transition.OutputSymbol<ORIGINAL_SYMBOL> os : origSymbols) {
			SYMBOL transformedSymbol = transformer.transform(os.getSymbol());
			symbols.add(new Transition.OutputSymbol<SYMBOL>(transformedSymbol, os.getDirection()));
		}
		return new Transition<SYMBOL>(transition.getTargetState(), symbols);
	}
		
	private int getMethodCodeFromOffset(int offset) {
		int i;
		for(i = 0; !methodFilter[i] || offset != methodOffsets[i]; ++i) { }// if the offset is not here, let IndexOutOfBounds be thrown
		return i;
	}		
}