/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.provision;

import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import org.ow2.dsrg.fm.tbplib.ltsa.DFA;

/** Dynamic implementation of {@link ProvisionAutomaton}.
 * 
 * <p>This implementation serves for building automaton (this)
 *    from several {@link CallSequence}s.
 *    It's supposed to be compacted before using for fast traversal. 
 * </p>
 * 
 * <p>The process of construction of this automata from {@link CallSequence}s
 *    (subsequently from {@link DFA}s) is similar to process of determinization
 *    of several DFAs combined using alternative.
 *    In addition, the result is enriched with state symbols (CSQ) and transition
 *    symbols (CSQ coupled with direction UP/DOWN). 
 * </p> 
 * 
 * <p>The state symbols are formed by original DFAs and their states, which
 *    form the resulting state.
 *    The transition symbols are formed by state symbols coupled with the
 *    local transition context.
 * </p>
 *    
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt
 */
public class DynamicProvisionAutomaton<CSQ extends DFAHandler> implements
		ProvisionAutomaton<CSQ> {
	
	private ArrayList< State<CSQ> > states;
	private State<CSQ> initialState;
	private Collection< State<CSQ> > finalStates;
	private Comparator<CSQ> comparator;
	
	/** Builds new instance from several {@link CallSequence}s.
	 * 
	 * @param callSequences Collection of {@link DFAHandler} extensions as a source of the automaton. 
	 * @param factory {@link MergingFactory} for creating state symbols.
	 * @param comparator {@link Comparator} for comparing state symbols.
	 */
	public DynamicProvisionAutomaton(Collection<CSQ> callSequences, MergingFactory<CSQ> factory, Comparator<CSQ> comparator) {
		this.states = new ArrayList<State<CSQ>>();
		this.finalStates = new ArrayList<State<CSQ>>();		
		this.initialState = null;
		this.comparator = comparator;
		buildFromDFAs(callSequences, factory);
	}
				
	@Override
	public Map<Long, Transition<CSQ>> getTransitions(int stateId) {
		State<CSQ> state = getState(stateId);
		Map<Long, Transition<CSQ>> result = new HashMap<Long, Transition<CSQ>>(state.getEdges().size());
		for(Long key : state.getEdges().keySet()) {
			result.put(key, getTransition(stateId, key.longValue()));
		}
		return result;
	}

	@Override
	public CSQ getStateSymbol(int state) {
		return getState(state).getSource();
	}

	@Override
	public Transition<CSQ> getTransition(int stateId, long event) {
		State<CSQ> sourceState = getState(stateId);
		int targetStateId = sourceState.getEdges().get(event);
		if (targetStateId == DFA.NO_SUCH_TRANSITION) {
			return null;
		}
		State<CSQ> targetState = getState(targetStateId);
		LinkedList<Transition.OutputSymbol<CSQ>> outputSymbols = new LinkedList<Transition.OutputSymbol<CSQ>>();
		
		if (sourceState.id == initialState()) {
			outputSymbols.add(new Transition.OutputSymbol<CSQ>(targetState.getSource(), Transition.OutputSymbolDirection.Up));
			if (isFinal(targetState)) {
				outputSymbols.add(new Transition.OutputSymbol<CSQ>(targetState.getSource(), Transition.OutputSymbolDirection.Down));				
			}			
		} else if (isFinal(sourceState) || isFinal(targetState)) {
			if (isFinal(sourceState)) {
				outputSymbols.add(new Transition.OutputSymbol<CSQ>(targetState.getSource(), Transition.OutputSymbolDirection.Up));				
			}
			if (isFinal(targetState)) {
				outputSymbols.add(new Transition.OutputSymbol<CSQ>(sourceState.getSource(), Transition.OutputSymbolDirection.Down));								
			}
		} else if (comparator.compare(sourceState.getSource(), targetState.getSource()) != 0 ) {
			// source is greater than target
			outputSymbols.add(new Transition.OutputSymbol<CSQ>(sourceState.getSource(), Transition.OutputSymbolDirection.Down));
			outputSymbols.add(new Transition.OutputSymbol<CSQ>(targetState.getSource(), Transition.OutputSymbolDirection.Up));						
		}				
		return new Transition<CSQ>(targetStateId, outputSymbols);
	}
	
	@Override
	public int getNumberOfState() {
		return states.size();
	}

	@Override
	public int initialState() {
		return initialState.getId();
	}

	@Override
	public boolean isFinal(int stateId) {
		State<CSQ> state = getState(stateId);
		return isFinal(state);
	}
	
	@Override
	public Map<Long, Integer> getEdges(int state) {
		Map<Long, Transition<CSQ>> transitions = getTransitions(state);
		Map<Long, Integer> result = new HashMap<Long, Integer>(transitions.size());
		for(Long key : transitions.keySet()) {
			result.put(key, transitions.get(key).getTargetState());
		}
		return result;
	}

	@Override
	public int nextState(int stateId, long event) {
		Transition<CSQ> t = getTransition(stateId, event);
		return t == null ? DFA.NO_SUCH_TRANSITION : t.getTargetState(); 
	}
	
	static protected class SourceStatePair<SOURCE extends DFAHandler> {
		private DFA dfa;
		private SOURCE source;
		private int state;
		protected SourceStatePair(SOURCE source, int state) {
			this.source = source; 
			this.dfa = source.getDFA();
			this.state = state;
		}
		protected DFA getDFA() {
			return dfa;
		}
		protected SOURCE getSource() {
			return source;
		}
		protected int getState() {
			return state;
		}
		
		@Override
		public int hashCode() {
			return dfa != null ? dfa.hashCode() : state;
		}
		@Override
		public boolean equals(Object that) {
			return	(this == that)
					||
					(	(that instanceof SourceStatePair<?>)
						&&
						this.hashCode() == that.hashCode()
						&&
						(this.dfa == ((SourceStatePair<?>)that).dfa)
						&&
						(this.state == ((SourceStatePair<?>)that).state)
					);
		}
	}
	
	static protected class State<SOURCE extends DFAHandler> {
		private int id;
		private Set<SourceStatePair<SOURCE>> coveredStates;
		private SOURCE combinedSource;
		private Map<Long, Integer> edges;
		protected State(int id, Set<SourceStatePair<SOURCE>> coveredStates, SOURCE source) {
			this.id = id;
			this.edges = new HashMap<Long, Integer>();
			this.coveredStates = new HashSet<SourceStatePair<SOURCE>>(coveredStates);
			this.combinedSource = source;
		}
		protected int getId() {
			return id;
		}
		protected Map<Long, Integer> getEdges() {
			return edges;
		}
		protected void addEdge(long event, int target) {
			if (edges.containsKey(event)) {
				throw new UnsupportedOperationException("Duplicate event.");
			}
			edges.put(event, target);
		}
		protected Set<SourceStatePair<SOURCE>> getCoveredStates() {
			return coveredStates;
		}
		protected SOURCE getSource() {
			return combinedSource;
		}
		protected void copyEdges(State<?> state) {
			edges.clear();
			edges.putAll(state.getEdges());
		}
	}
	
	static protected class SourceSelectIterable<SOURCE extends DFAHandler> implements Iterable<SOURCE> {
		protected class SourceSelectIterator implements Iterator<SOURCE> {
			private Iterator<SourceStatePair<SOURCE>> underlayedIterator;
			protected SourceSelectIterator(Iterable<SourceStatePair<SOURCE>> underlayedIterable) {
				this.underlayedIterator = underlayedIterable.iterator();
			}
			@Override
			public boolean hasNext() {
				return underlayedIterator.hasNext();
			}
			@Override
			public SOURCE next() {
				SourceStatePair<SOURCE> wrapper = underlayedIterator.next();
				return wrapper.getSource();
			}
			@Override
			public void remove() {
				throw new UnsupportedOperationException("Removal not supported.");
			}		
		}
		private Iterable<SourceStatePair<SOURCE>> underlayedIterable;
		protected SourceSelectIterable(Iterable<SourceStatePair<SOURCE>> iterable) {
			this.underlayedIterable = iterable;
		}
		@Override
		public Iterator<SOURCE> iterator() {
			return new SourceSelectIterator(this.underlayedIterable);
		}
	}
	
	private State<CSQ> findStateByCoveredStates(Set<SourceStatePair<CSQ>> coveredStates) {
		for(State<CSQ> state: states) {
			if (state.getCoveredStates().equals(coveredStates)) {
				return state;
			}
		}
		return null;
	}

	
	private void buildFromDFAs(Collection<CSQ> sources, MergingFactory<CSQ> factory) {
		LinkedList<State<CSQ>> stack = new LinkedList<State<CSQ>>();
		
		// create first state and push it onto stack
		Set<SourceStatePair<CSQ>> coveredStates = new HashSet<SourceStatePair<CSQ>>();
		boolean coversFinal = false;
		boolean containsInitialCycle = false;
		for(CSQ s : sources) {
			
			SourceStatePair<CSQ> tmpSSP = new SourceStatePair<CSQ>(s, s.getDFA().initialState());
			coversFinal = coversFinal || tmpSSP.getDFA().isFinal(tmpSSP.getState());
			coveredStates.add(tmpSSP);
		}
		State<CSQ> newState = new State<CSQ>(states.size(), coveredStates, factory.createCombined(new SourceSelectIterable<CSQ>(coveredStates)));
		State<CSQ> startState = newState;
		stack.add(newState);
		states.add(newState);
		if (coversFinal) {
			finalStates.add(newState);
		}
				
		// create other states and add transitions...
		Set<Long> availableEvents = new HashSet<Long>();
		while(!stack.isEmpty()) {
			State<CSQ> currentState = stack.pop();
			availableEvents.clear();
			// collect available events
			for(SourceStatePair<CSQ> dfasp : currentState.getCoveredStates()) {
				availableEvents.addAll( dfasp.getDFA().getEdges(dfasp.getState()).keySet() );
			}
			// create or find viable targets and add appropriate edges
			for(Long event : availableEvents) {
				coveredStates.clear();
				coversFinal = false;
				
				// create covered states
				for(SourceStatePair<CSQ> ssp: currentState.getCoveredStates()) {
					int newPartStateId = ssp.getDFA().nextState(ssp.getState(), event);
					if (newPartStateId != DFA.NO_SUCH_TRANSITION) {
						SourceStatePair<CSQ> tmpSSP = new SourceStatePair<CSQ>(ssp.getSource(), newPartStateId);
						coversFinal = coversFinal || tmpSSP.getDFA().isFinal(tmpSSP.getState());
						coveredStates.add(tmpSSP);
					}
				}
				
				// create or find state
				newState = findStateByCoveredStates(coveredStates);
				if (newState == null) {
					newState = new State<CSQ>(states.size(), coveredStates, factory.createCombined(new SourceSelectIterable<CSQ>(coveredStates)));
					stack.add(newState);
					states.add(newState);
					if (coversFinal) {
						if (isFinal(currentState)) {
							throw new InvalidParameterException("DFAs are not capable to be used as source for provision automata.");
						}
						finalStates.add(newState);
					}
				}
				
				// add edge
				currentState.addEdge(event, newState.getId());
				if (newState == startState) {
					containsInitialCycle = true;
				}
			}
		}
		
		// now we have deterministic machine, so "pop" the initial state if necessary 
		if (containsInitialCycle) {
			initialState = new State<CSQ>(states.size(), startState.getCoveredStates(), factory.createCombined(new SourceSelectIterable<CSQ>(coveredStates)));
			initialState.copyEdges(startState);
			if (isFinal(startState)) {
				finalStates.add(initialState);
			}
		} else {
			initialState = startState;
		}		
	}
	
	private State<CSQ> getState(int stateId) {
		return states.get(stateId);
	}
	private boolean isFinal(State<CSQ> state) {
		return finalStates.contains(state);		
	}	
	
}
