/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.provision.actuator.visitor;

import java.util.ArrayList;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.LTS;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSImpl;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPAtomic;
import org.ow2.dsrg.fm.tbplib.node.TBPImperativeNull;
import org.ow2.dsrg.fm.tbplib.node.TBPImperativeSequence;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;

/** Provides instantiation of <i>closed</i> and <i>limited</i> thread scheme into LTS.
 * 
 * <ul>
 * 	<li><i>Closed</i> scheme means that there are nonempty actions after the bunch of thread ends.</li>
 * 	<li><i>Limited</i> scheme means that there is certain (limited) number of threads.
 * </ul>
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class ClosedLimitedScheme implements ThreadCooperationScheme {
	
	@Override
	public <CODE_REFERENCE, REFERENCE> void instanciateScheme(
			VariableFactory<REFERENCE> variableFactory,
			List<LTSAHandler<CODE_REFERENCE, REFERENCE>> threadBodies,
			int atLeastExecuted,
			LTSAHandler<CODE_REFERENCE, REFERENCE> controllerLTS,
			List<LTSAHandler<CODE_REFERENCE, REFERENCE>> threads) {
		
		assert(threadBodies != null && !threadBodies.isEmpty());
		assert(controllerLTS.getEntryPoints().isEmpty() && controllerLTS.getEndPoints().isEmpty());
		
		// compute number of threads
		int numberOfThreads = 0;
		for(LTSAHandler<CODE_REFERENCE, REFERENCE> h : threadBodies) {
			numberOfThreads += h.getCopies();
		}
		
		// create variables
		VariableDescriptor<REFERENCE> exitFlag = variableFactory.getSharedStateVariable(EXIT_FLAG_NAME, 2);
		VariableDescriptor<REFERENCE> progressCounter = variableFactory.createStateVariable("_p", numberOfThreads + 1);
		VariableDescriptor<REFERENCE> minProgressCounter = variableFactory.createStateVariable("_m", atLeastExecuted + 1);
		
		assert(!progressCounter.getValues().isEmpty());
		assert(!minProgressCounter.getValues().isEmpty());
		
		fillController(controllerLTS, progressCounter, minProgressCounter);
		
		for(LTSAHandler<CODE_REFERENCE, REFERENCE> threadBody : threadBodies) {
			if (threadBody != null) {
				threads.add(createThread(threadBody, exitFlag, progressCounter, minProgressCounter));
			}
		}
	}
	
	private <CODE_REFERENCE, REFERENCE> void fillController(
			LTSAHandler<CODE_REFERENCE, REFERENCE> controller,
			VariableDescriptor<REFERENCE> progress,   /* domain >= 1 */
			VariableDescriptor<REFERENCE> minProgress /* domain >= 1 */) {
		
		// create transitions
		List<TBPAssignment<REFERENCE>> assignments = new ArrayList<TBPAssignment<REFERENCE>>(2);
		assignments.add(new TBPAssignment<REFERENCE>(progress.getIdentifier(), LTSAManipulationHelper.createTBPValue(progress.getValues().get(progress.getValues().size() - 1))));
		assignments.add(new TBPAssignment<REFERENCE>(minProgress.getIdentifier(), LTSAManipulationHelper.createTBPValue(minProgress.getValues().get(0))));
		LTSA<CODE_REFERENCE, REFERENCE> firstTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, null, assignments);
		LTSA<CODE_REFERENCE, REFERENCE> lastTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createFirstValueTBPCondition(progress), null);
		
		// connect each other and add them to LTSHandler
		LTSAManipulationHelper.connectAll(firstTransition, lastTransition);
		controller.addEntryPoint(firstTransition);
		controller.addEndPoint(lastTransition);
	}
	
	private <CODE_REFERENCE, REFERENCE> LTSAHandler<CODE_REFERENCE, REFERENCE> createThread(
			LTSAHandler<CODE_REFERENCE, REFERENCE> thread,
			VariableDescriptor<REFERENCE> exitFlag,   /* |domain| == 2 */
			VariableDescriptor<REFERENCE> progress,   /* |domain| >= 1 */
			VariableDescriptor<REFERENCE> minProgress /* |domain| >= 1 */) {
	
		
		// create transitions
		LTSA<CODE_REFERENCE, REFERENCE> enterTransition = 
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createFirstValueTBPCondition(progress).createNegation(), null);
		List<LTSA<CODE_REFERENCE, REFERENCE>> enterBodyTransitions =
			LTSAManipulationHelper.createVariableAdjustmentLTSATransitions(minProgress, 1, true);
		LTSA<CODE_REFERENCE, REFERENCE> leaveBodyTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>();
		LTSA<CODE_REFERENCE, REFERENCE> skipBodyTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createLastValueTBPCondition(minProgress), null);
		List<LTSA<CODE_REFERENCE, REFERENCE>> leaveTransitions =
			LTSAManipulationHelper.createVariableAdjustmentLTSATransitions(progress, -1, false);
		LTSA<CODE_REFERENCE, REFERENCE> repeatTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createFirstValueTBPCondition(progress), null);
		LTSA<CODE_REFERENCE, REFERENCE> exitTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createLastValueTBPCondition(exitFlag), null);
		
		// connect the transitions
		
		LTSAManipulationHelper.connectAll(enterTransition, enterBodyTransitions);
		LTSAManipulationHelper.connectAll(enterTransition, skipBodyTransition);
		
		LTSAManipulationHelper.connectAll(enterBodyTransitions, thread.getEntryPoints());
		
		LTSAManipulationHelper.connectAll(thread.getEndPoints(), leaveBodyTransition);
		
		LTSAManipulationHelper.connectAll(skipBodyTransition, leaveTransitions);
		LTSAManipulationHelper.connectAll(leaveBodyTransition, leaveTransitions);		

		LTSAManipulationHelper.connectAll(leaveTransitions, repeatTransition);
		
		LTSAManipulationHelper.connectAll(repeatTransition, enterTransition);
		LTSAManipulationHelper.connectAll(repeatTransition, exitTransition);
		
		// insert into LTSHandler
		
		LTSAHandler<CODE_REFERENCE, REFERENCE> ltsa = new LTSAHandler<CODE_REFERENCE, REFERENCE>();
		ltsa.setCopies(thread.getCopies());
		ltsa.addEntryPoint(enterTransition);
		ltsa.addEntryPoint(exitTransition);
		ltsa.addEndPoint(exitTransition);
		
		return ltsa;
	}	
	


	/** @deprecated Replaced by {@link #instanciateScheme(VariableFactory, List, int, LTSAHandler, List)} */ 
	@Override
	public <REFERENCE> void instanciateScheme(
			VariableFactory<REFERENCE> variableFactory,
			List<LTSHandler<REFERENCE>> threadBodies,
			int atLeastExecuted,
			LTSHandler<REFERENCE> controllerLTS,
			List<LTSHandler<REFERENCE>> threads) {
		
		assert(threadBodies != null && !threadBodies.isEmpty());
		assert(controllerLTS.getEntryPoints().isEmpty() && controllerLTS.getEndPoints().isEmpty());
		
		VariableDescriptor<REFERENCE> exitFlag = variableFactory.getSharedStateVariable(EXIT_FLAG_NAME, 2);
		VariableDescriptor<REFERENCE> progressCounter = variableFactory.createStateVariable("_p", threadBodies.size() + 1);
		VariableDescriptor<REFERENCE> minProgressCounter = variableFactory.createStateVariable("_m", atLeastExecuted + 1);
		
		assert(!progressCounter.getValues().isEmpty());
		assert(!minProgressCounter.getValues().isEmpty());
		
		fillController(controllerLTS, progressCounter, minProgressCounter);
		
		for(LTSHandler<REFERENCE> threadBody : threadBodies) {
			if (threadBody != null) {
				threads.add(createThread(threadBody, exitFlag, progressCounter, minProgressCounter));
			}
		}
	}

	/** @deprecated */
	private <REFERENCE> void fillController(
			LTSHandler<REFERENCE> controller,
			VariableDescriptor<REFERENCE> progress,   /* domain >= 1 */
			VariableDescriptor<REFERENCE> minProgress /* domain >= 1 */) {
		
		// create transitions
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> firstTransition = 
			LTSImpl.createInstance(	new TBPAtomic<REFERENCE>(
										new TBPImperativeSequence<REFERENCE>(
											new TBPAssignment<REFERENCE>(progress.getIdentifier(), LTSAManipulationHelper.createTBPValue(progress.getValues().get(progress.getValues().size() - 1))),
											new TBPAssignment<REFERENCE>(minProgress.getIdentifier(), LTSAManipulationHelper.createTBPValue(minProgress.getValues().get(0))))),
									null /* no guard */);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> lastTransition = 
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createFirstValueTBPCondition(progress));
		
		// connect each other and add them to LTSHandler
		LTSAManipulationHelper.connectAllLTSTransitions(firstTransition, lastTransition);
		controller.addEntryPoint(firstTransition);
		controller.addEndPoint(lastTransition);
	}
	
	/** @deprecated */
	private <REFERENCE> LTSHandler<REFERENCE> createThread(
			LTSHandler<REFERENCE> thread,
			VariableDescriptor<REFERENCE> exitFlag,   /* |domain| == 2 */
			VariableDescriptor<REFERENCE> progress,   /* |domain| >= 1 */
			VariableDescriptor<REFERENCE> minProgress /* |domain| >= 1 */) {
	
		
		// create transitions
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> enterTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createFirstValueTBPCondition(progress).createNegation());
		List<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> enterBodyTransitions =
			LTSAManipulationHelper.createVariableAdjustmentLTSTransitions(minProgress, 1);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> leaveBodyTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									null);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> skipBodyTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createLastValueTBPCondition(minProgress));
		List<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> leaveTransitions =
			LTSAManipulationHelper.createVariableAdjustmentLTSTransitions(progress, -1);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> repeatTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createFirstValueTBPCondition(progress));
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> exitTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createLastValueTBPCondition(exitFlag));
		
		// connect the transitions
		
		LTSAManipulationHelper.connectAllLTSTransitions(enterTransition, enterBodyTransitions);
		LTSAManipulationHelper.connectAllLTSTransitions(enterTransition, skipBodyTransition);
		
		LTSAManipulationHelper.connectAllLTSTransitions(enterBodyTransitions, thread.getEntryPoints());
		
		LTSAManipulationHelper.connectAllLTSTransitions(thread.getEndPoints(), leaveBodyTransition);
		
		LTSAManipulationHelper.connectAllLTSTransitions(skipBodyTransition, leaveTransitions);
		LTSAManipulationHelper.connectAllLTSTransitions(leaveBodyTransition, leaveTransitions);		

		LTSAManipulationHelper.connectAllLTSTransitions(leaveTransitions, repeatTransition);
		
		LTSAManipulationHelper.connectAllLTSTransitions(repeatTransition, enterTransition);
		LTSAManipulationHelper.connectAllLTSTransitions(repeatTransition, exitTransition);
		
		// insert into LTSHandler
		
		LTSHandler<REFERENCE> lts = new LTSHandler<REFERENCE>();
		lts.addEntryPoint(enterTransition);
		lts.addEntryPoint(exitTransition);
		lts.addEndPoint(exitTransition);
		
		return lts;
	}	
}
