/*
 * $Id$
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.provision.actuator.visitor;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.LTS;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSImpl;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPCondition;
import org.ow2.dsrg.fm.tbplib.node.TBPImperativeNull;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPValue;
import org.ow2.dsrg.fm.tbplib.node.guard.ITBPCondition;
import org.ow2.dsrg.fm.tbplib.node.guard.TBPEquals;

/** Provides set of methods for easier manipulation with LTS transitions.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class LTSAManipulationHelper {
	
	/** Creates {@link TBPValue} from REFERENCE.
	 * 
	 * @param <REFERENCE>
	 * @param valueRef REFERENCE
	 * @return {@link TBPValue}
	 */
	public static <REFERENCE> TBPValue<REFERENCE> createTBPValue(REFERENCE valueRef) {
		return new TBPValue<REFERENCE>(valueRef);
	}
	
	/** Creates {@link TBPCondition} that holds if given <i>variable</i> has its first (the least) value.
	 * 
	 * @param <REFERENCE>
	 * @param var variable
	 * @return {@link TBPCondition}
	 */
	public static <REFERENCE> ITBPCondition<REFERENCE> createFirstValueTBPCondition(VariableDescriptor<REFERENCE> var) {
		ITBPCondition<REFERENCE> result = null;
		if (var != null && var.getValues() != null && !var.getValues().isEmpty()) {
			result = new TBPEquals<REFERENCE>(var.getIdentifier(), var.getValues().get(0));
		}
		return result;
	}

	/** Creates {@link TBPCondition} that holds if given <i>variable</i> has its last (the greatest) value.
	 * 
	 * @param <REFERENCE>
	 * @param var variable
	 * @return {@link TBPCondition}
	 */
	public static <REFERENCE> ITBPCondition<REFERENCE> createLastValueTBPCondition(VariableDescriptor<REFERENCE> var) {
		ITBPCondition<REFERENCE> result = null;
		if (var != null && var.getValues() != null && !var.getValues().isEmpty()) {
			result = new TBPEquals<REFERENCE>(var.getIdentifier(), var.getValues().get(var.getValues().size() - 1));
		}
		return result;
	}
	
	/** Creates transitions that adjusts given variable linearly (adds given difference).
	 * 
	 * @param <REFERENCE>
	 * @param variable variable to be adjusted
	 * @param difference difference (additive)
	 * @param includeEdge include edge transitions?
	 *        <br/>e.g.: <i>difference = 2, domain = {0, 1 ,2}</i>; include transtions <i>(?v=1, v&larr;2)</i> and <i>(?v=2, v&larr;2 )</i>
	 * @return {@link List} of transitions
	 */
	public static <CODE_REFERENCE, REFERENCE> List<LTSA<CODE_REFERENCE, REFERENCE>> createVariableAdjustmentLTSATransitions(VariableDescriptor<REFERENCE> variable, int difference, boolean includeEdge) {
		LinkedList<LTSA<CODE_REFERENCE, REFERENCE>> result =
			new LinkedList<LTSA<CODE_REFERENCE, REFERENCE>>();
		int size = variable.getValues().size();
		for(int i = 0; i < size; ++i) {
			int targetIndex = i + difference;
			REFERENCE source = variable.getValues().get(i);
			REFERENCE target = null;
			if(targetIndex >= 0 && targetIndex < size) {
				target = variable.getValues().get(targetIndex);
			} else if (includeEdge && targetIndex < 0) {
				target = variable.getValues().get(0);
			} else if (includeEdge && targetIndex >= size) {
				target = variable.getValues().get(size - 1);
			}
			if (target != null) {
				ITBPCondition<REFERENCE> guard = new TBPEquals<REFERENCE>(variable.getIdentifier(), source);
				TBPAssignment<REFERENCE> assignment = new TBPAssignment<REFERENCE>(variable.getIdentifier(), createTBPValue(target));
				
				result.add(new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, guard, Collections.singletonList(assignment)));
			}
		}
		return result;
	}
	
	/** Connects all <i>successors</i> to the <i>predecessors</i> 
	 * 
	 * @param <REFERENCE>
	 * @param predecessors
	 * @param successors
	 */
	public static <CODE_REFERENCE, REFERENCE> void connectAll(Collection<LTSA<CODE_REFERENCE, REFERENCE>> predecessors, Collection<LTSA<CODE_REFERENCE, REFERENCE>> successors) {
		assert(!predecessors.isEmpty() && !successors.isEmpty());
		if (predecessors.size() > 1 && successors.size() > 1) {
			// many to many connection
			LTSA<CODE_REFERENCE, REFERENCE> bridge = new LTSAImpl<CODE_REFERENCE, REFERENCE>();
			for(LTSA<CODE_REFERENCE, REFERENCE> src : predecessors) {
				src.connectTransition(bridge);
			}
			for(LTSA<CODE_REFERENCE, REFERENCE> tgt : successors) {
				bridge.connectTransition(tgt);
			}			
		} else {
			// 1 to ? or ? to 1 connection
			for(LTSA<CODE_REFERENCE, REFERENCE> src : predecessors) {
				for(LTSA<CODE_REFERENCE, REFERENCE> tgt : successors) {
					src.connectTransition(tgt);
				}
			}
		}
	}

	/** Connects all <i>successors</i> to <i>predecessor</i>. 
	 * 
	 * @param <REFERENCE>
	 * @param predecessor
	 * @param successors
	 */
	public static <CODE_REFERENCE, REFERENCE> void connectAll(LTSA<CODE_REFERENCE, REFERENCE> predecessor, Collection<LTSA<CODE_REFERENCE, REFERENCE>> successors) {
		connectAll(Collections.singletonList(predecessor), successors);
	}
	
	/** Connects <i>predecessor</i> to all <i>successors</i>.
	 * 
	 * @param <REFERENCE>
	 * @param predecessors
	 * @param successor
	 */
	public static <CODE_REFERENCE, REFERENCE> void connectAll(Collection<LTSA<CODE_REFERENCE, REFERENCE>> predecessors, LTSA<CODE_REFERENCE, REFERENCE> successor) {
		connectAll(predecessors, Collections.singletonList(successor));
	}
	
	/** Connects <i>predecessor</i> to <i>successor</i>.
	 * 
	 * @param <REFERENCE>
	 * @param predecessor
	 * @param successor
	 */
	public static <CODE_REFERENCE, REFERENCE> void connectAll(LTSA<CODE_REFERENCE, REFERENCE> predecessor, LTSA<CODE_REFERENCE, REFERENCE> successor) {
		predecessor.connectTransition(successor);
	}
	
	//
	//
	// Deprecated methods (manipulation with LTS)
	//
	//
	
	/** Creates transitions that adjusts given variable linearly (adds given difference).
	 * 
	 * @param <REFERENCE>
	 * @param variable variable to be adjusted
	 * @param difference difference (additive)
	 * @return {@link List} of transitions 
	 * @deprecated Replaced by {@link #createVariableAdjustmentLTSATransitions(VariableDescriptor, int, boolean)}
	 */
	public static <REFERENCE> List<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> createVariableAdjustmentLTSTransitions(VariableDescriptor<REFERENCE> variable, int difference) {
		LinkedList<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> result =
			new LinkedList<LTS<? extends TBPNode<REFERENCE>,REFERENCE>>();
		int size = variable.getValues().size();
		for(int i = 0; i < size; ++i) {
			int targetIndex = i + difference;
			if(targetIndex >= 0 && targetIndex < size) {
				REFERENCE source = variable.getValues().get(i);
				REFERENCE target = variable.getValues().get(targetIndex);
				
				ITBPCondition<REFERENCE> guard = new TBPEquals<REFERENCE>(variable.getIdentifier(), source);
				TBPAssignment<REFERENCE> assignment = new TBPAssignment<REFERENCE>(variable.getIdentifier(), createTBPValue(target));
				
				result.add(LTSImpl.createInstance(assignment, guard));
			}
		}
		return result;
	}
	
	/** Connects all <i>successors</i> to the <i>predecessors</i> 
	 * 
	 * @param <REFERENCE>
	 * @param predecessors
	 * @param successors
	 * @deprecated Replaced by {@link #connectAll(Collection, Collection)}
	 */
	public static <REFERENCE> void connectAllLTSTransitions(Collection<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> predecessors, Collection<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> successors) {
		assert(!predecessors.isEmpty() && !successors.isEmpty());
		if (predecessors.size() > 1 && successors.size() > 1) {
			// many to many connection
			LTS<? extends TBPNode<REFERENCE>, REFERENCE> bridge = LTSImpl.createInstance(new TBPImperativeNull<REFERENCE>(), null);
			for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> src : predecessors) {
				src.connectTransition(bridge);
			}
			for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> tgt : successors) {
				bridge.connectTransition(tgt);
			}			
		} else {
			// 1 to ? or ? to 1 connection
			for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> src : predecessors) {
				for(LTS<? extends TBPNode<REFERENCE>, REFERENCE> tgt : successors) {
					src.connectTransition(tgt);
				}
			}
		}
	}

	/** Connects all <i>successors</i> to <i>predecessor</i>. 
	 * 
	 * @param <REFERENCE>
	 * @param predecessor
	 * @param successors
	 * @deprecated Replaced by {@link #connectAll(LTSA, Collection)}
	 */
	public static <REFERENCE> void connectAllLTSTransitions(LTS<? extends TBPNode<REFERENCE>, REFERENCE> predecessor, Collection<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> successors) {
		LinkedList<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> tmp = new LinkedList<LTS<? extends TBPNode<REFERENCE>,REFERENCE>>();
		tmp.add(predecessor);
		connectAllLTSTransitions(tmp, successors);
	}
	
	/** Connects <i>predecessor</i> to all <i>successors</i>.
	 * 
	 * @param <REFERENCE>
	 * @param predecessors
	 * @param successor
	 * @deprecated {@link #connectAll(LTSA, Collection)}
	 */
	public static <REFERENCE> void connectAllLTSTransitions(Collection<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> predecessors, LTS<? extends TBPNode<REFERENCE>, REFERENCE> successor) {
		LinkedList<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> tmp = new LinkedList<LTS<? extends TBPNode<REFERENCE>,REFERENCE>>();
		tmp.add(successor);
		connectAllLTSTransitions(predecessors, tmp);
	}
	
	/** Connects <i>predecessor</i> to <i>successor</i>.
	 * 
	 * @param <REFERENCE>
	 * @param predecessor
	 * @param successor
	 * @deprecated {@link #connectAll(LTSA, LTSA)}
	 */
	public static <REFERENCE> void connectAllLTSTransitions(LTS<? extends TBPNode<REFERENCE>, REFERENCE> predecessor, LTS<? extends TBPNode<REFERENCE>, REFERENCE> successor) {
		predecessor.connectTransition(successor);
	}
	
}
