/*
 * $Id$
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.provision.actuator.visitor;

import java.util.ArrayList;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.LTS;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSImpl;
import org.ow2.dsrg.fm.tbplib.node.TBPAssignment;
import org.ow2.dsrg.fm.tbplib.node.TBPAtomic;
import org.ow2.dsrg.fm.tbplib.node.TBPImperativeNull;
import org.ow2.dsrg.fm.tbplib.node.TBPImperativeSequence;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;

/**
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 *
 */
public class OpenUnlimitedScheme implements ThreadCooperationScheme {

	@Override
	public <CODE_REFERENCE, REFERENCE> void instanciateScheme(
			VariableFactory<REFERENCE> variableFactory,
			List<LTSAHandler<CODE_REFERENCE, REFERENCE>> threadBodies,
			int atLeastExecuted,
			LTSAHandler<CODE_REFERENCE, REFERENCE> controllerLTS,
			List<LTSAHandler<CODE_REFERENCE, REFERENCE>> threads) {

		assert(threadBodies != null && !threadBodies.isEmpty());
		assert(controllerLTS.getEntryPoints().isEmpty() && controllerLTS.getEndPoints().isEmpty());
		
		VariableDescriptor<REFERENCE> exitFlag = variableFactory.getSharedStateVariable(EXIT_FLAG_NAME, 2);		
		VariableDescriptor<REFERENCE> readyFlag = variableFactory.createStateVariable("_r", 2);
		VariableDescriptor<REFERENCE> minProgressCounter = variableFactory.createStateVariable("_m", atLeastExecuted + 1);
		
		assert(exitFlag.getValues().size() == 2);
		assert(readyFlag.getValues().size() == 2);
		assert(!minProgressCounter.getValues().isEmpty());
		
		fillController(controllerLTS, readyFlag, minProgressCounter);
		
		for(LTSAHandler<CODE_REFERENCE, REFERENCE> threadBody : threadBodies) {
			if (threadBody != null) {
				threads.add(createThread(threadBody, exitFlag, readyFlag, minProgressCounter));
			}
		}
		
	}
	
	private <CODE_REFERENCE, REFERENCE> void fillController(
			LTSAHandler<CODE_REFERENCE, REFERENCE> controller,
			VariableDescriptor<REFERENCE> readyFlag,  /* |domain| == 2 */
			VariableDescriptor<REFERENCE> minProgress /* |domain| >= 1 */) {
		
		// create transitions
		List<TBPAssignment<REFERENCE>> assignments = new ArrayList<TBPAssignment<REFERENCE>>(2);
		assignments.add(new TBPAssignment<REFERENCE>(readyFlag.getIdentifier(), LTSAManipulationHelper.createTBPValue(readyFlag.getValues().get(1))));
		assignments.add(new TBPAssignment<REFERENCE>(minProgress.getIdentifier(), LTSAManipulationHelper.createTBPValue(minProgress.getValues().get(0))));
		LTSA<CODE_REFERENCE, REFERENCE> transition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, null, assignments);
		
		// insert to LTSHandler
		controller.addEntryPoint(transition);
		controller.addEndPoint(transition);
	}
	
	private <CODE_REFERENCE, REFERENCE> LTSAHandler<CODE_REFERENCE, REFERENCE> createThread(
			LTSAHandler<CODE_REFERENCE, REFERENCE> thread,
			VariableDescriptor<REFERENCE> exitFlag,   /* |domain| == 2 */			
			VariableDescriptor<REFERENCE> readyFlag,  /* |domain| == 2 */
			VariableDescriptor<REFERENCE> minProgress /* |domain| >= 1 */) {
	
		
		// create transitions
		LTSA<CODE_REFERENCE, REFERENCE> enterTransition = 
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createLastValueTBPCondition(readyFlag), null);
		List<LTSA<CODE_REFERENCE, REFERENCE>> enterBodyTransitions =
			LTSAManipulationHelper.createVariableAdjustmentLTSATransitions(minProgress, 1, true);
		LTSA<CODE_REFERENCE, REFERENCE> leaveBodyTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>();
		LTSA<CODE_REFERENCE, REFERENCE> skipBodyTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createLastValueTBPCondition(minProgress), null);
		LTSA<CODE_REFERENCE, REFERENCE> leaveTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>();
		LTSA<CODE_REFERENCE, REFERENCE> immediateExitTransition =
			new LTSAImpl<CODE_REFERENCE, REFERENCE>(null, LTSAManipulationHelper.createLastValueTBPCondition(exitFlag), null);

		// connect the transitions
		LTSAManipulationHelper.connectAll(enterTransition, enterBodyTransitions);
		LTSAManipulationHelper.connectAll(enterTransition, skipBodyTransition);
		
		LTSAManipulationHelper.connectAll(enterBodyTransitions, thread.getEntryPoints());
		
		LTSAManipulationHelper.connectAll(thread.getEndPoints(), leaveBodyTransition);
		
		LTSAManipulationHelper.connectAll(leaveBodyTransition, leaveTransition);
		LTSAManipulationHelper.connectAll(skipBodyTransition, leaveTransition);
				
		// insert into LTSHandler
		LTSAHandler<CODE_REFERENCE, REFERENCE> ltsa = new LTSAHandler<CODE_REFERENCE, REFERENCE>();
		ltsa.setCopies(thread.getCopies());		
		ltsa.addEntryPoint(enterTransition);
		ltsa.addEntryPoint(immediateExitTransition);
		ltsa.addEndPoint(leaveTransition);
		ltsa.addEndPoint(immediateExitTransition);
		
		return ltsa;
	}
	

	/** @deprecated Replaced by {@link #instanciateScheme(VariableFactory, List, int, LTSAHandler, List)} */
	@Override
	public <REFERENCE> void instanciateScheme(
			VariableFactory<REFERENCE> variableFactory,
			List<LTSHandler<REFERENCE>> threadBodies,
			int atLeastExecuted,
			LTSHandler<REFERENCE> controllerLTS,
			List<LTSHandler<REFERENCE>> threads) {
		
		assert(threadBodies != null && !threadBodies.isEmpty());
		assert(controllerLTS.getEntryPoints().isEmpty() && controllerLTS.getEndPoints().isEmpty());
		
		VariableDescriptor<REFERENCE> exitFlag = variableFactory.getSharedStateVariable(EXIT_FLAG_NAME, 2);		
		VariableDescriptor<REFERENCE> readyFlag = variableFactory.createStateVariable("_r", 2);
		VariableDescriptor<REFERENCE> minProgressCounter = variableFactory.createStateVariable("_m", atLeastExecuted + 1);
		
		assert(exitFlag.getValues().size() == 2);
		assert(readyFlag.getValues().size() == 2);
		assert(!minProgressCounter.getValues().isEmpty());
		
		fillController(controllerLTS, readyFlag, minProgressCounter);
		
		for(LTSHandler<REFERENCE> threadBody : threadBodies) {
			if (threadBody != null) {
				threads.add(createThread(threadBody, exitFlag, readyFlag, minProgressCounter));
			}
		}
	}
	/** @deprecated */
	private <REFERENCE> void fillController(
			LTSHandler<REFERENCE> controller,
			VariableDescriptor<REFERENCE> readyFlag,  /* |domain| == 2 */
			VariableDescriptor<REFERENCE> minProgress /* |domain| >= 1 */) {
		
		// create transitions
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> transition = 
			LTSImpl.createInstance(	new TBPAtomic<REFERENCE>(
										new TBPImperativeSequence<REFERENCE>(
											new TBPAssignment<REFERENCE>(readyFlag.getIdentifier(), LTSAManipulationHelper.createTBPValue(readyFlag.getValues().get(1))),
											new TBPAssignment<REFERENCE>(minProgress.getIdentifier(), LTSAManipulationHelper.createTBPValue(minProgress.getValues().get(0))))),
									null /* no guard */);
		
		// insert to LTSHandler
		controller.addEntryPoint(transition);
		controller.addEndPoint(transition);
	}
	
	/** @deprecated */
	private <REFERENCE> LTSHandler<REFERENCE> createThread(
			LTSHandler<REFERENCE> thread,
			VariableDescriptor<REFERENCE> exitFlag,   /* |domain| == 2 */			
			VariableDescriptor<REFERENCE> readyFlag,  /* |domain| == 2 */
			VariableDescriptor<REFERENCE> minProgress /* |domain| >= 1 */) {
	
		
		// create transitions
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> enterTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createLastValueTBPCondition(readyFlag));
		List<LTS<? extends TBPNode<REFERENCE>, REFERENCE>> enterBodyTransitions =
			LTSAManipulationHelper.createVariableAdjustmentLTSTransitions(minProgress, 1);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> leaveBodyTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									null);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> skipBodyTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createLastValueTBPCondition(minProgress));

		LTS<? extends TBPNode<REFERENCE>, REFERENCE> leaveTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									null);
		LTS<? extends TBPNode<REFERENCE>, REFERENCE> immediateExitTransition =
			LTSImpl.createInstance(	new TBPImperativeNull<REFERENCE>(),
									LTSAManipulationHelper.createLastValueTBPCondition(exitFlag));


		// connect the transitions
		LTSAManipulationHelper.connectAllLTSTransitions(enterTransition, enterBodyTransitions);
		LTSAManipulationHelper.connectAllLTSTransitions(enterTransition, skipBodyTransition);
		
		LTSAManipulationHelper.connectAllLTSTransitions(enterBodyTransitions, thread.getEntryPoints());
		
		LTSAManipulationHelper.connectAllLTSTransitions(thread.getEndPoints(), leaveBodyTransition);
		
		LTSAManipulationHelper.connectAllLTSTransitions(leaveBodyTransition, leaveTransition);
		LTSAManipulationHelper.connectAllLTSTransitions(skipBodyTransition, leaveTransition);
				
		// insert into LTSHandler
		LTSHandler<REFERENCE> lts = new LTSHandler<REFERENCE>();
		lts.addEntryPoint(enterTransition);
		lts.addEntryPoint(immediateExitTransition);
		lts.addEndPoint(leaveTransition);
		lts.addEndPoint(immediateExitTransition);
		
		return lts;
	}

}
