/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.provision.actuator.visitor;

import java.util.LinkedList;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.ltsa.LTSA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThread;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSAThreadImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.Label;
import org.ow2.dsrg.fm.tbplib.node.TBPNode;
import org.ow2.dsrg.fm.tbplib.node.TBPVariableDefinition;
import org.ow2.dsrg.fm.tbplib.provision.MalformedProvisionException;
import org.ow2.dsrg.fm.tbplib.provision.actuator.ActionSeqNode;
import org.ow2.dsrg.fm.tbplib.provision.actuator.BranchAction;
import org.ow2.dsrg.fm.tbplib.provision.actuator.EmitEventAction;
import org.ow2.dsrg.fm.tbplib.provision.actuator.EmitThreadsAction;
import org.ow2.dsrg.fm.tbplib.provision.actuator.NullAction;
import org.ow2.dsrg.fm.tbplib.provision.actuator.RepeatAction;

/** Translates actuator to {@link LTSA}.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public class TranslateToLTSAVisitor<REFERENCE> implements
		ActionSeqNodeVisitor<LTSAHandler<TBPNode<REFERENCE>, REFERENCE>, REFERENCE> {
	
	/** Representation of infinity */
	public static final int INFINITY = -1;
	private List<LTSAThread<TBPNode<REFERENCE>, REFERENCE>> threads;
	private VariableFactory<REFERENCE> variableFactory;
	private boolean metSubThreads;
	
	/** Creates this visitor.
	 * 
	 * @param factory for creation of variables for thread synchronization 
	 */
	public TranslateToLTSAVisitor(VariableFactory<REFERENCE> factory) {
		this.variableFactory = factory;
		this.threads = new LinkedList<LTSAThread<TBPNode<REFERENCE>, REFERENCE>>();
		this.metSubThreads = false;
	}
	
	/** Translates graph of {@link ActionSeqNode} into {@link LTSSpecification}.
	 * 
	 * @param node initial node of the graph
	 * @return resulting {@link LTSSpecification}
	 */
	public LTSASpecification<TBPNode<REFERENCE>, REFERENCE> translate(ActionSeqNode<REFERENCE> node) {
		LTSAThread<TBPNode<REFERENCE>, REFERENCE> mainThread = createThread(node.visit(this), "main", 1);
		LTSASpecification<TBPNode<REFERENCE>, REFERENCE> s = new LTSASpecification<TBPNode<REFERENCE>, REFERENCE>();
		
		// add threads
		s.addThread(mainThread);
		for(LTSAThread<TBPNode<REFERENCE>, REFERENCE> auxThread: threads) {
			s.addThread(auxThread);
		}
		
		return s;
	}

	@Override
	public LTSAHandler<TBPNode<REFERENCE>, REFERENCE> visitBranchAction(
			BranchAction<REFERENCE> node) {
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> alternative = new LTSAHandler<TBPNode<REFERENCE>, REFERENCE>();
		
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> succ = node.getSuccessor() != null ? node.getSuccessor().visit(this) : null;
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> leftBranch = node.getLeftBranch().visit(this);
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> rightBranch = node.getRightBranch().visit(this);
		
		alternative.addEntryPoints(leftBranch.getEntryPoints());
		alternative.addEndPoints(leftBranch.getEndPoints());
		
		alternative.addEntryPoints(rightBranch.getEntryPoints());
		alternative.addEndPoints(rightBranch.getEndPoints());
		
		return concat(alternative, succ);
	}

	@Override
	public LTSAHandler<TBPNode<REFERENCE>, REFERENCE> visitEmitEventAction(
			EmitEventAction<REFERENCE> node) {
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> tmp = new LTSAHandler<TBPNode<REFERENCE>, REFERENCE>();		
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> succ = node.getSuccessor() != null ? node.getSuccessor().visit(this) : null;
		
		// create LTS transition
		LinkedList<Label<REFERENCE>> labels = new LinkedList<Label<REFERENCE>>();
		labels.add(Label.createCallLabel(node.getEmit().getMethodCall().getInterface(),
				node.getEmit().getMethodCall().getMethod(),
				node.getEmit().getMethodCall().getParamDecl(),
				node.getEmit().getMethodCall().getFullname()));
		labels.add(Label.createReturnLabel(node.getEmit().getMethodCall().getInterface(),
				node.getEmit().getMethodCall().getMethod(),
				node.getEmit().getMethodCall().getFullname()));		
		LTSA<TBPNode<REFERENCE>, REFERENCE> transition =
			new LTSAImpl<TBPNode<REFERENCE>, REFERENCE>(labels, null, null);
		
		tmp.addEntryPoint(transition);
		tmp.addEndPoint(transition);
						
		return concat(tmp, succ);
	}

	@Override
	public LTSAHandler<TBPNode<REFERENCE>, REFERENCE> visitEmitThreadsAction(
			EmitThreadsAction<REFERENCE> node) {
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> tmp = new LTSAHandler<TBPNode<REFERENCE>, REFERENCE>();	
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> succ = node.getSuccessor() != null ? node.getSuccessor().visit(this) : null;
		
		// create temporary wrapper
		ThreadLauncherWrapper tWrapper = new ThreadLauncherWrapper();
		tWrapper.threadLauncher = node;
		tWrapper.rightOpen = succ == null;
				
		// create thread controller
		
		// collect thread bodies
		List<LTSAHandler<TBPNode<REFERENCE>, REFERENCE>> threadBodies = new LinkedList<LTSAHandler<TBPNode<REFERENCE>, REFERENCE>>();
		if (tWrapper.isUnlimited()) {
			
			metSubThreads = false;
			for(ActionSeqNode<REFERENCE> t : node.getEntryPointsOfThreads()) {
				LTSAHandler<TBPNode<REFERENCE>, REFERENCE> handler = t.visit(this);
				handler.setCopies(LTSAThread.INFINITY);
				threadBodies.add(handler);
			}
			if (metSubThreads) {
				throw new MalformedProvisionException("Any parallel operator within unlimited parallel are not translatable to provision actuator.");
			}
		} else {
			assert(tWrapper.threadLauncher.getMaxThreadReplicas() > 0);
			int copies = tWrapper.threadLauncher.getMaxThreadReplicas();
			
			for(ActionSeqNode<REFERENCE> t : node.getEntryPointsOfThreads()) {
				for(int i = 0; i < copies; ++i) {
					metSubThreads = false;
					LTSAHandler<TBPNode<REFERENCE>, REFERENCE> handler = t.visit(this);					
					threadBodies.add(handler);
					if (!metSubThreads) {
						handler.setCopies(copies);
						break;
					}
				}
			}
		}
		
		List<LTSAHandler<TBPNode<REFERENCE>, REFERENCE>> resultingThreadBodies = new LinkedList<LTSAHandler<TBPNode<REFERENCE>, REFERENCE>>();
		ThreadCooperationScheme scheme = getScheme(tWrapper);
		scheme.instanciateScheme(variableFactory, threadBodies, tWrapper.threadLauncher.getMinThreadLaunches(), tmp, resultingThreadBodies);
		
		for(LTSAHandler<TBPNode<REFERENCE>, REFERENCE> resultingThreadBody: resultingThreadBodies) {
			threads.add(createThread(resultingThreadBody, "worker", resultingThreadBody.getCopies()));
		}
		
		this.metSubThreads = true;
		return concat(tmp, succ);

	}

	@Override
	public LTSAHandler<TBPNode<REFERENCE>, REFERENCE> visitNullAction(NullAction<REFERENCE> node) {
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> tmp = new LTSAHandler<TBPNode<REFERENCE>, REFERENCE>();		
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> succ = node.getSuccessor() != null ? node.getSuccessor().visit(this) : null;
		
		// create LTS transition
		LTSA<TBPNode<REFERENCE>, REFERENCE> transition = createNullTransition();
		
		tmp.addEntryPoint(transition);
		tmp.addEndPoint(transition);
						
		return concat(tmp, succ);
	}

	@Override
	public LTSAHandler<TBPNode<REFERENCE>, REFERENCE> visitRepeatAction(
			RepeatAction<REFERENCE> node) {		
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> tmp = new LTSAHandler<TBPNode<REFERENCE>, REFERENCE>();
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> repeated = node.getRepeatedSequenceEntryPoint().visit(this);		
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> succ = node.getSuccessor() != null ? node.getSuccessor().visit(this) : null;
		
		// create LTS transition
		//LTS<? extends TBPNode<REFERENCE>, REFERENCE> repeatTransition = createNullTransition();
		
		assert(repeated != null && repeated.getEntryPoints().size() >= 1 && repeated.getEndPoints().size() >= 1);
		
		// end of backward edge
		LTSA<TBPNode<REFERENCE>, REFERENCE> repeatTransition =  createNullTransition();
		for(LTSA<TBPNode<REFERENCE>, REFERENCE> e : repeated.getEntryPoints()) {
			repeatTransition.connectTransition(e);
		}
		
		// start of backward edge, end of repetition
		LTSA<TBPNode<REFERENCE>, REFERENCE> repetitionEnd = createNullTransition();		
		for(LTSA<TBPNode<REFERENCE>, REFERENCE> e : repeated.getEndPoints()) {
			e.connectTransition(repetitionEnd);
			e.connectTransition(repeatTransition);
		}
				
		tmp.addEntryPoints(repeated.getEntryPoints());
		tmp.addEndPoint(repetitionEnd);
						
		return concat(tmp, succ);
	}
	
	
	//
	// Auxiliary types
	//
	private class ThreadLauncherWrapper {
		public EmitThreadsAction<REFERENCE> threadLauncher;
		public boolean rightOpen = false;
		
		public boolean isUnlimited() {
			return threadLauncher.getMaxThreadReplicas() == EmitThreadsAction.INFINITY;
		}
	}
	
	//
	// Auxiliary methods
	//
	
	private LTSA<TBPNode<REFERENCE>, REFERENCE> createNullTransition() {
		return new LTSAImpl<TBPNode<REFERENCE>, REFERENCE>();
	}
	
	private LTSAThread<TBPNode<REFERENCE>, REFERENCE> createThread(LTSAHandler<TBPNode<REFERENCE>, REFERENCE> body, String name, int replicationFactor) {
		
		LTSAThread<TBPNode<REFERENCE>, REFERENCE> thr = new LTSAThreadImpl<TBPNode<REFERENCE>, REFERENCE>(
				variableFactory.createThreadName(name), new LinkedList<TBPVariableDefinition<REFERENCE>>(), replicationFactor);
		
		LTSAManipulationHelper.connectAll(thr, body.getEntryPoints());
		
		return thr;
	}
	
	private ThreadCooperationScheme getScheme(ThreadLauncherWrapper tWrapper) {
		if (!tWrapper.isUnlimited()) {
			// limited scheme
			// currently we have just closed limited scheme,
			// which is however sufficient for both; closed and open limited cases
			// (but it is not really optimal for closed case)
			return new ClosedLimitedScheme();
			
		} else if (tWrapper.rightOpen){
			// unlimited scheme
			// we have just a scheme for (at least right)closed unlimited case
			// unfortunately, the open unlimited scheme seems to be not realizable
			return new OpenUnlimitedScheme();
			
		} else {
			throw new RuntimeException("Unable to create thread-cooperation meachanism for unlimited parallel followed by non-empty suffix.");
		}
	}	
	
	private LTSAHandler<TBPNode<REFERENCE>, REFERENCE> concat(LTSAHandler<TBPNode<REFERENCE>, REFERENCE> head, LTSAHandler<TBPNode<REFERENCE>, REFERENCE> tail) {
		assert(head != tail);
		
		if (head == null) {
			return tail;
		}
		if (tail == null) {
			return head;
		}
		
		assert(head != null && head.getEntryPoints().size() >= 1 && head.getEndPoints().size() >= 1);
		assert(tail != null && tail.getEntryPoints().size() >= 1 && tail.getEndPoints().size() >= 1);
		
		LTSAHandler<TBPNode<REFERENCE>, REFERENCE> result = new LTSAHandler<TBPNode<REFERENCE>, REFERENCE>();
		
		// entry
		result.addEntryPoints(head.getEntryPoints());
		
		// connect
		LTSAManipulationHelper.connectAll(head.getEndPoints(), tail.getEntryPoints());
		
		// end
		result.addEndPoints(tail.getEndPoints());
		
		return result;
	}
}