/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.provision.actuator.visitor;

import java.util.Collection;
import java.util.List;

import org.ow2.dsrg.fm.tbplib.node.TBPVariableDefinition;
import org.ow2.dsrg.fm.tbplib.reference.MethodCall;

/** Forms basic interface for creating variables.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt;
 */
public interface VariableFactory<REFERENCE> {
	/** Creates reference to thread name.
	 * 
	 * @param suggestedName suggested (human-readable) name of thread
	 * @return REFERENCE to the name of thread
	 */
	public REFERENCE createThreadName(String suggestedName);
	/** Creates new unique variable. 
	 * 
	 * @param suggestedName Suggested name. The actual name does not have to be this name.
	 * @param domainSize Type domain size.
	 * @return {@link VariableDescriptor}
	 */
	public VariableDescriptor<REFERENCE> createStateVariable(String suggestedName, int domainSize);
	/** Creates new unique variable.
	 * 
	 * @param suggestedName Suggested name. The actual name does not have to be this name.
	 * @param domainSize Type domain size.
	 * @param initialValue Initial value.
	 * @return {@link VariableDescriptor}
	 */
	public VariableDescriptor<REFERENCE> createStateVariable(String suggestedName, int domainSize, int initialValue);	
	/** Gets shared variable.
	 * 
	 * <p>Behavior of this method depends on context.
	 *    On first call of this method with fixed <i>name</i> <b>N</b> (1st parameter),
	 *    this methods behaves as {@link #createStateVariable(String, int)}.
	 *    On every consequent call of this method with the same 1st parameter,
	 *    the return value is reference-equal to the return value of the first call
	 *    with the same 1st parameter.
	 * </p>
	 * <p>Note that if there are two call of this method with same <i>name</i>
	 *    and different <i>domainSize</i>, the behavior of second call is not defined
	 *    (an exception is recommended).
	 * </p>
	 * 
	 * @param name 
	 * @param domainSize
	 * @return {@link VariableDescriptor}
	 */
	public VariableDescriptor<REFERENCE> getSharedStateVariable(String name, int domainSize)
		throws Error;
	/** Notices method call used within given the component managed by this factory.
	 * 
	 * <p>Might be useful for implementations that collect information for creation of component.</p>
	 *
	 * @param mc method call
	 */
	public void noticeMethodCall(MethodCall<REFERENCE> mc);
	
	/** Creates local-variable definition from <b>name</b> and <b>type</b> and avoid conflicts with <b>localNames</b>.
	 * 
	 * <p>Note that the initial value depends on implementation.
	 * </p>
	 * 
	 * @param name
	 * @param type
	 * @param owner method or thread
	 * @param localNames 
	 * @return {@link TBPVariableDefinition}
	 */
	public TBPVariableDefinition<REFERENCE> createLocalVariable(REFERENCE name, REFERENCE type, REFERENCE owner, Collection<REFERENCE> localNames);
	/** Creates local-variable definition from <b>name</b> and <b>typeSize</b>.
	 * 
	 * @param name
	 * @param typeSize
	 * @param initialValue
	 * @param localNames
	 * @return {@link TBPVariableDefinition}
	 */
	public TBPVariableDefinition<REFERENCE> createLocalVariable(String name, int typeSize, int initialValue, Collection<REFERENCE> localNames);
	
	/** Returns list of all state variables.
	  * 
	  * <p>In order to prevent name clashing,
	  *    the implementing class should require list of all state variables
	  *    by its initialization.
	  *    This method returns list of these state variables
	  *    plus state variables created by the factory.
	  * </p>
	  * 
	  * @return {@link List} of {@link TBPVariableDefinition}s
	  */
	public List<TBPVariableDefinition<REFERENCE>> getStateVariables();

}
