/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2010       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.provision.actuator.visitor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.ow2.dsrg.fm.tbplib.architecture.Autonomous;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.architecture.Interface;
import org.ow2.dsrg.fm.tbplib.architecture.Method;
import org.ow2.dsrg.fm.tbplib.node.TBPVariableDefinition;
import org.ow2.dsrg.fm.tbplib.reference.Constant;
import org.ow2.dsrg.fm.tbplib.reference.EnumerationType;
import org.ow2.dsrg.fm.tbplib.reference.LocalVariable;
import org.ow2.dsrg.fm.tbplib.reference.MethodCall;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.reference.StateVariable;

/** Common implementation of {@link VariableFactory}.
 * 
 * @author Ondrej Cerny &lt;ondrej.cerny@email.cz&gt
 */
public class VariableFactoryImpl implements VariableFactory<Reference> {
		
	// constants
	private static final Comparator<Constant> CONSTANT_CODE_COMPARATOR = new Comparator<Constant>() {
		@Override
		public int compare(Constant o1, Constant o2) {
			if (	(o1 == o2)
				||	(o1 != null && o2 != null && (o1.getCode() == o2.getCode()))) {
				return 0;
			} else if (o1 == null || (o2 != null && o1.getCode() < o2.getCode())) {
				return -1;
			} else {
				return 1;
			}
		}
	};
	
	// state fields
	private static Set<String> usedNames = new HashSet<String>();
	
	private Component result;
	
	// data fields
	private List<TBPVariableDefinition<Reference>> stateVariablesDefinitions;
	private List<StateVariable> releasedStateVariables;
	private List<EnumerationType> releasedTypes;
	private List<Autonomous> threadNames;
	private Set<Reference> calledInterfaces;

	private Map<String, Map<Integer, VariableDescriptor<Reference>>> sharedVariables;
		
	private Map<Integer, EnumerationType> typeMap;
	
	public VariableFactoryImpl(List<TBPVariableDefinition<Reference>> usedVariables) {
		this.stateVariablesDefinitions = new LinkedList<TBPVariableDefinition<Reference>>();
		this.releasedStateVariables = new LinkedList<StateVariable>();
		this.releasedTypes = new LinkedList<EnumerationType>();
		this.threadNames = new LinkedList<Autonomous>();
		this.calledInterfaces = new HashSet<Reference>();
		this.sharedVariables = new HashMap<String, Map<Integer,VariableDescriptor<Reference>>>();
		this.result = null;
		this.typeMap = new HashMap<Integer, EnumerationType>();
		for(TBPVariableDefinition<Reference> variableDef: usedVariables) {
			if (variableDef.getName() != null && variableDef.getName() instanceof StateVariable) {
				stateVariablesDefinitions.add(variableDef);
			}
		}
	}
	@Override
	public Reference createThreadName(String suggestedName) {
		assert(result == null);
		
		String name = makeNameUnique(suggestedName);
		Autonomous thread = new Autonomous(name);
		
		threadNames.add(thread);
		return thread;
	}
	
	private EnumerationType createType(int domainSize) {
		EnumerationType type = this.typeMap.get(domainSize);
		
		if (type == null) {
			ArrayList<String> enums = new ArrayList<String>(domainSize);
			
			for(int i = 0; i < domainSize; ++i) {
				enums.add(makeNameUnique("_V_" + i + "_" + domainSize + "_"));
			}
			
			type = new EnumerationType(makeNameUnique("_T_" + domainSize + "_"), enums);
			this.releasedTypes.add(type);
			this.typeMap.put(domainSize, type);
		}
		return type;		
	}
	
	/* Returns list of constants (enumeration of type) sorted by constants' codes. */ 
	private List<Constant> getSortedTypeConstants(EnumerationType type) {
		ArrayList<Constant> constants = new ArrayList<Constant>(type.getEnums().values());
		Collections.sort(constants, CONSTANT_CODE_COMPARATOR);
		return constants;
	}
	
	private Constant getNthOrFirstConstant(List<Constant> constants, int n) {
		if (n < 0 || n >= constants.size()) {
			return constants.get(0);
		} else {
			return constants.get(n);
		}
	}

	@Override
	public VariableDescriptor<Reference> createStateVariable(String suggestedName, int domainSize) {
		return createStateVariable(suggestedName, domainSize, 0);
	}

	
	@Override
	public VariableDescriptor<Reference> createStateVariable(String suggestedName,
			int domainSize, int initialValue) {
		assert(result == null && domainSize > 0 && domainSize > initialValue && initialValue >= 0);
		 
		VariableDescriptor<Reference> result = new VariableDescriptor<Reference>();
		EnumerationType type = createType(domainSize);
		
		
		List<Constant> values = getSortedTypeConstants(type);
		
		for(Constant constant: values) {
			result.addValue(constant);
		}
		
		// create state variable reference and set it to result
		StateVariable stateVar = new StateVariable(type, makeNameUnique(suggestedName), getNthOrFirstConstant(values, initialValue), null);
		result.setIdentifier(stateVar);

		// store the values
		registerStateVariable(stateVar);
		releasedTypes.add(type);
		return result;
	}

	@Override
	public VariableDescriptor<Reference> getSharedStateVariable(String name,
			int domainSize) throws Error {
		assert(result == null);
		
		VariableDescriptor<Reference> result = null;
		Map<Integer, VariableDescriptor<Reference>> sizeMap = null;
		if (sharedVariables.containsKey(name)) {
			sizeMap = sharedVariables.get(name);
			if (sizeMap.containsKey(domainSize)) {
				result = sizeMap.get(domainSize); 
			}
		} else {
			sizeMap = new HashMap<Integer, VariableDescriptor<Reference>>();
			sharedVariables.put(name, sizeMap);
		}
		if (result == null) {
			result = createStateVariable(name, domainSize);
			sizeMap.put(domainSize, result);
		}
		return result;
	}
	
	@Override
	public void noticeMethodCall(MethodCall<Reference> mc) {
		calledInterfaces.add(mc.getInterface());
	}
	@Override
	public TBPVariableDefinition<Reference> createLocalVariable(String name, int typeSize, int initialValue, Collection<Reference> localNames) {
		EnumerationType type = createType(typeSize);
		List<Constant> values = getSortedTypeConstants(type);
		Constant initialConst = getNthOrFirstConstant(values, initialValue);
		return createLocalVariable(name, type, initialConst, null, localNames);
	}
	
	@Override
	public TBPVariableDefinition<Reference> createLocalVariable(
			Reference name, Reference type, Reference autonomous,
			Collection<Reference> localNames) {
		
		// pick a initial value
		EnumerationType variableType = null;
		Constant initialValue = null;
		if (name != null && name instanceof LocalVariable) {
			LocalVariable lv = (LocalVariable)name;
			initialValue = lv.getInitialValue();
			variableType = lv.getType();
			if (!variableType.equals(type)) {
				variableType = null;
				initialValue = null;
			}
		} 
		if (variableType == null || initialValue == null) {
			if (type != null && type instanceof EnumerationType) {
				variableType = (EnumerationType)type;
				for (Entry<String, Constant> mapEntry: variableType.getEnums().entrySet()) {
					initialValue = mapEntry.getValue();
					break;
				}
			} else {
				throw new ClassCastException("Type of variable is invalid.");
			}
		}
		
		assert(variableType != null && initialValue != null);
		
		return createLocalVariable(name.getName(), variableType, initialValue, autonomous, localNames);				
	}
	
	@Override
	public List<TBPVariableDefinition<Reference>> getStateVariables() {
		return Collections.unmodifiableList(stateVariablesDefinitions);
	}
	
	private TBPVariableDefinition<Reference> createLocalVariable(
			String name, EnumerationType type, Constant initialValue, Reference autonomous,
			Collection<Reference> localNames) {
		
		// pick a name
		String newName = getUniqueLocalName(name, localNames);
		
		Method m = null;
		if (autonomous != null) {
			if (autonomous instanceof Method) {
				m = (Method)autonomous;
			} else {
				throw new ClassCastException("Owner code (method/thread).");
			}
		} 
		
		LocalVariable nameRef = new LocalVariable(m, type, newName, initialValue);
		return new TBPVariableDefinition<Reference>(type, nameRef, initialValue);		
	}

	
	
	//
	private String getUniqueLocalName(String name, Collection<Reference> localNames) {
		boolean nameAlreadyExists = false;
		String result = null;
		for( Reference tmpName : localNames ) {
			String n = tmpName.getName();
			if (name.equals(n)) {
				nameAlreadyExists = true;
			}
		}
		if (nameAlreadyExists) {
			int copies = 0;
			Pattern p = Pattern.compile("_[0-9]+$");
			for( Reference varName: localNames) {
				String tmpName = varName.getName();
				if (tmpName.startsWith(name)) {
					name.length();
					Matcher m = p.matcher(tmpName);
					if (m.find(name.length()) && m.start() == name.length()) {
						String tmpNumber = tmpName.substring(tmpName.length() + 1);
						if (!tmpNumber.isEmpty()) {
							int number = 0;
							try {
								number = Integer.parseInt(tmpNumber);
							} catch (NumberFormatException nfe) {
								number = 0;
							}
							if (number > copies) {
								copies = number;
							}
						}
					}
				}
			}
			result = getUniqueLocalName(name + "_" + Integer.toString(copies), localNames);
		} else {
			result = name;
		}
		return result;
	}
	
	
	public Component toComponent(String name) {
		if (result == null) {
			result = new Component(name, this.releasedTypes, this.releasedStateVariables);
			for(StateVariable v : releasedStateVariables) {
				v.setComponent(result);
			}
			for(Autonomous thread : threadNames) {
				result.addThread(thread);
			}
			for(Reference itf: calledInterfaces) {
				result.addRequiredInterface(new Interface(itf.getName(), result, Interface.Direction.REQUIRED));
			}
		}
		return result;
	}
	
	private void registerStateVariable(StateVariable sv) {
		releasedStateVariables.add(sv);
		stateVariablesDefinitions.add(new TBPVariableDefinition<Reference>(sv.getType(), sv, sv.getInitialValue()));
	}
	
	//
	// AUX
	//
	
	private String makeNameUnique(String name) {
		String uniqueName = name;
		int counter = 0;
		synchronized (usedNames) {
			while (usedNames.contains(uniqueName)) {
				uniqueName = name + "_" + counter;
				++counter;
			}
			usedNames.add(uniqueName);			
		}
		return uniqueName;
	}	
}
