/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.tbplib.reference;

import org.ow2.dsrg.fm.tbplib.TBPResolvingException;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * Representation of a type. Contains special type for mutex variables as a static
 * field.
 * 
 * @author caitt3am
 *
 */
public class EnumerationType implements Reference {

    private static final String UNLOCKED_STR = "UNLOCKED", LOCKED_STR = "LOCKED";
	public static final EnumerationType MUTEX_TYPE =
            new EnumerationType("__MUTEX", Arrays.asList(UNLOCKED_STR, LOCKED_STR));
	public static final Constant UNLOCKED = MUTEX_TYPE.getEnums().get(UNLOCKED_STR);
	public static final Constant LOCKED = MUTEX_TYPE.getEnums().get(LOCKED_STR);
	
	protected final Map<String, Constant> enums;
	private final String name;

    /**
     * Creates an empty EnumerationType for use (and fill) by ancestors.
     *
     * @param name Name of the enumeration type.
     */
	protected EnumerationType(String name) {
        this.name = name;
        this.enums = new HashMap<String, Constant>();
    }

	public EnumerationType(String name, Collection<String> enums) {
        this.enums = new HashMap<String, Constant>(enums.size());
        for (String item : enums) {
            Constant c = new Constant(this, item, (byte)this.enums.size());

            if(this.enums.containsKey(item))
                throw new TBPResolvingException("Duplicate constant \'" +
                        c.getName() + "\' in the type: " + name);
            
            this.enums.put(item, c);
        }
		this.name = name;
	}

	/**
	 * Returns name of this type.
	 */
    @Override
	public String getName() {
		return name;
	}

	/**
	 * Returns list of names of constants that constitutes this type
	 */
	public Map<String, Constant> getEnums() {
		return enums;
	}

	@Override
	public String toString() {
		return name;
	}

	@Override
	public int hashCode() {
		return name.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof EnumerationType))
			return false;
		final EnumerationType other = (EnumerationType) obj;
		if (!enums.equals(other.enums))
			return false;
		if (!name.equals(other.name))
			return false;
		return true;
	}
}