/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.reference;

import org.ow2.dsrg.fm.tbplib.TBPResolvingException;
import org.ow2.dsrg.fm.tbplib.architecture.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * ReferenceTransformer that resolves textual identifier and method names int
 * Implementations of the interface Reference.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class ReferenceResolver implements ReferenceTransformer<String, Reference> {

    protected Architecture architecture;
    protected Component component;
    protected Method method;
    protected Autonomous thread;

    public ReferenceResolver(Architecture architecture) {
        this.architecture = architecture;
    }

    @Override
    public void enterComponent(Component component) {
        this.component = component;
        this.method = null;
        this.thread = null;
    }

    @Override
    public void leaveComponent() {
        this.component = null;
        this.method = null;
        this.thread = null;
    }

    @Override
    public void enterMethod(Reference method) {
        this.method = (Method) method;
        this.thread = null;
    }

    @Override
    public void leaveMethod() {
        this.method = null;
        this.thread = null;
    }

    @Override
    public void enterThread(Reference thread) {
        this.method = null;
        this.thread = (Autonomous) thread;
    }

    @Override
    public void leaveThread() {
        this.method = null;
        this.thread = null;
    }

    @Override
    public MethodSignature<Reference> resolveMethodSignature(
            MethodSignature<String> methodSignature, Reference method) {
        Method m = (Method) method;
        MethodSignature<Reference> ms = m.getMethodSignature();
        assert (ms != null);
        return ms;
    }

    @Override
    public MethodCall<Reference> resolveMethodCall(MethodCall<String> methodCall,
            boolean remote) {
        // NOTE: The method call will refer to the interface that isbound 
        // to the corresponding interface on the current component.
        Interface localIface;
        
        if (remote) {
            localIface = component.getRequiredInterfaces().get(methodCall.getInterface());
        } else {
            localIface = component.getProvidedInterfaces().get(methodCall.getInterface());
        }
        assert(localIface != null);
        Method localMethod = localIface.findSymbol(methodCall.getMethod());
        assert(localMethod != null);
        
        // FIXME: Unbound interfaces are not considered yet.
        if (remote && !localIface.isBound())
            throw new UnsupportedOperationException("Cannot resolve unbound interface (yet to be implemented): " +
                    localIface.getName());

        Interface targetIface;
        Method targetMethod;

        if (remote) {
            targetIface = localIface.getBoundInterface();
            targetMethod = targetIface.findSymbol(methodCall.getMethod());
            if (targetMethod == null) {
            	throw new RuntimeException("Method " + methodCall.getFullname() + " is called but not declared.");
            }
        } else {
            targetIface = localIface;
            targetMethod = localMethod;
        }

        List<String> oldParams = methodCall.getParamDecl();
        List<Reference> newParams = new ArrayList<Reference>(oldParams.size());

        for (String par : oldParams) {
            // FIXME: Maybe we might deduce the types of constants?
            newParams.add(resolveIdentifier(par));
        }

        MethodCall<Reference> newMethodCall = new MethodCall<Reference>(
                targetIface, targetMethod, newParams);

        return newMethodCall;
    }

    @Override
    public Reference resolveIdentifier(String identifier) {
        // FIXME: No conflicts are checked here!!!
        // NOTE: No typing rules are enforced (only constants in switch branches...).
        // The typing rules should be enforced later during the semanticCheck

        // Try local variable / method parameter
        Variable var = null;
        if (method != null) {
            var = method.getLocals().get(identifier);
            if (var != null) {
                return var;
            }
            var = method.getParameters().get(identifier);
        } else if (thread != null) {
            var = thread.getLocals().get(identifier);
        }
        if (var != null) {
            return var;
        }

        // Try state variable
        var = component.getStateVariables().get(identifier);
        if (var != null) {
            return var;
        }

        // Try a constant
        for (EnumerationType type : architecture.getTypes().values()) {
            Constant c = type.getEnums().get(identifier);

            if (c != null) {
                return c;
            }
        }

        if (identifier.equals("?")){
            return ArbitraryValue.instance();
        }

        throw new TBPResolvingException("Undefined identifier: " + identifier);
    }

    @Override
    public Component resolveComponent(String identifier) {
        for (Component comp : architecture.getComponents()) {
            if (comp.getName().equals(identifier)) {
                return comp;
            }
        }
        throw new TBPResolvingException("Undefined component: " + identifier);
    }

    @Override
    public Interface resolveInterface(String identifier) {
        Interface iface = component.getProvidedInterfaces().get(identifier);
        if (iface == null) {
            throw new TBPResolvingException("Undefined provided interface: " + identifier);
        }
        return iface;
    }

    @Override
    public Method resolveMethod(String identifier, Reference iface) {
        Interface i = (Interface) iface;
        Method m = i.findSymbol(identifier);
        if (m == null) {
            throw new TBPResolvingException("Method \'" + identifier +
                    "\' not found in the interface: " + identifier);
        }
        return m;
    }

    @Override
    public Reference resolveThread(String identifier) {
        Autonomous t = component.getThreads().get(identifier);
        if (t == null) {
            throw new TBPResolvingException("Undefined thread: " + identifier);
        }
        return t;
    }

    @Override
    public Reference resolveVariable(String identifier) {
        Variable var;
        if (method != null) {
            var = method.getLocals().get(identifier);
        } else if (thread != null) {
            var = thread.getLocals().get(identifier);
        } else {
            var = component.getStateVariables().get(identifier);
        }
        if (var == null) {
            throw new TBPResolvingException("Undefined variable: " + identifier);
        }
        return var;
    }

    @Override
    public Reference resolveType(String identifier) {
        EnumerationType type = architecture.getTypes().get(identifier);

        if (type == null) {
            throw new TBPResolvingException("Undefined type: " + identifier);
        }
        return type;
    }

    @Override
    public Reference resolveConstant(String identifier, Reference type) {
        EnumerationType realType = (EnumerationType) type;
        Constant c = realType.getEnums().get(identifier);

        if (c == null) {
            throw new TBPResolvingException("Identifier \'" + identifier +
                    "\' is not member of the type: " + realType.getName());
        }

        return c;
    }

    @Override
    public void resolveFilterItem(String identifier, Set<Reference> toStoreResults){
        String[] parts = identifier.split("\\.");
        String ifaceStr = parts[0];

        String mStr;

        if ((parts.length==1) && parts[0].equals("*")){
            mStr= "*";
        } else {
            mStr=parts[1];
        }


        if (mStr.equals("*")){
            List<Interface> ifaces;
            if (ifaceStr.equals("*")){

                ifaces = new ArrayList<Interface>(component.getProvidedInterfaces().values());
            } else {
                Interface iface = resolveInterface(ifaceStr);
                ifaces = new ArrayList<Interface>(1);
                ifaces.add(iface);
            }
            for (Interface i:ifaces){
                Map<String, Method> allMethods = i.getMethods();
                toStoreResults.addAll(allMethods.values());
            }
            return;
        } else {
            Interface iface = resolveInterface(ifaceStr);
            Reference method = resolveMethod(mStr, iface);
            toStoreResults.add(method);
        }

    }
}
