/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.reference;

import org.ow2.dsrg.fm.tbplib.architecture.Component;

import java.util.Set;

/**
 * Interface for resolving references in the parse tree.
 *
 * TODO: Add local variables!!!
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public interface ReferenceTransformer<FROM, TO> {

    public void enterComponent(Component component);

    public void leaveComponent();

    public void enterMethod(TO method);

    public void leaveMethod();

    public void enterThread(TO thread);

    public void leaveThread();

    /**
     * Resolves references in a method signature.
     *
     * @param methodSignature Method signature with references in the source type.
     * @param method Identifier of a related method in the destination type.
     * @return Method signature with references in the destination type.
     */
    MethodSignature<TO> resolveMethodSignature(MethodSignature<FROM> methodSignature, TO method);

    /**
     * Resolves references in a method call.
     *
     * @param methodCall Method call with references in the source type.
     * @param remote Pass <code>true</code> to store remote (bound) interafce.
     *      Local interface is stored otherwise.
     * @return Method call with references in the destination type.
     */
    MethodCall<TO> resolveMethodCall(MethodCall<FROM> methodCall,
            boolean remote);

    /**
     * Resolves a general identifier. The identifier may represent either
     * a state variable, a constant, or a formal parameter.
     *
     * @param identifier Identifier in the source type.
     * @return Identifier in the destination type.
     */
    TO resolveIdentifier(FROM identifier);

    /**
     * Resolves a component identifier.
     *
     * @param identifier Identifier in the source type.
     * @return Identifier in the destination type.
     */
    TO resolveComponent(FROM identifier);

    /**
     * Resolves a interface identifier.
     *
     * @param identifier Identifier in the source type.
     * @return Identifier in the destination type.
     */
    TO resolveInterface(FROM identifier);

    /**
     * Resolves a method identifier.
     *
     * @param identifier Identifier in the source type.
     * @param iface Identifier of a related interface in the destination type.
     * @return Identifier in the destination type.
     */
    TO resolveMethod(FROM identifier, TO iface);

    /**
     * Resolves a thread identifier.
     *
     * @param identifier Identifier in the source type.
     * @return Identifier in the destination type.
     */
    TO resolveThread(FROM identifier);

    /**
     * Resolves a variable (either local or state variable) identifier.
     *
     * @param identifier Identifier in the source type.
     * @return Identifier in the destination type.
     */
    TO resolveVariable(FROM identifier);

    /**
     * Resolves a type identifier.
     *
     * @param identifier Identifier in the source type.
     * @return Identifier in the destination type.
     */
    TO resolveType(FROM identifier);

    /**
     * Resolves a constant of a known type.
     *
     * @param identifier Identifier in the source type.
     * @param type Identifier of the type of the constant in the destination type.
     * @return Identifier in the destination type.
     */
    TO resolveConstant(FROM identifier, TO type);

    void resolveFilterItem(FROM identifier, Set<TO> toStoreResults);
}
