/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009-2010  DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */
package org.ow2.dsrg.fm.tbplib.util;

import org.antlr.runtime.RecognitionException;
import org.ow2.dsrg.fm.tbplib.TBPLibLog;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.architecture.Interface;
import org.ow2.dsrg.fm.tbplib.architecture.Method;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl.MethodEncoderImpl;
import org.ow2.dsrg.fm.tbplib.ltsa.CompactDFA;
import org.ow2.dsrg.fm.tbplib.ltsa.DynamicDFA;
import org.ow2.dsrg.fm.tbplib.ltsa.DynamicNFA;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSASpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.LTSSpecification;
import org.ow2.dsrg.fm.tbplib.ltsa.Observer;
import org.ow2.dsrg.fm.tbplib.ltsa.visitor.LTSToLTSAConversionVisitor;
import org.ow2.dsrg.fm.tbplib.node.*;
import org.ow2.dsrg.fm.tbplib.node.visitor.*;
import org.ow2.dsrg.fm.tbplib.parser.ParserFacade;
import org.ow2.dsrg.fm.tbplib.provision.actuator.ActionSeqNode;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.TranslateToLTSAVisitor;
import org.ow2.dsrg.fm.tbplib.provision.actuator.visitor.VariableFactory;
import org.ow2.dsrg.fm.tbplib.reference.EnumerationType;
import org.ow2.dsrg.fm.tbplib.reference.MethodSignature;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.reference.ReferenceResolver;

import java.io.IOException;
import java.util.*;

/**
 * Set of utility static methods for performing selected transformations from
 * the transformation chain.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class TransformationChain {

    /**
     * Parses a set of TBP specifications from the given files.
     *
     * @param arch An architecture to which the parsed specifications should be
     *      added. It is modified in place.
     * @param files List of filenames to be loaded and parsed.
     * @return List of TBP specifications
     * @throws java.io.IOException On I/O errors when parsing
     * @throws org.antlr.runtime.RecognitionException On parsing errors.
     */
    public static List<TBPSpecification<String>> parseFiles(
            Architecture arch, String... files) throws IOException,
            RecognitionException {

        List<TBPSpecification<String>> tbps =
                new ArrayList<TBPSpecification<String>>(files.length);

        for (String filename : files) {
            TBPLibLog.logger.info("Parsing file: \"" + filename + "\"");

            ParserResult pr = ParserFacade.parseFile(filename);
            TBPSpecification<String> tbp = pr.getSpecification();
            arch.addComponent(new Component(tbp.getComponent(),pr.getTypes()));
            tbps.add(tbp);
        }

        return tbps;
    }

    public static void renameBoundInterfaces(Architecture architecture) {
        // FIXME: Implement it
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /**
     * Resolves references in the given TBP Specifications, based on the given
     * architectural information.
     * FIXME: Should we pair MethodCalls and Method definitions - renaming
     * has to take place before this ransformation.
     *
     * @param tbps Ordered list of TBP Specifications to be resolved.
     * @param architecture Architectural information to be referenced to.
     * @return Ordered list of the resolved TBP Specifications.
     */
    public static List<TBPSpecification<Reference>> resolveArchitecture(
            List<TBPSpecification<String>> tbps, Architecture architecture) {

        ReferenceResolver resolver = new ReferenceResolver(architecture);
        List<Component> comps = architecture.getComponents();
        List<TBPSpecification<Reference>> result = new ArrayList<TBPSpecification<Reference>>(tbps.size());
        assert (comps.size() == tbps.size());

        for (int i = 0; i < tbps.size(); i++) {
            TBPSpecification<String> tbp = tbps.get(i);
            Component comp = comps.get(i);

            resolver.enterComponent(comp);
            result.add(tbp.resolve(resolver));
            resolver.leaveComponent();
        }
        return result;
    }

    /**
     * Collect definitions from the given TBP specifications and build up
     * an architectural information with these definitions. Information
     * concerning interfaces, methods, state variables, local variables,
     * and parameters is extracted.
     *
     * @param tbps An ordered list of TBP specifications to be sought for
     *      definitions.
     * @param architecture The holder of architectural information to be
     *      modified in place. The information should already contain
     *      ordered list of components and a list of defined types.
     */
    public static void collectSymbolsFromArchitecture(
            List<TBPSpecification<String>> tbps, Architecture architecture) {

        List<Component> comps = architecture.getComponents();
        SymbolCollectingVisitor visitor = new SymbolCollectingVisitor(architecture);

        assert (comps.size() == tbps.size());

        for (int i = 0; i < tbps.size(); i++) {
            TBPSpecification<String> tbp = tbps.get(i);
            Component comp = comps.get(i);
            visitor.setComponent(comp);

            tbp.visit(visitor);
        }
    }

    /**
     * Resolves references in the given TBP Specification, based on the given
     * component information.
     * FIXME: Should we pair MethodCalls and Method definitions - renaming
     * has to take place before this ransformation.
     *
     * @param tbp TBP Specifications to be resolved.
     * @param component Component information to be referenced to.
     * @return Ordered list of the resolved TBP Specifications.
     */
    public static TBPSpecification<Reference> resolveTBP(
            TBPSpecification<String> tbp,
            Component component) {

        ReferenceResolver resolver = new ReferenceResolver(component.getParent());

        resolver.enterComponent(component);

        // Resolve the rest
        return tbp.resolve(resolver);
    }

    /**
     * This method accepts a list of TBP specification plus an architectural
     * information, and performs a set of semantic checks to assure
     * well-formedness of the specification with respect to the architectural
     * information. Violation of a semantic rule is signalized by rising
     * a runtime exception. The specifications should already be resolved
     * with the corresponding interfaces bound.
     * The following checks are performed:
     * <ul>
     * <li>ReturnCheck - all reactions with a defined type do return a value
     * along every finit execution branch.</li>
     * <li>VarInitCheck - no uninitialized local variable is read from.</li>
     * <li>ReferenceTypeCheck - references of a correct type are on a correct
     * place (switch labels). FIXME: anything else?</li>
     * <li>TypeCheck - no type mismatches in conditions, switch, method call,
     * and assignment.</li>
     * <li>RecursionCheck - there is no recursion in the architecture.</li>
     * </ul>
     *
     * @param tbps
     * @param architecture
     */
    public static void performSemanticCheck(
            List<TBPSpecification<Reference>> tbps, Architecture architecture) {
        // FIXME: Implement it
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /**
     * Unbound required interfaces of components will be bound to
     * the identically named provided interfaces of the same
     * component (if present).
     * 
     * @param architecture The holder of architectural information to be
     *      modified in place.
     */
    public static void performImplicitBinding(Architecture architecture) {

        for (Component comp : architecture.getComponents()) {

            // Perform the implicit binding
            Map<String, Interface> provided = comp.getProvidedInterfaces();

            for (Component c2 : architecture.getComponents()) {
                for (Interface iface : c2.getRequiredInterfaces().values()) {
                    if (iface.isBound()) {
                        continue;
                    }

                    Interface target = provided.get(iface.getName());

                    if (target != null) {
                        iface.bindTo(target);
                    }
                }
            }
        }
    }

    /**
     * Creates a composition of the given list of TBP specifications.
     *
     * @param tbps A list of TBP specifications to be composed.
     * @param name Reference/name of the resulting composition.
     * @return The resulting composed TBP specification.
     */
    public static <REFERENCE> TBPSpecification<REFERENCE> performComposition(
            List<TBPSpecification<REFERENCE>> tbps, REFERENCE name) {

        TBPComposition<REFERENCE> tbpComp = new TBPComposition<REFERENCE>();
        tbpComp.addMultipleTBP(tbps);

        return tbpComp.getComposition(name);
    }

    /**
     * Prints the given TBP specification into the given stream.
     *
     * @param output A CodeWriter to recieve the textual form of the TBP
     *      specification.
     * @param tbp TBP Specification to be printed.
     * @param component Component information to be used.
     */
    public static <REFERENCE> void printTBP(CodeWriter output,
            TBPSpecification<REFERENCE> tbp, Component component) {

        PrintComponentVisitor<REFERENCE> visitor = new PrintComponentVisitor<REFERENCE>(output);

        visitor.printComponentSpecification(tbp, component);

        output.flush();
    }

    /**
     * Fills the given event table by the symbols found in reactions of
     * the given architecture. This method should be called after performing
     * all bindings and renaming. This requiremnt follows from the textual
     * representation in the event table.
     *
     * @param arch Architecture to be used for filling the event table in.
     * @param eventTable EventTable to be filled in based on reaction method
     *      signatures.
     */
    public static void fillEventTable(Architecture arch, EventTable eventTable) {
        for (Component c : arch.getComponents()) {
            for (Interface iface : c.getProvidedInterfaces().values()) {
                for (Method method : iface.getMethods().values()) {
                    // Extract domains from the method signature
                    MethodSignature<Reference> sig = method.getMethodSignature();
                    EnumerationType retType = (EnumerationType) sig.getReturnType();
                    int retDomain = retType == null ? 0 : retType.getEnums().size();
                    LinkedHashMap<Reference, Reference> params = sig.getParams();
                    int[] parDomains = new int[params.size()];
                    int i = 0;

                    for (Reference ref : params.values()) {
                        parDomains[i] = ((EnumerationType) ref).getEnums().size();
                        i++;
                    }

                    eventTable.addMethodEncoder(new MethodEncoderImpl(
                            method.getFullname(), retDomain, parDomains));
                }
            }
        }
    }

    /**
     * Transforms a single provision into its compact DFA form.
     *
     * @param provision A resolved provision to be transformed into a
     *      corresponding automaton.
     * @param eventTable An EventTable to be used for event translation into
     *      their long codes.
     * @return Compact representation of the provision.
     */
    public static <REFERENCE extends Reference> Observer transformProvision(
            TBPProvisionContainerNode<REFERENCE> provision,
            EventTable eventTable) {

        TranslateProvisionVisitor<REFERENCE> visitor = new TranslateProvisionVisitor<REFERENCE>(eventTable);

        DynamicNFA nfa = provision.visit(visitor);
        DynamicDFA dfa = nfa.determinize();
        nfa = null;
        dfa = dfa.makeMinimal();

        Set<Integer> filter = new HashSet<Integer>();

        for (REFERENCE ref : provision.getFilter()) {
            filter.add(eventTable.encodeMethod(((Method) ref).getFullname()));
        }

        return CompactDFA.makeCompact(dfa, eventTable, filter);
    }

    public static <REFERENCE extends Reference> List<Observer> transformProvisions(
            List<TBPSpecification<REFERENCE>> tbps, EventTable eventTable) {

        List<Observer> dfas = new ArrayList<Observer>();

        for (TBPSpecification<REFERENCE> tbp : tbps) {
            for (TBPProvisionContainerNode<REFERENCE> provision : tbp.getProvisions()) {
                dfas.add(transformProvision(provision, eventTable));
            }
        }

        return dfas;
    }

    public static void replaceArbitraryValuesInImperativePartsByAlternatives(TBPSpecification<Reference> tbp){
        ReplaceArbitraryParameterVisitor vis = new ReplaceArbitraryParameterVisitor();

        List<TBPMethodDeclaration<Reference>> reactions = tbp.getReactions();
        for (TBPMethodDeclaration<Reference> md:reactions){
            TBPNode<Reference> child = md.getChild();
            TBPImperativeNode<Reference> childCopy = child.visit(vis);
            md.setChild(childCopy);
        }

        List<TBPThreadContainerNode<Reference>> threads = tbp.getThreads();
        for (TBPThreadContainerNode<Reference> md:threads){
            TBPNode<Reference> child = md.getChild();
            TBPImperativeNode<Reference> childCopy = child.visit(vis);
            md.setChild(childCopy);
        }
    }
    
    /** Transforms {@link LTSSpecification LTS} to {@link LTSASpecification LTSA}
     * 
     * @param <R> reference type
     * @param ltsSpecification {@link LTSSpecification}
     * @return {@link LTSASpecification}
     */
    public static <R> LTSASpecification<TBPNode<R>, R> transformToLTSA(LTSSpecification<R> ltsSpecification)
    {
        LTSToLTSAConversionVisitor<R> ltsConvertor = new LTSToLTSAConversionVisitor<R>();
        return ltsConvertor.convertToLTSASpecification(ltsSpecification);
    }

   /** Searches architecture for unbound interfaces.
    * 
    * <p>Unbound provided and required interfaces will be added to <b>unboundProvidedInterfaces</b>
    *    and <b>unboundRequiredInterfaces</b> collections respectively.
    * </p>
    *
    * @param arch architecture
    * @param unboundProvidedInterfaces a collection, where all unbound provided interfaces will be added to
    * @param unboundRequiredInterfaces a list, where all unbound required interfaces will be added to
    */
    public static void findUnboundInterfaces(
    		Architecture arch,
    		Collection<Interface> unboundProvidedInterfaces,
    		Collection<Interface> unboundRequiredInterfaces) {
    	
		Set<Interface> boundProvidedInterfaces = new java.util.HashSet<Interface>();
		for(Component component : arch.getComponents()) {
			TBPLibLog.logger.finest(String.format("Inspecting component %s.", component.getName()));
			for(Interface iface: component.getRequiredInterfaces().values()) {
				if (!iface.isBound()) {
					TBPLibLog.logger.finer(String.format("Required inteface %s.%s is not bound.", component.getName(), iface.getName()));
					unboundRequiredInterfaces.add(iface);
				} else {
					TBPLibLog.logger.finest(String.format("Required interface %s.%s already bound to %s.%s.", component.getName(), iface.getName(), iface.getBoundInterface().getParent().getName(), iface.getBoundInterface().getName()));
					boundProvidedInterfaces.add(iface.getBoundInterface());
				}
			}
		}
		for(Component component : arch.getComponents()) {
			for(Interface iface: component.getProvidedInterfaces().values()) {
				if (!iface.isBound()) {
					if (!boundProvidedInterfaces.contains(iface)) {
						unboundProvidedInterfaces.add(iface);
						TBPLibLog.logger.finer(String.format("Provided inteface %s.%s is not bound.", component.getName(), iface.getName()));
					} else {
						TBPLibLog.logger.finer(String.format("Provided interface %s.%s marked as unbound, but another required interface is bound to it.", component.getName(), iface.getName()));
					}
				} else {
					TBPLibLog.logger.finest(String.format("Provided interface %s.%s already bound to %s.%s.", component.getName(), iface.getName(), iface.getBoundInterface().getParent().getName(), iface.getBoundInterface().getName()));
				}
			}
		}
    }
    
    /** Couples interfaces with appropriate provisions.
     * 
     * @param specifications a list of TBPSpecifications 
     * @param interfaces a list of interfaces
     * @return mapping of provisions to interfaces associated with them 
     */
    public static <REFERENCE extends Reference> Map<TBPProvisionContainerNode<REFERENCE>, Collection<Interface>> findProvisionsForInterfaces(
    		Collection<TBPSpecification<REFERENCE>> specifications,
    		Collection<Interface> interfaces) {
			   
		Map<TBPProvisionContainerNode<REFERENCE>, Collection<Interface>> provisionInterfaceMap
			= new HashMap<TBPProvisionContainerNode<REFERENCE>, Collection<Interface>>();
			   
		List<Interface> filteredInterfaces = new LinkedList<Interface>();
		for (TBPSpecification<REFERENCE> specification: specifications) {
			filteredInterfaces.clear();
			String componentName = specification.getComponent().getName();
			for(Interface iface : interfaces) {
				if (componentName.equals(iface.getParent().getName())) {
					filteredInterfaces.add(iface);
				}
			}
			   	
			//filtered interfaces contains just interfaces of components
			// with same name as component of *specification*
			for(TBPProvisionContainerNode<REFERENCE> provision : specification.getProvisions()) {
				boolean doesProvisionReferUnboundInterface = false;
				boolean doesProvisionReferBoundInterface = false;
				Collection<Interface> provisionInterfaces = new LinkedList<Interface>();
				for(Reference filterMemer : provision.getFilter()) {
					Method m;
					if (filterMemer instanceof Method) {
						m = (Method)filterMemer;
						Interface i = m.getParent();
						if (interfaces.contains(i)) {
							doesProvisionReferUnboundInterface = true;
						} else {
							doesProvisionReferBoundInterface = true;
						}
						provisionInterfaces.add(i);
					}
					if (doesProvisionReferUnboundInterface && doesProvisionReferBoundInterface) {
						String provisionName = provision.getName();
						provisionName = provisionName == null ? "" : provisionName;
						throw new UnsupportedOperationException(
								String.format(
										provisionName.isEmpty()
										? "A provision refers both bound and unbound interfaces.%s"
										: "Provision %s refers both bound and unbound interfaces which is not supported.",
											provisionName)
								);
					} else if (doesProvisionReferUnboundInterface) {
						provisionInterfaceMap.put(provision, provisionInterfaces);
					}
				}
			}	
		}
		return provisionInterfaceMap; 
    }
    
	/** Transform provisions to actuators. 
	 *   
	 * @param provisionsMap map of provisions to interfaces
	 * @param variableFactory factory for creating aux. variables within actuators  
	 * @return map of provision actuators (LTSA) to interfaces
	 */
    public static <REFERENCE> Map<LTSASpecification<TBPNode<REFERENCE>, REFERENCE>, Collection<Interface>> transformProvisionsToActuators(
    		Map<TBPProvisionContainerNode<REFERENCE>, Collection<Interface>> provisionsMap,
    		VariableFactory<REFERENCE> variableFactory) { 
    
    
		TranslateProvisionToSymbolicActuatorVisitor<REFERENCE> translatorToSA
			= new TranslateProvisionToSymbolicActuatorVisitor<REFERENCE>();
		TranslateToLTSAVisitor<REFERENCE> translatorToLTSA 
			= new TranslateToLTSAVisitor<REFERENCE>(variableFactory);
		Map<LTSASpecification<TBPNode<REFERENCE>, REFERENCE>, Collection<Interface>> actuatorsMap
			= new HashMap<LTSASpecification<TBPNode<REFERENCE>, REFERENCE>, Collection<Interface>>();
   
		for (Map.Entry<TBPProvisionContainerNode<REFERENCE>, Collection<Interface>> provisionEntry : provisionsMap.entrySet()) {
			TBPProvisionContainerNode<REFERENCE> provision = provisionEntry.getKey();        	
			ActionSeqNode<REFERENCE> a = provision.visit(translatorToSA);
		 	LTSASpecification<TBPNode<REFERENCE>, REFERENCE> actuatorLTSA = translatorToLTSA.translate(a);
		 	actuatorsMap.put(actuatorLTSA, provisionEntry.getValue());
		}
		return actuatorsMap;
    }
    
    /** Removes <b>provisions</b> from <p>specifications</b>.
     * 
     * <p>Current implementation removes the provisions in place
     *    (it does not perform any copying of specifications).
     *    But keep in mind that it must not hold in general.
     * </p> 
     * 
     * @param <REFERENCE>
     * @param specifications {@link TBPSpecification specification}s
     * @param provisions {@link TBPProvisionContainerNode provision}s
     * @return {@link TBPSpecification specification}s where given provisions has been removed
     */
    public static <REFERENCE> List<TBPSpecification<REFERENCE>> removeProvisionsFromSpecifications(
    		List<TBPSpecification<REFERENCE>> specifications,
    		Collection<TBPProvisionContainerNode<REFERENCE>> provisions) {
    	LinkedList<TBPProvisionContainerNode<REFERENCE>> provisionsList = new LinkedList<TBPProvisionContainerNode<REFERENCE>>(provisions);
    	for(TBPSpecification<REFERENCE> specification: specifications) {
    		Iterator<TBPProvisionContainerNode<REFERENCE>> it = provisionsList.iterator();
    		while(it.hasNext()) {
    			TBPProvisionContainerNode<REFERENCE> provision = it.next();
    			if (specification.getProvisions().remove(provision)) {
    				it.remove();
    			}
    		}
    	}
    	return specifications;
    }
}
