/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.ltsa;

import org.junit.Test;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl;

import java.util.Arrays;
import java.util.HashSet;

import static org.junit.Assert.*;

/**
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
public class AutomatonTest {

    EventTable et;

    public AutomatonTest() {
        et = new EventTableImpl();
        et.addMethodEncoder(new EventTableImpl.MethodEncoderImpl("i.m1", 0));
        et.addMethodEncoder(new EventTableImpl.MethodEncoderImpl("i.m2", 0));
        et.addMethodEncoder(new EventTableImpl.MethodEncoderImpl("i.m3", 2, 5, 2));
        et.addMethodEncoder(new EventTableImpl.MethodEncoderImpl("i.m4", 10, 10, 10));
        et.addMethodEncoder(new EventTableImpl.MethodEncoderImpl("i.m5", 3, 3, 3));
    }

    /**
     * -> o --"0"-> x
     *      \-"0"-> o
     */
    @Test
    public void testAutomata1() {
        DynamicNFA nfa = new DynamicNFA();

        int init = nfa.allocateState();
        int s1 = nfa.allocateState();
        int s2 = nfa.allocateState();
        long e1 = et.encodeMethodCall(0);
        long e2 = et.encodeMethodCall(1);

        nfa.setInitial(init);
        nfa.markFinal(s1);

        nfa.addEdge(init, e1, s1);
        nfa.addEdge(init, e1, s2);

        DynamicDFA result = nfa.determinize();
        assertEquals(result.getNumberOfState(), 2);
        result = result.makeMinimal();
        assertEquals(result.getNumberOfState(), 2);

        assertFalse(acceptsWord(result, e2));
        assertTrue(acceptsWord(result, e1));
        assertFalse(acceptsWord(result, e1, e1));
        assertFalse(acceptsWord(result, e2, e2));

        // Compact the automaton and examine again
        CompactDFA compact = CompactDFA.makeCompact(result, et, new HashSet<Integer>(Arrays.asList(0, 1)));

        assertEquals(compact.getNumberOfState(), 2);

        assertFalse(acceptsWord(compact, e2));
        assertTrue(acceptsWord(compact, e1));
        assertFalse(acceptsWord(compact, e1, e1));
        assertFalse(acceptsWord(compact, e2, e2));
    }

    /**
     * -> o --"0"-> x
     *      \-"1"-> o
     */
    @Test
    public void testAutomata2() {
        DynamicNFA nfa = new DynamicNFA();

        int init = nfa.allocateState();
        int s1 = nfa.allocateState();
        int s2 = nfa.allocateState();
        long e1 = et.encodeMethodCall(0);
        long e2 = et.encodeMethodCall(1);

        nfa.setInitial(init);
        nfa.markFinal(s1);

        nfa.addEdge(init, e1, s1);
        nfa.addEdge(init, e2, s2);

        DynamicDFA result = nfa.determinize();
        assertEquals(result.getNumberOfState(), 3);
        result = result.makeMinimal();
        assertEquals(result.getNumberOfState(), 2);

        assertFalse(acceptsWord(result, e2));
        assertTrue(acceptsWord(result, e1));
        assertFalse(acceptsWord(result, e1, e1));

        // Compact the automaton and examine again
        CompactDFA compact = CompactDFA.makeCompact(result, et, new HashSet<Integer>(Arrays.asList(0, 1)));

        assertEquals(compact.getNumberOfState(), 2);

        assertFalse(acceptsWord(compact, e2));
        assertTrue(acceptsWord(compact, e1));
        assertFalse(acceptsWord(compact, e1, e1));
    }

    @Test
    public void testAutomata3() {
        DynamicNFA nfa = new DynamicNFA();

        int init = nfa.allocateState();
        int de1 = nfa.allocateState();
        int de2 = nfa.allocateState();
        int s1 = nfa.allocateState();
        int s2 = nfa.allocateState();
        int s3 = nfa.allocateState();
        int s4 = nfa.allocateState();
        int s5 = nfa.allocateState();
        int s6 = nfa.allocateState();
        int s7 = nfa.allocateState();
        int s8 = nfa.allocateState();
        int s9 = nfa.allocateState();
        long e1 = et.encodeMethodCall(0);
        long e2 = et.encodeMethodCall(1);

        nfa.setInitial(init);
        nfa.markFinal(s4);
        nfa.markFinal(s9);
        nfa.markFinal(de2);

        nfa.addEdge(de1, e1, de2);
        nfa.addEdge(de2, e2, de1);

        nfa.addEdge(init, e1, s1);
        nfa.addEdge(init, e1, s5);
        nfa.addEdge(init, e1, s7);
        nfa.addEdge(s1, e2, init);
        nfa.addEdge(s1, e1, s2);
        nfa.addEdge(s1, e1, s6);
        nfa.addEdge(s2, e1, s3);
        nfa.addEdge(s2, e2, s4);
        nfa.addEdge(s5, e2, init);
        nfa.addEdge(s5, e1, s6);
        nfa.addEdge(s6, e2, s4);
        nfa.addEdge(s7, e2, init);
        nfa.addEdge(s7, e1, s8);
        nfa.addEdge(s8, e2, s7);
        nfa.addEdge(s8, e2, s9);

        DynamicDFA result = nfa.determinize();
        assertEquals(7, result.getNumberOfState());
        result = result.makeMinimal();
        assertEquals(4, result.getNumberOfState());

        assertFalse(acceptsWord(result, e2, e1, e1));
        assertFalse(acceptsWord(result, e1, e1, e1));
        assertFalse(acceptsWord(result, e1, e1, e2, e2));
        assertFalse(acceptsWord(result, e1, e1, e2, e1, e2, e1, e2, e2));
        assertFalse(acceptsWord(result, e1, e1, e2, e1, e2, e1, e2, e1));
        assertTrue(acceptsWord(result, e1, e1, e2));
        assertTrue(acceptsWord(result, e1, e1, e2, e1, e2, e1, e2));

        // Compact the automaton and examine again
        CompactDFA compact = CompactDFA.makeCompact(result, et, new HashSet<Integer>(Arrays.asList(0, 1)));

        assertEquals(4, compact.getNumberOfState());

        assertFalse(acceptsWord(compact, e2, e1, e1));
        assertFalse(acceptsWord(compact, e1, e1, e1));
        assertFalse(acceptsWord(compact, e1, e1, e2, e2));
        assertFalse(acceptsWord(compact, e1, e1, e2, e1, e2, e1, e2, e2));
        assertFalse(acceptsWord(compact, e1, e1, e2, e1, e2, e1, e2, e1));
        assertTrue(acceptsWord(compact, e1, e1, e2));
        assertTrue(acceptsWord(compact, e1, e1, e2, e1, e2, e1, e2));
    }


    @Test
    public void testAutomata4() {
        DynamicNFA nfa = new DynamicNFA();

        int init = nfa.allocateState();
        int s1 = nfa.allocateState();
        int s2 = nfa.allocateState();
        int s3 = nfa.allocateState();
        int s4 = nfa.allocateState();
        long e1 = et.encodeMethodCall(1);
        long e2 = et.encodeMethodCall(2, (byte)1, (byte)1);
        long e3 = et.encodeMethodReturn(2, (byte)1);
        long e4 = et.encodeMethodCall(3, (byte)7, (byte)7);
        long e5 = et.encodeMethodCall(4, (byte)2, (byte)2);
        long e6 = et.encodeMethodCall(4, (byte)0, (byte)1);
        long e7 = et.encodeMethodReturn(4, (byte)1);

        nfa.setInitial(init);
        nfa.markFinal(s4);

        nfa.addEdge(init, e2, s1);
        nfa.addEdge(init, e5, s1);
        nfa.addEdge(init, e6, s1);

        nfa.addEdge(s1, e3, s2);
        nfa.addEdge(s1, e7, s2);

        nfa.addEdge(s2, e5, s3);
        nfa.addEdge(s2, e6, s3);

        nfa.addEdge(s3, e7, s4);

        DynamicDFA result = nfa.determinize();
        assertEquals(5, result.getNumberOfState());
        result = result.makeMinimal();
        assertEquals(5, result.getNumberOfState());

        assertFalse(acceptsWord(result, e1, e3, e5, e7));
        assertFalse(acceptsWord(result, e2, e2, e5, e7));
        assertFalse(acceptsWord(result, e2, e3));
        assertFalse(acceptsWord(result, e2, e4));
        assertFalse(acceptsWord(result));
        assertTrue(acceptsWord(result, e2, e3, e5, e7));
        assertTrue(acceptsWord(result, e6, e7, e6, e7));

        // Compact the automaton and examine again
        CompactDFA compact = CompactDFA.makeCompact(result, et, new HashSet<Integer>(Arrays.asList(2,4)));

        assertEquals(5, compact.getNumberOfState());

        assertFalse(compact.observesEvent(e1));
        assertFalse(compact.observesEvent(e4));
        assertTrue(compact.observesEvent(e2));

        assertFalse(compact.observesMethod(0));
        assertFalse(compact.observesMethod(3));
        assertTrue(compact.observesMethod(2));
        assertTrue(compact.observesMethod(4));

        assertFalse(acceptsWord(compact, e2, e2, e5, e7));
        assertFalse(acceptsWord(compact, e2, e3));
        assertFalse(acceptsWord(compact));
        assertTrue(acceptsWord(compact, e2, e3, e5, e7));
        assertTrue(acceptsWord(compact, e6, e7, e6, e7));
    }

    @Test
    public void testAutomata5() {
        DynamicNFA nfa = new DynamicNFA();

        int init = nfa.allocateState();
        int s1 = nfa.allocateState();
        int s2 = nfa.allocateState();
        int s3 = nfa.allocateState();
        long e1 = et.encodeMethodCall(2, (byte)1, (byte)1);
        long e2 = et.encodeMethodReturn(2, (byte)1);
        long e3 = et.encodeMethodCall(4, (byte)2, (byte)2);
        long e4 = et.encodeMethodReturn(4, (byte)1);

        nfa.setInitial(init);
        nfa.markFinal(s2);
        nfa.markFinal(init);

        nfa.addEdge(init, e1, s1);
        nfa.addEdge(init, e3, s3);

        nfa.addEdge(s1, e2, s2);
        nfa.addEdge(s3, e4, s2);

        nfa.addEdge(s2, e1, s1);
        nfa.addEdge(s2, e3, s3);

        DynamicDFA result = nfa.determinize();
        assertEquals(4, result.getNumberOfState());
        result = result.makeMinimal();
        assertEquals(3, result.getNumberOfState());

        assertFalse(acceptsWord(result, e1, e3));
        assertFalse(acceptsWord(result, e3, e4, e1));
        assertFalse(acceptsWord(result, e1));
        assertTrue(acceptsWord(result));
        assertTrue(acceptsWord(result, e1, e2, e3, e4));

        // Compact the automaton and examine again
        CompactDFA compact = CompactDFA.makeCompact(result, et, new HashSet<Integer>(Arrays.asList(2,4)));

        assertEquals(3, compact.getNumberOfState());

        assertFalse(acceptsWord(compact, e1, e3));
        assertFalse(acceptsWord(compact, e1, e1));
        assertFalse(acceptsWord(compact, e3, e4, e1));
        assertFalse(acceptsWord(compact, e1));
        assertTrue(acceptsWord(compact));
        assertTrue(acceptsWord(compact, e1, e2, e3, e4));
    }

    private boolean acceptsWord(DFA dfa, long... word) {
        int current = dfa.initialState();

        for (long event : word) {
            current = dfa.nextState(current, event);

            if (current == DFA.NO_SUCH_TRANSITION)
                return false;
        }
        return dfa.isFinal(current);
    }
}