/*
 *
 * Threaded Behavior Protocols  - Parsers, Transformations
 * Copyright (C) 2008   DSRG, Charles University in Prague
 *                      http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.tbplib.parser;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.ow2.dsrg.fm.tbplib.TBPResolvingException;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl;
import org.ow2.dsrg.fm.tbplib.node.ParserResult;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.util.TransformationChain;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Test cases for the TBP Parser.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
@RunWith(Parameterized.class)
public class ParserTest {

    private static String testPrefix = "test/input/simple/";

    @Parameterized.Parameters
    public static Collection<Object[]> testInputs() {
        return Arrays.asList(new Object[][]{
                    {"variables/01_ok.tbp", null}, 
                    {"variables/02_ok.tbp", null},
                    {"variables/03_ok.tbp", null},
                    {"variables/04_missing_type.tbp", TBPResolvingException.class},
                    {"variables/05_enum_conflict.tbp", TBPResolvingException.class},
                    {"variables/06_ok.tbp", null},
                    {"variables/07_type_missmatch.tbp", TBPResolvingException.class},
                    {"variables/08_type_missmatch.tbp", TBPResolvingException.class},
                    {"variables/09_type_missmatch.tbp", TBPResolvingException.class},
                    {"variables/10_type_missmatch.tbp", TBPResolvingException.class},
                    {"provisions/01_ok.tbp", null}
                });
    }
    ;
    public String file;
    public Class exClass;

    public ParserTest(String file, Class exClass) {
        this.file = file;
        this.exClass = exClass;
    }

    private void parseFile(String filename) throws Exception {
        EventTable eventTable = new EventTableImpl();

        System.out.println(" (II) Analyzing file '" + filename + "'");
        // Parsing
        Architecture arch = new Architecture("testing-arch");
        List<TBPSpecification<String>> tbps = new ArrayList<TBPSpecification<String>>(1);
        ParserResult pr = ParserFacade.parseFile(filename);
        TBPSpecification<String> tbp =pr.getSpecification();
        arch.addComponent(new Component(tbp.getComponent(),pr.getTypes()));

        tbps.add(tbp);
        System.out.println(" Parsing: OK");

        // Symbol collection
        TransformationChain.collectSymbolsFromArchitecture(tbps, arch);
        System.out.println(" Symbol collection: OK");

        // Implicit bindings
        TransformationChain.performImplicitBinding(arch);
        System.out.println(" Implicit binding: OK");

        // Fill in event table
        TransformationChain.fillEventTable(arch, eventTable);
        System.out.println(" Filling event table: OK");

        // Resolving
        List<TBPSpecification<Reference>> rtbps =
                TransformationChain.resolveArchitecture(tbps, arch);
        System.out.println(" Resolving: OK");

        // Translating provisions
        TransformationChain.transformProvisions(rtbps, eventTable);
        System.out.println(" Provisions --> DFA: OK");
    }

    @Test
    public void simpleTest() throws Exception {
        String testFile = testPrefix + file;
        try {
            parseFile(testFile);
        } catch (Exception ex) {
            if (ex.getClass().equals(exClass)) {
                System.out.println("An exception " + exClass.getCanonicalName() + " thrown as expected.");
                return;
            } else {
                System.out.println("An unexpected exception " + ex.getClass().getCanonicalName() + " thrown.");
                throw ex;
            }
        }
        if (exClass != null) {
            Assert.fail("An exception " + exClass.getCanonicalName() + " expected.");
        }
    }
}
