/*
 *  $Id$
 * 
 *   Threaded Behavior Protocols  - Parsers, Transformations
 *   Copyright (C) 2009       DSRG, Charles University in Prague
 *                            http://dsrg.mff.cuni.cz/
 * 
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 * 
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 * 
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *   MA 02110-1301 USA
 * 
 */

package org.ow2.dsrg.fm.tbplib.parser;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.ow2.dsrg.fm.tbplib.architecture.Architecture;
import org.ow2.dsrg.fm.tbplib.architecture.Component;
import org.ow2.dsrg.fm.tbplib.event.EventTable;
import org.ow2.dsrg.fm.tbplib.event.EventTableImpl;
import org.ow2.dsrg.fm.tbplib.node.ParserResult;
import org.ow2.dsrg.fm.tbplib.node.TBPSpecification;
import org.ow2.dsrg.fm.tbplib.reference.Reference;
import org.ow2.dsrg.fm.tbplib.util.CodeWriter;
import org.ow2.dsrg.fm.tbplib.util.TransformationChain;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Test that parses a TBP, prints it, and parses again.
 *
 * @author Ondrej Sery <ondrej.sery@dsrg.mff.cuni.cz>
 */
@RunWith(Parameterized.class)
public class PrintTest {

    private static String testPrefix = "test/input/simple/";

    @Parameterized.Parameters
    public static Collection<Object[]> testInputs() {
        return Arrays.asList(new Object[][]{
                    {"variables/01_ok.tbp", null},
                    {"variables/02_ok.tbp", null},
                    {"variables/03_ok.tbp", null},
                    {"variables/06_ok.tbp", null},
                    {"provisions/01_ok.tbp", null}
                });
    }
    ;
    public String file;
    public Class exClass;

    public PrintTest(String file, Class exClass) {
        this.file = file;
        this.exClass = exClass;
    }

    private void parseFile(String filename) throws Exception {
        EventTable eventTable = new EventTableImpl();

        System.out.println(" (II) Analyzing file '" + filename + "'");
        // Parsing
        Architecture arch = new Architecture("testing-arch");
        List<TBPSpecification<String>> tbps = new ArrayList<TBPSpecification<String>>(1);
        ParserResult pr = ParserFacade.parseFile(filename);
        TBPSpecification<String> tbp =pr.getSpecification();
        arch.addComponent(new Component(tbp.getComponent(),pr.getTypes()));
        tbps.add(tbp);
        System.out.println(" Parsing: OK");

        // Symbol collection
        TransformationChain.collectSymbolsFromArchitecture(tbps, arch);
        System.out.println(" Symbol collection: OK");

        // Implicit bindings
        TransformationChain.performImplicitBinding(arch);
        System.out.println(" Implicit binding: OK");

        // Fill in event table
        TransformationChain.fillEventTable(arch, eventTable);
        System.out.println(" Filling event table: OK");

        // Resolving
        List<TBPSpecification<Reference>> rtbps =
                TransformationChain.resolveArchitecture(tbps, arch);
        System.out.println(" Resolving: OK");

        // Translating provisions
        TransformationChain.transformProvisions(rtbps, eventTable);
        System.out.println(" Provisions --> DFA: OK");

        ByteArrayOutputStream bfr = new ByteArrayOutputStream();
        TransformationChain.printTBP(new CodeWriter(bfr), tbp, arch.getComponents().get(0));
        ByteArrayInputStream in = new ByteArrayInputStream(bfr.toByteArray());
        System.out.println(" Printing TBP in Memory: OK");

        System.out.print(bfr.toString());

        // Parsing again
        eventTable = new EventTableImpl();
        arch = new Architecture("testing-arch");
        tbps = new ArrayList<TBPSpecification<String>>(1);

        ParserResult pr2 = ParserFacade.parseFile(in);
        TBPSpecification<String> tbp2 =pr.getSpecification();
        arch.addComponent(new Component(tbp.getComponent(),pr.getTypes()));

        tbps.add(tbp);
        System.out.println(" Parsing in Memory: OK");

        // Symbol collection
        TransformationChain.collectSymbolsFromArchitecture(tbps, arch);
        System.out.println(" Symbol collection: OK");

        // Implicit bindings
        TransformationChain.performImplicitBinding(arch);
        System.out.println(" Implicit binding: OK");

        // Fill in event table
        TransformationChain.fillEventTable(arch, eventTable);
        System.out.println(" Filling event table: OK");

        // Resolving
        rtbps =
                TransformationChain.resolveArchitecture(tbps, arch);
        System.out.println(" Resolving: OK");

        // Translating provisions
        TransformationChain.transformProvisions(rtbps, eventTable);
        System.out.println(" Provisions --> DFA: OK");
    }

    @Test
    public void simpleTest() throws Exception {
        String testFile = testPrefix + file;
        parseFile(testFile);
    }
}
